@@ -1,7 +1,7 @@
 /*
  * binary.c - Binary I/O routines
  *
- *   Copyright (c) 2004-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2004-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -527,15 +527,3 @@ void Scm_PutBinaryF64(ScmUVector *uv, int off, ScmObj val, ScmSymbol *e)
     SWAP_D(e, v);
     inject(uv, v.buf, off, 8);
 }
-
-/*
- * Init
- */
-extern void Scm_Init_binarylib(ScmModule *mod);
-
-SCM_EXTENSION_ENTRY void Scm_Init_binary__io(void)
-{
-    ScmModule *mod_io = SCM_FIND_MODULE("binary.io", SCM_FIND_MODULE_CREATE);
-    SCM_INIT_EXTENSION(binary__io);
-    Scm_Init_binarylib(mod_io);
-}
