@@ -3,7 +3,7 @@
 (use gauche.test)
 (test-start "collections and sequences")
 
-(use util.queue)
+(use data.queue)
 (use srfi-1)
 (use srfi-13)
 (use gauche.collection)
@@ -558,6 +558,55 @@
        (group-sequence '(1 1 1 2 3 4 4 2 2 3 1 1 3)
                        :test (^[x y] (= (modulo x 2) (modulo y 2)))))
 
+(test* "sequence-contains" '(#f 0 9 #f)
+       (let1 pat "abrabrabre"
+         (list (sequence-contains "abracadabra" pat)
+               (sequence-contains "abrabrabrebrea" pat)
+               (sequence-contains "abrabrabrabrabrabrebra" pat)
+               (sequence-contains "abrabrabrabrabrabrabra" pat))))
+(test* "sequence-contains" 0
+       (sequence-contains "abracadabra" '()))
+
+(let ([data
+       ;; (input needle before after)
+       '(((1 2 3 4 5 4 5 6 7 8 9) (4 5 4 5 6) (1 2 3) (4 5 4 5 6 7 8 9))
+         ((4 5 4 5 6)             (4 5 4 5 6) ()      (4 5 4 5 6))
+         ((4 5 4 5 5 4 5 4 5 6)   (4 5 4 5 6) (4 5 4 5 5) (4 5 4 5 6))
+         (()                      (4 5 4 5 6) ()      ())
+         ((4 5 6 4 5 4 5 1)       (4 5 4 5 6) (4 5 6 4 5 4 5 1) ())
+         ((4 5 6 4 5 4 5 1)       ()          () (4 5 6 4 5 4 5 1))
+         )])
+  (define (tester !? in needle before after)
+    (test* (format "break-list-by-sequence~a ~s" (if !? '! "") in)
+           (list before after)
+           (receive (a b) ((if !?
+                             break-list-by-sequence!
+                             break-list-by-sequence)
+                           in needle)
+             (and (or (not !?)
+                      (null? a)
+                      (eq? a in))  ; check if head part is shared
+                  (list a b)))))
+    
+  (dolist [datum data]
+    (apply tester #f datum)
+    (apply tester #t datum)))
+  
+(test* "common-prefix" '(a b c)
+       (common-prefix-to <list> '(a b c d e) '(a b c e d)))
+(test* "common-prefix" '(a b c)
+       (common-prefix '(a b c d e) '#(a b c e d)))
+(test* "common-prefix" '#(a b c)
+       (common-prefix '#(a b c e d) '(a b c d e)))
+(test* "common-prefix" "xyz"
+       (common-prefix "xyz" "xyzw"))
+(test* "common-prefix" ""
+       (common-prefix "xyz" "abc"))
+(test* "common-prefix" '()
+       (common-prefix '(a b c) ""))
+(test* "common-prefix" "ABC"
+       (common-prefix "ABCE" "abcd" :test char-ci=?))
+
 (define (permute-tester msg expected source order . fallback)
   (define (unit type elt-coercer order-type)
     (test* #"permute~msg ~type by ~order-type"
