@@ -1,6 +1,8 @@
 (include "test-colseq.scm")
 (include "test-hook.scm")
 (include "test-parameter.scm")
+(include "test-pp.scm")
 (include "test-record.scm")
 (include "test-generator.scm")
 (include "test-lazy.scm")
+(include "test-unicode.scm")
