@@ -2,7 +2,7 @@
  * mt-random.c - implements MT19937 random number generation algorithm
  *
  * Mersenne Twister algorithm invented by Makoto Matsumoto & Takuji Nishimura.
- *   http://www.math.keio.ac.jp/~matumoto/emt.html
+ *   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
  * This code is based on mt18837ar.c (2002/1/16).   It is released under
  * BSD license.
  * I modified the code in the following parts.
@@ -52,8 +52,8 @@
 
 
    Any feedback is very welcome.
-   http://www.math.keio.ac.jp/matumoto/emt.html
-   email: matumoto@math.keio.ac.jp
+   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
+   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
 */
 
 #include <math.h>
@@ -302,15 +302,11 @@ static ScmObj mt_allocate(ScmClass *klass, ScmObj initargs)
     return SCM_OBJ(mt);
 }
 
-extern void Scm_Init_mt_lib(ScmModule*);
-
-SCM_EXTENSION_ENTRY void Scm_Init_math__mt_random(void)
+void Scm_Init_mt_random(void)
 {
     ScmModule *mod = SCM_FIND_MODULE("math.mt-random", SCM_FIND_MODULE_CREATE);
-    SCM_INIT_EXTENSION(math__mtrandom);
     Scm_InitStaticClass(&Scm_MersenneTwisterClass, "<mersenne-twister>",
                         mod, NULL, 0);
     key_seed = SCM_MAKE_KEYWORD("seed");
-    Scm_Init_mt_lib(mod);
 }
 
