@@ -1,7 +1,7 @@
 /*
  * net.c - network interface
  *
- *   Copyright (c) 2000-2012  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -50,6 +50,9 @@ static void socket_finalize(ScmObj obj, void *data)
        be GC-ed and finalized, so we don't flush them here.
        Clearing them would help them to be collected earlier, though. */
     if (!(SOCKET_CLOSED(sock->fd))) {
+#if defined(GAUCHE_WINDOWS)
+        if (sock->cfd >= 0) { close(sock->cfd); sock->cfd = -1; }
+#endif /*GAUCHE_WINDOWS*/
         closeSocket(sock->fd);
         sock->fd = INVALID_SOCKET;
         sock->status = SCM_SOCKET_STATUS_CLOSED;
@@ -97,6 +100,9 @@ ScmSocket *make_socket(Socket fd, int type)
     s->address = NULL;
     s->name = NULL;
     s->type = type;
+#if defined(GAUCHE_WINDOWS)
+    s->cfd = -1;
+#endif /*GAUCHE_WINDOWS*/
     Scm_RegisterFinalizer(SCM_OBJ(s), socket_finalize, NULL);
     return s;
 }
@@ -149,6 +155,9 @@ ScmObj Scm_SocketClose(ScmSocket *s)
         Scm_ClosePort(s->outPort); /* ignore errors */
         s->outPort = NULL;
     }
+#if defined(GAUCHE_WINDOWS)
+    if (s->cfd >= 0) { close(s->cfd); s->cfd = -1; }
+#endif /*GAUCHE_WINDOWS*/
     closeSocket(s->fd);
     s->fd = INVALID_SOCKET;
     s->status = SCM_SOCKET_STATUS_CLOSED;
@@ -165,26 +174,29 @@ static void sockport_err(ScmSocket *sock, const char *io)
 ScmObj Scm_SocketInputPort(ScmSocket *sock, int buffering)
 {
     if (sock->inPort == NULL) {
-	ScmObj sockname;
-	int infd;
+        int infd;
         if (sock->type != SOCK_DGRAM &&
             sock->status < SCM_SOCKET_STATUS_CONNECTED) {
             sockport_err(sock, "input");
         }
 #ifndef GAUCHE_WINDOWS
-	infd = sock->fd;
+        infd = sock->fd;
 #else  /*GAUCHE_WINDOWS*/
-	infd = _open_osfhandle(sock->fd, O_RDONLY);
+        /* cfd will be closed when this socket is closed. */
+        if (sock->cfd < 0) {
+            sock->cfd = _open_osfhandle(sock->fd, 0);
+        }
+        infd = sock->cfd;
 #endif /*GAUCHE_WINDOWS*/
         if (infd == INVALID_SOCKET) sockport_err(sock, "input");
 
         /* NB: I keep the socket itself in the port name, in order to avoid
            the socket from GCed prematurely if application doesn't keep
            pointer to the socket. */
-	sockname = SCM_LIST2(SCM_MAKE_STR("socket input"), SCM_OBJ(sock));
+        ScmObj sockname = SCM_LIST2(SCM_MAKE_STR("socket input"),
+                                    SCM_OBJ(sock));
         sock->inPort = SCM_PORT(Scm_MakePortWithFd(sockname, SCM_PORT_INPUT,
-						   infd, buffering,
-						   FALSE));
+                                                   infd, buffering, FALSE));
     }
     return SCM_OBJ(sock->inPort);
 }
@@ -192,25 +204,29 @@ ScmObj Scm_SocketInputPort(ScmSocket *sock, int buffering)
 ScmObj Scm_SocketOutputPort(ScmSocket *sock, int buffering)
 {
     if (sock->outPort == NULL) {
-	ScmObj sockname;
-	int outfd;
+        int outfd;
         if (sock->type != SOCK_DGRAM &&
             sock->status < SCM_SOCKET_STATUS_CONNECTED) {
             sockport_err(sock, "output");
         }
 #ifndef GAUCHE_WINDOWS
-	outfd = sock->fd;
+        outfd = sock->fd;
 #else  /*GAUCHE_WINDOWS*/
-	outfd = _open_osfhandle(sock->fd, 0);
+        /* cfd will be closed when this socket is closed. */
+        if (sock->cfd < 0) {
+            sock->cfd = _open_osfhandle(sock->fd, 0);
+        }
+        outfd = sock->cfd;
 #endif /*GAUCHE_WINDOWS*/
         if (outfd == INVALID_SOCKET) sockport_err(sock, "output");
 
         /* NB: I keep the socket itself in the port name, in order to avoid
            the socket from GCed prematurely if application doesn't keep
            pointer to the socket. */
-	sockname = SCM_LIST2(SCM_MAKE_STR("socket output"), SCM_OBJ(sock));
+        ScmObj sockname = SCM_LIST2(SCM_MAKE_STR("socket output"),
+                                    SCM_OBJ(sock));
         sock->outPort = SCM_PORT(Scm_MakePortWithFd(sockname, SCM_PORT_OUTPUT,
-						    outfd, buffering, FALSE));
+                                                    outfd, buffering, FALSE));
     }
     return SCM_OBJ(sock->outPort);
 }
@@ -228,7 +244,6 @@ ScmObj Scm_SocketOutputPort(ScmSocket *sock, int buffering)
 
 ScmObj Scm_SocketBind(ScmSocket *sock, ScmSockAddr *addr)
 {
-    ScmSockAddr *naddr;
     int r;
     CLOSE_CHECK(sock->fd, "bind", sock);
     SCM_SYSCALL(r, bind(sock->fd, &addr->addr, addr->addrlen));
@@ -238,8 +253,8 @@ ScmObj Scm_SocketBind(ScmSocket *sock, ScmSockAddr *addr)
     /* The system may assign different address than <addr>, especially when
        <addr> contains some 'wild card' (e.g. port=0).  We call getsockname
        to obtain the exact address.   Patch provided by ODA Hideo */
-    naddr = SCM_SOCKADDR(Scm_MakeSockAddr(SCM_CLASS_OF(addr),
-                                          &addr->addr, addr->addrlen));
+    ScmSockAddr *naddr = SCM_SOCKADDR(
+        Scm_MakeSockAddr(SCM_CLASS_OF(addr), &addr->addr, addr->addrlen));
     SCM_SYSCALL(r, getsockname(sock->fd, &naddr->addr, &naddr->addrlen));
     if (r < 0) {
         Scm_SysError("getsockname failed to %S", addr);
@@ -266,7 +281,6 @@ ScmObj Scm_SocketAccept(ScmSocket *sock)
     Socket newfd;
     struct sockaddr_storage addrbuf;
     socklen_t addrlen = sizeof(addrbuf);
-    ScmSocket *newsock;
     ScmClass *addrClass = Scm_ClassOf(SCM_OBJ(sock->address));
 
     CLOSE_CHECK(sock->fd, "accept from", sock);
@@ -278,7 +292,7 @@ ScmObj Scm_SocketAccept(ScmSocket *sock)
             Scm_SysError("accept(2) failed");
         }
     }
-    newsock = make_socket(newfd, sock->type);
+    ScmSocket *newsock = make_socket(newfd, sock->type);
     newsock->address =
         SCM_SOCKADDR(Scm_MakeSockAddr(addrClass,
                                       (struct sockaddr*)&addrbuf,
@@ -344,10 +358,10 @@ static const char *get_message_body(ScmObj msg, u_int *size)
 
 ScmObj Scm_SocketSend(ScmSocket *sock, ScmObj msg, int flags)
 {
-    int r; u_int size;
-    const char *cmsg;
+    int r;
+    u_int size;
     CLOSE_CHECK(sock->fd, "send to", sock);
-    cmsg = get_message_body(msg, &size);
+    const char *cmsg = get_message_body(msg, &size);
     SCM_SYSCALL(r, send(sock->fd, cmsg, size, flags));
     if (r < 0) Scm_SysError("send(2) failed");
     return SCM_MAKE_INT(r);
@@ -356,10 +370,10 @@ ScmObj Scm_SocketSend(ScmSocket *sock, ScmObj msg, int flags)
 ScmObj Scm_SocketSendTo(ScmSocket *sock, ScmObj msg, ScmSockAddr *to,
                         int flags)
 {
-    int r; u_int size;
-    const char *cmsg;
+    int r;
+    u_int size;
     CLOSE_CHECK(sock->fd, "send to", sock);
-    cmsg = get_message_body(msg, &size);
+    const char *cmsg = get_message_body(msg, &size);
     SCM_SYSCALL(r, sendto(sock->fd, cmsg, size, flags,
                           &SCM_SOCKADDR(to)->addr, SCM_SOCKADDR(to)->addrlen));
     if (r < 0) Scm_SysError("sendto(2) failed");
@@ -369,25 +383,24 @@ ScmObj Scm_SocketSendTo(ScmSocket *sock, ScmObj msg, ScmSockAddr *to,
 ScmObj Scm_SocketSendMsg(ScmSocket *sock, ScmObj msg, int flags)
 {
 #if !GAUCHE_WINDOWS
-    int r; u_int size;
-    const char *cmsg;
+    int r;
+    u_int size;
     CLOSE_CHECK(sock->fd, "send to", sock);
-    cmsg = get_message_body(msg, &size);
+    const char *cmsg = get_message_body(msg, &size);
     SCM_SYSCALL(r, sendmsg(sock->fd, (struct msghdr*)cmsg, flags));
     if (r < 0) Scm_SysError("sendmsg(2) failed");
     return SCM_MAKE_INT(r);
 #else  /*GAUCHE_WINDOWS*/
     Scm_Error("sendmsg is not implemented on this platform.");
-    return SCM_UNDEFINED;	/* dummy */
+    return SCM_UNDEFINED;       /* dummy */
 #endif /*GAUCHE_WINDOWS*/
 }
 
 ScmObj Scm_SocketRecv(ScmSocket *sock, int bytes, int flags)
 {
     int r;
-    char *buf;
     CLOSE_CHECK(sock->fd, "recv from", sock);
-    buf = SCM_NEW_ATOMIC2(char*, bytes);
+    char *buf = SCM_NEW_ATOMIC2(char*, bytes);
     SCM_SYSCALL(r, recv(sock->fd, buf, bytes, flags));
     if (r < 0) {
         Scm_SysError("recv(2) failed");
@@ -408,9 +421,8 @@ ScmObj Scm_SocketRecvX(ScmSocket *sock, ScmUVector *buf, int flags)
 {
     int r;
     u_int size;
-    char *z;
     CLOSE_CHECK(sock->fd, "recv from", sock);
-    z = get_message_buffer(buf, &size);
+    char *z = get_message_buffer(buf, &size);
     SCM_SYSCALL(r, recv(sock->fd, z, size, flags));
     if (r < 0) {
         Scm_SysError("recv(2) failed");
@@ -421,11 +433,10 @@ ScmObj Scm_SocketRecvX(ScmSocket *sock, ScmUVector *buf, int flags)
 ScmObj Scm_SocketRecvFrom(ScmSocket *sock, int bytes, int flags)
 {
     int r;
-    char *buf;
     struct sockaddr_storage from;
     socklen_t fromlen = sizeof(from);
     CLOSE_CHECK(sock->fd, "recv from", sock);
-    buf = SCM_NEW_ATOMIC2(char*, bytes);
+    char *buf = SCM_NEW_ATOMIC2(char*, bytes);
     SCM_SYSCALL(r, recvfrom(sock->fd, buf, bytes, flags,
                             (struct sockaddr*)&from, &fromlen));
     if (r < 0) {
@@ -446,18 +457,18 @@ ScmObj Scm_SocketRecvFromX(ScmSocket *sock, ScmUVector *buf,
 {
     int r;
     u_int size;
-    char *z;
     struct sockaddr_storage from;
     socklen_t fromlen = sizeof(from);
-    ScmObj cp, addr = SCM_FALSE;
+    ScmObj addr = SCM_FALSE;
 
     CLOSE_CHECK(sock->fd, "recv from", sock);
-    z = get_message_buffer(buf, &size);
+    char *z = get_message_buffer(buf, &size);
     SCM_SYSCALL(r, recvfrom(sock->fd, z, size, flags,
                             (struct sockaddr*)&from, &fromlen));
     if (r < 0) {
         Scm_SysError("recvfrom(2) failed");
     }
+    ScmObj cp;
     SCM_FOR_EACH(cp, addrs) {
         ScmObj a = SCM_CAR(cp);
         if (Scm_SockAddrP(a)) {
@@ -506,7 +517,6 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
     }
 
     if (iov != NULL) {
-        int i;
         int iovsiz = SCM_VECTOR_SIZE(iov) * sizeof(struct iovec);
         msg->msg_iovlen = SCM_VECTOR_SIZE(iov);
         if (bufsiz >= iovsiz) {
@@ -515,7 +525,7 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
         } else {
             msg->msg_iov = SCM_NEW_ARRAY(struct iovec, msg->msg_iovlen);
         }
-        for (i=0; i < msg->msg_iovlen; i++) {
+        for (int i=0; i < msg->msg_iovlen; i++) {
             ScmObj elt = SCM_VECTOR_ELEMENT(iov, i);
             u_int iovlen;
             msg->msg_iov[i].iov_base = (char*)get_message_body(elt, &iovlen);
@@ -529,7 +539,6 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
     if (SCM_PAIRP(control)) {
         ScmObj cp;
         int ctrllen = 0;
-        struct cmsghdr *cmsg;
 
         SCM_FOR_EACH(cp, control) {
             u_int clen;
@@ -550,7 +559,7 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
         } else {
             msg->msg_control = SCM_NEW_ATOMIC_ARRAY(char, ctrllen);
         }
-        cmsg = CMSG_FIRSTHDR(msg);
+        struct cmsghdr *cmsg = CMSG_FIRSTHDR(msg);
         SCM_FOR_EACH(cp, control) {
             u_int clen;
             ScmObj c = SCM_CAR(cp);
@@ -625,31 +634,105 @@ ScmObj Scm_SocketGetOpt(ScmSocket *s, int level, int option, int rsize)
     }
 }
 
-/* Low-level ioctl. */
-ScmObj Scm_SocketIoctl(ScmSocket *s, int request, ScmObj data)
-{
-    int r = 0;
 #if HAVE_STRUCT_IFREQ
+static void ioctl_by_ifr_name(int fd, struct ifreq *ifr, ScmObj data,
+                              unsigned long req, const char *req_name)
+{
+    if (!SCM_STRINGP(data)) {
+        Scm_Error("string expected for %s ioctl argument, but got %s",
+                  req_name, data);
+    }
+    strncpy(ifr->ifr_name, Scm_GetStringConst(SCM_STRING(data)), IFNAMSIZ-1);
+    int r;
+    SCM_SYSCALL(r, ioctl(fd, req, ifr));
+    if (r < 0) Scm_SysError("ioctl(%s) failed", req_name);
+}
+#endif
+
+/* Low-level ioctl. */
+ScmObj Scm_SocketIoctl(ScmSocket *s, u_long request, ScmObj data)
+{
+#if HAVE_STRUCT_IFREQ
+    int r;
     struct ifreq ifreq_pkt;
 
     CLOSE_CHECK(s->fd, "ioctl on", s);
     memset(&ifreq_pkt, 0, sizeof(ifreq_pkt));
     switch (request) {
+#if defined(SIOCGIFNAME)
+    case SIOCGIFNAME:
+        if (!SCM_UINTEGERP(data))
+            Scm_TypeError("SIOCGIFNAME ioctl argument" , "unsigned integer", data);
+#if   HAVE_STRUCT_IFREQ_IFR_IFINDEX
+        ifreq_pkt.ifr_ifindex = Scm_GetIntegerU(data);
+#elif HAVE_STRUCT_IFREQ_IFR_INDEX
+        ifreq_pkt.ifr_index = Scm_GetIntegerU(data);
+#endif /*HAVE_STRUCT_IFREQ_IFR_INDEX*/
+        SCM_SYSCALL(r, ioctl(s->fd, SIOCGIFNAME, &ifreq_pkt));
+        if (r < 0)
+            Scm_SysError("ioctl(SIOCGIFNAME) failed");
+        return Scm_MakeString(ifreq_pkt.ifr_name, -1, -1, SCM_STRING_COPYING);
+#endif /*SIOCGIFNAME*/
 #if defined(SIOCGIFINDEX)
     case SIOCGIFINDEX:
-        if (!SCM_STRINGP(data)) {
-            Scm_TypeError("SIOCGIFINDEX ioctl argument", "string", data);
-        }
-        strncpy(ifreq_pkt.ifr_name, Scm_GetStringConst(SCM_STRING(data)),
-                IFNAMSIZ-1);
-        SCM_SYSCALL(r, ioctl(s->fd, SIOCGIFINDEX, &ifreq_pkt));
-        if (r < 0) Scm_SysError("ioctl(SIOCGIFINDEX) failed");
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFINDEX, "SIOCGIFINDEX");
 #if HAVE_STRUCT_IFREQ_IFR_IFINDEX
         return Scm_MakeInteger(ifreq_pkt.ifr_ifindex);
 #elif HAVE_STRUCT_IFREQ_IFR_INDEX
-       return Scm_MakeInteger(ifreq_pkt.ifr_index);
+        return Scm_MakeInteger(ifreq_pkt.ifr_index);
 #endif /*HAVE_STRUCT_IFREQ_IFR_INDEX*/
 #endif /*SIOCGIFINDEX*/
+#if defined(SIOCGIFFLAGS)
+    case SIOCGIFFLAGS:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFFLAGS, "SIOCGIFFLAGS");
+        return Scm_MakeInteger(ifreq_pkt.ifr_flags);
+#endif /*SIOCGIFFLAGS*/
+#if defined(SIOCGIFADDR)
+    case SIOCGIFADDR:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFADDR, "SIOCGIFADDR");
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_addr,
+                                sizeof(ifreq_pkt.ifr_addr));
+#endif /*SIOCGIFADDR*/
+#if defined(SIOCGIFDSTADDR)
+    case SIOCGIFDSTADDR:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFDSTADDR, "SIOCGIFDSTADDR");
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_dstaddr,
+                                sizeof(ifreq_pkt.ifr_dstaddr));
+#endif  /*SIOCGIFDSTADDR*/
+#if defined(SIOCGIFBRDADDR)
+    case SIOCGIFBRDADDR:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFBRDADDR, "SIOCGIFBRDADDR");
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_broadaddr,
+                                sizeof(ifreq_pkt.ifr_broadaddr));
+#endif  /*SIOCGIFBRDADDR*/
+#if defined(SIOCGIFNETMASK)
+    case SIOCGIFNETMASK:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFNETMASK, "SIOCGIFNETMASK");
+#if defined(HAVE_STRUCT_IFREQ_IFR_NETMASK)
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_netmask,
+                                sizeof(ifreq_pkt.ifr_netmask));
+#else
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_addr,
+                                sizeof(ifreq_pkt.ifr_addr));
+#endif /*HAVE_STRUCT_IFREQ_IFR_NETMASK*/
+#endif  /*SIOCGIFNETMASK*/
+#if defined(SIOCGIFMTU)
+    case SIOCGIFMTU:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFMTU, "SIOCGIFMTU");
+        return Scm_MakeInteger(ifreq_pkt.ifr_mtu);
+#endif  /*SIOCGIFMTU*/
     default:
         Scm_Error("unsupported ioctl operation: %d", request);
     }
@@ -667,54 +750,55 @@ ScmObj Scm_SocketIoctl(ScmSocket *s, int request, ScmObj data)
 int inet_pton(int af, const char *src, void *dst)
 {
     TCHAR *str = SCM_MBS2WCS(src);
-    int r;
-    INT addrsize;
-    struct sockaddr_in sa;
-    struct sockaddr_in6 sa6;
 
     switch (af) {
-    case AF_INET:
-        addrsize = (INT)sizeof(sa);
-        r = WSAStringToAddress(str, af, NULL, (LPSOCKADDR)&sa, &addrsize);
+    case AF_INET: {
+        struct sockaddr_in sa;
+        INT addrsize = (INT)sizeof(sa);
+        int r = WSAStringToAddress(str, af, NULL, (LPSOCKADDR)&sa, &addrsize);
         if (r != 0) return -1;
         memcpy(dst, &sa.sin_addr, sizeof(struct in_addr));
         return 1;
-    case AF_INET6:
-        addrsize = (INT)sizeof(sa6);
-        r = WSAStringToAddress(str, af, NULL, (LPSOCKADDR)&sa6, &addrsize);
+    }
+    case AF_INET6: {
+        struct sockaddr_in6 sa6;
+        INT addrsize = (INT)sizeof(sa6);
+        int r = WSAStringToAddress(str, af, NULL, (LPSOCKADDR)&sa6, &addrsize);
         if (r != 0) return -1;
         memcpy(dst, &sa6.sin6_addr, sizeof(struct in6_addr));
         return 1;
     }
+    }
     return -1;
 }
 
 const char *inet_ntop(int af, const void *src, char *dst, socklen_t size)
 {
 #define ADDR_MAXLEN 64
-    struct sockaddr_in sa;
-    struct sockaddr_in6 sa6;
     TCHAR buf[ADDR_MAXLEN];
-    int r;
-    size_t ressize;
     DWORD tressize = ADDR_MAXLEN-1;
     const char *res;
+    int r = 0;
 
     switch (af) {
-    case AF_INET:
+    case AF_INET: {
+        struct sockaddr_in sa;
         memset(&sa, 0, sizeof(sa));
         sa.sin_family = AF_INET;
         memcpy(&sa.sin_addr, src, sizeof(struct in_addr));
         r = WSAAddressToString((LPSOCKADDR)&sa, (DWORD)sizeof(sa), NULL,
                                buf, &tressize);
         break;
-    case AF_INET6:
+    }
+    case AF_INET6: {
+        struct sockaddr_in6 sa6;
         memset(&sa6, 0, sizeof(sa6));
         sa6.sin6_family = AF_INET6;
         memcpy(&sa6.sin6_addr, src, sizeof(struct in6_addr));
         r = WSAAddressToString((LPSOCKADDR)&sa6, (DWORD)sizeof(sa6), NULL,
                                buf, &tressize);
         break;
+    }
     default:
         return NULL;
     }
@@ -722,7 +806,7 @@ const char *inet_ntop(int af, const void *src, char *dst, socklen_t size)
     if (r != 0) return NULL;
     buf[tressize] = 0;
     res = SCM_WCS2MBS(buf);
-    ressize = strlen(res);
+    size_t ressize = strlen(res);
     if (size <= (int)ressize) return NULL;
     memcpy(dst, res, ressize+1);
     return dst;
@@ -744,10 +828,8 @@ extern void Scm_Init_netaux(void);
 
 SCM_EXTENSION_ENTRY void Scm_Init_gauche__net(void)
 {
-    ScmModule *mod;
-
     SCM_INIT_EXTENSION(gauche__net);
-    mod = SCM_FIND_MODULE("gauche.net", SCM_FIND_MODULE_CREATE);
+    ScmModule *mod = SCM_FIND_MODULE("gauche.net", SCM_FIND_MODULE_CREATE);
 #ifdef HAVE_IPV6
     Scm_AddFeature("gauche.net.ipv6", NULL);
 #endif
