@@ -138,7 +138,7 @@
                #x00000000000000000000000000000001 AF_INET6)
     (addr-test "v6-3" "::0.1.0.2"
                #x00000000000000000000000000010002 AF_INET6)
-    (addr-test "v6-4" "::1:2:3:4:5:6:7"
+    (addr-test "v6-4" "0:1:2:3:4:5:6:7"
                #x00000001000200030004000500060007 AF_INET6)
     (addr-test "v6-5" "ffe0::"
                #xffe00000000000000000000000000000 AF_INET6)
