@@ -2,7 +2,7 @@
 ;;; peg.scm - Parser Expression Grammar Parser
 ;;;
 ;;;   Copyright (c) 2006 Rui Ueyama (rui314@gmail.com)
-;;;   Copyright (c) 2008-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2008-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -33,13 +33,11 @@
 ;;;
 
 (define-module parser.peg
-  (use srfi-1)
   (use srfi-13)
   (use srfi-14)
   (use gauche.collection)
   (use gauche.generator)
   (use gauche.lazy)
-  (use util.list)
   (use util.match)
   (export <parse-error>
           make-peg-parse-error
@@ -312,7 +310,7 @@
  (define-cproc rope->string (obj)
    (let* ([p (Scm_MakeOutputStringPort TRUE)])
      (rope2string_int obj p)
-     (result (Scm_GetOutputString (SCM_PORT p) 0))))
+     (return (Scm_GetOutputString (SCM_PORT p) 0))))
  )
 
 ;;;============================================================
