@@ -34,6 +34,7 @@
 (use gauche.test)
 (use srfi-1)
 (use gauche.generator)
+(use gauche.parameter)
 
 (test-start "parser.peg")
 
@@ -51,13 +52,13 @@
 (define-syntax test-succ
   (syntax-rules ()
     [(_ label expect parse input)
-     (test* #`",label (success)" expect
+     (test* #"~label (success)" expect
             (peg-parse-string parse input))]))
 
 (define-syntax test-fail
   (syntax-rules ()
     [(_ label expect parse input)
-     (test* #`",label (failure)" expect
+     (test* #"~label (failure)" expect
             (guard (e [(<parse-error> e)
                        (list (ref e 'position) (ref e 'objects))]
                       [else e])
@@ -240,6 +241,21 @@
                 ($string (rope-finalize v)))
            "foobar")
 
+;; $lift and $lift*
+(test-succ "$lift" '(#\a . #\b)
+           ($lift cons anychar anychar)
+           "abc")
+(test-fail "$lift" '(1 #\z)
+           ($lift cons anychar ($char #\z))
+           "abc")
+
+(test-succ "$lift*" "abc"
+           ($lift* list->string anychar anychar anychar)
+           "abc")
+(test-fail "$lift" '(2 #\z)
+           ($lift* list->string anychar anychar ($char #\z))
+           "abc")
+
 ;; $fold-parsers and $fold-parsers-right
 (test-succ "$fold-parsers" '()                  ; base case
            ($fold-parsers cons '() '())
@@ -535,8 +551,8 @@
        (p        ($do [number?] [eof] ($return #t))))
   (define (%test str . fails?)
     (if (null? fails?)
-      (test-succ #`"number(1) \",str\"" #t p str)
-      (test-fail #`"number(1) \",str\"" (car fails?) p str)))
+      (test-succ #"number(1) \"~str\"" #t p str)
+      (test-fail #"number(1) \"~str\"" (car fails?) p str)))
   (%test "27652") (%test "-27652") (%test "+27652")
   (%test "" '(0 #[0-9])) (%test "-" '(1 #[0-9])) (%test "+" '(1 #[0-9]))
   (%test "+27.46") (%test "0.46")
@@ -661,7 +677,7 @@
 
 (let ()
   (define (t str)
-    (test* #`"parse error ,str" (test-error <json-parse-error>)
+    (test* #"parse error ~str" (test-error <json-parse-error>)
            (parse-json-string str)))
   (t "{\"x\": 100")
   (t "{x : 100}}")
@@ -733,8 +749,18 @@
 
 (test* "Parsing sequence of json objects"
        '((("a" . 1)("b" . 2)) (("c" . 3) ("d" . 4)))
-       (with-input-from-string "{\"a\":1, \"b\":2} {\"c\":3, \"d\":4} "
-         (^[] (generator->list (peg-parser->generator json-parser read-char)))))
+       (with-input-from-string "{\"a\":1, \"b\":2}{\"c\":3, \"d\":4}"
+         parse-json*))
+
+(test* "Customizing consturctors"
+       '(object ("x" array 1 2 3) ("y" array #f #t null))
+       (parameterize ([json-array-handler (^[elts] (cons 'array elts))]
+                      [json-object-handler (^[pairs] (cons 'object pairs))]
+                      [json-special-handler (^y (case y
+                                                  [(false) #f]
+                                                  [(true) #t]
+                                                  [(null) 'null]))])
+         (parse-json-string "{\"x\":[1,2,3],\"y\":[false,true,null]}")))
 
 (let ()
   (define (test-writer name obj)
@@ -773,13 +799,36 @@
                    ("Country"   . "US"))))
   )
 
+(cond-expand
+ [gauche.ces.utf8
+  (let1 data `(("[\"\\u03bb\"]" #("\x3bb;"))
+               ("[\"\\ud800\"]" ,(test-error <json-parse-error>))
+               ("[\"\\ud867\\ude3d\\u03bb\"]" #("\x29e3d;\x3bb;"))
+               ("[\"\\ude3d\\ud867\"]" ,(test-error <json-parse-error>))
+               ("[\"\\uf020\\u03bb\"]"  #("\xf020;\x3bb;")))
+    (dolist [d data]
+      (test* (format "unicode escape reading (~s)" (car d))
+             (cadr d)
+             (parse-json-string (car d)))
+      (when (vector? (cadr data))
+        (test* (format "unicode escape writing (~s)" (cadr d))
+               (car d)
+               (construct-json-string (cadr d))))))]
+ [else])
+
 (let ()
   (define (t obj)
-    (test* #`"writer error ,obj" (test-error <json-construct-error>)
+    (test* #"writer error ~obj" (test-error <json-construct-error>)
            (construct-json-string obj)))
   (t "a")
   (t '#(1 2 x))
-  (t '(("a" . 2) (3 . "b")))
   (t '(("a" . 2) 9)))
 
+(test* "generalized array" "[1,2,3]"
+       (construct-json-string '#u8(1 2 3)))
+(test* "generalized object" (test-one-of "{\"a\":1,\"b\":2}"
+                                         "{\"b\":2,\"a\":1}")
+       (construct-json-string (hash-table 'eq? '(a . 1) '(b . 2))))
+
+
 (test-end)
