@@ -559,6 +559,12 @@
        (xsubstring "abc"
                    30000000000000000000000000000000
                    30000000000000000000000000000007))
+(test* "xsubstring" "abcde"
+       (xsubstring "abcde" -10 -5))
+(test* "xsubstring" "eabcdea"
+       (xsubstring "abcde" -11 -4))
+(test* "xsubstring" "eabcdeabcdeabc"
+       (xsubstring "abcde" -11 3))
 (test* "xsubstring" "defdefd"
        (xsubstring "abcdefg" 0 7 3 6))
 (test* "xsubstring" ""
@@ -868,23 +874,7 @@
 (assert-error "vector-unfold-right 3"
               (vector-unfold-right values -1))
 
-(assert-equal "vector-copy 0"
-              (vector-copy '#(a b c d e f g h i))
-              '#(a b c d e f g h i))
-(assert-equal "vector-copy 1"
-              (vector-copy '#(a b c d e f g h i) 6)
-              '#(g h i))
-(assert-equal "vector-copy 2"
-              (vector-copy '#(a b c d e f g h i) 3 6)
-              '#(d e f))
-(assert-equal "vector-copy 3"
-              (vector-copy '#(a b c d e f g h i) 6 12 'x)
-              '#(g h i x x x))
-(assert-equal "vector-copy 4"
-              (vector-copy '#(a b c d e f g h i) 6 6)
-              '#())
-(assert-error "vector-copy 5"
-              (vector-copy '#(a b c d e f g h i) 4 2))
+;; vector-copy - now built-in
 
 (assert-equal "vector-reverse-copy 0"
               (vector-reverse-copy '#(a b c d e))
@@ -899,21 +889,7 @@
               (vector-reverse-copy '#(a b c d e) 2 1))
 
 
-(assert-equal "vector-append 0"
-              (vector-append '#(x) '#(y))
-              '#(x y))
-(assert-equal "vector-append 1"
-              (let ((v '#(x y)))
-                (vector-append v v v))
-              '#(x y x y x y))
-(assert-equal "vector-append 2"
-              (vector-append '#(x) '#() '#(y))
-              '#(x y))
-(assert-equal "vector-append 3"
-              (vector-append)
-              '#())
-(assert-error "vector-append 4"
-              (vector-append '#() 'b 'c))
+;; vector-append - now built-in
 
 (assert-equal "vector-concatenate 0"
               (vector-concatenate '(#(a b) #(c d)))
@@ -924,6 +900,7 @@
 (assert-error "vector-concatenate 2"
               (vector-concatenate '(#(a b) c)))
 
+
 ;;;
 ;;; Predicates
 ;;;
@@ -1281,43 +1258,8 @@
 (assert-error "vector-reverse! e0" (vector-reverse! (vector 'a 'b) 0 3))
 (assert-error "vector-reverse! e1" (vector-reverse! (vector 'a 'b) 2 1))
 (assert-error "vector-reverse! e2" (vector-reverse! (vector 'a 'b) -1 1))
-(assert-error "vector-reverse! e3" (vector-reverse! (vector) 0 0))
 
-(assert-equal "vector-copy! 0"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 0 '#(1 2 3)))
-              '#(1 2 3 d e))
-(assert-equal "vector-copy! 1"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 2 '#(1 2 3)))
-              '#(a b 1 2 3))
-(assert-equal "vector-copy! 2"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 2 '#(1 2 3) 1))
-              '#(a b 2 3 e))
-(assert-equal "vector-copy! 3"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 2 '#(1 2 3 4 5) 2 5))
-              '#(a b 3 4 5))
-(assert-equal "vector-copy! 4"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 2 '#(1 2 3) 1 1))
-              '#(a b c d e))
-(assert-equal "vector-copy! self0"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 0 v 1 3))
-              '#(b c c d e))
-(assert-equal "vector-copy! self1"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 2 v 1 4))
-              '#(a b b c d))
-(assert-equal "vector-copy! self2"
-              (rlet1 v (vector 'a 'b 'c 'd 'e)
-                (vector-copy! v 0 v 0))
-              '#(a b c d e))
-(assert-error "vector-copy! e0" (vector-copy! (vector 1 2) 3 '#(1 2 3)))
-(assert-error "vector-copy! e1" (vector-copy! (vector 1 2) 0 '#(1 2 3)))
-(assert-error "vector-copy! e2" (vector-copy! (vector 1 2) 1 '#(1 2 3) 1))
+;; vector-copy! - now built-in
 
 (assert-equal "vector-reverse-copy! 0"
               (rlet1 v (vector 'a 'b 'c 'd 'e)
@@ -1443,4 +1385,17 @@
 (assert-error "reverse-list->vector e2"
               (reverse-list->vector '(0 1 2 3) 2 1))
 
+;;
+;; Test srfi-133
+;;
+
+(test-section "srfi-133")
+
+(use compat.chibi-test)
+(use srfi-133)
+(use srfi-11)
+
+(chibi-test
+ (include "vectors-test.scm"))
+
 (test-end)
