@@ -1,7 +1,7 @@
 /*
  * mutex.c - Scheme-level synchronization devices
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -61,8 +61,7 @@ static void mutex_finalize(ScmObj obj, void* data)
 
 static ScmObj mutex_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmMutex *mutex = SCM_ALLOCATE(ScmMutex, klass);
-    SCM_SET_CLASS(mutex, klass);
+    ScmMutex *mutex = SCM_NEW_INSTANCE(ScmMutex, klass);
     SCM_INTERNAL_MUTEX_INIT(mutex->mutex);
     SCM_INTERNAL_COND_INIT(mutex->cv);
     Scm_RegisterFinalizer(SCM_OBJ(mutex), mutex_finalize, NULL);
@@ -170,12 +169,12 @@ ScmObj Scm_MakeMutex(ScmObj name)
 ScmObj Scm_MutexLock(ScmMutex *mutex, ScmObj timeout, ScmVM *owner)
 {
 #ifdef GAUCHE_HAS_THREADS
-    struct timespec ts;
+    ScmTimeSpec ts;
     ScmObj r = SCM_TRUE;
     ScmVM *abandoned = NULL;
     int intr = FALSE;
 
-    struct timespec *pts = Scm_GetTimeSpec(timeout, &ts);
+    ScmTimeSpec *pts = Scm_GetTimeSpec(timeout, &ts);
     SCM_INTERNAL_MUTEX_SAFE_LOCK_BEGIN(mutex->mutex);
     while (mutex->locked) {
         if (mutex->owner && mutex->owner->state == SCM_VM_TERMINATED) {
@@ -212,10 +211,10 @@ ScmObj Scm_MutexUnlock(ScmMutex *mutex, ScmConditionVariable *cv, ScmObj timeout
 {
     ScmObj r = SCM_TRUE;
 #ifdef GAUCHE_HAS_THREADS
-    struct timespec ts;
+    ScmTimeSpec ts;
     int intr = FALSE;
 
-    struct timespec *pts = Scm_GetTimeSpec(timeout, &ts);
+    ScmTimeSpec *pts = Scm_GetTimeSpec(timeout, &ts);
     SCM_INTERNAL_MUTEX_SAFE_LOCK_BEGIN(mutex->mutex);
     mutex->locked = FALSE;
     mutex->owner = NULL;
@@ -287,8 +286,7 @@ static void cv_finalize(ScmObj obj, void *data)
 
 static ScmObj cv_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmConditionVariable *cv = SCM_ALLOCATE(ScmConditionVariable, klass);
-    SCM_SET_CLASS(cv, klass);
+    ScmConditionVariable *cv = SCM_NEW_INSTANCE(ScmConditionVariable, klass);
     SCM_INTERNAL_COND_INIT(cv->cv);
     Scm_RegisterFinalizer(SCM_OBJ(cv), cv_finalize, NULL);
     cv->name = SCM_FALSE;
