@@ -1,7 +1,7 @@
 /*
  * thread.c - Scheme thread API
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -48,18 +48,6 @@
  * Thread interface
  */
 
-static ScmObj thread_error_handler(ScmObj *args, int nargs, void *data)
-{
-    /* For now, uncaptured error causes thread termination with
-       setting <uncaught-exception> to the resultException field.
-       It is handled in thread_entry(), so we don't need to do anything
-       here. */
-    return SCM_UNDEFINED;
-}
-
-static SCM_DEFINE_STRING_CONST(thread_error_handler_NAME, "thread-error-handler", 20, 20);
-static SCM_DEFINE_SUBR(thread_error_handler_STUB, 1, 0, SCM_OBJ(&thread_error_handler_NAME), thread_error_handler, NULL, NULL);
-
 /* Creation.  In the "NEW" state, a VM is allocated but actual thread
    is not created. */
 ScmObj Scm_MakeThread(ScmProcedure *thunk, ScmObj name)
@@ -71,7 +59,6 @@ ScmObj Scm_MakeThread(ScmProcedure *thunk, ScmObj name)
     }
     ScmVM *vm = Scm_NewVM(current, name);
     vm->thunk = thunk;
-    vm->defaultEscapeHandler = SCM_OBJ(&thread_error_handler_STUB);
     return SCM_OBJ(vm);
 }
 
@@ -127,11 +114,12 @@ static SCM_INTERNAL_THREAD_PROC_RETTYPE thread_entry(void *data)
                     Scm_MakeError(SCM_MAKE_STR("stale continuation thrown"));
                 break;
             case SCM_VM_ESCAPE_ERROR: {
-                ScmObj exc = Scm_MakeThreadException(
-                    SCM_CLASS_UNCAUGHT_EXCEPTION, vm);
+                /* Error handling is delegated to the thread that calls
+                   thread-join!. */
+                ScmObj exc =
+                    Scm_MakeThreadException(SCM_CLASS_UNCAUGHT_EXCEPTION, vm);
                 SCM_THREAD_EXCEPTION(exc)->data = SCM_OBJ(vm->escapeData[1]);
                 vm->resultException = exc;
-                Scm_ReportError(SCM_OBJ(vm->escapeData[1]));
                 break;
             }
             default:
@@ -215,11 +203,11 @@ ScmObj Scm_ThreadStart(ScmVM *vm)
 ScmObj Scm_ThreadJoin(ScmVM *target, ScmObj timeout, ScmObj timeoutval)
 {
 #ifdef GAUCHE_HAS_THREADS
-    struct timespec ts;
+    ScmTimeSpec ts;
     ScmObj result = SCM_FALSE, resultx = SCM_FALSE;
     int intr = FALSE, tout = FALSE;
 
-    struct timespec *pts = Scm_GetTimeSpec(timeout, &ts);
+    ScmTimeSpec *pts = Scm_GetTimeSpec(timeout, &ts);
     SCM_INTERNAL_MUTEX_SAFE_LOCK_BEGIN(target->vmlock);
     while (target->state != SCM_VM_TERMINATED) {
         if (pts) {
@@ -248,7 +236,7 @@ ScmObj Scm_ThreadJoin(ScmVM *target, ScmObj timeout, ScmObj timeoutval)
     }
     return result;
 #else  /*!GAUCHE_HAS_THREADS*/
-    Scm_Error("not implemented!\n");
+    Scm_Error("not implemented!");
     return SCM_UNDEFINED;
 #endif /*!GAUCHE_HAS_THREADS*/
 }
@@ -280,12 +268,12 @@ ScmObj Scm_ThreadJoin(ScmVM *target, ScmObj timeout, ScmObj timeoutval)
 ScmObj Scm_ThreadStop(ScmVM *target, ScmObj timeout, ScmObj timeoutval)
 {
 #ifdef GAUCHE_HAS_THREADS
-    struct timespec ts;
+    ScmTimeSpec ts;
     ScmVM *vm = Scm_VM();
     ScmVM *taker = NULL;
     int timedout;
     int invalid_state = FALSE;
-    struct timespec *pts = Scm_GetTimeSpec(timeout, &ts);
+    ScmTimeSpec *pts = Scm_GetTimeSpec(timeout, &ts);
 
  retry:
     timedout = FALSE;
@@ -332,7 +320,7 @@ ScmObj Scm_ThreadStop(ScmVM *target, ScmObj timeout, ScmObj timeoutval)
     if (timedout == SCM_INTERNAL_COND_TIMEDOUT) return timeoutval;
     return SCM_OBJ(target);
 #else  /*!GAUCHE_HAS_THREADS*/
-    Scm_Error("not implemented!\n");
+    Scm_Error("not implemented!");
     return SCM_UNDEFINED;
 #endif /*!GAUCHE_HAS_THREADS*/
 }
@@ -363,7 +351,7 @@ ScmObj Scm_ThreadCont(ScmVM *target)
                                     target, stopped_by_other);
     return SCM_OBJ(target);
 #else  /*!GAUCHE_HAS_THREADS*/
-    Scm_Error("not implemented!\n");
+    Scm_Error("not implemented!");
     return SCM_UNDEFINED;
 #endif /*!GAUCHE_HAS_THREADS*/
 }
@@ -372,14 +360,14 @@ ScmObj Scm_ThreadCont(ScmVM *target)
 ScmObj Scm_ThreadSleep(ScmObj timeout)
 {
 #ifdef GAUCHE_HAS_THREADS
-    struct timespec ts;
+    ScmTimeSpec ts;
     ScmInternalCond dummyc;
     ScmInternalMutex dummym;
     int intr = FALSE;
 
     SCM_INTERNAL_COND_INIT(dummyc);
     SCM_INTERNAL_MUTEX_INIT(dummym);
-    struct timespec *pts = Scm_GetTimeSpec(timeout, &ts);
+    ScmTimeSpec *pts = Scm_GetTimeSpec(timeout, &ts);
     if (pts == NULL) Scm_Error("thread-sleep! can't take #f as a timeout value");
     SCM_INTERNAL_MUTEX_LOCK(dummym);
     if (SCM_INTERNAL_COND_TIMEDWAIT(dummyc, dummym, pts) == SCM_INTERNAL_COND_INTR) {
@@ -390,7 +378,7 @@ ScmObj Scm_ThreadSleep(ScmObj timeout)
     SCM_INTERNAL_COND_DESTROY(dummyc);
     if (intr) Scm_SigCheck(Scm_VM());
 #else  /*!GAUCHE_HAS_THREADS*/
-    Scm_Error("not implemented!\n");
+    Scm_Error("not implemented!");
 #endif /*!GAUCHE_HAS_THREADS*/
     return SCM_UNDEFINED;
 }
@@ -410,7 +398,7 @@ ScmObj Scm_ThreadSleep(ScmObj timeout)
    gracefully. */
 static int wait_for_termination(ScmVM *target)
 {
-    struct timespec ts;
+    ScmTimeSpec ts;
     int r;
     ScmObj t = Scm_MakeFlonum(0.001); /* 1ms. somewhat arbitrary */
     Scm_GetTimeSpec(t, &ts);
