@@ -43,7 +43,21 @@
 
 #if defined(GAUCHE_USE_AXTLS)
 #include "axTLS/ssl/ssl.h"
-#endif /*GAUCHE_USE_AXTLS*/
+#else /*!GAUCHE_USE_AXTLS*/
+#define SSL_CLIENT_AUTHENTICATION               0x00010000
+#define SSL_SERVER_VERIFY_LATER                 0x00020000
+#define SSL_NO_DEFAULT_KEY                      0x00040000
+#define SSL_DISPLAY_STATES                      0x00080000
+#define SSL_DISPLAY_BYTES                       0x00100000
+#define SSL_DISPLAY_CERTS                       0x00200000
+#define SSL_DISPLAY_RSA                         0x00400000
+#define SSL_CONNECT_IN_PARTS                    0x00800000
+#define SSL_OBJ_X509_CERT                       1
+#define SSL_OBJ_X509_CACERT                     2
+#define SSL_OBJ_RSA_KEY                         3
+#define SSL_OBJ_PKCS8                           4
+#define SSL_OBJ_PKCS12                          5
+#endif /*!GAUCHE_USE_AXTLS*/
 
 SCM_DECL_BEGIN
 
@@ -62,9 +76,13 @@ SCM_CLASS_DECL(Scm_TLSClass);
 #define SCM_TLS(obj)    ((ScmTLS*)obj)
 #define SCM_TLSP(obj)   SCM_XTYPEP(obj, SCM_CLASS_TLS)
 
-extern ScmObj Scm_MakeTLS(void);
+extern ScmObj Scm_MakeTLS(uint32_t options, int num_sessions);
 extern ScmObj Scm_TLSDestroy(ScmTLS* t);
+extern ScmObj Scm_TLSLoadObject(ScmTLS* t, ScmObj obj_type,
+                                const char *filename,
+                                const char *password);
 extern ScmObj Scm_TLSConnect(ScmTLS* t, int fd);
+extern ScmObj Scm_TLSAccept(ScmTLS* t, int fd);
 extern ScmObj Scm_TLSClose(ScmTLS* t);
 
 /*
