@@ -40,7 +40,6 @@ SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_TLSClass, tls_print);
 
 static void tls_print(ScmObj obj, ScmPort* port, ScmWriteContext* ctx)
 {
-    ScmTLS* t = SCM_TLS(obj);
     Scm_Printf(port, "#<TLS");
     /* at the moment there's not much to print, so we leave this hole
        for future development. */
@@ -73,16 +72,12 @@ static void close_check(ScmTLS* tls, const char* op)
 #endif /*GAUCHE_USE_AXTLS*/
 }
 
-ScmObj Scm_MakeTLS(void)
+ScmObj Scm_MakeTLS(uint32_t options, int num_sessions)
 {
     ScmTLS* t = SCM_NEW(ScmTLS);
     SCM_SET_CLASS(t, SCM_CLASS_TLS);
 #if defined(GAUCHE_USE_AXTLS)
-    /* NB: we don't support certificate validation/trust. future work
-       will have to take care of this if anyone cares about it at the
-       policy level. (it should be noted that axTLS does support this;
-       we just don't use it at the moment) */
-    t->ctx = ssl_ctx_new(SSL_SERVER_VERIFY_LATER, 0);
+    t->ctx = ssl_ctx_new(options, num_sessions);
     t->conn = NULL;
     t->in_port = t->out_port = 0;
 #endif /*GAUCHE_USE_AXTLS*/
@@ -113,19 +108,41 @@ ScmObj Scm_TLSClose(ScmTLS* t)
     return SCM_TRUE;
 }
 
+ScmObj Scm_TLSLoadObject(ScmTLS* t, ScmObj obj_type,
+                         const char *filename, const char *password)
+{
+#if defined(GAUCHE_USE_AXTLS)
+    uint32_t type = Scm_GetIntegerU32Clamp(obj_type, SCM_CLAMP_ERROR, NULL);
+    if (ssl_obj_load(t->ctx, type, filename, password) == SSL_OK)
+        return SCM_TRUE;
+#endif /*GAUCHE_USE_AXTLS*/
+    return SCM_FALSE;
+}
+
 ScmObj Scm_TLSConnect(ScmTLS* t, int fd)
 {
 #if defined(GAUCHE_USE_AXTLS)
     context_check(t, "connect");
     if (t->conn) Scm_SysError("attempt to connect already-connected TLS %S", t);
     t->conn = ssl_client_new(t->ctx, fd, 0, 0);
-    if (SSL_OK != ssl_handshake_status(t->conn)) {
-        Scm_SysError("TLS handshake failed");
+    int r = ssl_handshake_status(t->conn);
+    if (r != SSL_OK) {
+        Scm_Error("TLS handshake failed: %d", r);
     }
 #endif /*GAUCHE_USE_AXTLS*/
     return SCM_OBJ(t);
 }
 
+ScmObj Scm_TLSAccept(ScmTLS* t, int fd)
+{
+#if defined(GAUCHE_USE_AXTLS)
+    context_check(t, "accept");
+    if (t->conn) Scm_SysError("attempt to connect already-connected TLS %S", t);
+    t->conn = ssl_server_new(t->ctx, fd);
+#endif /*GAUCHE_USE_AXTLS*/
+    return SCM_OBJ(t);
+}
+
 ScmObj Scm_TLSRead(ScmTLS* t)
 {
 #if defined(GAUCHE_USE_AXTLS)
