@@ -2,10 +2,8 @@
 ;; Generates uvect.c from uvect.c.tmpl
 ;;
 
-(use srfi-1)
 (use srfi-13)
 (use text.tree)
-(use util.list)
 (use file.util)
 (use gauche.sequence)
 (use gauche.parseopt)
@@ -56,7 +54,8 @@
              (p "\"" (regexp-replace-all #/[\\\"]/ line "\\\\\\0") "\"")
              (loop (read-line))])))
 
-  (receive (out name) (sys-mkstemp tmpl-file)
+  (receive (out name)
+      (sys-mkstemp (build-path (temporary-directory) (sys-basename tmpl-file)))
     (unwind-protect
      (begin
        (with-output-to-port out
