@@ -1,7 +1,7 @@
 /*
  * vport.c - 'virtual port'
  *
- *   Copyright (c) 2004-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2004-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -106,7 +106,7 @@ static int vport_getb(ScmPort *p)
                to pushback multiple bytes. */
             Scm_UngetbUnsafe(buf[i], p);
         }
-        return buf[0];
+        return (unsigned char)buf[0];
     } else {
         ScmObj b = Scm_ApplyRec(data->getb_proc, SCM_NIL);
         if (!SCM_INTP(b)) return EOF;
@@ -694,8 +694,6 @@ static ScmClassStaticSlotSpec boport_slots[] = {
  * Initialization
  */
 
-void Scm_Init_vportlib(ScmModule*);
-
 SCM_EXTENSION_ENTRY void Scm_Init_gauche__vport(void)
 {
     SCM_INIT_EXTENSION(gauche__vport);
@@ -711,6 +709,4 @@ SCM_EXTENSION_ENTRY void Scm_Init_gauche__vport(void)
                         "<buffered-output-port>", mod, boport_slots, 0);
 
     key_bufsize = SCM_MAKE_KEYWORD("buffer-size");
-
-    Scm_Init_vportlib(mod);
 }
