@@ -1,7 +1,7 @@
 /*
  * init.c - MS Windows specific utilities
  *
- *   Copyright (c) 2010-2012  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2010-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -53,5 +53,9 @@ SCM_EXTENSION_ENTRY void Scm_Init_os__windows(void)
     Scm_Init_console(mod);
 }
 
-#endif /*defined(GAUCHE_WINDOWS)*/
+#else  /*!defined(GAUCHE_WINDOWS)*/
+SCM_EXTENSION_ENTRY void Scm_Init_os__windows(void)
+{
+}
+#endif /*!defined(GAUCHE_WINDOWS)*/
 
