@@ -95,8 +95,7 @@ SCM_DEFINE_BASE_CLASS(Scm_ZlibVersionErrorClass, ScmZlibVersionError,
 
 static ScmObj zliberror_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmZlibError *e = SCM_ALLOCATE(ScmZlibError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmZlibError *e = SCM_NEW_INSTANCE(ScmZlibError, klass);
     e->message = SCM_FALSE;
     return SCM_OBJ(e);
 }
@@ -163,14 +162,13 @@ void Scm_ZlibError(int error_code, const char *msg, ...)
         e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
     }
     SCM_END_PROTECT;
-    Scm_VMThrowException(vm, e);
+    Scm_VMThrowException2(vm, e, 0);
     Scm_Panic("Scm_Error: Scm_VMThrowException returned.  something wrong.");
 }
 
 static ScmObj porterror_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmPortError *e = SCM_ALLOCATE(ScmPortError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmPortError *e = SCM_NEW_INSTANCE(ScmPortError, klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->port = NULL;                /* set by initialize */
     return SCM_OBJ(e);
@@ -556,13 +554,8 @@ ScmObj Scm_InflateSync(ScmPort *port)
 /*
  * Module initialization function.
  */
-extern void Scm_Init_zliblib(ScmModule*);
-
-SCM_EXTENSION_ENTRY void Scm_Init_rfc__zlib(void)
+void Scm_Init_zlib(void)
 {
-    /* Register this DSO to Gauche */
-    SCM_INIT_EXTENSION(rfc__zlib);
-
     /* Create the module if it doesn't exist yet. */
     ScmModule *mod = SCM_MODULE(SCM_FIND_MODULE("rfc.zlib", TRUE));
 
@@ -596,7 +589,4 @@ SCM_EXTENSION_ENTRY void Scm_Init_rfc__zlib(void)
                                 "<zlib-version-error>",
                                 mod, cond_meta, SCM_FALSE,
                                 zliberror_slots, 0);
-
-    /* Register stub-generated procedures */
-    Scm_Init_zliblib(mod);
 }
