@@ -1,7 +1,7 @@
 ;;;
 ;;; dbm - abstract base class for dbm interface
 ;;;
-;;;   Copyright (c) 2000-2012  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -34,6 +34,7 @@
 (define-module dbm
   (use gauche.collection)
   (use gauche.dictionary)
+  (use gauche.generator)
   (export <dbm> <dbm-meta>
           dbm-open    dbm-close   dbm-closed? dbm-get
           dbm-put!    dbm-delete! dbm-exists?
@@ -86,7 +87,7 @@
       (sys-link to1 to2))
     (begin
       (copy-file from1 to1 :safe #t :if-exists if-exists)
-      (guard (e (else (sys-unlink to1) (sys-unlink to2) (raise e)))
+      (guard (e [else (sys-unlink to1) (sys-unlink to2) (raise e)])
         (copy-file from2 to2 :safe #t :if-exists if-exists)))))
 
 (define (%dbm-rename2 from1 to1 from2 to2 :key (if-exists :error))
@@ -108,15 +109,15 @@
 
 (define-method dbm-open ((self <dbm>))
   (define (pick-proc slot default custom)
-    (let ((spec (slot-ref self slot)))
-      (cond ((eq? spec #f) identity)
-            ((eq? spec #t) default)
-            ((and (pair? spec)
+    (let1 spec (slot-ref self slot)
+      (cond [(eq? spec #f) identity]
+            [(eq? spec #t) default]
+            [(and (pair? spec)
                   (null? (cddr spec))
                   (procedure? (car spec))
                   (procedure? (cadr spec)))
-             (custom spec))
-            (else (errorf "bad value for ~s: has to be boolean or a list of two procedures, but got ~s" slot spec)))))
+             (custom spec)]
+            [else (errorf "bad value for ~s: has to be boolean or a list of two procedures, but got ~s" slot spec)])))
 
   (slot-set! self 'k2s (pick-proc 'key-convert write-to-string car))
   (slot-set! self 's2k (pick-proc 'key-convert read-from-string cadr))
@@ -157,60 +158,43 @@
 
 (define-method dbm-for-each ((dbm <dbm>) proc)
   (when (dbm-closed? dbm) (errorf "dbm-for-each: dbm already closed: ~s" dbm))
-  (dbm-fold dbm (lambda (key value r) (proc key value)) #f))
+  (dbm-fold dbm (^[key value r] (proc key value)) #f))
 
 (define-method dbm-map ((dbm <dbm>) proc)
   (when (dbm-closed? dbm) (errorf "dbm-map: dbm already closed: ~s" dbm))
   (reverse
-   (dbm-fold dbm (lambda (key value r) (cons (proc key value) r)) '())))
+   (dbm-fold dbm (^[key value r] (cons (proc key value) r)) '())))
 
+;;
 ;; Collection framework
-;;   This is a fallback, using "iterator inversion" technique to obtain
-;;   a generator from dbm-fold.  The subclass may directly implement
-;;   them if they have underlying generators.
-;; NB: this doesn't work due to the bug of call/cc handling.
-;(define-method call-with-iterator ((dbm <dbm>) proc . options)
-;  (define restart #f)
-;  (define buf #f)
-;  (define (fetch)
-;    (cond
-;     ((eq? restart 'end) #f) ;; already finished
-;     ((not restart)          ;; initial setup
-;      (let/cc return
-;        (dbm-fold dbm
-;                  (lambda (k v r)
-;                    (let/cc res
-;                      (set! restart res)
-;                      (return (cons k v))))
-;                  #f)
-;        (set! restart 'end)
-;        'end))
-;     (else (restart #f))))
-;  (set! buf (fetch))
-;  (proc (lambda () (eq? buf 'end))
-;        (lambda () (begin0 buf (set! buf (fetch))))))
+;;
+(define-method call-with-iterator ((dbm <dbm>) proc . options)
+  (let* ([g (generate
+             (^[yield] (dbm-fold dbm (^[k v r] (yield (cons k v))) #f)))]
+         [buf (g)])
+    (proc (^[] (eof-object? buf))
+          (^[] (begin0 buf (set! buf (g)))))))
 
+(define-method coerce-to ((target <list-meta>) (dbm <dbm>))
+  (dbm-map dbm cons))
+
+;;
 ;; Dictionary framework
-(define-method dict-get ((dbm <dbm>) key . maybe-default)
-  (apply dbm-get dbm key maybe-default))
+;;
+(define-dict-interface <dbm>
+  :get       dbm-get
+  :put!      dbm-put!
+  :exists?   dbm-exists?
+  :delete!   dbm-delete!
+  :fold      dbm-fold
+  :map       dbm-map
+  :for-each  dbm-for-each)
 
-(define-method dict-put! ((dbm <dbm>) key val)
-  (dbm-put! dbm key val))
-
-(define-method dict-exists? ((dbm <dbm>) key)
-  (dbm-exists? dbm key))
-
-(define-method dict-delete! ((dbm <dbm>) key)
-  (dbm-delete! dbm key))
-
-(define-method dict-fold ((dbm <dbm>) proc seed)
-  (dbm-fold dbm proc seed))
-
-(define-method dict-for-each ((dbm <dbm>) proc)
-  (dbm-for-each dbm proc))
-
-(define-method dict-map ((dbm <dbm>) proc)
-  (dbm-map dbm proc))
+(define-method dict-comparator ((dbm <dbm>))
+  (let1 k2s (~ dbm'k2s)
+    (make-comparator #t
+                     (^[a b] (equal? (k2s a) (k2s b)))
+                     #f #f)))
 
 ;;
 ;; Meta-operations
@@ -229,17 +213,17 @@
   ;; but it opens from-db first with read mode, so if the implementation
   ;; has sane locking, the to-db opening with create option would fail.
   ;; (That's why we're using let* here.)
-  (let* ((from-db (dbm-open class :path from :rw-mode :read))
-         (to-db   (dbm-open class :path to   :rw-mode :create)))
-    (dbm-for-each from-db (lambda (k v) (dbm-put! to-db k v)))
+  (let* ([from-db (dbm-open class :path from :rw-mode :read)]
+         [to-db   (dbm-open class :path to   :rw-mode :create)])
+    (dbm-for-each from-db (^[k v] (dbm-put! to-db k v)))
     (dbm-close to-db)
     (dbm-close from-db)))
 
 (define-method dbm-db-move ((class <dbm-meta>) from to)
   ;; generic one - see above.
-  (let* ((from-db (dbm-open class :path from :rw-mode :read))
-         (to-db   (dbm-open class :path to   :rw-mode :create)))
-    (dbm-for-each from-db (lambda (k v) (dbm-put! to-db k v)))
+  (let* ([from-db (dbm-open class :path from :rw-mode :read)]
+         [to-db   (dbm-open class :path to   :rw-mode :create)])
+    (dbm-for-each from-db (^[k v] (dbm-put! to-db k v)))
     (dbm-close to-db)
     (dbm-close from-db)
     (dbm-db-remove class from)))
@@ -251,8 +235,8 @@
 ;; Returns #f if it couldn't retrieve the named dbm module.
 
 (define (dbm-type->class dbmtype)
-  (let ([module-name (string->symbol #`"dbm.,dbmtype")]
-        [class-name (string->symbol #`"<,|dbmtype|>")])
+  (let ([module-name (string->symbol #"dbm.~dbmtype")]
+        [class-name (string->symbol #"<~|dbmtype|>")])
     (and (library-exists? module-name :strict? #t)
          (guard (e [else #f])
            ((with-module gauche.internal %require)
