@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.cgen.precomp - Precompile Scheme into C data
 ;;;
-;;;   Copyright (c) 2004-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -46,14 +46,11 @@
   (use gauche.experimental.lamb)
   (use file.util)
   (use util.match)
-  (use util.list)
   (use util.toposort)
   (use text.tr)
   (export cgen-precompile cgen-precompile-multi))
 (select-module gauche.cgen.precomp)
 
-;; Do not load optimizer unless explicitly requested.  To compile
-;; 0.9.4 with 0.9.3 compiler, we can't depend on the optimizer module.
 (autoload gauche.cgen.optimizer optimize-compiled-code)
 
 ;;================================================================
@@ -164,7 +161,7 @@
     (setup ext-initializer sub-initializers)
     (with-input-from-file src
       (cut emit-toplevel-executor
-           (reverse (port-fold compile-toplevel-form '() read))))
+           (reverse (generator-fold compile-toplevel-form '() read))))
     (finalize sub-initializers)
     (cgen-emit-c (cgen-current-unit))))
 
@@ -197,7 +194,7 @@
                             :extra-optimization extra-optimization
                             :ext-initializer (and (equal? src main)
                                                   ext-initializer)
-                            :initializer-name #`"Scm_Init_,initname"))))]
+                            :initializer-name #"Scm_Init_~initname"))))]
     ))
 
 ;; Common stuff -- process single source
@@ -306,6 +303,7 @@
   (with-module gauche.internal vm-eval-situation))
 (define global-eq?? (with-module gauche.internal global-eq??))
 (define make-identifier (with-module gauche.internal make-identifier))
+(define pair-attribute-get (with-module gauche.internal pair-attribute-get))
 
 (define-constant SCM_VM_COMPILING 2) ;; must match with vm.h
 
@@ -318,7 +316,7 @@
          [safe-name (string-tr base "-+" "__")])
     (rlet1 u (make <cgen-stub-unit>
                :name base :c-name-prefix safe-name
-               :preamble `(,#`"/* Generated automatically from ,|src|.  DO NOT EDIT */")
+               :preamble `(,#"/* Generated automatically from ~|src|.  DO NOT EDIT */")
                :init-prologue (format "~avoid Scm_Init_~a() {"
                                       (if ext-init? "SCM_EXTENSION_ENTRY " "")
                                       safe-name)
@@ -387,21 +385,20 @@
 
 (define (setup ext-init? subinits)
   (cgen-decl "#include <gauche/code.h>")
-  (cgen-decl "#include <gauche/macro.h>") ; for MakeMacroTransformerOld. temporary.
   (cond [(and ext-init? (ext-module-file))
          => (^[extm]
               (cgen-decl "#include <gauche/extend.h>")
               (let* ([extname ($ path-sans-extension
                                  $ sys-basename $ port-name extm)]
                      [safe-extname (regexp-replace-all #/\W/ extname "_")])
-                (cgen-init #`"SCM_INIT_EXTENSION(,safe-extname);")))])
+                (cgen-init #"SCM_INIT_EXTENSION(~safe-extname);")))])
   (dolist [init subinits]
-    (cgen-decl #`"extern void Scm_Init_,init(void);"))
+    (cgen-decl #"extern void Scm_Init_~init(void);"))
   )
 
 (define (finalize subinits)
   (dolist [init subinits]
-    (cgen-init #`"  Scm_Init_,init();")))
+    (cgen-init #"  Scm_Init_~init();")))
 
 ;;================================================================
 ;; Compiler stuff
@@ -454,7 +451,6 @@
 (define-global-pred =export-if-defined? export-if-defined)
 (define-global-pred =provide?         provide)
 (define-global-pred =lambda?          lambda)
-(define-global-pred =declare?         declare)
 
 ;; A parameter that holds the list of 'omitted' #<compiled-code> - for
 ;; example, the cliche of (CLOSURE #<compiled-code> DEFINE #<identifier> RET)
@@ -505,15 +501,6 @@
       [((? =export-all?)) (compile-module-exports #t)]
       [((? =export-if-defined?) . _) (write-ext-module form) seed]
       [((? =provide?) arg) (write-ext-module form) seed]
-      ;; TODO - we need more general framework supporting various declarations.
-      ;; for the time being, this ad-hoc solution suffice our needs.
-      [((? =declare?) decls ...)
-       (dolist [x decls]
-         (match x
-           [('keep-private-macro . macros)
-            (private-macros-to-keep (append (private-macros-to-keep) macros))]
-           [other (error "Unknown declaration:" other)]))
-       seed]
       ;; Finally, ordinary expressions.
       [else
        (let* ([compiled-code (compile-in-current-tmodule form)]
@@ -548,22 +535,25 @@
              (cadr (~ toplevel-code 2))))) ; define flags
 
 (define (emit-toplevel-definition inner-code id flags)
-  (let ([sym  (cgen-literal (~ id'name))]
+  (let ([sym  (cgen-literal (unwrap-syntax id))]
         ;; NB: Currently, the main tmodule has no name.  It's better to give
         ;; it a proper name.  Then the following Scm_CurrentModule hack
         ;; will be unnecessary.
         [mod  (and-let* ([n (module-name (~ id'module))])
                 (find-tmodule n))]
-        [code (cgen-literal inner-code)])
+        [code (cgen-literal inner-code)]
+        [tmp (gensym)])
+    (cgen-init (format "  ScmObj ~a = Scm_MakeClosure(~a, NULL);\n"
+                       tmp (cgen-cexpr code)))
     (cgen-init (format "  Scm_MakeBinding(SCM_MODULE(~a) /* ~a */, \
                                           SCM_SYMBOL(~a) /* ~a */, \
-                                          Scm_MakeClosure(~a, NULL),\
+                                          ~a,\
                                           ~a);\n"
                        (if mod (tmodule-cname mod) "Scm_CurrentModule()")
                        (if mod (cgen-safe-comment (~ mod'name)) "")
                        (cgen-cexpr sym)
-                       (cgen-safe-comment (~ id'name))
-                       (cgen-cexpr code)
+                       (cgen-safe-comment (unwrap-syntax id))
+                       tmp
                        (case flags
                          [(2) 'SCM_BINDING_CONST]
                          [(4) 'SCM_BINDING_INLINABLE]
@@ -595,6 +585,11 @@
 ;; The current approach still has an issue when the compiled source
 ;; overrides these special forms.  Such sources should be very unusual,
 ;; so we don't support them for the time being.
+;;
+;; TODO: Reconstructing define-cproc etc. would lose the source information
+;; attached to the original form.  We can't fix it right now, for define-macro
+;; doesn't provide access to the original form.  Maybe after we replace
+;; it with er-macro-expander...
 
 (define *special-handlers*
   '((define-macro (current-module)
@@ -609,14 +604,35 @@
       ((with-module gauche.cgen.stub cgen-stub-parse-form)
        (unwrap-syntax (cons 'define-cproc args)))
       (undefined))
-    ;; (define-macro (define . f)
-    ;;   ((with-module gauche.cgen.precomp handle-define) f))
+    (define-macro (define-enum . args)
+      ((with-module gauche.cgen.stub cgen-stub-parse-form)
+       (unwrap-syntax (cons 'define-enum args)))
+      (undefined))
+    (define-macro (define-enum-conditionally . args)
+      ((with-module gauche.cgen.stub cgen-stub-parse-form)
+       (unwrap-syntax (cons 'define-enum-conditionally args)))
+      (undefined))
     (define-macro (define-constant . f)
       ((with-module gauche.cgen.precomp handle-define-constant) f))
     (define-macro (define-syntax . f)
       ((with-module gauche.cgen.precomp handle-define-syntax) f))
     (define-macro (define-macro . f)
       ((with-module gauche.cgen.precomp handle-define-macro) f))
+    ;; TODO - we need more general framework supporting various declarations.
+    ;; for the time being, this ad-hoc solution suffice our needs.
+    (define-macro (declare . f)
+      ((with-module gauche.cgen.precomp handle-declare) f))
+    ;; A special directive not to precompile; the given forms are
+    ;; emitted to *.sci file as they are.  It is useful if you delay
+    ;; macro-expansion until the load time (e.g. cond-expand).  The
+    ;; forms are not evaluated at all in the compiling environment,
+    ;; so they are not avaialble for macro expansion of the forms
+    ;; to be precompiled.  (The case can be handled more generally by
+    ;; 'eval-when' mechanism, but properly support eval-when needs more
+    ;; work.)
+    (define-macro (without-precompiling . forms)
+      ((with-module gauche.cgen.precomp handle-without-compiling) forms)
+      (undefined))
     ))
 
 ;; Macros are "consumed" by the Gauche's compiler---that is, it is
@@ -624,8 +640,6 @@
 ;; For exported macros, we should include the macro itself in the compiled
 ;; file, so we intercept define-macro and define-syntax.
 
-;; For the time being, we only compile the legacy macros into C file.
-;; R5RS macros are put in ext-module file as is.
 (define (handle-define-macro form)
   (define %define (make-identifier 'define (find-module 'gauche) '()))
   (define %define-syntax (make-identifier 'define-syntax (find-module 'gauche) '()))
@@ -648,14 +662,28 @@
     [(name expr) (do-handle name expr)]
     [_ (error "Malformed define-macro" form)]))
 
+;; At this moment, we can only precompile macros that have closure
+;; in its transformer.
 (define (handle-define-syntax form)
   (match form
-    [(name . _)
+    [(name xformer-spec)
+     ;; This inserts sets-up compile-time environment.
+     (eval-in-current-tmodule
+      `((with-module gauche define-syntax) ,@form))
+     ;; If the macro needs to exported, check if we can put it in
+     ;; precompiled file (it is, if the transformer is a closure).
+     ;; Othewise, we emit the form to *.sci file.
      (when (or (symbol-exported? name)
                (memq name (private-macros-to-keep)))
-       (write-ext-module `(define-syntax . ,form)))]
-    [_ #f])
-  (cons '(with-module gauche define-syntax) form))
+       (let1 val (global-variable-ref (~ (current-tmodule)'module) name #f)
+         (or (and-let* ([ ((with-module gauche.internal macro?) val) ]
+                        [tx ((with-module gauche.internal macro-transformer) val)]
+                        [ (closure? tx) ])
+               `((with-module gauche.internal %insert-syntax-binding)
+                 (current-module) ',name ,xformer-spec))
+             (write-ext-module `(define-syntax . ,form))
+             #f)))]
+    [_ (error "Malformed define-syntax" form)]))
 
 (define (handle-define-constant form)
   (match form
@@ -665,6 +693,25 @@
     [_ #f])
   (cons '(with-module gauche define-constant) form))
 
+(define (handle-without-compiling forms)
+  (for-each write-ext-module forms))
+
+;; Handle declaration.
+;; At this moment, we only recognize (keep-private-macro name ...) form
+;; Caveat:
+;;  - The decalation must be seen before define-macro
+;;  - The name is treated unhygienic and global, no matter where it appears.
+;; It should really be a metadata of macro binding and must be folded
+;; into define-macro form.  The keep-private-macro trick should strictly
+;; be considered temporary hack.
+(define (handle-declare decls)
+  (dolist [x decls]
+    (match x
+      [('keep-private-macro . macros)
+       (private-macros-to-keep (append (private-macros-to-keep) macros))]
+      [other (error "Unknown declaration:" other)]))
+  (undefined))
+
 ;; check to see if the symbol is exported
 (define (symbol-exported? sym)
   (or (eq? (compile-module-exports) #t)
@@ -674,9 +721,10 @@
 ;; Compiler-specific literal handling definitions
 ;;
 (define-cgen-literal <cgen-scheme-code> <compiled-code>
-  ((code-name   :init-keyword :code-name)
-   (code-vector-c-name :init-keyword :code-vector-c-name)
-   (literals    :init-keyword :literals)
+  ([code-name          :init-keyword :code-name]
+   [code-vector-c-name :init-keyword :code-vector-c-name]
+   [literals           :init-keyword :literals]
+   [signature-info     :init-keyword :signature-info]
    )
   (make (value)
     (let* ([code (if (run-extra-optimization-passes)
@@ -686,7 +734,7 @@
            [lv  (extract-literals cv)]
            [cvn (allocate-code-vector cv lv (~ code'full-name))]
            [code-name (cgen-literal (~ code'name))]
-           [arg-info (cgen-literal (~ code'arg-info))]
+           [signature-info (cgen-literal (serializable-signature-info code))]
            [inliner (check-packed-inliner code)])
       (define (init-thunk)
         (format #t "    SCM_COMPILED_CODE_CONST_INITIALIZER(  /* ~a */\n"
@@ -700,7 +748,9 @@
                 (if (cgen-literal-static? code-name)
                   (cgen-cexpr code-name)
                   "SCM_FALSE")
-                (cgen-cexpr arg-info))
+                (if (cgen-literal-static? signature-info)
+                  (cgen-cexpr signature-info)
+                  "SCM_FALSE"))
         (format #t "            ~a, ~a)"
                 (let1 parent-code (~ code'parent)
                   (if (memq parent-code (omitted-code))
@@ -714,16 +764,37 @@
                                                 init-thunk)
             :code-vector-c-name cvn
             :code-name code-name
+            :signature-info signature-info
             :literals lv)))
   (init (self)
     (unless (cgen-literal-static? (~ self'code-name))
       (print "  SCM_COMPILED_CODE("(~ self'c-name)")->name = "
              (cgen-cexpr (~ self'code-name))";"
              "/* "(cgen-safe-comment (~ self'value'full-name))" */"))
+    (unless (cgen-literal-static? (~ self'signature-info))
+      (print "  SCM_COMPILED_CODE("(~ self'c-name)")->signatureInfo = "
+             (cgen-cexpr (~ self'signature-info)) ";"))
     (fill-code self))
   (static (self) #t)
   )
 
+;; Construct serializable signature info.  cgen-literal doesn't emit
+;; ExtendedPair as it is, and we need to use <serialiable-extended-pair>.
+;; See literal.scm.
+(define (serializable-signature-info code)
+  ;; TRANSIENT: This code needs to run with 0.9.4 as well, so we use old
+  ;; slot name 'arg-info'.  Replace it to 'signature-info' after 0.9.5
+  ;; release.
+  (and-let* ([sig (~ code'arg-info)])
+    (if-let1 si (and (pair? sig)
+                     (pair? (car sig))
+                     (pair-attribute-get (car sig) 'source-info #f))
+      (cons (make-serializable-extended-pair (unwrap-syntax (caar sig))
+                                             (unwrap-syntax (cdar sig))
+                                             `((source-info . ,si)))
+            (unwrap-syntax (cdr sig)))
+      (unwrap-syntax sig))))
+
 ;; Returns a list of the same length of CODE, which includes the
 ;; <cgen-literal>s corresponding to the literal values in CODE.
 ;; #f is filled in the places that don't have corresponding litaral value.
@@ -805,14 +876,18 @@
 
   (loop cv lv 0 #f))
 
-(define (fill-code code)
-  (let ([cvn  (~ code'code-vector-c-name)]
-        [lv   (~ code'literals)])
+(define (fill-code code-literal)
+  (let ([cvn  (~ code-literal'code-vector-c-name)]
+        [lv   (~ code-literal'literals)]
+        [si   (~ code-literal'signature-info)])
     (for-each-with-index
      (^[index lit] (when (and lit (not (cgen-literal-static? lit)))
                      (format #t "  ((ScmWord*)~a)[~a] = SCM_WORD(~a);\n"
                              cvn index (cgen-cexpr lit))))
-     lv)))
+     lv)
+    (unless (cgen-literal-static? si)
+      (format #t "  SCM_COMPILED_CODE(~a)->signatureInfo = SCM_OBJ(~a);\n"
+              (cgen-cexpr code-literal) (cgen-cexpr si)))))
 
 ;; If the compiled-code has packed IForm for inliner, translate it for
 ;; the target VM insns and returns the packed IForm.
@@ -834,7 +909,7 @@
       (error "identifier with compiler environment can't be compiled" value))
     (make <cgen-scheme-identifier> :value value
           :c-name (cgen-allocate-static-datum)
-          :id-name (cgen-literal (~ value'name))
+          :id-name (cgen-literal (unwrap-syntax value))
           :mod-name (cond [(module-name-fix (~ value'module))
                            => cgen-literal]
                           [(current-tmodule) => (^m (cgen-literal (~ m'name)))]
@@ -843,7 +918,7 @@
     (let ([name (cgen-cexpr (~ self'id-name))]
           [cname (~ self'c-name)])
       (or (and-let* ([modnam (~ self'mod-name)])
-            (format #t "  ~a = Scm_MakeIdentifier(SCM_SYMBOL(~a), \
+            (format #t "  ~a = Scm_MakeIdentifier(~a, \
                                   Scm_FindModule(SCM_SYMBOL(~a), \
                                                  SCM_FIND_MODULE_CREATE),
                                   SCM_NIL); /* ~a#~a */\n"
@@ -852,7 +927,7 @@
                     (cgen-safe-comment (~ self'id-name'value)))
             #t)
           (let1 mod-cname (current-tmodule-cname)
-            (format #t "  ~a = Scm_MakeIdentifier(SCM_SYMBOL(~a), \
+            (format #t "  ~a = Scm_MakeIdentifier(~a, \
                                                   SCM_MODULE(~a), \
                                                   SCM_NIL); /* ~a#~a */\n"
                     cname name mod-cname
