@@ -1,7 +1,7 @@
 ;;;
 ;;; process.scm - process interface
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -37,8 +37,8 @@
 ;; as an object on top of basic system interface.
 
 (define-module gauche.process
+  (use gauche.generator)
   (use srfi-1)
-  (use srfi-2)
   (use srfi-13)
   (use srfi-14)
   (export <process> <process-abnormal-exit>
@@ -47,6 +47,7 @@
           process-wait process-wait-any process-exit-status
           process-send-signal process-kill process-stop process-continue
           process-list
+          run-process-pipeline
           ;; process ports
           open-input-process-port   open-output-process-port
           call-with-input-process   call-with-output-process
@@ -54,7 +55,7 @@
           call-with-process-io
           process-output->string    process-output->string-list
           ;; shell utilities
-          shell-escape-string
+          shell-escape-string shell-tokenize-string
           ))
 (select-module gauche.process)
 
@@ -173,10 +174,10 @@
            [argv (if host (%prepare-remote host argv directory) argv)]
            [dir  (if host #f directory)])
       (%check-directory dir)
-      (receive (iomap toclose ipipes opipes)
+      (receive (iomap toclose ipipes opipes tmpfiles)
           (if (pair? redirs)
             (%setup-iomap proc redirs)
-            (values #f '() '() '()))
+            (values #f '() '() '() '()))
         (set! (~ proc'in-pipes) ipipes)
         (set! (~ proc'out-pipes) opipes)
         (if fork
@@ -186,14 +187,15 @@
                                        :detached detached)
             (push! (ref proc 'processes) proc)
             (set!  (ref proc 'pid) pid)
-            (dolist (p toclose)
+            (dolist [p toclose]
               (if (input-port? p)
                 (close-input-port p)
                 (close-output-port p)))
             (when (and wait (not detached))
               ;; the following expr waits until the child exits
               (set! (ref proc 'status) (values-ref (sys-waitpid pid) 1))
-              (update! (ref proc 'processes) (cut delete proc <>)))
+              (update! (ref proc 'processes) (cut delete proc <>))
+              (for-each sys-unlink tmpfiles))
             proc)
           (sys-exec (car argv) argv
                     :iomap iomap :directory dir
@@ -206,7 +208,20 @@
         ,@(if in  `((< 0 ,(if (eq? in  :pipe) 'stdin  in))) '())
         ,@(if out `((> 1 ,(if (eq? out :pipe) 'stdout out))) '())
         ,@(if err `((> 2 ,(if (eq? err :pipe) 'stderr err))) '()))
-    (for-each %check-redirects redirs)))
+    ;; Reject if the same pipe appears more than once in the reirect list.
+    ;; We do allow the same file appears more than once; e.g. redirecting
+    ;; both 1 and 2 to "/dev/null".
+    (fold (^[redir seen]
+            (%check-redirects redir)
+            (let1 source-sink (caddr redir)
+              (if (and (symbol? source-sink)
+                       (not (memq source-sink '(:null :pipe))))
+                (if (memq source-sink seen)
+                  (errorf "Pipe name `~s' appears more than once in the \
+                           redirection source or target" source-sink)
+                  (cons source-sink seen))
+                seen)))
+          '() redirs)))
 
 (define (%check-redirects arg)
   (unless (and (pair? arg) (pair? (cdr arg)) (pair? (cddr arg))
@@ -277,7 +292,7 @@
   (let1 val #f
     (lambda ()
       (or val
-          (rlet1 v (sys-normalize-pathname #`",(sys-tmpdir)/gauche"
+          (rlet1 v (sys-normalize-pathname #"~(sys-tmpdir)/gauche"
                                            :canonicalize #t)
             (set! val v))))))
 
@@ -290,6 +305,7 @@
   (define opipes '())   ;list of (name . port)
   (define iomap '())    ;list of (fd . port/fd)
   (define seen '())     ;list of seen fds
+  (define tmpfiles '()) ;list of temporary files
 
   (define (do-file dir fd arg)
     (let1 p (case dir
@@ -350,13 +366,20 @@
                                            [else (write-block arg o)])
                                    (close-output-port o)))])
            (cond-expand
-            [gauche.sys.pthreads
+            [gauche.sys.threads
              (receive (in out) (sys-pipe)
                (push! iomap `(,fd . ,in))
                (thread-start! (make-thread (cut write-arg out))))]
             [else
+             ;; If we can't use threads, we fall back to using temporary
+             ;; files.  It has some drawbacks: (1) If we 'exec'-ing
+             ;; (not forking), we don't have chance to remove them, and
+             ;; (2) On Windows we won't be able to remove temp file unless
+             ;; we wait the child process to complete.  We'll do our best
+             ;; anyway.
              (receive (out nam) (sys-mkstemp (%temp-path-prefix))
                (write-arg out)
+               (push! tmpfiles nam)
                (do-file '< fd nam))]))]
         [(<& >&) (push! todup r)] ;; process dups later
         [else (error "invalid redirection" dir)])))
@@ -370,7 +393,7 @@
   (unless (assv 1 iomap) (push! iomap '(1 . 1)))
   (unless (assv 2 iomap) (push! iomap '(2 . 2)))
 
-  (values iomap toclose ipipes opipes))
+  (values iomap toclose ipipes opipes tmpfiles))
 
 (define (%ensure-mask mask)
   (cond
@@ -446,6 +469,41 @@
    [gauche.os.windows (undefined)]
    [else (process-send-signal process SIGCONT)]))
 
+;;-----------------------------------------------------------------
+;; pipeline
+;;
+
+;; We might adopt scsh-like process forms eventually, but finding an
+;; optimal DSL takes time.  Meanwhile, this intermediate-level API
+;; would cover typical use case...
+(define (run-process-pipeline commands
+                              :key (input #f) (output #f) (error #f)
+                              (wait #f)
+                              (sigmask #f) (directory #f)
+                              (detached #f))
+  (when (null? commands)
+    (error "At least one command is required to run-command-pipeline"))
+  (and-let1 offending (any (^c (and (not (pair? c)) (list c))) commands)
+    (errorf "Command list contains non-list command line '~s': ~s"
+            (car offending) commands))
+  (let* ([pipe-pairs (map (^_ (call-with-values sys-pipe cons))
+                          (cdr commands))]
+         [cmds (map (^[cmdline in out w]
+                      `(,cmdline :input ,in :output ,out :error ,error
+                                 :wait ,w :sigmask ,sigmask
+                                 :directory ,directory :detached ,detached))
+                    commands
+                    (cons input (map car pipe-pairs))
+                    (fold-right cons (list output) (map cdr pipe-pairs))
+                    `(,@(make-list (- (length commands) 1) #f) ,wait))]
+         ;; We have to close output pipe after spawning the process, for
+         ;; the process that reads from the pipe would wait until all the
+         ;; ends are closed.  We have to treat the last command separately,
+         ;; for we might :wait #t for it.
+         [ps (map (cut apply run-process <>) (drop-right cmds 1))])
+    (dolist [p pipe-pairs] (close-output-port (cdr p)))
+    (append ps (list (apply run-process (last cmds))))))
+
 ;;===================================================================
 ;; Process ports
 ;;
@@ -532,20 +590,148 @@
 ;; Shell utility
 ;;
 
-(define (shell-escape-string str)
-  (cond-expand
-   [gauche.os.windows
+(define (shell-escape-string str :optional (flavor
+                                            (cond-expand
+                                             [gauche.os.windows 'windows]
+                                             [else 'posix])))
+  (ecase flavor
+   [(windows)
     ;; This is supported in src/libsys.scm.  See the comment in it.
     (%sys-escape-windows-command-line str)]
-   [else
+   [(posix)
     ;; We follow standard unix shell convention: if STR contains special
     ;; chars, we quote the entire STR by single-quotes.  If STR contains
     ;; a single quote, we replace it with '"'"'.
     (cond [(string-null? str) "''"]
-          [(string-index str #[\s\\\"\'*?$<>!\[\](){}])
+          [(string-index str #[\s\\\"\'`*?$<>!\[\](){}])
            (string-append "'" (regexp-replace-all #/'/ str "'\"'\"'") "'")]
           [else str])]))
 
+;; Aux procedure to tokenize command-line string into command name and
+;; arglist.   We don't deal with shell metacharacters; just recognize
+;; quotes.  For the time being, we reject any unquoted shell metacharacters.
+;; POSIX shell syntax:
+;; http://pubs.opengroup.org/onlinepubs/009695399/utilities/xcu_chap02.html
+;; About wicked MS shell syntax, see the comment of
+;; %sys-escape-windows-command-line in src/libsys.scm.
+;; NB: We don't expect STR to be too long, so just recurse.
+(define (shell-tokenize-string str :optional (flavor
+                                              (cond-expand
+                                               [gauche.os.windows 'windows]
+                                               [else 'posix])))
+  (define (skip-ws)
+    (let loop ([c (peek-char)])
+      (when (and (char? c) (char-whitespace? c))
+        (begin (read-char) (loop (peek-char))))))
+  (define (err fmt . args)
+    (apply errorf #"Bad shell syntax - ~fmt" args))
+  (define (err-meta c)
+    (err "shell metacharacter ~s not allowed, in ~s" c str))
+  (define (err-backslash)
+    (err "stray backslash in ~s" str))
+
+  (define (tokenize-posix)
+    (with-input-from-string str
+      (^[]
+        (define (read-next)
+          (let1 c (read-char)
+            (cond [(eof-object? c) '()]
+                  [(char-whitespace? c) (skip-ws) '()]
+                  [(eqv? c #\') (read-sq)]
+                  [(eqv? c #\") (read-dq)]
+                  [(eqv? c #\\) (read-escaped)]
+                  [(#[|&\;<>()$`] c) (err-meta c)]
+                  [else (read-unquoted c)])))
+        (define (read-escaped)
+          (let1 c (read-char)
+            (if (eof-object? c)
+              (err-backslash)
+              (read-unquoted c))))
+        (define (read-unquoted c0)
+          (let1 c (peek-char)
+            (cond [(eof-object? c) (list c0)]
+                  [(char-whitespace? c) (skip-ws) (list c0)]
+                  [(eqv? c #\') (cons c0 (read-sq))]
+                  [(eqv? c #\") (cons c0 (read-dq))]
+                  [(eqv? c #\\) (read-char) (cons c0 (read-escaped))]
+                  [(#[|&\;<>()$`] c) (err-meta c)]
+                  [else (read-char) (cons c0 (read-unquoted c))])))
+        (define (read-sq) 
+          (let loop ([c (read-char)] [cs '()]) 
+            (cond [(eof-object? c) (err "unclosed single quote: ~s" str)]
+                  [(eqv? c #\') (reverse cs (read-next))]
+                  [else (loop (read-char) (cons c cs))])))
+        (define (read-dq)
+          (let loop ([c (read-char)] [cs '()]) 
+            (cond [(eof-object? c) (err "unclosed double quote: ~s" str)]
+                  [(eqv? c #\") (reverse cs (read-next))]
+                  [(eqv? c #\\) (let1 c (read-char) ; section 2.2.3
+                                  (cond [(eof-object? c) (err-backslash)]
+                                        [(#[$`\"\\\n] c)
+                                         (loop (read-char) (cons c cs))]
+                                        [else
+                                         (loop (read-char) (cons* c #\\ cs))]))]
+                  [(#[$`] c) (err-meta c)]
+                  [else (loop (read-char) (cons c cs))])))
+
+        ;; body
+        (skip-ws)
+        (let loop ([words '()])
+          (if (eof-object? (peek-char))
+            (reverse words)
+            (loop (cons (list->string (read-next)) words)))))))
+
+  (define (tokenize-windows-cmd)
+    (with-input-from-string str
+      (^[]
+        (define (read-next)
+          (let1 c (read-char)
+            (cond [(eof-object? c) '()]
+                  [(char-whitespace? c) (skip-ws) '()]
+                  [(eqv? c #\") (read-quoted)]
+                  [else (read-unquoted c)])))
+        (define (read-unquoted c)
+          (let loop ([c c] [cs '()])
+            (cond [(eof-object? c) (reverse cs)]
+                  [(char-whitespace? c) (skip-ws) (reverse cs)]
+                  [(eqv? c #\\) (let1 cs (read-backslash cs)
+                                  (loop (read-char) cs))]
+                  [(eqv? c #\") (reverse cs (read-quoted))]
+                  [else (loop (read-char) (cons c cs))])))
+        (define (read-quoted)
+          (let loop ([c (read-char)] [cs '()])
+            (cond [(eof-object? c) (reverse cs)] ;implicitly closed
+                  [(eqv? c #\\) (let1 cs (read-backslash cs)
+                                  (loop (read-char) cs))]
+                  [(eqv? c #\") (read-after-closing cs)]
+                  [else (loop (read-char) (cons c cs))])))
+        (define (read-backslash cs)
+          (let loop ([n 1] [cs cs])
+            (let1 c (peek-char)
+              (cond [(eqv? c #\\) (read-char) (loop (+ n 1) cs)]
+                    [(eqv? c #\")
+                     (if (even? n)
+                       (append (make-list (ash n -1) #\\) cs)
+                       (cons (read-char)
+                             (append (make-list (ash n -1) #\\) cs)))]
+                    [else (append (make-list n #\\) cs)]))))
+        (define (read-after-closing cs)
+          (let1 c (peek-char)
+            (if (eqv? c #\") ; this is treated as literal dq
+              (begin (read-char)
+                     (reverse (cons c cs) (read-next)))
+              (reverse cs (read-next)))))
+        ;; body
+        (skip-ws)
+        (let loop ([words '()])
+          (if (eof-object? (peek-char))
+            (reverse words)
+            (loop (cons (list->string (read-next)) words)))))))
+
+  (ecase flavor
+    [(windows) (tokenize-windows-cmd)]
+    [(posix) (tokenize-posix)]))
+  
 ;;----------------------------------------------------------------------
 ;; Internal utilities for process ports
 ;;
