@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.test - test framework
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -83,15 +83,27 @@
 
 (define-module gauche.test
   (export test test* test-start test-end test-section test-log
-          test-module test-error test-one-of
+          test-module test-script
+          test-error test-one-of test-none-of
           test-check test-record-file test-summary-check
           *test-error* *test-report-error* test-error? prim-test))
 (select-module gauche.test)
 
-;; An object to represent error.
+;; An object to represent error.  This class isn't exported; the user
+;; must use `test-error' procedure to create an instance.
+;;
+;; This object is used in both the expected result and the actual result
+;; of test expression.   For the actual result, this object holds the
+;; raised condition in `condition' slot, and its class and message in
+;; the `class' and `message' slots.
+;; For the expected result, class slot must be set as one of <condition>
+;; classes, or #f.  If it's a <condition> class, then it is used to test
+;; the actual result condition has the condition type.  If it's #f,
+;; then any <test-error> object matches.
 (define-class <test-error> ()
-  ((message :init-keyword :message :init-value #f)
-   (class   :init-keyword :class   :init-value #f)))
+  ((condition :init-keyword :condition :init-value #f)
+   (class     :init-keyword :class     :init-value #f)
+   (message   :init-keyword :message   :init-value #f)))
 
 (define-method write-object ((obj <test-error>) out)
   (let1 cname (if (ref obj'class) (class-name (ref obj'class)) 'error)
@@ -104,38 +116,51 @@
 (define (test-error :optional (class #f) (message #f))
   (make <test-error> :class class :message message))
 
-;; An object to represet any one of
+;; An object to represent "any one of Xs"
 (define-class <test-one-of> ()
   ((choices :init-keyword :choices)))
 
 (define-method write-object ((obj <test-one-of>) out)
   (format out "#<test-one-of: any one of ~s>" (slot-ref obj'choices)))
 
+;; API
 (define (test-one-of . choices) (make <test-one-of> :choices choices))
 
+;; An object to represent "none of Xx"
+(define-class <test-none-of> ()
+  ((choices :init-keyword :choices)))
+
+(define-method write-object ((obj <test-none-of>) out)
+  (format out "#<test-none-of: none of ~s>" (slot-ref obj'choices)))
+
+;; API
+(define (test-none-of . choices) (make <test-none-of> :choices choices))
+
+;; API
+;; We don't use generic function dispatch (at least for the time being),
+;; to make it easy to troubleshoot when object system gets messed up.
+;; In future we'll make use of generic functions.
 (define (test-check expected result :optional (fallback equal?))
   (cond [(test-error? expected)
          (and (test-error? result)
-              (let ([ex (slot-ref expected'class)]
-                    [ey (slot-ref result'class)])
-                (or (not ex)
-                    (memq (slot-ref expected'class)
-                          (class-precedence-list (slot-ref result'class))))))]
+              (let ([c (slot-ref expected'class)]
+                    [e (slot-ref result'condition)])
+                (or (not c)
+                    (condition-has-type? e c))))]
         [(is-a? expected <test-one-of>)
-         ;; NB: not using srfi-1's any to avoid dependency
-         (let loop ([choices (slot-ref expected 'choices)])
-           (cond [(null? choices) #f]
-                 [(test-check (car choices) result fallback)]
-                 [else (loop (cdr choices))]))]
+         (any (lambda (choice) (test-check choice result fallback))
+              (slot-ref expected 'choices))]
+        [(is-a? expected <test-none-of>)
+         (every (lambda (choice) (not (test-check choice result fallback)))
+                (slot-ref expected 'choices))]
         [else (fallback expected result)]))
 
 (define *test-error* (make <test-error>)) ;DEPRECATED
-
 (define *test-report-error* (sys-getenv "GAUCHE_TEST_REPORT_ERROR"))
-
 (define *test-record-file* (sys-getenv "GAUCHE_TEST_RECORD_FILE"))
 
-(define (test-record-file file) (set! *test-record-file* file)) ;public API
+;; API
+(define (test-record-file file) (set! *test-record-file* file))
 
 ;; List of discrepancies
 (define *discrepancy-list* '())
@@ -208,17 +233,24 @@
            (guard (e [else
                       (when *test-report-error*
                         (report-error e))
-                      (make <test-error>
-                        :class (class-of e)
-                        :message (if (is-a? e <message-condition>)
-                                   (ref e 'message)
-                                   e))])
+                      (make <test-error> :condition e
+                            :class (class-of e)
+                            :message
+                            (if (condition-has-type? e <message-condition>)
+                              (slot-ref e 'message)
+                              e))])
              (thunk)))
          compare))
 
 ;; A convenient macro version
-(define-macro (test* msg expect form . compare)
-  `(test ,msg ,expect (lambda () ,form) ,@compare))
+;; We use er-macro-transformer, so test* should be used after macro
+;; subsystem is tested with more primitive framework.
+(define-syntax test*
+  (er-macro-transformer
+   (lambda (f r c)
+     (apply (lambda (_ msg expect form . compare)
+              `(,(r 'test) ,msg ,expect (,(r 'lambda) () ,form) ,@compare))
+            f))))
 
 ;; Toplevel binding sanity check ----------------------------------
 
@@ -238,74 +270,91 @@
                    (error "test-module requires module or symbol, but got"
                           module)])
     (format #t "testing bindings in ~a ... " mod) (flush)
-    (let ([bad-autoload '()]
-          [bad-export '()]
-          [bad-gref '()]
-          [bad-arity '()]
-          [report '()])
-      ;; 1. Check if there's no dangling autoloads.
-      (hash-table-for-each (module-table mod)
-                           (lambda (sym val)
-                             (guard (_ (else (push! bad-autoload sym)))
-                               (global-variable-ref mod sym))))
-      ;; 2. Check if all exported symbols are properly defined.
-      (when (pair? (module-exports mod))
-        (for-each (lambda (sym)
-                    (guard (_ [else (push! bad-export sym)])
-                      (global-variable-ref mod sym)))
-                  (module-exports mod)))
-      ;; 3. Check if all global references are resolvable, and if it is
-      ;; called, gets valid number of arguments.
-      (for-each
-       (lambda (closure)
-         (for-each (lambda (arg)
-                     (let ([gref (car arg)]
-                           [numargs (cadr arg)])
-                       (cond [(memq (slot-ref gref 'name) allow-undefined)]
-                             [(dangling-gref? gref closure)
-                              => (lambda (bad) (push! bad-gref bad))]
-                             [(memq (slot-ref gref 'name) bypass-arity-check)]
-                             [(arity-invalid? gref numargs closure)
-                              => (lambda (bad) (push! bad-arity bad))])))
-                   (closure-grefs closure)))
-       (toplevel-closures mod))
-      ;; report discrepancies
-      (unless (null? bad-autoload)
-        (push! report (format "found dangling autoloads: ~a" bad-autoload)))
-      (unless (null? bad-export)
-        (unless (null? report) (push! report " AND "))
-        (push! report
-               (format "symbols exported but not defined: ~a" bad-export)))
-      (unless (null? bad-gref)
-        (unless (null? report) (push! report " AND "))
-        (push! report
-               (format "symbols referenced but not defined: ~a"
-                       (string-join (map (lambda (z)
-                                           (format "~a(~a)" (car z) (cdr z)))
-                                         bad-gref)
-                                    ", "))))
-      (unless (null? bad-arity)
-        (unless (null? report) (push! report " AND "))
-        (push! report
-               (format "procedures received wrong number of argument: ~a"
-                       (string-join (map (lambda (z)
-                                           (format "~a(~a) got ~a"
-                                                   (car z) (cadr z) (caddr z)))
-                                         bad-arity)
-                                    ", "))))
-      (cond
-       [(null? report) (test-pass++) (format #t "ok\n")]
-       [else
-        (test-fail++)
-        (let ([s (apply string-append report)])
-          (format #t "ERROR: ~a\n" s)
-          (set! *discrepancy-list*
-                (cons (list (format #f "bindings in module ~a" (module-name mod))
-                            '() s)
-                      *discrepancy-list*)))])
-      )
+    (test-module-common mod module allow-undefined bypass-arity-check)))
+
+;; Check toplevel bindings of a script file.  Script files usually don't
+;; have their own modules.  For this test, we load the script file into
+;; an anonymous module then use test-module on it.  
+(define (test-script file :key (allow-undefined '()) (bypass-arity-check '()))
+  (test-count++)
+  (let1 m (make-module #f)
+    (format #t "testing bindings in script ~a ... " file) (flush)
+    ;; *program-name* and *argv* are defined in #<module user> by gosh,
+    ;; but we'll load the script in a temporary module, so we fake them.
+    (eval `(define *program-name* ',file) m)
+    (eval `(define *argv* '()) m)
+    (load file :environment m)
+    (test-module-common m file allow-undefined bypass-arity-check)))
+
+;; Common op for test-module and test-script.  Returns 
+(define (test-module-common mod name allow-undefined bypass-arity-check)
+  (let ([bad-autoload '()]
+        [bad-export '()]
+        [bad-gref '()]
+        [bad-arity '()]
+        [report '()])
+    ;; 1. Check if there's no dangling autoloads.
+    (hash-table-for-each (module-table mod)
+                         (lambda (sym val)
+                           (guard (_ (else (push! bad-autoload sym)))
+                             (global-variable-ref mod sym))))
+    ;; 2. Check if all exported symbols are properly defined.
+    (when (pair? (module-exports mod))
+      (for-each (lambda (sym)
+                  (guard (_ [else (push! bad-export sym)])
+                    (global-variable-ref mod sym)))
+                (module-exports mod)))
+    ;; 3. Check if all global references are resolvable, and if it is
+    ;; called, gets valid number of arguments.
+    (for-each
+     (lambda (closure)
+       (for-each (lambda (arg)
+                   (let ([gref (car arg)]
+                         [numargs (cadr arg)])
+                     (cond [(memq (slot-ref gref 'name) allow-undefined)]
+                           [(dangling-gref? gref closure)
+                            => (lambda (bad) (push! bad-gref bad))]
+                           [(memq (slot-ref gref 'name) bypass-arity-check)]
+                           [(arity-invalid? gref numargs closure)
+                            => (lambda (bad) (push! bad-arity bad))])))
+                 (closure-grefs closure)))
+     (toplevel-closures mod))
+    ;; report discrepancies
+    (unless (null? bad-autoload)
+      (push! report (format "found dangling autoloads: ~a" bad-autoload)))
+    (unless (null? bad-export)
+      (unless (null? report) (push! report " AND "))
+      (push! report
+             (format "symbols exported but not defined: ~a" bad-export)))
+    (unless (null? bad-gref)
+      (unless (null? report) (push! report " AND "))
+      (push! report
+             (format "symbols referenced but not defined: ~a"
+                     (string-join (map (lambda (z)
+                                         (format "~a(~a)" (car z) (cdr z)))
+                                       bad-gref)
+                                  ", "))))
+    (unless (null? bad-arity)
+      (unless (null? report) (push! report " AND "))
+      (push! report
+             (format "procedures received wrong number of argument: ~a"
+                     (string-join (map (lambda (z)
+                                         (format "~a(~a) got ~a"
+                                                 (car z) (cadr z) (caddr z)))
+                                       bad-arity)
+                                  ", "))))
+    (cond
+     [(null? report) (test-pass++) (format #t "ok\n")]
+     [else
+      (test-fail++)
+      (let ([s (apply string-append report)])
+        (format #t "ERROR: ~a\n" s)
+        (set! *discrepancy-list*
+              (cons (list (format #f "bindings in ~a" name) '() s)
+                    *discrepancy-list*)))])
     ))
 
+
 ;; Auxiliary funcs to catch dangling grefs.  We use the fact that
 ;; an identifier embedded within the vm code is almost always an
 ;; operand of GREF or GSET.  (This is because identifiers are
@@ -355,9 +404,11 @@
 
 (define (arity-invalid? gref numargs closure)
   (and-let* ([ numargs ]
+             ;; TODO: What if GREF is nested identifier?
              [proc (global-variable-ref
-                    (slot-ref gref 'module)
-                    (slot-ref gref 'name))]
+                    (slot-ref gref'module)
+                    (unwrap-syntax gref)
+                    #f)]
              ;; We exclude <generic> with no methods.  Such "placeholder"
              ;; generic function may be used in the base module, expecting
              ;; the other module adds methods to it.
@@ -368,11 +419,9 @@
           numargs)))
 
 (define (dangling-gref? ident closure)
-  (and (not ((with-module gauche.internal find-binding)
-             (slot-ref ident 'module)
-             (slot-ref ident 'name)
-             #f))
-       (cons (slot-ref ident 'name) (slot-ref closure 'info))))
+  (let1 name (unwrap-syntax ident)
+    (and (not ((with-module gauche.internal id->bound-gloc) ident))
+         (cons name (slot-ref closure 'info)))))
 
 ;; Logging and bookkeeping -----------------------------------------
 (define (test-section msg)
