@@ -1,7 +1,7 @@
 ;;;
 ;;; collection.scm - collection generics
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -47,7 +47,7 @@
   )
 (select-module gauche.collection)
 
-;; utility - we can't depend on util.queue, so this is a simple
+;; utility - we can't depend on data.queue, so this is a simple
 ;; alternative.
 (define (make-queue)   (let1 anchor (list #f) (cons anchor anchor)))
 (define (enqueue! q x) (set! (cddr q) (list x)) (set! (cdr q) (cddr q)))
@@ -178,8 +178,9 @@
           (^[] (get-output-string s)))))
 
 (define-method call-with-builder ((class <hash-table-meta>) proc
-                                  :key (type 'eq?) :allow-other-keys)
-  (let1 h (make-hash-table type)
+                                  :key (comparator #f) (type #f)
+                                  :allow-other-keys)
+  (let1 h (make-hash-table (or type comparator 'eq?))
     (proc (^[item]
             (unless (pair? item)
               (error "pair required to build a hashtable, but got" item))
@@ -187,9 +188,12 @@
           (^[] h))))
 
 (define-method call-with-builder ((class <tree-map-meta>) proc
-                                  :key (key=? =) (key<? <)
+                                  :key (comparator default-comparator)
+                                       (key=? #f) (key<? #f)
                                   :allow-other-keys)
-  (let1 tree (make-tree-map key=? key<?)
+  (let1 tree (if (and key=? key<?)
+               (make-tree-map key=? key<?)
+               (make-tree-map comparator))
     (proc (^[item]
             (unless (pair? item)
               (error "pair required to build a tree-map, but got" item))
