@@ -1,7 +1,7 @@
 /*
  * bignum.c - multiple precision exact integer arithmetic
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -186,9 +186,14 @@ ScmObj Scm_MakeBignumFromUIArray(int sign, const u_long *values, int size)
 
 ScmObj Scm_MakeBignumFromDouble(double val)
 {
-    if (val >= LONG_MIN && val <= LONG_MAX) {
+    if (LONG_MIN <= val
+#if SIZEOF_LONG == 4
+        && val <= LONG_MAX
+#else
+        && val <= nextafter((double)LONG_MAX, 0.0)
+#endif
+        )
         return Scm_MakeBignumFromSI((long)val);
-    }
 
     int exponent, sign;
     ScmObj mantissa = Scm_DecodeFlonum(val, &exponent, &sign);
