@@ -1,7 +1,7 @@
 /*
  * char.c - character and character set operations
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -136,7 +136,7 @@ void Scm__InstallCharconvHooks(ScmChar (*u2c)(int), int (*c2u)(ScmChar))
     char2ucs_hook = c2u;
 }
 
-/* NB: These two variables are no longer used, but kept here for the
+/* TRANSIENT: These two variables are no longer used, but kept here for the
    ABI compatibility.  Remove them on 1.0 release. */
 ScmChar (*Scm_UcsToCharHook)(int ucs4) = NULL;
 int (*Scm_CharToUcsHook)(ScmChar ch) = NULL;
@@ -220,7 +220,7 @@ static void charset_print_ch(ScmPort *out, ScmChar ch, int firstp)
 {
     if (ch != 0 && ch < 0x80
         && (strchr("[]-\\", ch) != NULL || (ch == '^' && firstp))) {
-        Scm_Printf(out, "\\%c", ch);
+        Scm_Printf(out, "\\%C", ch);
     } else {
         switch (Scm_CharGeneralCategory(ch)) {
         case SCM_CHAR_CATEGORY_Mn:
@@ -231,8 +231,8 @@ static void charset_print_ch(ScmPort *out, ScmChar ch, int firstp)
         case SCM_CHAR_CATEGORY_Cs:
         case SCM_CHAR_CATEGORY_Co:
         case SCM_CHAR_CATEGORY_Cn:
-            if (ch < 0x10000) Scm_Printf(out, "\\u%04x", ch);
-            else              Scm_Printf(out, "\\U%08x", ch);
+            if (ch < 0x10000) Scm_Printf(out, "\\u%04lx", ch);
+            else              Scm_Printf(out, "\\U%08lx", ch);
             break;
         default:
             Scm_Putc(ch, out);
@@ -276,7 +276,7 @@ static void charset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
             charset_print_ch(out, (int)e->value, FALSE);
         }
     }
-    Scm_Printf(out, "]", obj);
+    Scm_Printf(out, "]");
 }
 
 /*-----------------------------------------------------------------
@@ -608,7 +608,7 @@ ScmObj Scm_CharSetRead(ScmPort *input, int *complement_p,
                 case 'e': ch = 0x1b; goto ordchar;
                 case 'x': case 'u': case 'U':
                     ch = Scm_ReadXdigitsFromString(cp, end-cp, ch,
-                                                   Scm_ReaderLexicalMode(),
+                                                   Scm_GetPortReaderLexicalMode(input),
                                                    TRUE, &cp);
                     if (ch == SCM_CHAR_INVALID) goto err;
                     goto ordchar;
