@@ -1,7 +1,7 @@
 /*
  * code.c - compiled code builder/handler
  *
- *   Copyright (c) 2005-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2005-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -36,6 +36,7 @@
 #include "gauche/class.h"
 #include "gauche/code.h"
 #include "gauche/vminsn.h"
+#include "gauche/priv/codeP.h"
 #include "gauche/priv/builtin-syms.h"
 
 /*===============================================================
@@ -65,14 +66,6 @@ ScmObj Scm_CompiledCodeFullName(ScmCompiledCode *cc)
     }
 }
 
-static void compiled_code_print(ScmObj obj, ScmPort *out, ScmWriteContext *c)
-{
-    Scm_Printf(out, "#<compiled-code %S@%p>",
-               Scm_CompiledCodeFullName(SCM_COMPILED_CODE(obj)), obj);
-}
-
-SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_CompiledCodeClass, compiled_code_print);
-
 static ScmCompiledCode *make_compiled_code(void)
 {
     ScmCompiledCode *cc = SCM_NEW(ScmCompiledCode);
@@ -80,8 +73,8 @@ static ScmCompiledCode *make_compiled_code(void)
     cc->code = NULL;
     cc->constants = NULL;
     cc->maxstack = -1;
-    cc->info = SCM_NIL;
-    cc->argInfo = SCM_FALSE;
+    cc->debugInfo = SCM_NIL;
+    cc->signatureInfo = SCM_FALSE;
     cc->name = SCM_FALSE;
     cc->parent = SCM_FALSE;
     cc->builder = NULL;
@@ -144,11 +137,11 @@ void Scm_CompiledCodeDump(ScmCompiledCode *cc)
     print_header("main_code", SCM_MAKE_STR(""), cc);
     do {
         ScmWord *p = cc->code;
-        Scm_Printf(SCM_CUROUT, "args: %S\n", cc->argInfo);
+        Scm_Printf(SCM_CUROUT, "signatureInfo: %S\n", cc->signatureInfo);
         for (int i=0; i < cc->codeSize; i++) {
             ScmWord insn = p[i];
             ScmPort *out = SCM_PORT(Scm_MakeOutputStringPort(TRUE));
-            ScmObj info = Scm_Assq(SCM_MAKE_INT(i), cc->info);
+            ScmObj info = Scm_Assq(SCM_MAKE_INT(i), cc->debugInfo);
             u_int code = SCM_VM_INSN_CODE(insn);
             const char *insn_name = Scm_VMInsnName(code);
 
@@ -283,13 +276,13 @@ static ScmObj check_lifted_closure(ScmWord *p, ScmObj lifted)
     };
 
     if (!SCM_IDENTIFIERP(p[1])) return lifted;
-    ScmIdentifier *id = SCM_IDENTIFIER(p[1]);
+    ScmIdentifier *id = Scm_OutermostIdentifier(SCM_IDENTIFIER(p[1]));
     if (SCM_SYMBOL_INTERNED(id->name)) return lifted;
 
     for (int i=0; i < sizeof(gref_insns)/sizeof(ScmWord); i++) {
         if (code == gref_insns[i]) {
-            ScmObj g = Scm_GlobalVariableRef(id->module, id->name,
-                                      SCM_BINDING_STAY_IN_MODULE);
+            ScmObj g = Scm_GlobalVariableRef(id->module, SCM_SYMBOL(id->name),
+                                             SCM_BINDING_STAY_IN_MODULE);
             if (SCM_CLOSUREP(g)) {
                 if (SCM_FALSEP(Scm_Assq(SCM_CLOSURE(g)->code, lifted))) {
                     return Scm_Acons(SCM_CLOSURE(g)->code,
@@ -340,7 +333,7 @@ typedef struct cc_builder_rec {
     ScmObj labelDefs;           /* alist of (name . offset) */
     ScmObj labelRefs;           /* alist of (name . offset-to-fill) */
     int labelCount;             /* counter to generate unique labels */
-    ScmObj info;                /* alist of (offset (source-info obj)) */
+    ScmObj debugInfo;           /* alist of (offset (source-info obj)) */
 } cc_builder;
 
 /* Indicates that there's no pending instruction. */
@@ -377,7 +370,7 @@ static cc_builder *make_cc_builder(void)
     b->currentState = -1;
     b->labelDefs = b->labelRefs = SCM_NIL;
     b->labelCount = 0;
-    b->info = SCM_NIL;
+    b->debugInfo = SCM_NIL;
     return b;
 }
 
@@ -404,10 +397,10 @@ static void cc_builder_add_constant(cc_builder *b, ScmObj obj)
 static void cc_builder_add_info(cc_builder *b)
 {
     if (SCM_FALSEP(b->currentInfo)) return;
-    b->info = Scm_Acons(SCM_MAKE_INT(b->currentIndex),
-                        SCM_LIST1(Scm_Cons(SCM_SYM_SOURCE_INFO,
-                                           b->currentInfo)),
-                        b->info);
+    b->debugInfo = Scm_Acons(SCM_MAKE_INT(b->currentIndex),
+                             SCM_LIST1(Scm_Cons(SCM_SYM_SOURCE_INFO,
+                                                b->currentInfo)),
+                             b->debugInfo);
     b->currentInfo = SCM_FALSE;
 }
 
@@ -563,6 +556,19 @@ void Scm_CompiledCodeSetLabel(ScmCompiledCode *cc, ScmObj label)
                              b->labelDefs);
 }
 
+/* Push arbitrary debug-info into the builder.
+   <debug-info> : (<insn-offset> <item> ...)
+   <insn-offset> : <integer> or 'definition
+   Currently supported item:
+   <item> : (<source-info> . source)
+*/
+void Scm_CompiledCodePushInfo(ScmCompiledCode *cc, ScmObj info)
+{
+    cc_builder *b;
+    CC_BUILDER_GET(b, cc);
+    b->debugInfo = Scm_Cons(info, b->debugInfo);
+}
+
 /* Pack the code accumulated in the builder into a code vector.
    Perform label resolution and jump optimization. */
 void Scm_CompiledCodeFinishBuilder(ScmCompiledCode *cc, int maxstack)
@@ -621,7 +627,7 @@ void Scm_CompiledCodeFinishBuilder(ScmCompiledCode *cc, int maxstack)
     cc_builder_jumpopt(cc);
 
     /* record debug info */
-    cc->info = b->info;
+    cc->debugInfo = b->debugInfo;
 
     /* set max stack depth */
     cc->maxstack = maxstack;
@@ -819,7 +825,7 @@ void Scm_CompiledCodeEmit(ScmCompiledCode *cc,
 }
 
 /*----------------------------------------------------------------
- * CompiledCode - Scheme interface
+ * CompiledCode - introspection
  */
 
 /* Converts the code vector into a list.
@@ -873,65 +879,6 @@ ScmObj Scm_CompiledCodeToList(ScmCompiledCode *cc)
     return h;
 }
 
-static ScmObj code_size_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->codeSize);
-}
-
-static ScmObj code_maxstack_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->maxstack);
-}
-
-static ScmObj code_info_get(ScmObj cc)
-{
-    return SCM_COMPILED_CODE(cc)->info;
-}
-
-static ScmObj code_arginfo_get(ScmObj cc)
-{
-    return SCM_COMPILED_CODE(cc)->argInfo;
-}
-
-static ScmObj code_reqargs_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->requiredArgs);
-}
-
-static ScmObj code_optargs_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->optionalArgs);
-}
-
-static ScmObj code_name_get(ScmObj cc)
-{
-    return SCM_COMPILED_CODE(cc)->name;
-}
-
-static ScmObj code_parent_get(ScmObj cc)
-{
-    return SCM_OBJ(SCM_COMPILED_CODE(cc)->parent);
-}
-
-static ScmObj code_iform_get(ScmObj cc)
-{
-    return SCM_OBJ(SCM_COMPILED_CODE(cc)->intermediateForm);
-}
-
-static ScmClassStaticSlotSpec code_slots[] = {
-    SCM_CLASS_SLOT_SPEC("parent", code_parent_get, NULL),
-    SCM_CLASS_SLOT_SPEC("arg-info", code_arginfo_get, NULL),
-    SCM_CLASS_SLOT_SPEC("info", code_info_get, NULL),
-    SCM_CLASS_SLOT_SPEC("required-args", code_reqargs_get, NULL),
-    SCM_CLASS_SLOT_SPEC("optional-args", code_optargs_get, NULL),
-    SCM_CLASS_SLOT_SPEC("name", code_name_get, NULL),
-    SCM_CLASS_SLOT_SPEC("full-name", Scm_CompiledCodeFullName, NULL),
-    SCM_CLASS_SLOT_SPEC("size", code_size_get, NULL),
-    SCM_CLASS_SLOT_SPEC("max-stack", code_maxstack_get, NULL),
-    SCM_CLASS_SLOT_SPEC("intermediate-form", code_iform_get, NULL),
-    SCM_CLASS_SLOT_SPEC_END()
-};
-
 /*===========================================================
  * VM Instruction introspection
  */
@@ -1037,11 +984,3 @@ ScmWord Scm_VMInsnBuild(ScmObj obj)
     return 0;       /* dummy */
 }
 
-/*===========================================================
- * Initialization
- */
-void Scm__InitCode(void)
-{
-    Scm_InitStaticClass(SCM_CLASS_COMPILED_CODE, "<compiled-code>",
-                        Scm_GaucheModule(), code_slots, 0);
-}
