@@ -1,7 +1,7 @@
 /*
  * compaux.c - C API bridge for the compiler
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -145,7 +145,8 @@ static void identifier_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
        module-based sandbox implementation, so further consideration is
        required.
     */
-    Scm_Printf(port, "#<identifier %S#%S>", id->module->name, id->name);
+    Scm_Printf(port, "#<identifier %S#%S.%x>",
+               id->module->name, id->name, SCM_WORD(id));
 }
 
 SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_IdentifierClass, identifier_print);
@@ -156,13 +157,16 @@ static ScmObj get_binding_frame(ScmObj var, ScmObj env)
     SCM_FOR_EACH(frame, env) {
         if (!SCM_PAIRP(SCM_CAR(frame))) continue;
         SCM_FOR_EACH(fp, SCM_CDAR(frame)) {
-            if (SCM_CAAR(fp) == var) return frame;
+            if (SCM_CAAR(fp) == var) {
+                SCM_ASSERT(SCM_LISTP(frame));
+                return frame;
+            }
         }
     }
     return SCM_NIL;
 }
 
-ScmObj Scm_MakeIdentifier(ScmSymbol *name, ScmModule *mod, ScmObj env)
+ScmObj Scm_MakeIdentifier(ScmObj name, ScmModule *mod, ScmObj env)
 {
     ScmIdentifier *id = SCM_NEW(ScmIdentifier);
     SCM_SET_CLASS(id, SCM_CLASS_IDENTIFIER);
@@ -172,6 +176,28 @@ ScmObj Scm_MakeIdentifier(ScmSymbol *name, ScmModule *mod, ScmObj env)
     return SCM_OBJ(id);
 }
 
+ScmIdentifier *Scm_OutermostIdentifier(ScmIdentifier *id)
+{
+    while (SCM_IDENTIFIERP(id->name)) {
+        id = SCM_IDENTIFIER(SCM_IDENTIFIER(id)->name);
+    }
+    return id;
+}
+
+ScmSymbol *Scm_UnwrapIdentifier(ScmIdentifier *id)
+{
+    ScmObj z = Scm_OutermostIdentifier(id)->name;
+    SCM_ASSERT(SCM_SYMBOLP(z));
+    return SCM_SYMBOL(z);
+}
+
+/* returns global binding of the identifier */
+ScmGloc *Scm_IdentifierGlobalBinding(ScmIdentifier *id)
+{
+    ScmIdentifier *z = Scm_OutermostIdentifier(id);
+    return Scm_FindBinding(z->module, SCM_SYMBOL(z->name), 0);
+}
+
 /* returns true if SYM has the same binding with ID in ENV. */
 int Scm_IdentifierBindingEqv(ScmIdentifier *id, ScmSymbol *sym, ScmObj env)
 {
@@ -179,11 +205,11 @@ int Scm_IdentifierBindingEqv(ScmIdentifier *id, ScmSymbol *sym, ScmObj env)
     return (bf == id->env);
 }
 
-ScmObj Scm_CopyIdentifier(ScmIdentifier *orig)
+ScmObj Scm_WrapIdentifier(ScmIdentifier *orig)
 {
     ScmIdentifier *id = SCM_NEW(ScmIdentifier);
     SCM_SET_CLASS(id, SCM_CLASS_IDENTIFIER);
-    id->name = orig->name;
+    id->name = SCM_OBJ(orig);
     id->module = orig->module;
     id->env = orig->env;
     return SCM_OBJ(id);
@@ -196,10 +222,10 @@ static ScmObj identifier_name_get(ScmObj obj)
 
 static void   identifier_name_set(ScmObj obj, ScmObj val)
 {
-    if (!SCM_SYMBOLP(val)) {
-        Scm_Error("symbol required, but got %S", val);
+    if (!SCM_SYMBOLP(val) && !SCM_IDENTIFIERP(val)) {
+        Scm_Error("symbol or identifier required, but got %S", val);
     }
-    SCM_IDENTIFIER(obj)->name = SCM_SYMBOL(val);
+    SCM_IDENTIFIER(obj)->name = val;
 }
 
 static ScmObj identifier_module_get(ScmObj obj)
@@ -242,7 +268,7 @@ static ScmClassStaticSlotSpec identifier_slots[] = {
 /* Convert all identifiers in form into a symbol.
    This keeps linear history to avoid entering infinite loop if
    the given form is circular; but it doens't recover the shared
-   substricture. */
+   substructure. */
 static ScmObj unwrap_rec(ScmObj form, ScmObj history)
 {
     if (!SCM_PTRP(form)) return form;
@@ -262,7 +288,7 @@ static ScmObj unwrap_rec(ScmObj form, ScmObj history)
         }
     }
     if (SCM_IDENTIFIERP(form)) {
-        return SCM_OBJ(SCM_IDENTIFIER(form)->name);
+        return SCM_OBJ(Scm_UnwrapIdentifier(SCM_IDENTIFIER(form)));
     }
     if (SCM_VECTORP(form)) {
         int len = SCM_VECTOR_SIZE(form);
