@@ -1,7 +1,7 @@
 ;;;
 ;;; comp-macros.scm - Utility macros for the compiler
 ;;;
-;;;   Copyright (c) 2004-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -88,7 +88,7 @@
 ;;     (define *foo-dispatch-table* (generate-dispatch-table foo))
 (define-macro (define/case proc dispatch-expr . clauses)
   (define (generate-handler key clause)
-    (let1 name (string->symbol #`",(car proc)/,key")
+    (let1 name (string->symbol #"~(car proc)/~key")
       `(define (,name ,@(cdr proc)) ,@clause)))
   (define (find-clause key)
     (if-let1 c (find (^c (memq key (car c))) clauses)
@@ -97,7 +97,7 @@
         (cdr c)
         (errorf "No dispatch clause for ~a found during expanding ~a"
                 key proc))))
-  (define dispatch-table (string->symbol #`"*,(car proc)-dispatch-table*"))
+  (define dispatch-table (string->symbol #"*~(car proc)-dispatch-table*"))
 
   `(begin
      (define-inline ,proc
@@ -105,7 +105,7 @@
      ,@(map (^t (generate-handler (car t) (find-clause (car t))))
             .intermediate-tags.)
      (define ,dispatch-table
-       (vector ,@(map (^t (string->symbol #`",(car proc)/,(car t)"))
+       (vector ,@(map (^t (string->symbol #"~(car proc)/~(car t)"))
                       .intermediate-tags.)))))
 
 ;; Inlining map.  Combined with closure optimization, we can avoid
@@ -159,7 +159,7 @@
 
 ;; Generate dispatch table
 (define-macro (generate-dispatch-table prefix)
-  `(vector ,@(map (lambda (p) (string->symbol #`",|prefix|/,(car p)"))
+  `(vector ,@(map (lambda (p) (string->symbol #"~|prefix|/~(car p)"))
                   .intermediate-tags.)))
 
 
@@ -222,14 +222,15 @@
       ))
   `(begin
      ,@(if constructor
-         `(,(make-constructor))
+         `((declare (keep-private-macro ,constructor))
+           ,(make-constructor))
          '())
      ,@(let loop ((s slot-defs) (i (if tag 1 0)) (r '()))
          (if (null? s)
            (reverse r)
            (let* ([slot-name (if (pair? (car s)) (caar s) (car s))]
-                  [acc (string->symbol #`",|name|-,|slot-name|")]
-                  [mod (string->symbol #`",|name|-,|slot-name|-set!")])
+                  [acc (string->symbol #"~|name|-~|slot-name|")]
+                  [mod (string->symbol #"~|name|-~|slot-name|-set!")])
              (loop (cdr s)
                    (+ i 1)
                    (list*
@@ -237,6 +238,7 @@
                        `(vector-ref ,obj ,,i))
                     `(define-macro (,mod obj val)
                        `(vector-set! ,obj ,,i ,val))
+                    `(declare (keep-private-macro ,acc ,mod))
                     r))))))
   )
 
