@@ -1,7 +1,7 @@
 /*
  * core.c - core kernel interface
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -64,7 +64,7 @@ static struct {
  */
 
 extern void Scm__InitModule(void);
-extern void Scm__InitModulePost(void);
+extern void Scm__InitHash(void);
 extern void Scm__InitSymbol(void);
 extern void Scm__InitNumber(void);
 extern void Scm__InitChar(void);
@@ -81,7 +81,6 @@ extern void Scm__InitRegexp(void);
 extern void Scm__InitRead(void);
 extern void Scm__InitSignal(void);
 extern void Scm__InitSystem(void);
-extern void Scm__InitCode(void);
 extern void Scm__InitVM(void);
 extern void Scm__InitAutoloads(void);
 extern void Scm__InitCollection(void);
@@ -105,6 +104,7 @@ extern void Scm_Init_libnum(void);
 extern void Scm_Init_libobj(void);
 extern void Scm_Init_libproc(void);
 extern void Scm_Init_librx(void);
+extern void Scm_Init_libsrfis(void);
 extern void Scm_Init_libstr(void);
 extern void Scm_Init_libsym(void);
 extern void Scm_Init_libsys(void);
@@ -120,11 +120,16 @@ static void init_cond_features(void);
 static int (*ptr_pthread_create)(void) = NULL;
 #endif
 
+/* flag to see if Scheme infrastructure is fully initialized or not */
+static int scheme_initialized = FALSE;
+
 /*
  * Entry point of initlalizing Gauche runtime
  */
 void Scm_Init(const char *signature)
 {
+    if (scheme_initialized) return;
+    
     /* make sure the main program links the same version of libgauche */
     if (strcmp(signature, GAUCHE_SIGNATURE) != 0) {
         Scm_Panic("libgauche ABI version mismatch: libgauche %s, expected %s",
@@ -148,18 +153,17 @@ void Scm_Init(const char *signature)
        rely on the other components to be initialized. */
     Scm__InitParameter();
     Scm__InitVM();
+    Scm__InitHash();
     Scm__InitSymbol();
     Scm__InitModule();
     Scm__InitNumber();
     Scm__InitChar();
     Scm__InitClass();
-    Scm__InitModulePost();
     Scm__InitCollection();
     Scm__InitExceptions();
     Scm__InitProc();
     Scm__InitPort();
     Scm__InitWrite();
-    Scm__InitCode();
     Scm__InitMacro();
     Scm__InitLoad();
     Scm__InitRegexp();
@@ -186,6 +190,7 @@ void Scm_Init(const char *signature)
     Scm_Init_libobj();
     Scm_Init_libproc();
     Scm_Init_librx();
+    Scm_Init_libsrfis();
     Scm_Init_libstr();
     Scm_Init_libsym();
     Scm_Init_libsys();
@@ -207,6 +212,13 @@ void Scm_Init(const char *signature)
     /* a trick to make sure the gc thread object is linked */
     ptr_pthread_create = (int (*)(void))GC_pthread_create;
 #endif
+
+    scheme_initialized = TRUE;
+}
+
+int Scm_InitializedP()
+{
+    return scheme_initialized;
 }
 
 /*=============================================================
@@ -508,52 +520,6 @@ init_cond_features()
         { "gauche.net.tls", "rfc.tls" },
         { "gauche.net.tls.openssl", "rfc.tls" },
 #endif
-
-        /* SRFIs */
-        { "srfi-0", NULL },         /* autoloaded */
-        { "srfi-1", "srfi-1" },
-        { "srfi-2", NULL },         /* builtin */
-        { "srfi-4", "gauche.uvector" },
-        { "srfi-5", "srfi-5" },
-        { "srfi-6", NULL },         /* builtin */
-        { "srfi-7", NULL },         /* autoloaded */
-        { "srfi-8", NULL },         /* builtin */
-        { "srfi-9", "gauche.record" },
-        { "srfi-10", NULL },
-        { "srfi-11", "srfi-11" },
-        { "srfi-13", "srfi-13" },
-        { "srfi-14", "srfi-14" },
-        { "srfi-16", NULL },
-        { "srfi-17", NULL },
-        { "srfi-18", "gauche.threads" },
-        { "srfi-19", "srfi-19" },
-        { "srfi-22", NULL },
-        { "srfi-23", NULL },
-        { "srfi-25", "gauche.array" },
-        { "srfi-26", NULL },
-        { "srfi-27", "srfi-27" },
-        { "srfi-28", NULL },
-        { "srfi-29", "srfi-29" },
-        { "srfi-30", NULL },
-        { "srfi-31", NULL },
-        { "srfi-34", NULL },
-        { "srfi-35", NULL },
-        { "srfi-36", NULL },
-        { "srfi-37", "srfi-37" },
-        { "srfi-38", NULL },
-        { "srfi-39", "gauche.parameter" },
-        { "srfi-40", "util.stream" },
-        { "srfi-42", "srfi-42" },
-        { "srfi-43", "srfi-43" },
-        { "srfi-45", NULL },
-        { "srfi-55", NULL },
-        { "srfi-61", NULL },
-        { "srfi-62", NULL },
-        { "srfi-87", NULL },
-        { "srfi-95", NULL },
-        { "srfi-98", "srfi-98" },
-        { "srfi-99", "gauche.record" },
-        { "srfi-106", "srfi-106" },
         { NULL, NULL }
     };
 
