@@ -1,7 +1,7 @@
 /*
  * error.c - error handling
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -42,6 +42,7 @@
 #include <string.h>
 #include <ctype.h>
 
+static ScmObj condition_allocate(ScmClass *klass, ScmObj initargs);
 static void   message_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx);
 static ScmObj message_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs);
@@ -75,14 +76,19 @@ static ScmClass *condition_cpl[] = {
 };
 
 SCM_DEFINE_BASE_CLASS(Scm_ConditionClass, ScmInstance,
-                      NULL, NULL, NULL,
-                      Scm_ObjectAllocate, SCM_CLASS_DEFAULT_CPL);
+                      NULL, NULL, NULL, 
+                      condition_allocate, SCM_CLASS_DEFAULT_CPL);
 SCM_DEFINE_BASE_CLASS(Scm_MessageConditionClass, ScmMessageCondition,
                       message_print, NULL, NULL,
                       message_allocate, condition_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_SeriousConditionClass, ScmSeriousCondition,
                       NULL, NULL, NULL,
-                      Scm_ObjectAllocate, condition_cpl);
+                      condition_allocate, condition_cpl);
+
+static ScmObj condition_allocate(ScmClass *klass, ScmObj initargs)
+{
+    return SCM_OBJ(SCM_NEW_INSTANCE(ScmCondition, klass));
+}
 
 static void message_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 {
@@ -94,8 +100,7 @@ static void message_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 
 static ScmObj message_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmError *e = SCM_ALLOCATE(ScmError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmError *e = SCM_NEW_INSTANCE(ScmError, klass);
     e->message = SCM_FALSE;     /* would be set by initialize */
     return SCM_OBJ(e);
 }
@@ -213,8 +218,7 @@ SCM_DEFINE_BASE_CLASS(Scm_IOUnitErrorClass, ScmIOUnitError,
 
 static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmSystemError *e = SCM_ALLOCATE(ScmSystemError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmSystemError *e = SCM_NEW_INSTANCE(ScmSystemError, klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->error_number = 0;           /* set by initialize */
     return SCM_OBJ(e);
@@ -222,8 +226,8 @@ static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs)
 
 static ScmObj sigerror_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmUnhandledSignalError *e = SCM_ALLOCATE(ScmUnhandledSignalError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmUnhandledSignalError *e = SCM_NEW_INSTANCE(ScmUnhandledSignalError,
+                                                  klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->signal = 0;                 /* set by initialize */
     return SCM_OBJ(e);
@@ -231,8 +235,7 @@ static ScmObj sigerror_allocate(ScmClass *klass, ScmObj initargs)
 
 static ScmObj readerror_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmReadError *e = SCM_ALLOCATE(ScmReadError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmReadError *e = SCM_NEW_INSTANCE(ScmReadError, klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->port = NULL;                /* set by initialize */
     e->line = -1;                  /* set by initialize */
@@ -241,8 +244,7 @@ static ScmObj readerror_allocate(ScmClass *klass, ScmObj initargs)
 
 static ScmObj porterror_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmPortError *e = SCM_ALLOCATE(ScmPortError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmPortError *e = SCM_NEW_INSTANCE(ScmPortError, klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->port = NULL;                /* set by initialize */
     return SCM_OBJ(e);
@@ -364,17 +366,28 @@ static ScmClass *compound_cpl[] = {
     NULL
 };
 
+static void compound_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+{
+    ScmClass *k = Scm_ClassOf(obj);
+    Scm_Printf(port, "#<%A", Scm__InternalClassName(k));
+    ScmCompoundCondition *c = SCM_COMPOUND_CONDITION(obj);
+    ScmObj cp;
+    SCM_FOR_EACH(cp, c->conditions) {
+        Scm_Printf(port, " %A", SCM_CAR(cp));
+    }
+    Scm_Printf(port, ">");
+}
+
 SCM_DEFINE_BASE_CLASS(Scm_CompoundConditionClass, ScmCompoundCondition,
-                      NULL, NULL, NULL,
+                      compound_print, NULL, NULL,
                       compound_allocate, compound_cpl+2);
 SCM_DEFINE_BASE_CLASS(Scm_SeriousCompoundConditionClass, ScmCompoundCondition,
-                      NULL, NULL, NULL,
+                      compound_print, NULL, NULL,
                       compound_allocate, compound_cpl);
 
 static ScmObj compound_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmCompoundCondition *e = SCM_ALLOCATE(ScmCompoundCondition, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmCompoundCondition *e = SCM_NEW_INSTANCE(ScmCompoundCondition, klass);
     e->conditions = SCM_NIL;
     return SCM_OBJ(e);
 }
@@ -426,6 +439,26 @@ ScmObj Scm_MakeCompoundCondition(ScmObj conditions)
     return cond;
 }
 
+/* Extract condition of type TYPE from a compound condition CONDITION.
+   The returned condition may be of subtype of TYPE.
+   CONDITION may be simple condition; in that case, CONDITION itself
+   is returned iff condition is of a subtype of TYPE.
+   This isn't the same as `extract-condition' of SRFI-35, which always
+   returns a condition of the specified type; it also has different
+   rule to refer to the slots.
+ */
+ScmObj Scm_ExtractSimpleCondition(ScmObj condition, ScmClass *type)
+{
+    ScmObj cs = (SCM_COMPOUND_CONDITION_P(condition)
+                 ? SCM_COMPOUND_CONDITION(condition)->conditions
+                 : SCM_LIST1(condition));
+    ScmObj cp;
+    SCM_FOR_EACH(cp, cs) {
+        if (SCM_ISA(SCM_CAR(cp), type)) return SCM_CAR(cp);
+    }
+    return SCM_FALSE;
+}
+
 static ScmObj conditions_get(ScmCompoundCondition *obj)
 {
     return obj->conditions;
@@ -522,7 +555,7 @@ ScmObj Scm_ConditionMessage(ScmObj c)
    suitable for the error message.  Because of personal preference
    and backward compatibility, I upcase the class name of the condition
    sans brackets.  If it is a composite condition, the component's typenames
-   are joind with commas.
+   are joind with commas, excluding mixin conditions.
 */
 ScmObj Scm_ConditionTypeName(ScmObj c)
 {
@@ -538,6 +571,7 @@ ScmObj Scm_ConditionTypeName(ScmObj c)
         ScmObj h = SCM_NIL, t = SCM_NIL, cp;
         SCM_FOR_EACH(cp, SCM_COMPOUND_CONDITION(c)->conditions) {
             ScmObj cc = SCM_CAR(cp);
+            if (SCM_MIXIN_CONDITION_P(cc)) continue;
             SCM_APPEND1(h, t, Scm__InternalClassName(Scm_ClassOf(cc)));
         }
         if (SCM_NULLP(h)) {
@@ -547,7 +581,15 @@ ScmObj Scm_ConditionTypeName(ScmObj c)
             sname = Scm_StringJoin(h, &cond_name_delim, SCM_STRING_JOIN_INFIX);
         }
     }
-    return sname;
+
+    ScmDString ds;
+    Scm_DStringInit(&ds);
+    ScmObj p = Scm_MakeInputStringPort(SCM_STRING(sname), TRUE);
+    int ch;
+    while ((ch = Scm_Getc(SCM_PORT(p))) != EOF) {
+        Scm_DStringPutc(&ds, Scm_CharUpcase(ch));
+    }
+    return Scm_DStringGet(&ds, 0);
 }
 
 /*================================================================
@@ -568,33 +610,60 @@ ScmObj Scm_ConditionTypeName(ScmObj c)
  *   seem to use over 2^31 for the time being).
  */
 
+/* Double fault check
+ * In order to aviod infinite loop when error throwing routine
+ * throws an error, we use vm flag SCM_ERROR_BEING_HANDLED to
+ * check that.  Ideally a common single API should handle it,   
+ * but for the time being, we add the check at the beginning
+ * of Scm_*Error APIs.
+ * The SCM_ERROR_BEING_HANDLED flag is cleared in Scm_VMThrowException2().
+ */
+#define SCM_ERROR_DOUBLE_FAULT_CHECK(vm)                                \
+    do {                                                                \
+        if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_ERROR_BEING_HANDLED)) {  \
+            ScmObj e =                                                  \
+                Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler")); \
+            Scm_VMThrowException2(vm, e, SCM_RAISE_NON_CONTINUABLE);    \
+        }                                                               \
+        SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_HANDLED);           \
+    } while (0)
+
+/* Common part to format error message passed as varargs.  
+ * ostr must be declared as ScmObj, and msg must be the last
+ * arg of function before '...'.
+ */
+#define SCM_ERROR_MESSAGE_FORMAT(ostr, msg)             \
+    do {                                                \
+        ostr = Scm_MakeOutputStringPort(TRUE);          \
+        va_list args__;                                 \
+        va_start(args__, msg);                          \
+        Scm_Vprintf(SCM_PORT(ostr), msg, args__, TRUE); \
+        va_end(args__);                                 \
+    } while (0)
+
+/* Like SCM_ERROR_MESSAGE_FORMAT, but we also append system error message.
+ * Need to pass errno, which should be fetched by get_errno() *before*
+ * calling Scm_VM().
+ */
+#define SCM_SYSERROR_MESSAGE_FORMAT(ostr, msg, en)      \
+    do {                                                \
+        SCM_ERROR_MESSAGE_FORMAT(ostr, msg);            \
+        ScmObj syserr = get_syserrmsg(en);              \
+        SCM_PUTZ(": ", -1, ostr);                       \
+        SCM_PUTS(syserr, ostr);                         \
+    } while (0)
+
 /*
  * C-like interface
  */
-
 void Scm_Error(const char *msg, ...)
 {
-    ScmObj e;
     ScmVM *vm = Scm_VM();
-    va_list args;
-
-    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_ERROR_BEING_HANDLED)) {
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
-        Scm_VMThrowException(vm, e);
-    }
-    SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_HANDLED);
-
-    SCM_UNWIND_PROTECT {
-        va_start(args, msg);
-        e = Scm_MakeError(Scm_Vsprintf(msg, args, TRUE));
-        va_end(args);
-    }
-    SCM_WHEN_ERROR {
-        /* TODO: should check continuation? */
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
-    }
-    SCM_END_PROTECT;
-    Scm_VMThrowException(vm, e);
+    SCM_ERROR_DOUBLE_FAULT_CHECK(vm);
+    ScmObj ostr;
+    SCM_ERROR_MESSAGE_FORMAT(ostr, msg);
+    ScmObj e = Scm_MakeError(Scm_GetOutputString(SCM_PORT(ostr), 0));
+    Scm_VMThrowException2(vm, e, SCM_RAISE_NON_CONTINUABLE);
     Scm_Panic("Scm_Error: Scm_VMThrowException returned.  something wrong.");
 }
 
@@ -609,17 +678,23 @@ static ScmObj get_syserrmsg(int en)
     syserr = SCM_MAKE_STR_COPYING(strerror(en));
 #else  /*GAUCHE_WINDOWS*/
     if (en < 0) {
-      LPTSTR msgbuf;
+      ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
+      LPTSTR msgbuf = NULL;
       const char *xmsgbuf;
-      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
-                    NULL,
-                    -en,
-                    0,
-                    (LPTSTR)&msgbuf,
-                    0, NULL);
-      xmsgbuf = SCM_WCS2MBS(msgbuf);
-      syserr = SCM_MAKE_STR_COPYING(xmsgbuf);
+      if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
+                        FORMAT_MESSAGE_FROM_SYSTEM |
+                        FORMAT_MESSAGE_IGNORE_INSERTS,
+                        NULL,
+                        -en,
+                        MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT),
+                        (LPTSTR)&msgbuf,
+                        0, NULL)) {
+        xmsgbuf = SCM_WCS2MBS(msgbuf);
+        SCM_PUTZ(xmsgbuf, -1, ostr);
+      }
       LocalFree(msgbuf);
+      Scm_Printf(SCM_PORT(ostr), "(error code = %d)", -en);
+      syserr = Scm_GetOutputString(SCM_PORT(ostr), 0);
     } else {
       syserr = SCM_MAKE_STR_COPYING(strerror(en));
     }
@@ -632,46 +707,35 @@ static int get_errno(void)
 #if !defined(GAUCHE_WINDOWS)
     return errno;
 #else  /*GAUCHE_WINDOWS*/
+    int en;
+    
     if (errno == 0) {
-      return -(int)GetLastError();
+        en = -(int)GetLastError();
     } else {
-      return errno;             /* NB: MSDN says we should use _get_errno,
+        en = errno;             /* NB: MSDN says we should use _get_errno,
                                    but MinGW doesn't seem to have it yet.*/
     }
-#endif /*GAUCHE_WINDOWS*/
-}
 
-void Scm_SysError(const char *msg, ...)
-{
-    ScmObj e;
-    va_list args;
-    ScmVM *vm = Scm_VM();
-    int en = get_errno();
-    ScmObj syserr = get_syserrmsg(en);
-
-#if defined(GAUCHE_WINDOWS)
     /* Reset the error code, so that we can find which is the actual
        error code in the next occasion. */
     errno = 0;                  /* NB: MSDN says we should use _set_errno,
                                    but MinGW doesn't seem to have it yet. */
     SetLastError(0);
-#endif /*GAUCHE_WINDOWS*/
 
-    SCM_UNWIND_PROTECT {
-        ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
-        va_start(args, msg);
-        Scm_Vprintf(SCM_PORT(ostr), msg, args, TRUE);
-        va_end(args);
-        SCM_PUTZ(": ", -1, ostr);
-        SCM_PUTS(syserr, ostr);
-        e = Scm_MakeSystemError(Scm_GetOutputString(SCM_PORT(ostr), 0), en);
-    }
-    SCM_WHEN_ERROR {
-        /* TODO: should check continuation */
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
-    }
-    SCM_END_PROTECT;
-    Scm_VMThrowException(vm, e);
+    return en;
+#endif /*GAUCHE_WINDOWS*/
+}
+
+void Scm_SysError(const char *msg, ...)
+{
+    int en = get_errno();       /* must take this before Scm_VM() */
+    ScmVM *vm = Scm_VM();
+    SCM_ERROR_DOUBLE_FAULT_CHECK(vm);
+
+    ScmObj ostr;
+    SCM_SYSERROR_MESSAGE_FORMAT(ostr, msg, en);
+    ScmObj e = Scm_MakeSystemError(Scm_GetOutputString(SCM_PORT(ostr), 0), en);
+    Scm_VMThrowException2(vm, e, SCM_RAISE_NON_CONTINUABLE);
     Scm_Panic("Scm_Error: Scm_VMThrowException returned.  something wrong.");
 }
 
@@ -687,7 +751,7 @@ void Scm_TypeError(const char *what, const char *expected, ScmObj got)
 
 
 /*
- * A convenience function to raise port-relates errors.
+ * A convenience function to raise port-related errors.
  * It creates either one of <port-error>, <io-read-error>,
  * <io-write-error>, <io-closed-error>, or <io-unit-error>,
  * depending on the 'reason' argument being
@@ -697,53 +761,39 @@ void Scm_TypeError(const char *what, const char *expected, ScmObj got)
  */
 void Scm_PortError(ScmPort *port, int reason, const char *msg, ...)
 {
-    ScmObj e;
+    int en = get_errno();       /* must take this before Scm_VM() */
     ScmVM *vm = Scm_VM();
-    va_list args;
-    int en = get_errno();
+    SCM_ERROR_DOUBLE_FAULT_CHECK(vm);
 
-    SCM_UNWIND_PROTECT {
-        ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
-        va_start(args, msg);
-        Scm_Vprintf(SCM_PORT(ostr), msg, args, TRUE);
-        va_end(args);
-        if (en != 0) {
-            ScmObj syserr = get_syserrmsg(en);
-            SCM_PUTZ(": ", -1, ostr);
-            SCM_PUTS(syserr, ostr);
-        }
-        ScmObj smsg = Scm_GetOutputString(SCM_PORT(ostr), 0);
-        ScmClass *peclass;
+    ScmObj ostr;
+    if (en != 0) SCM_SYSERROR_MESSAGE_FORMAT(ostr, msg, en);
+    else         SCM_ERROR_MESSAGE_FORMAT(ostr, msg);
+    ScmObj smsg = Scm_GetOutputString(SCM_PORT(ostr), 0);
 
-        switch (reason) {
-        case SCM_PORT_ERROR_INPUT:
-            peclass = SCM_CLASS_IO_READ_ERROR; break;
-        case SCM_PORT_ERROR_OUTPUT:
-            peclass = SCM_CLASS_IO_WRITE_ERROR; break;
-        case SCM_PORT_ERROR_CLOSED:
-            peclass = SCM_CLASS_IO_CLOSED_ERROR; break;
-        case SCM_PORT_ERROR_UNIT:
-            peclass = SCM_CLASS_IO_UNIT_ERROR; break;
-        default:
-            peclass = SCM_CLASS_PORT_ERROR; break;
-        }
-        ScmObj pe = porterror_allocate(peclass, SCM_NIL);
-        SCM_ERROR(pe)->message = SCM_LIST2(smsg, smsg);
-        SCM_PORT_ERROR(pe)->port = port;
-
-        if (en != 0) {
-            e = Scm_MakeCompoundCondition(SCM_LIST2(Scm_MakeSystemError(smsg, en),
-                                                    pe));
-        } else {
-            e = pe;
-        }
+    ScmClass *peclass;
+    switch (reason) {
+    case SCM_PORT_ERROR_INPUT:
+        peclass = SCM_CLASS_IO_READ_ERROR; break;
+    case SCM_PORT_ERROR_OUTPUT:
+        peclass = SCM_CLASS_IO_WRITE_ERROR; break;
+    case SCM_PORT_ERROR_CLOSED:
+        peclass = SCM_CLASS_IO_CLOSED_ERROR; break;
+    case SCM_PORT_ERROR_UNIT:
+        peclass = SCM_CLASS_IO_UNIT_ERROR; break;
+    default:
+        peclass = SCM_CLASS_PORT_ERROR; break;
     }
-    SCM_WHEN_ERROR {
-        /* TODO: should check continuation */
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
+
+    ScmObj pe = porterror_allocate(peclass, SCM_NIL);
+    SCM_ERROR(pe)->message = SCM_LIST2(smsg, smsg);
+    SCM_PORT_ERROR(pe)->port = port;
+
+    ScmObj e = pe;
+    if (en != 0) {
+        e = Scm_MakeCompoundCondition(SCM_LIST2(Scm_MakeSystemError(smsg, en),
+                                                pe));
     }
-    SCM_END_PROTECT;
-    Scm_VMThrowException(vm, e);
+    Scm_VMThrowException2(vm, e, SCM_RAISE_NON_CONTINUABLE);
     Scm_Panic("Scm_Error: Scm_VMThrowException returned.  something wrong.");
 }
 
@@ -754,6 +804,7 @@ void Scm_PortError(ScmPort *port, int reason, const char *msg, ...)
 
 void Scm_Warn(const char *msg, ...)
 {
+    if (Scm_GetEnv("GAUCHE_SUPPRESS_WARNING") != NULL) return;
     va_list args;
     va_start(args, msg);
     Scm_Printf(SCM_CURERR, "WARNING: %A\n", Scm_Vsprintf(msg, args, TRUE));
@@ -772,11 +823,24 @@ void Scm_FWarn(ScmString *fmt, ScmObj args)
  */
 
 /* An external API to hide Scm_VMThrowException. */
+#if  GAUCHE_API_0_95
+ScmObj Scm_Raise(ScmObj condition, u_long flags)
+{
+    return Scm_VMThrowException(Scm_VM(), condition, flags);
+}
+#else  /*!GAUCHE_API_0_95*/
 ScmObj Scm_Raise(ScmObj condition)
 {
-    return Scm_VMThrowException(Scm_VM(), condition);
+    return Scm_Raise2(condition, 0);
 }
 
+ScmObj Scm_Raise2(ScmObj condition, u_long flags)
+{
+    return Scm_VMThrowException2(Scm_VM(), condition, flags);
+}
+#endif /*!GAUCHE_API_0_95*/
+
+
 /* A convenient API---allows to call user-defined condition easily,
    even the condition type is defined in Scheme.  For example:
 
@@ -796,6 +860,13 @@ ScmObj Scm_Raise(ScmObj condition)
 
    This function isn't very efficient; but sometimes you want the convenience
    more, right?
+
+   The argument list format:
+
+     <condition-type> {<string-slot-name> <value>}* <terminator>
+
+     <terminator> : SCM_RAISE_CONDITION_MESSAGE <fmtstr> <fmtarg> ...
+                  | NULL
 */
 
 ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
@@ -841,147 +912,99 @@ ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
  *              frames of the handler itself.
  *   offset   - add this to the frame number.  Useful to show a middle part
  *              of frames only, by combining the skip parameter.
- *   format   - SCM_STACK_TRACE_FORMAT_* enum value.  EXPERIMENTAL.
+ *   format   - SCM_STACK_TRACE_FORMAT_* enum value.  No longer used.
  */
-
-#define STACK_DEPTH_LIMIT 30
-
-#define FMT_ORIG SCM_STACK_TRACE_FORMAT_ORIGINAL
-#define FMT_CC   SCM_STACK_TRACE_FORMAT_CC
-
-#define SHOW_EXPR(depth, expr) \
-    Scm_Printf(out, "%3d  %66.1S\n", (depth), Scm_UnwrapSyntax(expr));
-
 void Scm_ShowStackTrace(ScmPort *out, ScmObj stacklite,
                         int maxdepth, int skip, int offset, int format)
 {
-    ScmObj cp;
-    int depth = offset;
-
-    if (maxdepth == 0) maxdepth = STACK_DEPTH_LIMIT;
-
-    SCM_FOR_EACH(cp, stacklite) {
-        if (skip-- > 0) continue;
-        if (format == FMT_ORIG) {
-            SHOW_EXPR(depth++, SCM_CAR(cp));
-        }
-        if (SCM_PAIRP(SCM_CAR(cp))) {
-            ScmObj srci = Scm_PairAttrGet(SCM_PAIR(SCM_CAR(cp)),
-                                          SCM_SYM_SOURCE_INFO, SCM_FALSE);
-            if (SCM_PAIRP(srci) && SCM_PAIRP(SCM_CDR(srci))) {
-                switch (format) {
-                case FMT_ORIG:
-                    Scm_Printf(out, "        At line %S of %S\n",
-                               SCM_CADR(srci), SCM_CAR(srci));
-                    break;
-                case FMT_CC:
-                    Scm_Printf(out, "%A:%S:\n",
-                               SCM_CAR(srci), SCM_CADR(srci));
-                    break;
-                }
-            } else {
-                switch (format) {
-                case FMT_ORIG:
-                    Scm_Printf(out, "        [unknown location]\n");
-                    break;
-                case FMT_CC:
-                    Scm_Printf(out, "[unknown location]:\n");
-                    break;
-                }
-            }
-        } else {
-            Scm_Printf(out, "\n");
-        }
-        if (format == FMT_CC) {
-            SHOW_EXPR(depth++, SCM_CAR(cp));
-        }
-
-        if (maxdepth >= 0 && depth >= STACK_DEPTH_LIMIT) {
-            Scm_Printf(out, "... (more stack dump truncated)\n");
-            break;
-        }
-    }
+    static ScmObj show_stack_trace = SCM_UNDEFINED;
+    SCM_BIND_PROC(show_stack_trace,
+                  "%show-stack-trace",
+                  Scm_GaucheInternalModule());
+    Scm_ApplyRec5(show_stack_trace, stacklite, SCM_OBJ(out),
+                  SCM_MAKE_INT(maxdepth), SCM_MAKE_INT(skip),
+                  SCM_MAKE_INT(offset));
 }
 
-#undef SHOW_EXPR
-
+/* Dump stack trace.  Called from the default error reporter.
+   Also intended to be called from the debugger, so we allow vm to be NULL
+   to mean the current VM, and port to be NULL for the current error port. */
+void Scm_DumpStackTrace(ScmVM *vm, ScmPort *port)
+{
+    if (vm == NULL) vm = Scm_VM();
+    if (port == NULL) port = SCM_VM_CURRENT_ERROR_PORT(vm);
+    ScmObj stack = Scm_VMGetStackLite(vm);
+    SCM_PUTZ("Stack Trace:\n", -1, port);
+    SCM_PUTZ("_______________________________________\n", -1, port);
+    Scm_ShowStackTrace(port, stack, 0, 0, 0, 0);
+    SCM_FLUSH(port);
+}
 
 /*
  * Default error reporter
  */
 
 /* The default procedure to display the header of error message.
-   E is a thrown condition, not necessarily an error object. */
+   E is a thrown condition, not necessarily an error object.
+
+   The actual operation is written in Scheme (libexc.scm).  However,
+   we can't use that before the infrastructure is fully booted; so
+   this routine has a fallback which will be used only during initialization.
+*/
 static void Scm_PrintDefaultErrorHeading(ScmObj e, ScmPort *out)
 {
-    if (SCM_CONDITIONP(e)) {
-        char *heading = Scm_GetString(SCM_STRING(Scm_ConditionTypeName(e)));
-        /* TODO: considring that the class name may contain multibyte
-           characters, we should use string-upcase here. */
-        for (char *p=heading; *p; p++) {
-            *p = toupper(*p);
-        }
-        ScmObj msg = Scm_ConditionMessage(e);
-        if (!SCM_FALSEP(msg)) {
-            Scm_Printf(out, "*** %s: %A\n", heading, msg);
-        } else {
-            Scm_Printf(out, "*** %s\n", heading);
-        }
+    if (Scm_InitializedP()) {
+        static ScmObj print_default_error_heading = SCM_UNDEFINED;
+        SCM_BIND_PROC(print_default_error_heading,
+                      "print-default-error-heading",
+                      Scm_GaucheModule());
+        Scm_ApplyRec2(print_default_error_heading, e, SCM_OBJ(out));
     } else {
-        Scm_Printf(out, "*** ERROR: unhandled exception: %S\n", e);
+        /* Error during initialization. */
+        if (SCM_CONDITIONP(e)) {
+            Scm_Printf(out, "*** %A: %A\n",
+                       Scm_ConditionTypeName(e),
+                       Scm_ConditionMessage(e));
+        } else {
+            Scm_Printf(out, "*** ERROR: Unhandled condition: %S\n", e);
+        }
     }
 }
 
-static void report_error_inner(ScmVM *vm, ScmObj e)
-{
-    ScmObj stack = Scm_VMGetStackLite(vm);
-    ScmPort *err = SCM_VM_CURRENT_ERROR_PORT(vm);
-
-    Scm_PrintDefaultErrorHeading(e, err);
-    SCM_PUTZ("Stack Trace:\n", -1, err);
-    SCM_PUTZ("_______________________________________\n", -1, err);
-    Scm_ShowStackTrace(err, stack, 0, 0, 0, FMT_ORIG);
-    /* NB: stderr is autoflushed by default, but in case err is replaced
-       by some other port, we explicitly flush it. */
-    SCM_FLUSH(err);
-}
-
-void Scm_ReportError(ScmObj e)
+/* We treat out == #f or #t just like 'format' - #t for the current output
+   port, and #f for string port.  If it's output port, use it.  For any
+   other objects, we use current error port.  This permissive behavior is
+   intentional - report-error is usually called during error handling,
+   and raising an error there masks the original error.
+*/
+#if GAUCHE_API_0_95
+ScmObj Scm_ReportError(ScmObj e, ScmObj out)
+#else  /*!GAUCHE_API_0_95*/
+ScmObj Scm_ReportError2(ScmObj e, ScmObj out)
+#endif /*!GAUCHE_API_0_95*/
 {
     ScmVM *vm = Scm_VM();
-
-    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_ERROR_BEING_REPORTED)) {
-        /* An _uncaptured_ error occurred during reporting an error.
-           We can't proceed, for it will cause infinite loop.
-           Note that it is OK for an error to occur inside the error
-           reporter, as far as the error is handled by user-installed
-           handler.   The user-installed handler can even invoke a
-           continuation that is captured outside; the flag is reset
-           in such case.
-           Be careful that it is possible that stderr is no longer
-           available here (since it may be the very cause of the
-           recursive error).  All we can do is to abort. */
-        Scm_Abort("Unhandled error occurred during reporting an error.  Process aborted.\n");
+    ScmPort *port = SCM_VM_CURRENT_ERROR_PORT(vm);
+    if (SCM_FALSEP(out)) {
+        port = SCM_PORT(Scm_MakeOutputStringPort(TRUE));
+    } else if (SCM_TRUEP(out)) {
+        port = SCM_VM_CURRENT_OUTPUT_PORT(vm);
+    } else if (SCM_OPORTP(out)) {
+        port = SCM_PORT(out);
     }
-
-    SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_REPORTED);
-    SCM_UNWIND_PROTECT {
-        if (SCM_PROCEDUREP(vm->defaultEscapeHandler)) {
-            Scm_ApplyRec(vm->defaultEscapeHandler, SCM_LIST1(e));
-        } else {
-            report_error_inner(vm, e);
-        }
-    }
-    SCM_WHEN_ERROR {
-        /* NB: this is called when a continuation captured outside is
-           invoked inside the error reporter.   It may be invoked by
-           the user's error handler.  */
-        SCM_VM_RUNTIME_FLAG_CLEAR(vm, SCM_ERROR_BEING_REPORTED);
-    }
-    SCM_END_PROTECT;
-    SCM_VM_RUNTIME_FLAG_CLEAR(vm, SCM_ERROR_BEING_REPORTED);
+    Scm_PrintDefaultErrorHeading(e, port);
+    Scm_DumpStackTrace(vm, port);
+    if (SCM_FALSEP(out)) return Scm_GetOutputString(SCM_PORT(port), 0);
+    else return SCM_UNDEFINED;
 }
 
+#if !GAUCHE_API_0_95
+ScmObj Scm_ReportError(ScmObj e)
+{
+    return Scm_ReportError2(e, SCM_UNBOUND);
+}
+#endif /*!GAUCHE_API_0_95*/
+
 /*
  * Initialization
  */
@@ -1048,6 +1071,7 @@ void Scm__InitExceptions(void)
                                 mod, cond_meta, SCM_FALSE,
                                 porterror_slots, 0);
 
+
     Scm_InitStaticClassWithMeta(SCM_CLASS_COMPOUND_CONDITION,
                                 "<compound-condition>",
                                 mod, cond_meta, SCM_FALSE,
