@@ -1,7 +1,7 @@
 /*
  * gauche.h - Gauche scheme system header
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -125,7 +125,10 @@ SCM_DECL_BEGIN
 /* Define this to 0 to turn off lazy-pair feature. */
 #define GAUCHE_LAZY_PAIR 1
 
-#undef GAUCHE_UNIFY_SYMBOL_KEYWORD /* not yet */
+/* Enable an option to make keywords and symbols disjoint.
+   (Transient: Will be gone once we completely migrate to
+   unified keyword-symbol system */
+#define GAUCHE_KEEP_DISJOINT_KEYWORD_OPTION 1
 
 /* Include appropriate threading interface.  Threading primitives are
    abstracted with SCM_INTERNAL_* macros and ScmInternal* typedefs.
@@ -156,6 +159,13 @@ SCM_DECL_BEGIN
 #define SCM_ALIGN8  /*empty*/
 #endif /* !__GNUC__ */
 
+/* 'No return' attribute */
+#ifdef __GNUC__
+#define SCM_NORETURN  __attribute__((__noreturn__))
+#else  /*__GNUC__*/
+#define SCM_NORETURN  /*empty*/
+#endif /*__GNUC__*/
+
 /*-------------------------------------------------------------
  * BASIC TYPES
  */
@@ -301,7 +311,7 @@ SCM_EXTERN int Scm_EqualM(ScmObj x, ScmObj y, int mode);
 
 #define SCM_INTP(obj)        (SCM_TAG2(obj) == 1)
 #define SCM_INT_VALUE(obj)   (((signed long int)SCM_WORD(obj)) >> 2)
-#define SCM_MAKE_INT(obj)    SCM_OBJ(((intptr_t)(obj) << 2) + 1)
+#define SCM_MAKE_INT(obj)    SCM_OBJ(((uintptr_t)(obj) << 2) + 1)
 
 #define SCM_UINTP(obj)       (SCM_INTP(obj)&&((signed long int)SCM_WORD(obj)>=0))
 typedef long ScmSmallInt;    /* C integer type corresponds to Scheme fixnum
@@ -331,7 +341,7 @@ typedef struct ScmFlonumRec {
 #define SCM_CHAR(obj)           ((ScmChar)(obj))
 #define SCM_CHARP(obj)          ((SCM_WORD(obj)&0xff) == 3)
 #define SCM_CHAR_VALUE(obj)     SCM_CHAR(((unsigned long)SCM_WORD(obj)) >> 8)
-#define SCM_MAKE_CHAR(ch)       SCM_OBJ((((unsigned long)(ch))<<8) + 3)
+#define SCM_MAKE_CHAR(ch)       SCM_OBJ((intptr_t)(((unsigned long)(ch))<<8) + 3)
 
 #define SCM_CHAR_INVALID        ((ScmChar)(-1)) /* indicate invalid char */
 #define SCM_CHAR_MAX            (0xffffff)
@@ -497,9 +507,6 @@ typedef struct ScmTreeMapRec   ScmTreeMap;
 typedef struct ScmModuleRec    ScmModule;
 typedef struct ScmSymbolRec    ScmSymbol;
 typedef struct ScmGlocRec      ScmGloc;
-#if !GAUCHE_UNIFY_SYMBOL_KEYWORD
-typedef struct ScmKeywordRec   ScmKeyword;
-#endif /*!GAUCHE_UNIFY_SYMBOL_KEYWORD*/
 typedef struct ScmProcedureRec ScmProcedure;
 typedef struct ScmClosureRec   ScmClosure;
 typedef struct ScmSubrRec      ScmSubr;
@@ -511,9 +518,13 @@ typedef struct ScmMacroRec     ScmMacro;
 typedef struct ScmPromiseRec   ScmPromise;
 typedef struct ScmRegexpRec    ScmRegexp;
 typedef struct ScmRegMatchRec  ScmRegMatch;
-typedef struct ScmWriteContextRec ScmWriteContext;
+typedef struct ScmWriteControlsRec  ScmWriteControls;  /* see writer.h */
+typedef struct ScmWriteContextRec   ScmWriteContext;   /* see writerP.h */
+typedef struct ScmWriteStateRec     ScmWriteState;     /* see wrtierP.h */
 typedef struct ScmAutoloadRec  ScmAutoload;
 typedef struct ScmComparatorRec ScmComparator;
+typedef struct ScmDLObjRec     ScmDLObj;           /* see load.c */
+typedef struct ScmReadContextRec ScmReadContext;   /* see read.c */
 
 typedef ScmObj ScmSubrProc(ScmObj *, int, void*);
 
@@ -549,17 +560,11 @@ typedef struct ScmEvalPacketRec {
     ScmModule *module;          /* 'Current module' after evaluation */
 } ScmEvalPacket;
 
-#if !defined(GAUCHE_API_PRE_0_9)
 SCM_EXTERN int Scm_Eval(ScmObj form, ScmObj env, ScmEvalPacket *packet);
 SCM_EXTERN int Scm_EvalCString(const char *form, ScmObj env,
                                ScmEvalPacket *packet);
 SCM_EXTERN int Scm_Apply(ScmObj proc, ScmObj args,
                          ScmEvalPacket *packet);
-#else  /*GAUCHE_API_PRE_0_9*/
-#define Scm_Eval(f, e)        Scm_EvalRec(f, e)
-#define Scm_EvalCString(f, e) Scm_EvalCStringRec(f, e)
-#define Scm_Apply(p, a)       Scm_ApplyRec(p, a)
-#endif /*GAUCHE_API_PRE_0_9*/
 
 /* Calls VM recursively to evaluate the Scheme code.  These
    ones does not capture exceptions. */
@@ -613,13 +618,6 @@ SCM_EXTERN ScmObj Scm_VMWithGuardHandler(ScmObj handler, ScmObj thunk);
 SCM_EXTERN ScmObj Scm_VMWithExceptionHandler(ScmObj handler, ScmObj thunk);
 
 /* Miscellaneous stuff */
-SCM_EXTERN ScmObj Scm_MakeMacroTransformer(ScmSymbol *name,
-                                           ScmObj proc);
-SCM_EXTERN ScmObj Scm_MakeMacroAutoload(ScmSymbol *name,
-                                        ScmAutoload *al);
-
-SCM_EXTERN ScmObj Scm_UnwrapSyntax(ScmObj form);
-
 SCM_EXTERN int    Scm_VMGetNumResults(ScmVM *vm);
 SCM_EXTERN ScmObj Scm_VMGetResult(ScmVM *vm);
 SCM_EXTERN ScmObj Scm_VMGetStackLite(ScmVM *vm);
@@ -724,8 +722,9 @@ typedef struct ScmClassStaticSlotSpecRec ScmClassStaticSlotSpec;
    SCM_CLASS_ABSTRACT 
        This class is defined in C, but doesn't allowed to create an
        instance by its own.  It is intended to be used as a mixin from
-       both C and Scheme-defined class.   This class shouldn't have
-       C members other than SCM_HEADER.   This class cannot be redefined.
+       both C and Scheme-defined class.   An instance of this class
+       shouldn't have C members other than SCM_HEADER.
+       This class cannot be redefined.
 
    SCM_CLASS_BASE
        This class is defined in C, and can be subclassed in Scheme.
@@ -777,10 +776,16 @@ enum {
        or class-slots becomes settable.
        We reset this flag at the end of class initialization, so that
        we can avoid the behavior of a class from being accidentally
-       chnaged.  The flag may be set during updating a class metaobject
+       changed.  The flag may be set during updating a class metaobject
        triggered by metaclass change (see lib/gauche/redefutil.scm).
      */
-    SCM_CLASS_MALLEABLE = 0x08
+    SCM_CLASS_MALLEABLE = 0x08,
+
+    /* This flag indicates the class is for the aggregate data type.
+       Currently the writer uses this info to determine when to stop
+       recursing (see print-level).  We may use this later for generic
+       data structure walker. */
+    SCM_CLASS_AGGREGATE = 0x10
 };
 
 #define SCM_CLASS_FLAGS(obj)        (SCM_CLASS(obj)->flags)
@@ -859,6 +864,7 @@ extern ScmClass *Scm_ObjectCPL[];
 
 /* Static definition of classes
  *   SCM_DEFINE_BUILTIN_CLASS
+ *   SCM_DEFINE_BUILTIN_CLASS_FLAGS
  *   SCM_DEFINE_BUILTIN_CLASS_SIMPLE
  *   SCM_DEFINE_ABSTRACT_CLASS
  *   SCM_DEFINE_BASE_CLASS
@@ -907,6 +913,11 @@ extern ScmClass *Scm_ObjectCPL[];
                              SCM_CLASS_BUILTIN,           \
                              printer, compare, serialize, allocate, cpa)
 
+#define SCM_DEFINE_BUILTIN_CLASS_FLAGS(cname, printer, compare, serialize, allocate, cpa, flags) \
+    SCM__DEFINE_CLASS_COMMON(cname, 0,                                  \
+                             SCM_CLASS_BUILTIN|(flags),                 \
+                             printer, compare, serialize, allocate, cpa)
+
 /* Define built-in class statically -- simpler version */
 #define SCM_DEFINE_BUILTIN_CLASS_SIMPLE(cname, printer)         \
     SCM_DEFINE_BUILTIN_CLASS(cname, printer, NULL, NULL, NULL, NULL)
@@ -935,12 +946,14 @@ typedef struct ScmForeignPointerRec {
                                    constructed by Scm_MakeForeignPointer. */
     ScmObj attributes;          /* alist.  useful to store e.g. callbacks.
                                    use accessor procedures. */
+    ScmWord flags;              /* used internally.  We use ScmWord to keep
+                                   ScmForeignPointer fit in 4 words. */
 } ScmForeignPointer;
 
 #define SCM_FOREIGN_POINTER_P(obj)   SCM_ISA(obj, SCM_CLASS_FOREIGN_POINTER)
 #define SCM_FOREIGN_POINTER(obj)     ((ScmForeignPointer*)(obj))
 #define SCM_FOREIGN_POINTER_REF(type, obj) \
-    ((type)(SCM_FOREIGN_POINTER(obj)->ptr))
+    ((type)(Scm_ForeignPointerRef(SCM_FOREIGN_POINTER(obj))))
 
 typedef void (*ScmForeignCleanupProc)(ScmObj);
 
@@ -952,8 +965,11 @@ SCM_EXTERN ScmClass *Scm_MakeForeignPointerClass(ScmModule *module,
 SCM_EXTERN ScmObj Scm_MakeForeignPointer(ScmClass *klass, void *ptr);
 SCM_EXTERN ScmObj Scm_MakeForeignPointerWithAttr(ScmClass *klass, void *ptr,
                                                  ScmObj attr);
+SCM_EXTERN void  *Scm_ForeignPointerRef(ScmForeignPointer *fp);
+SCM_EXTERN int    Scm_ForeignPointerInvalidP(ScmForeignPointer *fp);
+SCM_EXTERN void   Scm_ForeignPointerInvalidate(ScmForeignPointer *fp);
 
-/* foreign pointer flags */
+/* foreign pointer class flags */
 enum {
     SCM_FOREIGN_POINTER_KEEP_IDENTITY = (1L<<0),
          /* If set, a foreign pointer class keeps a weak hash table that maps
@@ -1141,6 +1157,7 @@ SCM_EXTERN ScmObj Scm_MonotonicMerge1(ScmObj sequences);
 SCM_EXTERN ScmObj Scm_Union(ScmObj list1, ScmObj list2);
 SCM_EXTERN ScmObj Scm_Intersection(ScmObj list1, ScmObj list2);
 
+SCM_EXTERN ScmObj Scm_MakeExtendedPair(ScmObj car, ScmObj cdr, ScmObj attrs);
 SCM_EXTERN ScmObj Scm_ExtendedCons(ScmObj car, ScmObj cdr);
 SCM_EXTERN ScmObj Scm_PairAttr(ScmPair *pair);
 SCM_EXTERN ScmObj Scm_PairAttrGet(ScmPair *pair, ScmObj key, ScmObj fallback);
@@ -1257,7 +1274,7 @@ struct ScmProcedureRec {
     unsigned int required : 16;    /* # of required args */
     unsigned int optional : 8;     /* >=1 if it takes opt args. see below.*/
     unsigned int type     : 3;     /* ScmProcedureType */
-    unsigned int locked   : 1;     /* setter locked? */
+    unsigned int locked   : 1;     /* setter locked? (see below) */
     unsigned int currying : 1;     /* autocurrying */
     unsigned int constant : 1;     /* constant procedure. see below. */
     unsigned int reserved : 2;     /* unused yet. */
@@ -1266,6 +1283,15 @@ struct ScmProcedureRec {
     ScmObj inliner;                /* inliner information.  see below. */
 };
 
+/* About locked slot:
+   For <procedure> and <generic>, it shows whether the setter is locked.
+   For <method>, it shows whether the alteration of the method is disallowed,
+   i.e. one can't redefine a method with matching signature.
+   (These two roles are reflected to the two macors,
+   SCM_PROCEDURE_SETTER_LOCKED and SCM_PROCEDURE_METHOD_LOCKED)
+   TODO: When we change ABI, maybe split these roles to different flags.
+ */
+
 /* About optional slot:
    If this slot is non-zero, the procedure takes optional arguments.
    For Standard Scheme procedures with 'rest' arguments, this slot is 1
@@ -1331,6 +1357,7 @@ enum ScmProcedureType {
 #define SCM_PROCEDURE_INFO(obj)     SCM_PROCEDURE(obj)->info
 #define SCM_PROCEDURE_SETTER(obj)   SCM_PROCEDURE(obj)->setter
 #define SCM_PROCEDURE_INLINER(obj)  SCM_PROCEDURE(obj)->inliner
+#define SCM_PROCEDURE_SETTER_LOCKED(obj) SCM_PROCEDURE(obj)->locked
 
 SCM_CLASS_DECL(Scm_ProcedureClass);
 #define SCM_CLASS_PROCEDURE    (&Scm_ProcedureClass)
@@ -1411,9 +1438,9 @@ struct ScmSubrRec {
 #define SCM_DEFINE_SUBRX(cvar, req, opt, cst, inf, flags, func, inliner, data) \
     SCM__DEFINE_SUBR_INT(cvar, req, opt, cst, inf, flags, func, inliner, data)
 
-/* old interface.  will be gone. */
+/* TRANSIENT: old interface.  will be gone. */
 #define SCM_DEFINE_SUBRI(cvar, req, opt, inf, func, inliner, data) \
-    SCM__DEFINE_SUBR_INT(cvar, req, opt, inf, SCM_SUBR_IMMEDIATE_ARG, \
+    SCM__DEFINE_SUBR_INT(cvar, req, opt, 0, inf, SCM_SUBR_IMMEDIATE_ARG, \
                          func, inliner, data)
 
 SCM_EXTERN ScmObj Scm_MakeSubr(ScmSubrProc *func,
@@ -1478,6 +1505,7 @@ SCM_CLASS_DECL(Scm_MethodClass);
 #define SCM_CLASS_METHOD           (&Scm_MethodClass)
 #define SCM_METHODP(obj)           SCM_ISA(obj, SCM_CLASS_METHOD)
 #define SCM_METHOD(obj)            ((ScmMethod*)obj)
+#define SCM_METHOD_LOCKED(obj)     SCM_METHOD(obj)->common.locked
 
 #define SCM_DEFINE_METHOD(cvar, gf, req, opt, specs, func, data)        \
     ScmMethod cvar = {                                                  \
@@ -1545,40 +1573,29 @@ SCM_EXTERN ScmObj Scm_Map(ScmObj proc, ScmObj arg1, ScmObj args);
  * MACROS AND SYNTAX
  */
 
-/* Syntax is a built-in procedure to compile given form. */
-struct ScmSyntaxRec {
-    SCM_HEADER;
-    ScmSymbol *name;            /* for debugging */
-    ScmObj     handler;         /* syntax handler.  (Sexpr, Env) -> IForm */
-};
+/* The actual definitions of ScmSyntax and ScmMacro are private.*/
 
 #define SCM_SYNTAX(obj)             ((ScmSyntax*)(obj))
 #define SCM_SYNTAXP(obj)            SCM_XTYPEP(obj, SCM_CLASS_SYNTAX)
-
 SCM_CLASS_DECL(Scm_SyntaxClass);
 #define SCM_CLASS_SYNTAX            (&Scm_SyntaxClass)
 
 SCM_EXTERN ScmObj Scm_MakeSyntax(ScmSymbol *name, ScmObj handler);
 
-/* Macro */
-struct ScmMacroRec {
-    SCM_HEADER;
-    ScmSymbol *name;            /* for debug */
-    ScmTransformerProc transformer; /* (Self, Sexpr, Env) -> Sexpr */
-    void *data;
-};
-
 #define SCM_MACRO(obj)             ((ScmMacro*)(obj))
 #define SCM_MACROP(obj)            SCM_XTYPEP(obj, SCM_CLASS_MACRO)
-
 SCM_CLASS_DECL(Scm_MacroClass);
 #define SCM_CLASS_MACRO            (&Scm_MacroClass)
 
-SCM_EXTERN ScmObj Scm_MakeMacro(ScmSymbol *name,
-                                ScmTransformerProc transformer,
-                                void *data);
+SCM_EXTERN ScmObj Scm_MakeMacro(ScmSymbol *name, ScmObj transformer);
+SCM_EXTERN ScmObj Scm_MacroTransformer(ScmMacro *mac);
 
-SCM_EXTERN ScmObj Scm_VMMacroExpand(ScmObj expr, ScmObj env, int oncep);
+SCM_EXTERN ScmObj Scm_MakeMacroTransformer(ScmSymbol *name,
+                                           ScmObj proc);
+SCM_EXTERN ScmObj Scm_MakeMacroAutoload(ScmSymbol *name,
+                                        ScmAutoload *al);
+
+SCM_EXTERN ScmObj Scm_UnwrapSyntax(ScmObj form);
 
 /*--------------------------------------------------------
  * PROMISE
@@ -1641,7 +1658,21 @@ SCM_EXTERN void Scm_PortError(ScmPort *port, int reason, const char *msg, ...);
 SCM_EXTERN void Scm_Warn(const char *msg, ...);
 SCM_EXTERN void Scm_FWarn(ScmString *fmt, ScmObj args);
 
+/* TRANSIENT: Scm_Raise2 is to keep ABI compatibility.  Will be gone
+   in 1.0.  */
+#if    GAUCHE_API_0_95
+SCM_EXTERN ScmObj Scm_Raise(ScmObj exception, u_long flags);
+#define Scm_Raise2(e, f)  Scm_Raise(e, f)
+#else  /*!GAUCHE_API_0_95*/
 SCM_EXTERN ScmObj Scm_Raise(ScmObj exception);
+SCM_EXTERN ScmObj Scm_Raise2(ScmObj exception, u_long flags);
+#endif /*!GAUCHE_API_0_95*/
+
+/* flags for Scm_Raise */
+enum {
+    SCM_RAISE_NON_CONTINUABLE = (1L<<0)
+};
+
 SCM_EXTERN ScmObj Scm_RaiseCondition(ScmObj conditionType, ...);
 
 /* A marker to insert between key-value pair and formatting string
@@ -1662,7 +1693,14 @@ SCM_EXTERN void Scm_ShowStackTrace(ScmPort *out, ScmObj stacklite,
                                    int maxdepth, int skip, int offset,
                                    int format);
 
-SCM_EXTERN void Scm_ReportError(ScmObj e);
+/* TRANSIENT: Scm_ReportErrr2 is to keep ABI compatibility.  Will be gone
+   in 1.0.  */
+#if    GAUCHE_API_0_95
+SCM_EXTERN ScmObj Scm_ReportError(ScmObj e, ScmObj out);
+#else  /*!GAUCHE_API_0_95*/
+SCM_EXTERN ScmObj Scm_ReportError(ScmObj e);
+SCM_EXTERN ScmObj Scm_ReportError2(ScmObj e, ScmObj out);
+#endif /*!GAUCHE_API_0_95*/
 
 /*--------------------------------------------------------
  * REGEXP
@@ -1771,10 +1809,11 @@ SCM_EXTERN void   Scm_ProfilerReset(void);
 /* Program start and termination */
 
 SCM_EXTERN void Scm_Init(const char *signature);
+SCM_EXTERN int  Scm_InitializedP(void);
 SCM_EXTERN void Scm_Cleanup(void);
-SCM_EXTERN void Scm_Exit(int code);
-SCM_EXTERN void Scm_Abort(const char *msg);
-SCM_EXTERN void Scm_Panic(const char *msg, ...);
+SCM_EXTERN void Scm_Exit(int code) SCM_NORETURN;
+SCM_EXTERN void Scm_Abort(const char *msg) SCM_NORETURN;
+SCM_EXTERN void Scm_Panic(const char *msg, ...) SCM_NORETURN;
 SCM_EXTERN ScmObj Scm_InitCommandLine(int argc, const char *argv[]);
 
 SCM_EXTERN void Scm_SimpleMain(int argc, const char *argv[],
