@@ -1,7 +1,7 @@
 /*
  * code.h - Virtual machine code
  *
- *   Copyright (c) 2005-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2005-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -60,17 +60,19 @@ struct ScmCompiledCodeRec {
                                    that takes rest arg.  Otherwise 0. */
     ScmObj name;                /* If this is the body of a closure, holds
                                    its name.  Otherwise #f. */
-    ScmObj info;                /* debug info.  alist of instruction offset
-                                   and info. */
-    ScmObj argInfo;             /* If this code is the body of the closure,
-                                   keeps a list of args.  #f otherwise. (*3) */
+    ScmObj debugInfo;           /* debug info, that associates instructions
+                                   and source code / other metainfo.  May be
+                                   () if no info is avaialble. (*3) */
+    ScmObj signatureInfo;       /* signature info, a metainfo related to the
+                                   interface of this closure.   Maybe #f
+                                   if no info is available. (*4) */
     ScmObj parent;              /* ScmCompiledCode if this code is compiled
                                    within other code chunk.  #f otherwise. */
     ScmObj intermediateForm;    /* A packed IForm of the body (see compile.scm
                                    for the details of IForm).  It is used
                                    to inline this procedure.  Only set if
                                    the procedure is defined with define-inline.
-                                   #f otherwise. (*4) */
+                                   #f otherwise. (*5) */
     void *builder;              /* An opaque data used during consturcting
                                    the code vector.  Usually NULL. */
 };
@@ -83,11 +85,29 @@ struct ScmCompiledCodeRec {
  *   *2) For the C-dumped code, the code vector is located in a static data
  *       area, subject to GC scanning.  In that case, the constants pointer
  *       is NULL.
- *   *3) This info isn't set for the time being.
- *   *4) This IForm is a direct result of Pass1, i.e. non-optimized form.
+ *   *3) ((<offset> <info> ...) ...)
+ *       <offset> is either an instruction offset or 'definition (for the
+ *       entire closure).
+ *       At this moment, only used <info> is (source-info . <source>).
+ *   *4) (<signature> <info> ...)
+ *       <signature> is (<procedure-name> <formal> ...)
+ *       <procedure-name> may be just a symbol, or a list (in case of internal
+ *       function).  There's no precise definition for the format yet---it's
+ *       for debug information.  See Scm_CompiledCodeFullName().
+ *       <formal> ... is the formal argument list, as appears in the original
+ *       lambda form, including :key, :optional etc.
+ *       At this moment we don't use <info> ... yet; the plan is to put
+ *       metainfo about closure interface, e.g. types.
+ *   *5) This IForm is a direct result of Pass1, i.e. non-optimized form.
  *       Pass2 scans it when IForm is inlined into the caller site.
  */
 
+/* TRANSIENT: ***WARNING - NAMESPACE POLLUTION*** Certain versions of
+   0.9.5_pre1 precompiler emits C code that refers to the old name
+   ScmCompiledCode.argInfo.  This is a dirty workaround to allow compiling
+   such precompiled code.   Make sure to remove this after 0.9.5 release.  */
+#define argInfo  signatureInfo
+
 SCM_CLASS_DECL(Scm_CompiledCodeClass);
 #define SCM_CLASS_COMPILED_CODE   (&Scm_CompiledCodeClass)
 
@@ -99,10 +119,10 @@ SCM_CLASS_DECL(Scm_CompiledCodeClass);
 #define SCM_COMPILED_CODE_OPTIONAL_ARGS(obj) \
     (SCM_COMPILED_CODE(obj)->optionalArgs)
 
-#define SCM_COMPILED_CODE_CONST_INITIALIZER(code, codesize, maxstack, reqargs, optargs, name, info, arginfo, parent, iform) \
+#define SCM_COMPILED_CODE_CONST_INITIALIZER(code, codesize, maxstack, reqargs, optargs, name, debuginfo, signatureinfo, parent, iform) \
     { { SCM_CLASS_STATIC_TAG(Scm_CompiledCodeClass) },   \
       (code), NULL, (codesize), 0, (maxstack),           \
-      (reqargs), (optargs), (name), (info), (arginfo),   \
+      (reqargs), (optargs), (name), (debuginfo), (signatureinfo),   \
       (parent), (iform) }
 
 SCM_EXTERN void   Scm_CompiledCodeCopyX(ScmCompiledCode *dest,
@@ -112,31 +132,6 @@ SCM_EXTERN ScmObj Scm_CompiledCodeToList(ScmCompiledCode *cc);
 SCM_EXTERN ScmObj Scm_CompiledCodeFullName(ScmCompiledCode *cc);
 SCM_EXTERN void   Scm_VMExecuteToplevels(ScmCompiledCode *cv[]);
 
-/* Builder API */
-SCM_EXTERN ScmObj Scm_MakeCompiledCodeBuilder(int reqargs, int optargs,
-                                              ScmObj name, ScmObj parent,
-                                              ScmObj intForm);
-SCM_EXTERN ScmObj Scm_CompiledCodeCurrentInsn(ScmCompiledCode *cc);
-SCM_EXTERN void   Scm_CompiledCodeReplaceInsn(ScmCompiledCode *cc,
-                                              ScmObj insn,
-                                              ScmObj operand,
-                                              ScmObj info);
-SCM_EXTERN void   Scm_CompiledCodeFlushInsn(ScmCompiledCode *cc);
-SCM_EXTERN void   Scm_CompiledCodePutInsn(ScmCompiledCode *cc,
-                                          ScmObj insn,
-                                          ScmObj operand,
-                                          ScmObj info);
-SCM_EXTERN ScmObj Scm_CompiledCodeNewLabel(ScmCompiledCode *cc);
-SCM_EXTERN void   Scm_CompiledCodeSetLabel(ScmCompiledCode *cc, ScmObj label);
-SCM_EXTERN void   Scm_CompiledCodeFinishBuilder(ScmCompiledCode *cc,
-                                                int maxstack);
-SCM_EXTERN void   Scm_CompiledCodeEmit(ScmCompiledCode *cc,
-                                       int code,
-                                       int arg0,
-                                       int arg1,
-                                       ScmObj operand,
-                                       ScmObj info);
-
 /*
  * VM instructions
  */
