@@ -1,7 +1,7 @@
 /*
  * hash.h - Public API for hashtables
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -105,18 +105,26 @@ struct ScmHashIterRec {
     void *next;
 };
 
-#if !defined(GAUCHE_API_PRE_0_9)
 SCM_EXTERN void Scm_HashIterInit(ScmHashIter *iter, ScmHashCore *core);
 SCM_EXTERN ScmDictEntry *Scm_HashIterNext(ScmHashIter *iter);
-#endif /* See the compatibility section below for old APIs */
 
 /*
  * Hash functions
  */
 SCM_EXTERN u_long Scm_EqHash(ScmObj obj);
 SCM_EXTERN u_long Scm_EqvHash(ScmObj obj);
-SCM_EXTERN u_long Scm_Hash(ScmObj obj);
 SCM_EXTERN u_long Scm_HashString(ScmString *str, u_long bound);
+SCM_EXTERN u_long Scm_PortableHash(ScmObj obj, u_long salt);
+SCM_EXTERN ScmSmallInt Scm_DefaultHash(ScmObj obj);
+SCM_EXTERN u_long Scm_CombineHashValue(u_long a, u_long b);
+
+SCM_EXTERN ScmSmallInt Scm_HashSaltRef(void);
+SCM_EXTERN ScmSmallInt Scm_HashSaltSet(ScmSmallInt);
+SCM_EXTERN ScmSmallInt Scm_PortableHashSalt(ScmObj newval);
+
+SCM_EXTERN ScmObj Scm_CurrentRecursiveHash(ScmObj);
+
+SCM_EXTERN u_long Scm_Hash(ScmObj obj); /* DEPRECATED */
 
 /*================================================================
  * ScmHashTable
@@ -135,7 +143,14 @@ SCM_CLASS_DECL(Scm_HashTableClass);
 
 #define SCM_HASH_TABLE_CORE(obj) (&SCM_HASH_TABLE(obj)->core)
 
-SCM_EXTERN ScmObj Scm_MakeHashTableSimple(ScmHashType type, int initSize);
+SCM_EXTERN ScmObj Scm_MakeHashTableSimple(ScmHashType type,
+                                          unsigned int initSize);
+SCM_EXTERN ScmObj Scm_MakeHashTableFull(ScmHashProc *hashfn,
+                                        ScmHashCompareProc *cmpfn,
+                                        unsigned int initSize,
+                                        void *data);
+
+SCM_EXTERN ScmHashType Scm_HashTableType(ScmHashTable *tab);
 
 SCM_EXTERN ScmObj Scm_HashTableCopy(ScmHashTable *tab);
 
@@ -156,10 +171,10 @@ SCM_EXTERN ScmObj Scm_HashTableStat(ScmHashTable *table);
  * For backward compatibility.  DEPRECATED.
  */
 
-#define SCM_HASHTABLE       SCM_HASH_TABLE
-#define SCM_HASHTABLEP      SCM_HASH_TABLE_P
-#define SCM_CLASS_HASHTABLE SCM_CLASS_HASH_TABLE
-#define SCM_HASH_ADDRESS    SCM_HASH_EQ
+#define SCM_HASHTABLE(x)       SCM_HASH_TABLE(x)
+#define SCM_HASHTABLEP(x)      SCM_HASH_TABLE_P(x)
+#define SCM_CLASS_HASHTABLE    SCM_CLASS_HASH_TABLE
+#define SCM_HASH_ADDRESS       SCM_HASH_EQ
 
 typedef struct ScmHashEntryRec {
     void *key;
@@ -176,14 +191,5 @@ SCM_EXTERN ScmObj Scm_MakeHashTable(ScmHashProc *hashfn,
                                     ScmHashCompareProc *cmpfn,
                                     unsigned int initSize);
 
-#if defined(GAUCHE_API_PRE_0_9)
-#define Scm_HashIterInit(table, iter) Scm__HashIterInitCompat(table, iter)
-#define Scm_HashIterNext(iter)        Scm__HashIterNextCompat(iter)
-
-SCM_EXTERN void          Scm__HashIterInitCompat(ScmHashTable *table,
-                                                 ScmHashIter *iter);
-SCM_EXTERN ScmHashEntry *Scm__HashIterNextCompat(ScmHashIter *iter);
-#endif /*GAUCHE_API_PRE_0_9*/
-
 #endif /* GAUCHE_HASH_H */
 
