@@ -1,7 +1,7 @@
 /*
  * gauche/port.h - Port API
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -147,7 +147,7 @@ struct ScmPortRec {
     ScmVM *lockOwner;           /* for port mutex; owner of the lock */
     int lockCount;              /* for port mutex; # of recursive locks */
 
-    ScmObj recursiveContext;    /* used internally */
+    ScmWriteState *writeState;  /* used internally */
 
     /* Input counters.  these doesn't take account of ungetting and
        seeking: Ungetting doesn't affect those counters (you can think
@@ -258,7 +258,7 @@ enum ScmPortICPolicy {
 #define SCM_PORT_FLAGS(obj)     (SCM_PORT(obj)->flags)
 #define SCM_PORT_ICPOLICY(obj)  (SCM_PORT(obj)->icpolicy)
 
-#define SCM_PORT_CASE_FOLD(obj) (SCM_PORT_FLAGS(obj)&SCM_PORT_CASE_FOLD)
+#define SCM_PORT_CASE_FOLDING(obj) (SCM_PORT_FLAGS(obj)&SCM_PORT_CASE_FOLD)
 
 #define SCM_PORT_CLOSED_P(obj)  (SCM_PORT(obj)->closed)
 #define SCM_PORT_OWNER_P(obj)   (SCM_PORT(obj)->ownerp)
@@ -309,6 +309,10 @@ SCM_EXTERN int    Scm_GetPortBufferingMode(ScmPort *port);
 SCM_EXTERN void   Scm_SetPortBufferingMode(ScmPort *port, int mode);
 SCM_EXTERN int    Scm_GetPortBufferSigpipeSensitive(ScmPort *port);
 SCM_EXTERN void   Scm_SetPortBufferSigpipeSensitive(ScmPort *port, int sensitive);
+SCM_EXTERN int    Scm_GetPortCaseFolding(ScmPort *port);
+SCM_EXTERN void   Scm_SetPortCaseFolding(ScmPort *port, int flag);
+SCM_EXTERN ScmObj Scm_GetPortReaderLexicalMode(ScmPort *port);
+SCM_EXTERN void   Scm_SetPortReaderLexicalMode(ScmPort *port, ScmObj obj);
 
 SCM_EXTERN void   Scm_FlushAllPorts(int exitting);
 
@@ -361,6 +365,8 @@ SCM_EXTERN int    Scm_Getc(ScmPort *port);
 SCM_EXTERN int    Scm_Getz(char *buf, int buflen, ScmPort *port);
 SCM_EXTERN ScmChar Scm_Peekc(ScmPort *port);
 SCM_EXTERN int    Scm_Peekb(ScmPort *port);
+SCM_EXTERN ScmObj Scm_UngottenChars(ScmPort *port);
+SCM_EXTERN ScmObj Scm_UngottenBytes(ScmPort *port);
 
 SCM_EXTERN void   Scm_UngetcUnsafe(ScmChar ch, ScmPort *port);
 SCM_EXTERN void   Scm_UngetbUnsafe(int b, ScmPort *port);
@@ -369,6 +375,8 @@ SCM_EXTERN int    Scm_GetcUnsafe(ScmPort *port);
 SCM_EXTERN int    Scm_GetzUnsafe(char *buf, int buflen, ScmPort *port);
 SCM_EXTERN ScmChar Scm_PeekcUnsafe(ScmPort *port);
 SCM_EXTERN int    Scm_PeekbUnsafe(ScmPort *port);
+SCM_EXTERN ScmObj Scm_UngottenCharsUnsafe(ScmPort *port);
+SCM_EXTERN ScmObj Scm_UngottenBytesUnsafe(ScmPort *port);
 
 SCM_EXTERN ScmObj Scm_ReadLine(ScmPort *port);
 SCM_EXTERN ScmObj Scm_ReadLineUnsafe(ScmPort *port);
@@ -412,22 +420,10 @@ SCM_EXTERN ScmObj Scm_SetCurrentErrorPort(ScmPort *port);
 SCM_EXTERN ScmObj Scm_MakeInputStringPort(ScmString *str, int privatep);
 SCM_EXTERN ScmObj Scm_MakeOutputStringPort(int privatep);
 
-#if !defined(GAUCHE_API_PRE_0_9)
 SCM_EXTERN ScmObj Scm_GetOutputString(ScmPort *port, int flags);
 SCM_EXTERN ScmObj Scm_GetOutputStringUnsafe(ScmPort *port, int flags);
 SCM_EXTERN ScmObj Scm_GetRemainingInputString(ScmPort *port, int flags);
 
-#else  /* GAUCHE_API_PRE_0_9 */
-#define Scm_GetOutputString(p) Scm__GetOutputStringCompat(p)
-#define Scm_GetOutputStringUnsafe(p) Scm__GetOutputStringUnsafeCompat(p)
-#define Scm_GetRemainingInputString(p) Scm__GetRemainingInputStringCompat(p)
-
-/* For backward compatibility */
-SCM_EXTERN ScmObj Scm__GetOutputStringCompat(ScmPort *port);
-SCM_EXTERN ScmObj Scm__GetOutputStringUnsafeCompat(ScmPort *port);
-SCM_EXTERN ScmObj Scm__GetRemainingInputStringCompat(ScmPort *port);
-#endif /* GAUCHE_API_PRE_0_9 */
-
 /*================================================================
  * Other type of ports
  */
