@@ -1,7 +1,7 @@
 /*
  * reader.h - Reader API
  *
- *   Copyright (c) 2000-2012  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -36,29 +36,34 @@
 #ifndef GAUCHE_READER_H
 #define GAUCHE_READER_H
 
-/*
- * ReadContext
- */
-typedef struct ScmReadContextRec {
-    SCM_HEADER;
-    int flags;                  /* see below */
-    ScmHashTable *table;        /* used internally. */
-    ScmObj pending;             /* used internally. */
-} ScmReadContext;
-
-enum ScmReadContextFlags {
-    SCM_READ_SOURCE_INFO = (1L<<0),  /* preserving souce file information */
-    SCM_READ_LITERAL_IMMUTABLE = (1L<<1), /* literal should be read as immutable */
-    SCM_READ_DISABLE_CTOR = (1L<<2), /* disable #,() */
-    SCM_READ_RECURSIVELY = (1L<<3)   /* used internally. */
-};
-
 SCM_CLASS_DECL(Scm_ReadContextClass);
 #define SCM_CLASS_READ_CONTEXT   (&Scm_ReadContextClass)
 #define SCM_READ_CONTEXT(obj)    ((ScmReadContext*)(obj))
 #define SCM_READ_CONTEXT_P(obj)  SCM_XTYPEP(obj, SCM_CLASS_READ_CONTEXT)
 
-ScmReadContext *Scm_MakeReadContext(ScmReadContext *proto);
+SCM_EXTERN ScmReadContext *Scm_MakeReadContext(ScmReadContext *proto);
+SCM_EXTERN int             Scm_ReadContextLiteralImmutable(ScmReadContext *);
+SCM_EXTERN ScmReadContext *Scm_CurrentReadContext();
+SCM_EXTERN ScmReadContext *Scm_SetCurrentReadContext(ScmReadContext *ctx);
+
+/* Reader-lexical-mode
+    permissive
+    legacy
+    warn-legacy
+    strict-r7
+   This is kept in a parameter, so it's thread-local.
+
+   NB: The actual implementation is in port.c, because ports require
+   this working at initialization, but reader is initialized after ports.
+   It is debatable which module these should belong to.  Reader *always*
+   reference port's reader lexical mode, and the global reader lexical
+   mode parameter is only used to initialize individual ports.  So these
+   can be in port's API.  But in that case, we may want other name
+   (e.g. DefaultReaderLexicalMode).
+ */
+SCM_EXTERN ScmObj Scm_ReaderLexicalMode(void);
+/* returns previous setting */
+SCM_EXTERN ScmObj Scm_SetReaderLexicalMode(ScmObj);
 
 /* An object to keep unrealized circular reference (e.g. #N=) during
  * 'read'.  It is replaced by the reference value before exitting 'read',
@@ -90,6 +95,24 @@ SCM_EXTERN ScmObj Scm_ReadFromCString(const char *string);
 
 SCM_EXTERN void   Scm_ReadError(ScmPort *port, const char *fmt, ...);
 
+/*
+ * Common utilities to handle hex-digit escapes
+ */
+SCM_EXTERN ScmChar Scm_ReadXdigitsFromString(const char *buf,
+                                             int buflen,
+                                             ScmChar key,
+                                             ScmObj mode,
+                                             int terminator,
+                                             const char**);
+SCM_EXTERN ScmObj  Scm_ReadXdigitsFromPort(ScmPort *port, int key, ScmObj mode,
+                                           int incompletep, ScmDString *buf);
+
+SCM_EXTERN long Scm_ReadDigitsAsLong(ScmPort *port, ScmChar ch, int radix,
+                                     ScmChar *next,
+                                     int *numread);
+SCM_EXTERN long Scm_ParseDigitsAsLong(const char *buf, size_t len, int radix,
+                                      int *numread);
+
 /*
  * SRFI-10 hash-comma syntax
  */
@@ -102,9 +125,5 @@ SCM_EXTERN ScmObj Scm_GetReaderCtor(ScmObj symbol, ScmObj module);
  */
 SCM_EXTERN ScmObj Scm_DefineReaderDirective(ScmObj symbol, ScmObj proc);
 
-/* Internal */
-SCM_EXTERN void   Scm__InstallReadUvectorHook(ScmObj (*)(ScmPort*, const char*, ScmReadContext*));
-
-
 #endif  /*GAUCHE_READER_H*/
 
