@@ -1,7 +1,7 @@
 ;;;
 ;;; libalpha.scm - Procedures needed by other lib*.scm
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -193,7 +193,7 @@
      (set! (-> packet min_reqargs) min-reqargs
            (-> packet max_optargs) max-optargs
            (-> packet dispatch_vector) dispatch-vector)
-     (result (Scm_MakeSubr case_lambda_dispatch packet
+     (return (Scm_MakeSubr case_lambda_dispatch packet
                            min_reqargs max_optargs
                            (SCM_LIST3 (?: (SCM_FALSEP name)
                                          'case-lambda-dispatcher
@@ -207,7 +207,6 @@
 (define-in-module gauche.internal (make-case-lambda minarg maxarg
                                                     formals closures
                                                     :optional (name #f))
-
   (define (fill-dispatch-vector! v formals closure)
     (define (%set n)
       (let1 i (- n minarg)
@@ -227,3 +226,26 @@
         (loop (cdr fs) (cdr cs))))
     ((with-module gauche.internal make-case-lambda-dispatcher) v minarg name)))
 
+;; Returns ((<required args> <optional arg> <procedure>) ...)
+;; This also depends on the structure of dispatch vector info.
+;; Programs that needs to deal with case-lambda should use this procedure
+;; instead of directly interpret dispatch vector info.
+(define-in-module gauche (case-lambda-info proc)
+  (and (subr? proc)
+       (let1 info (procedure-info proc)
+         (and (pair? info)
+              (pair? (cdr info))
+              (integer? (cadr info))
+              (pair? (cddr info))
+              (vector? (caddr info))
+              (let* ([min-args (cadr info)]
+                     [dispatch-vec (caddr info)]
+                     [len (vector-length dispatch-vec)])
+                (let loop ([r '()] [i 0])
+                  (if (= i (- len 1))
+                    (if-let1 optproc (vector-ref dispatch-vec i)
+                      (reverse r `((,(+ min-args i -1) #t ,optproc)))
+                      (reverse r))
+                    (if-let1 proc (vector-ref dispatch-vec i)
+                      (loop (cons `(,(+ min-args i) #f ,proc) r) (+ i 1))
+                      (loop r (+ i 1))))))))))
