@@ -1,7 +1,7 @@
 ;;;
 ;;; libbool.scm - builtin boolean/comparison procedures
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -60,7 +60,7 @@
 (select-module gauche)
 ;; a convenient coercer
 (define-cproc boolean (obj) ::<boolean> :constant
-  (result (not (SCM_FALSEP obj))))
+  (return (not (SCM_FALSEP obj))))
 
 ;; R7RS
 (define (boolean=? a b . args)
