@@ -1,7 +1,7 @@
 ;;;
 ;;; libdict.scm - dictionary-like data structures (hashtables etc.)
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -46,7 +46,7 @@
    [(_ dict referencer)
     `(let* ([v (,referencer ,dict key fallback)])
        (dict-check-entry ,dict key (SCM_UNBOUNDP v))
-       (result v))])
+       (return v))])
 
  (define-cise-expr dict-exists?
    [(_ dict referencer)
@@ -67,7 +67,7 @@
                 (cast void (SCM_DICT_SET_VALUE e fallback)))])
        (set! (aref data 0) (cast void* e))
        (Scm_VMPushCC ,cc data 1)
-       (result (Scm_VMApply1 proc (SCM_DICT_VALUE e))))])
+       (return (Scm_VMApply1 proc (SCM_DICT_VALUE e))))])
 
  (define-cise-stmt dict-push!
    [(_ dict searcher xtractor)
@@ -85,15 +85,16 @@
        (cond
         [(not e)
          (dict-check-entry ,dict key (SCM_UNBOUNDP fallback))
-         (result fallback)]
+         (return fallback)]
         [(not (SCM_PAIRP (SCM_DICT_VALUE e)))
          (when (SCM_UNBOUNDP fallback)
            (Scm_Error "%S's value for key %S is not a pair: %S"
                       ,dict key (SCM_DICT_VALUE e)))
-         (result fallback)]
+         (return fallback)]
         [else
-         (result (SCM_CAR (SCM_DICT_VALUE e)))
-         (cast void (SCM_DICT_SET_VALUE e (SCM_CDR (SCM_DICT_VALUE e))))]))])
+         (let* ([resultval_ (SCM_CAR (SCM_DICT_VALUE e))])
+           (cast void (SCM_DICT_SET_VALUE e (SCM_CDR (SCM_DICT_VALUE e))))
+           (return resultval_))]))])
  )
 
 ;;;
@@ -102,6 +103,8 @@
 
 (select-module gauche)
 (inline-stub
+ ;; this is called only for simple hash types, so don't need to handle
+ ;; SCM_HASH_GENEAL.
  (define-cise-stmt set-hash-type!
    [(_ cvar scmvar)
     `(cond [(SCM_EQ ,scmvar 'eq?)      (set! ,cvar SCM_HASH_EQ)]
@@ -113,29 +116,151 @@
  (define-cise-stmt get-hash-type
    [(_ expr)
     `(case ,expr
-       [(SCM_HASH_EQ)      (result 'eq?)]
-       [(SCM_HASH_EQV)     (result 'eqv?)]
-       [(SCM_HASH_EQUAL)   (result 'equal?)]
-       [(SCM_HASH_STRING)  (result 'string=?)]
-       [else (result '#f)]           ; TODO: need to think over
+       [(SCM_HASH_EQ)      (return 'eq?)]
+       [(SCM_HASH_EQV)     (return 'eqv?)]
+       [(SCM_HASH_EQUAL)   (return 'equal?)]
+       [(SCM_HASH_STRING)  (return 'string=?)]
+       [(SCM_HASH_GENERAL) (return 'general)]
+       [else (return '#f)]           ; TODO: need to think over
        )])
  )
 
+(select-module gauche.internal)
+(define-cproc %current-recursive-hash (:optional obj) Scm_CurrentRecursiveHash)
+  
+(select-module gauche)
+(define-cproc hash-salt () ::<fixnum> Scm_HashSaltRef)
+
 (define-cproc eq-hash (obj)  ::<ulong> :fast-flonum Scm_EqHash)
 (define-cproc eqv-hash (obj) ::<ulong> :fast-flonum Scm_EqvHash)
-(define-cproc hash (obj)     ::<ulong> :fast-flonum Scm_Hash)
+(define-cproc legacy-hash (obj) ::<ulong> :fast-flonum Scm_Hash)
+(define-cproc portable-hash (obj salt::<fixnum>) ::<ulong>
+  :fast-flonum Scm_PortableHash)
+(define-cproc default-hash (obj) ::<fixnum> :fast-flonum Scm_DefaultHash)
+(define-cproc combine-hash-value (a::<ulong> b::<ulong>) ::<ulong>
+  Scm_CombineHashValue)
+
+;; see object-hash in libomega.scm for the reason of this implementation
+(define (hash obj)
+  (let1 h ((with-module gauche.internal %current-recursive-hash))
+    (if h (h obj) (legacy-hash obj))))
+
 (define-cproc hash-table? (obj) ::<boolean> :fast-flonum SCM_HASH_TABLE_P)
 
-(define-cproc make-hash-table (:optional (type eq?) (init-size::<int> 0))
+(define-cproc %make-hash-table-simple (type init-size::<int>)
   (let* ([ctype::int 0])
     (set-hash-type! ctype type)
-    (result (Scm_MakeHashTableSimple ctype init-size))))
+    (return (Scm_MakeHashTableSimple ctype init-size))))
+
+(inline-stub
+(define-cfn generic-hashtable-hash (h::(const ScmHashCore*) key::intptr_t)
+  ::u_long :static
+  (let* ([c::ScmComparator* (cast ScmComparator* (-> h data))]
+         [v::ScmObj (Scm_ApplyRec1 (Scm_ComparatorHashFunction c)
+                                   (SCM_OBJ key))])
+    (unless (or (SCM_INTP v) (SCM_BIGNUMP v))
+      (Scm_Error "Comparator %S's hash function should return \
+                  an exact integer, but got: %S" c v))
+    (return (Scm_GetIntegerUMod v))))
+
+(define-cfn generic-hashtable-hash-typecheck (h::(const ScmHashCore*)
+                                              key::intptr_t)
+  ::u_long :static
+  (let* ([c::ScmComparator* (cast ScmComparator* (-> h data))]
+         [t::ScmObj (Scm_ApplyRec1 (-> c typeFn) (SCM_OBJ key))])
+    (when (SCM_FALSEP t)
+      (Scm_Error "Invalid key for hashtable: %S" (SCM_OBJ key)))
+    (return (generic-hashtable-hash h key))))
+
+(define-cfn generic-hashtable-eq (h::(const ScmHashCore*)
+                                  a::intptr_t b::intptr_t)
+  ::int :static
+  (let* ([c::ScmComparator* (cast ScmComparator* (-> h data))]
+         [e::ScmObj (Scm_ApplyRec2 (-> c eqFn) (SCM_OBJ a) (SCM_OBJ b))])
+    (return (not (SCM_FALSEP e)))))
+
+(define-cfn generic-hashtable-eq-typecheck (h::(const ScmHashCore*)
+                                            a::intptr_t b::intptr_t)
+  ::int :static
+  (let* ([c::ScmComparator* (cast ScmComparator* (-> h data))]
+         [t::ScmObj (Scm_ApplyRec1 (-> c typeFn) (SCM_OBJ a))])
+    ;; NB: a is the key given from outside, and b is the key that's already
+    ;; in the table, so we only need to check a.
+    ;; TODO: Currently we may perform typecheck of a multiple times if
+    ;; we have more than one items with the same hash value in the table;
+    ;; optimization required.
+    (when (SCM_FALSEP t)
+      (Scm_Error "Invalid key for hashtable: %S" (SCM_OBJ a)))
+    (return (generic-hashtable-eq h a b))))
+)
+
+(define-cproc %make-hash-table-from-comparator (comparator::<comparator>
+                                                init-size::<int>
+                                                has-type-check::<boolean>)
+  (if has-type-check
+    (return (Scm_MakeHashTableFull generic-hashtable-hash-typecheck
+                                   generic-hashtable-eq-typecheck
+                                   init-size
+                                   comparator))
+    (return (Scm_MakeHashTableFull generic-hashtable-hash
+                                   generic-hashtable-eq
+                                   init-size
+                                   comparator))))
+
+;; Comparator argument can be <comparator> or one of the symbols
+;; eq?, eqv?, equal? or string=?.
+(define (make-hash-table :optional (comparator 'eq?) (init-size 0))
+  (case comparator
+    [(eq? eqv? equal? string=?)
+     (%make-hash-table-simple comparator init-size)]
+    [else
+     (unless (comparator? comparator)
+       (error "make-hash-table requires a comparator or \
+               one of the symbols in eq?, eqv?, equal? or string=?, but got:"
+              comparator))
+     (cond
+      [(eq? comparator eq-comparator)
+       (make-hash-table 'eq? init-size)]
+      [(eq? comparator eqv-comparator)
+       (make-hash-table 'eqv? init-size)]
+      [(eq? comparator equal-comparator)
+       (make-hash-table 'equal? init-size)]
+      [(eq? comparator string-comparator)
+       (make-hash-table 'string=? init-size)]
+      [else
+       (unless (comparator-hashable? comparator)
+         (error "make-hash-table requires a comparator with hash function, \
+                 but got:" comparator))
+       ($ %make-hash-table-from-comparator
+          comparator init-size
+          (not (eq? (comparator-type-test-predicate comparator)
+                    (with-module gauche.internal default-type-test))))])]))
 
 (define-cproc hash-table-type (hash::<hash-table>)
   (get-hash-type (-> hash type)))
 
+(select-module gauche.internal)
+(define-cproc %hash-table-comparator-int (hash::<hash-table>)
+  (if (== (Scm_HashTableType hash) SCM_HASH_GENERAL)
+    (begin
+      (let* ([r (SCM_OBJ (-> (SCM_HASH_TABLE_CORE hash) data))])
+        (unless (SCM_COMPARATORP r)
+          (Scm_Error "Got some weird hashtable - possibly internal bug: %S"
+                     hash))
+        (return r)))
+    (return '#f)))
+(select-module gauche)
+(define (hash-table-comparator hash)
+  (case (hash-table-type hash)
+    [(eq?) eq-comparator]
+    [(eqv?) eqv-comparator]
+    [(equal?) equal-comparator]
+    [(string=?) string-comparator]
+    [(general) ((with-module gauche.internal %hash-table-comparator-int) hash)]
+    [else (error "unknown hashtable type:" hash)]))
+
 (define-cproc hash-table-num-entries (hash::<hash-table>) ::<int>
-  (result (Scm_HashCoreNumEntries (SCM_HASH_TABLE_CORE hash))))
+  (return (Scm_HashCoreNumEntries (SCM_HASH_TABLE_CORE hash))))
 
 (define-cproc hash-table-clear! (hash::<hash-table>) ::<void>
   (Scm_HashCoreClear (SCM_HASH_TABLE_CORE hash)))
@@ -149,10 +274,10 @@
 ;; this is hash-table-remove! in STk.  I use `delete' for
 ;; it's consistent with SRFI-1 and dbm-delete!.
 (define-cproc hash-table-delete! (hash::<hash-table> key) ::<boolean>
-  (result (not (SCM_UNBOUNDP (Scm_HashTableDelete hash key)))))
+  (return (not (SCM_UNBOUNDP (Scm_HashTableDelete hash key)))))
 
 (define-cproc hash-table-exists? (hash::<hash-table> key) ::<boolean>
-  (result (dict-exists? hash Scm_HashTableRef)))
+  (return (dict-exists? hash Scm_HashTableRef)))
 
 (inline-stub
  (define-cfn hash-table-update-cc (result (data :: void**)) :static
@@ -185,7 +310,7 @@
 (define-cproc %hash-table-iter (hash::<hash-table>)
   (let* ([iter::ScmHashIter* (SCM_NEW ScmHashIter)])
     (Scm_HashIterInit iter (SCM_HASH_TABLE_CORE hash))
-    (result (Scm_MakeSubr hash_table_iter iter 1 0 '"hash-table-iterator"))))
+    (return (Scm_MakeSubr hash_table_iter iter 1 0 '"hash-table-iterator"))))
 
 (define-cproc hash-table-copy (hash::<hash-table>)   Scm_HashTableCopy)
 (define-cproc hash-table-keys (hash::<hash-table>)   Scm_HashTableKeys)
@@ -193,8 +318,8 @@
 (define-cproc hash-table-stat (hash::<hash-table>)   Scm_HashTableStat)
 
 ;; conversion to/from hash-table
-(define (alist->hash-table a . opt-eq)
-  (rlet1 tb (apply make-hash-table opt-eq)
+(define (alist->hash-table a . opt-cmpr)
+  (rlet1 tb (apply make-hash-table opt-cmpr)
     (for-each (^x (hash-table-put! tb (car x) (cdr x))) a)))
 
 (define (hash-table->alist h)
@@ -206,15 +331,35 @@
 
 (select-module gauche)
 (inline-stub
+ ;; If tree-map-cmp is called, core->data should contain a comparator.
  (define-cfn tree-map-cmp (core::ScmTreeCore* x::intptr_t y::intptr_t)
    ::int :static
-   (let* ([cmp-proc (SCM_OBJ (-> core data))]
-          [r (Scm_ApplyRec2 cmp-proc (SCM_OBJ x) (SCM_OBJ y))])
-     (return (SCM_INT_VALUE r))))
+   (let* ([cmpr (SCM_OBJ (-> core data))])
+     (SCM_ASSERT (and cmpr (SCM_COMPARATORP cmpr)))
+     (let* ([r (Scm_ApplyRec2 (Scm_ComparatorComparisonProcedure
+                               (SCM_COMPARATOR cmpr))
+                              (SCM_OBJ x) (SCM_OBJ y))])
+       (unless (SCM_INTP r)
+         (Scm_Error "compare procedure of tree-map's comparator %S returned \
+                     non-integral value: %S" cmpr r))
+       (return (SCM_INT_VALUE r)))))
  )
 
-(define-cproc %make-tree-map (cmp-proc)
-  (result (Scm_MakeTreeMap tree_map_cmp cmp_proc)))
+(define-cproc %make-tree-map (comparator)
+  (begin
+    (SCM_ASSERT (SCM_COMPARATORP comparator))
+    (return (Scm_MakeTreeMap tree_map_cmp comparator))))
+
+;; TODO: We do want to return something even for tree-maps that aren't
+;; created from the Scheme world.  But how?
+(define-cproc tree-map-comparator (tm::<tree-map>)
+  (let* ([d::void* (-> (SCM_TREE_MAP_CORE tm) data)])
+    (if (or (== d NULL)
+            (!= (-> (SCM_TREE_MAP_CORE tm) cmp) tree-map-cmp))
+      (return SCM_FALSE)
+      (begin
+        (SCM_ASSERT (SCM_COMPARATORP d))
+        (return (SCM_OBJ d))))))
 
 (define-cproc tree-map-copy (tm::<tree-map>) Scm_TreeMapCopy)
 
@@ -227,7 +372,7 @@
   (Scm_TreeMapSet tm key val 0))
 
 (define-cproc tree-map-delete! (tm::<tree-map> key) ::<boolean>
-  (result (not (SCM_UNBOUNDP (Scm_TreeMapDelete tm key)))))
+  (return (not (SCM_UNBOUNDP (Scm_TreeMapDelete tm key)))))
 
 (inline-stub
  (define-cfn tree-map-update-cc (result data::void**) :static
@@ -246,10 +391,10 @@
   (dict-pop! tm Scm_TreeCoreSearch SCM_TREE_MAP_CORE))
 
 (define-cproc tree-map-exists? (tm::<tree-map> key) ::<boolean>
-  (result (dict-exists? tm Scm_TreeMapRef)))
+  (return (dict-exists? tm Scm_TreeMapRef)))
 
 (define-cproc tree-map-num-entries (tm::<tree-map>) ::<int>
-  (result (Scm_TreeCoreNumEntries (SCM_TREE_MAP_CORE tm))))
+  (return (Scm_TreeCoreNumEntries (SCM_TREE_MAP_CORE tm))))
 
 (define-cproc %tree-map-bound (tm::<tree-map> min::<boolean> pop::<boolean>)
   (let* ([op::ScmTreeCoreBoundOp (?: min SCM_TREE_CORE_MIN SCM_TREE_CORE_MAX)]
@@ -258,8 +403,8 @@
               (Scm_TreeCorePopBound (SCM_TREE_MAP_CORE tm) op)
               (Scm_TreeCoreGetBound (SCM_TREE_MAP_CORE tm) op))])
     (if e
-      (result (Scm_Cons (SCM_DICT_KEY e) (SCM_DICT_VALUE e)))
-      (result '#f))))
+      (return (Scm_Cons (SCM_DICT_KEY e) (SCM_DICT_VALUE e)))
+      (return '#f))))
 
 (inline-stub
  (define-cfn tree-map-iter (args::ScmObj* nargs::int data::void*) :static
@@ -275,11 +420,11 @@
 (define-cproc %tree-map-iter (tm::<tree-map>)
   (let* ([iter::ScmTreeIter* (SCM_NEW ScmTreeIter)])
     (Scm_TreeIterInit iter (SCM_TREE_MAP_CORE tm) NULL)
-    (result (Scm_MakeSubr tree_map_iter iter 2 0 '"tree-map-iterator"))))
+    (return (Scm_MakeSubr tree_map_iter iter 2 0 '"tree-map-iterator"))))
 
 (define-cproc %tree-map-check-consistency (tm::<tree-map>)
   (Scm_TreeCoreCheckConsistency (SCM_TREE_MAP_CORE tm))
-  (result '#t))
+  (return '#t))
 
 (define-cproc %tree-map-dump (tm::<tree-map>) ::<void>
   (Scm_TreeMapDump tm SCM_CUROUT))
@@ -306,9 +451,9 @@
         [(!= ,lh NULL) (,make-result (SCM_DICT_KEY ,lh) (SCM_DICT_VALUE ,lh))]
         [else          (,make-result key-fb val-fb)]))])
 
- (define-cise-stmt tree-map-closest-key-result [(_ k v) `(result ,k)])
- (define-cise-stmt tree-map-closest-val-result [(_ k v) `(result ,v)])
- (define-cise-stmt tree-map-closest-kv-result  [(_ k v) `(result ,k ,v)])
+ (define-cise-stmt tree-map-closest-key-result [(_ k v) `(return ,k)])
+ (define-cise-stmt tree-map-closest-val-result [(_ k v) `(return ,v)])
+ (define-cise-stmt tree-map-closest-kv-result  [(_ k v) `(return ,k ,v)])
  )
 
 (define-cproc tree-map-floor
@@ -346,6 +491,3 @@
 (define-cproc tree-map-successor-value
   (tm::<tree-map> key :optional (val-fb #f))
   (tree-map-closest-entry #f hi tree-map-closest-val-result))
-
-
-
