@@ -1,7 +1,7 @@
 ;;;
 ;;; libio.scm - builtin port and I/O procedures
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -35,7 +35,9 @@
 
 (inline-stub
  (declcode (.include <gauche/vminsn.h>
+                     <gauche/class.h>
                      <gauche/priv/portP.h>
+                     <gauche/priv/writerP.h>
                      <stdlib.h>
                      <fcntl.h>)))
 
@@ -63,36 +65,36 @@
 
 (define-cproc current-input-port (:optional newport)
   (cond [(SCM_IPORTP newport)
-         (result (Scm_SetCurrentInputPort (SCM_PORT newport)))]
+         (return (Scm_SetCurrentInputPort (SCM_PORT newport)))]
         [(not (SCM_UNBOUNDP newport))
          (Scm_TypeError "current-input-port" "input port" newport)
-         (result SCM_UNDEFINED)]
-        [else (result (SCM_OBJ SCM_CURIN))]))
+         (return SCM_UNDEFINED)]
+        [else (return (SCM_OBJ SCM_CURIN))]))
 
 (define-cproc current-output-port (:optional newport)
   (cond [(SCM_OPORTP newport)
-         (result (Scm_SetCurrentOutputPort (SCM_PORT newport)))]
+         (return (Scm_SetCurrentOutputPort (SCM_PORT newport)))]
         [(not (SCM_UNBOUNDP newport))
          (Scm_TypeError "current-output-port" "output port" newport)
-         (result SCM_UNDEFINED)]
-        [else (result (SCM_OBJ SCM_CUROUT))]))
+         (return SCM_UNDEFINED)]
+        [else (return (SCM_OBJ SCM_CUROUT))]))
 
 (select-module gauche)
 
 (define-cproc current-error-port (:optional newport)
   (cond
-   [(SCM_OPORTP newport) (result (Scm_SetCurrentErrorPort (SCM_PORT newport)))]
+   [(SCM_OPORTP newport) (return (Scm_SetCurrentErrorPort (SCM_PORT newport)))]
    [(not (SCM_UNBOUNDP newport))
     (Scm_TypeError "current-error-port" "output port" newport)
-    (result SCM_UNDEFINED)]
-   [else (result (SCM_OBJ SCM_CURERR))]))
+    (return SCM_UNDEFINED)]
+   [else (return (SCM_OBJ SCM_CURERR))]))
 
 (define-cproc standard-input-port (:optional (p::<input-port>? #f))
-  (result (?: p (Scm_SetStdin p) (Scm_Stdin))))
+  (return (?: p (Scm_SetStdin p) (Scm_Stdin))))
 (define-cproc standard-output-port (:optional (p::<output-port>? #f))
-  (result (?: p (Scm_SetStdout p) (Scm_Stdout))))
+  (return (?: p (Scm_SetStdout p) (Scm_Stdout))))
 (define-cproc standard-error-port (:optional (p::<output-port>? #f))
-  (result (?: p (Scm_SetStderr p) (Scm_Stderr))))
+  (return (?: p (Scm_SetStderr p) (Scm_Stderr))))
 
 
 ;;
@@ -106,7 +108,7 @@
 
 (define-cproc port-file-number (port::<port>)
   (let* ([i::int (Scm_PortFileNo port)])
-    (result (?: (< i 0) SCM_FALSE (Scm_MakeInteger i)))))
+    (return (?: (< i 0) SCM_FALSE (Scm_MakeInteger i)))))
 (define-cproc port-fd-dup! (dst::<port> src::<port>) ::<void> Scm_PortFdDup)
 
 (define-cproc port-attribute-set! (port::<port> key val)
@@ -125,10 +127,10 @@
 
 (define-cproc port-type (port::<port>)
   (case (SCM_PORT_TYPE port)
-    [(SCM_PORT_FILE) (result 'file)]
-    [(SCM_PORT_PROC) (result 'proc)]
-    [(SCM_PORT_OSTR SCM_PORT_ISTR) (result 'string)]
-    [else (result '#f)]))
+    [(SCM_PORT_FILE) (return 'file)]
+    [(SCM_PORT_PROC) (return 'proc)]
+    [(SCM_PORT_OSTR SCM_PORT_ISTR) (return 'string)]
+    [else (return '#f)]))
 
 (define-cproc port-buffering (port::<port>)
   (setter (port::<port> mode) ::<void>
@@ -136,7 +138,7 @@
             (Scm_Error "can't set buffering mode to non-buffered port: %S"port))
           (Scm_SetPortBufferingMode
            port (Scm_BufferingMode mode (-> port direction) -1)))
-  (result (Scm_GetPortBufferingModeAsKeyword port)))
+  (return (Scm_GetPortBufferingModeAsKeyword port)))
 
 (define-cproc port-case-fold-set! (port::<port> flag::<boolean>) ::<void>
   (if flag
@@ -195,7 +197,7 @@
                                 O_RDONLY bufmode 0)])
       (when (and (SCM_FALSEP o) (not (%open/allow-noexist? ignerr)))
         (Scm_SysError "couldn't open input file: %S" path))
-      (result o))))
+      (return o))))
 
 ;; Primitive open routine.  The Scheme wrapper handles other keyword args
 (define-cproc %open-output-file (path::<string>
@@ -234,7 +236,7 @@
                  (not (%open/allow-noexist? ignerr-noexist))
                  (not (%open/allow-exist? ignerr-exist)))
         (Scm_Error "couldn't open output file: %S" path))
-      (result o))))
+      (return o))))
 
 ;; Open port from fd
 (select-module gauche)
@@ -245,8 +247,8 @@
                                   (name #f))
   (let* ([bufmode::int (Scm_BufferingMode buffering SCM_PORT_INPUT
                                           SCM_PORT_BUFFER_FULL)])
-    (when (< fd 0) (Scm_Error "bad file descriptor: %d" fd))
-    (result (Scm_MakePortWithFd name SCM_PORT_INPUT fd bufmode ownerP))))
+    (when (< fd 0) (Scm_Error "bad file descriptor: %ld" fd))
+    (return (Scm_MakePortWithFd name SCM_PORT_INPUT fd bufmode ownerP))))
 
 (define-cproc open-output-fd-port (fd::<fixnum>
                                    :key (buffering #f)
@@ -255,7 +257,7 @@
   (let* ([bufmode::int (Scm_BufferingMode buffering SCM_PORT_OUTPUT
                                           SCM_PORT_BUFFER_FULL)])
     (when (< fd 0) (Scm_Error "bad file descriptor: %d" fd))
-    (result (Scm_MakePortWithFd name SCM_PORT_OUTPUT fd bufmode owner?))))
+    (return (Scm_MakePortWithFd name SCM_PORT_OUTPUT fd bufmode owner?))))
 
 ;; Buffered port
 (select-module gauche)
@@ -299,7 +301,7 @@
           (ref bufrec ready)   NULL
           (ref bufrec filenum) NULL
           (ref bufrec data)    (cast void* filler))
-    (result (Scm_MakeBufferedPort SCM_CLASS_PORT SCM_FALSE SCM_PORT_INPUT TRUE (& bufrec)))))
+    (return (Scm_MakeBufferedPort SCM_CLASS_PORT SCM_FALSE SCM_PORT_INPUT TRUE (& bufrec)))))
 
 (inline-stub
  (define-cfn bufport-flusher (p::ScmPort* cnt::int forcep::int) ::int :static
@@ -322,7 +324,7 @@
           (ref bufrec ready)   NULL
           (ref bufrec filenum) NULL
           (ref bufrec data)    (cast void* flusher))
-    (result (Scm_MakeBufferedPort SCM_CLASS_PORT SCM_FALSE SCM_PORT_OUTPUT
+    (return (Scm_MakeBufferedPort SCM_CLASS_PORT SCM_FALSE SCM_PORT_OUTPUT
                                   TRUE (& bufrec)))))
 
 ;; String ports (srfi-6)
@@ -335,13 +337,13 @@
   Scm_MakeOutputStringPort)
 
 (define-cproc get-output-string (oport::<output-port>) ;SRFI-6
-  (result (Scm_GetOutputString oport 0)))
+  (return (Scm_GetOutputString oport 0)))
 
 (define-cproc get-output-byte-string (oport::<output-port>)
-  (result (Scm_GetOutputString oport SCM_STRING_INCOMPLETE)))
+  (return (Scm_GetOutputString oport SCM_STRING_INCOMPLETE)))
 
 (define-cproc get-remaining-input-string (iport::<input-port>)
-  (result (Scm_GetRemainingInputString iport 0)))
+  (return (Scm_GetRemainingInputString iport 0)))
 
 ;; Coding aware port
 (select-module gauche)
@@ -393,10 +395,31 @@
             (logand= (-> port flags) (lognot SCM_PORT_WRITESS))))
   PORT_WRITESS_P)
 
-(define-cproc %port-recursive-context (port::<port>)
+(inline-stub
+ (define-cfn write_state_allocate (klass::ScmClass* initargs) :static
+   (return (SCM_OBJ (Scm_MakeWriteState NULL))))
+
+ (define-cfn write_state_print (obj port::ScmPort* ctx::ScmWriteContext*)
+   ::void :static
+   (Scm_Printf port "#<write-state %p>" obj))
+
+ (define-cclass <write-state>
+   "ScmWriteState*" "Scm_WriteStateClass"
+   ("Scm_TopClass")
+   ((shared-table   :type <hash-table>? :c-name "sharedTable")
+    (shared-counter :type <int> :c-name "sharedCounter"))
+   (allocator (c "write_state_allocate"))
+   (printer   (c "write_state_print")))
+ )
+
+(define-cproc %port-write-state (port::<port>)
   (setter (port::<port> obj) ::<void>
-          (set! (-> port recursiveContext) obj))
-  (result (-> port recursiveContext)))
+          (if (SCM_WRITE_STATE_P obj)
+            (set! (-> port writeState) (SCM_WRITE_STATE obj))
+            (set! (-> port writeState) NULL)))
+  (return (?: (-> port writeState)
+              (SCM_OBJ (-> port writeState))
+              SCM_FALSE)))
 
 (define-cproc %port-lock! (port::<port>) ::<void>
   (let* ([vm::ScmVM* (Scm_VM)])
@@ -421,16 +444,17 @@
   (unwind-protect
       (begin
         (%port-lock! port)
-        (when (%port-recursive-context port)
+        (when (%port-write-state port)
           (error "[internal] %with-2pass-setup called recursively on port:"
                  port))
-        (set! (%port-recursive-context port) (cons 0 (make-hash-table 'eq?)))
+        (set! (%port-write-state port)
+              (make <write-state> :shared-table (make-hash-table 'eq?)))
         (set! (%port-walking? port) #t)
         (apply walker args)
         (set! (%port-walking? port) #f)
         (apply emitter args))
     (set! (%port-walking? port) #f)
-    (set! (%port-recursive-context port) #f)
+    (set! (%port-write-state port) #f)
     (%port-unlock! port)))
 
 ;;;
@@ -440,18 +464,18 @@
 (select-module scheme)
 
 (define-cproc read (:optional (port::<input-port> (current-input-port)))
-  (result (Scm_Read (SCM_OBJ port))))
+  (return (Scm_Read (SCM_OBJ port))))
 
 (define-cproc read-char (:optional (port::<input-port> (current-input-port)))
   (inliner READ-CHAR)
   (let* ([ch::int])
     (SCM_GETC ch port)
-    (result (?: (== ch EOF) SCM_EOF (SCM_MAKE_CHAR ch)))))
+    (return (?: (== ch EOF) SCM_EOF (SCM_MAKE_CHAR ch)))))
 
 (define-cproc peek-char (:optional (port::<input-port> (current-input-port)))
   (inliner PEEK-CHAR)
   (let* ([ch::ScmChar (Scm_Peekc port)])
-    (result (?: (== ch SCM_CHAR_INVALID) SCM_EOF (SCM_MAKE_CHAR ch)))))
+    (return (?: (== ch SCM_CHAR_INVALID) SCM_EOF (SCM_MAKE_CHAR ch)))))
 
 (define-cproc eof-object? (obj) ::<boolean> :fast-flonum
   (inliner EOFP) SCM_EOFP)
@@ -462,7 +486,7 @@
 
 (select-module gauche)
 
-(define-cproc eof-object () :constant (result SCM_EOF)) ;R6RS
+(define-cproc eof-object () :constant (return SCM_EOF)) ;R6RS
 
 (define-cproc byte-ready? (:optional (port::<input-port> (current-input-port)))
   ::<boolean> Scm_ByteReady)
@@ -470,11 +494,11 @@
 (define-cproc read-byte (:optional (port::<input-port> (current-input-port)))
   (let* ([b::int])
     (SCM_GETB b port)
-    (result (?: (< b 0) SCM_EOF (SCM_MAKE_INT b)))))
+    (return (?: (< b 0) SCM_EOF (SCM_MAKE_INT b)))))
 
 (define-cproc peek-byte (:optional (port::<input-port> (current-input-port)))
   (let* ([b::int (Scm_Peekb port)])
-    (result (?: (< b 0) SCM_EOF (SCM_MAKE_INT b)))))
+    (return (?: (< b 0) SCM_EOF (SCM_MAKE_INT b)))))
 
 (define-cproc read-line (:optional (port::<input-port> (current-input-port))
                                    (allowbytestr #f))
@@ -483,7 +507,7 @@
                (SCM_STRINGP r)
                (SCM_STRING_INCOMPLETE_P r))
       (Scm_ReadError port "read-line: encountered illegal byte sequence: %S" r))
-    (result r)))
+    (return r)))
 
 (define (read-string n :optional (port (current-input-port)))
   (define o (open-output-string :private? #t))
@@ -497,25 +521,48 @@
             (get-output-string o))
           (begin (write-char c o) (loop (+ i 1))))))))
 
+;; Consume trailing whiespaces up to (including) first EOL.
+;; This is mainly intended for interactive REPL,
+;; where the input is buffered by line.  We want to ignore the
+;; trailing newline, so that when the user type (read-line) RET,
+;; we consume that RET and start reading from the fresh line.
+;
+;; We need to be careful not to block; that's why we use binary
+;; input here, since character input may block if the input stop
+;; between a multibyte character.
+;; Note that the 'whitespaces' here only inlucdes #\tab, #\space,
+;; #\return and #\newline.
+(define (consume-trailing-whitespaces :optional (port (current-input-port)))
+  (let loop ()
+    (when (byte-ready? port)
+      (let1 b (peek-byte port)
+        (cond [(memv b '(9 32)) (read-byte port) (loop)] ;tab, space
+              [(eqv? b 13)                               ;cr or crlf
+               (read-byte port)
+               (when (and (byte-ready? port)
+                          (eqv? (peek-byte port) 10))
+                 (read-byte port))]
+              [(eqv? b 10) (read-byte port)])))))        ;lf
+
 ;; DEPRECATED - read-uvector should be used
 (define-cproc read-block (bytes::<fixnum>
                           :optional (port::<input-port> (current-input-port)))
   (when (< bytes 0)
-    (Scm_Error "bytes must be non-negative integer: %d" bytes))
+    (Scm_Error "bytes must be non-negative integer: %ld" bytes))
   (if (== bytes 0)
-    (result (Scm_MakeString "" 0 0 0))
+    (return (Scm_MakeString "" 0 0 0))
     (let* ([buf::char* (SCM_NEW_ATOMIC2 (C: char*) (+ bytes 1))]
            [nread::int (Scm_Getz buf bytes port)])
-      (cond [(<= nread 0) (result SCM_EOF)]
+      (cond [(<= nread 0) (return SCM_EOF)]
             [else
              (SCM_ASSERT (<= nread bytes))
              (set! (aref buf nread) #\x00)
-             (result (Scm_MakeString buf nread nread SCM_STRING_INCOMPLETE))]
+             (return (Scm_MakeString buf nread nread SCM_STRING_INCOMPLETE))]
             ))))
 
 (define-cproc read-list (closer::<char>
                          :optional (port (current-input-port)))
-  (result (Scm_ReadList port closer)))
+  (return (Scm_ReadList port closer)))
 
 (define-cproc port->byte-string (port::<input-port>)
   (let* ([ds::ScmDString] [buf::(.array char (1024))])
@@ -523,22 +570,28 @@
     (loop (let* ([nbytes::int (Scm_Getz buf 1024 port)])
             (when (<= nbytes 0) (break))
             (Scm_DStringPutz (& ds) buf nbytes)))
-    (result (Scm_DStringGet (& ds) SCM_STRING_INCOMPLETE))))
+    (return (Scm_DStringGet (& ds) SCM_STRING_INCOMPLETE))))
 
 ;; Reader parameters
 (define-cproc reader-lexical-mode (:optional k)
   (if (SCM_UNBOUNDP k)
-    (result (Scm_ReaderLexicalMode))
-    (result (Scm_SetReaderLexicalMode k))))
+    (return (Scm_ReaderLexicalMode))
+    (return (Scm_SetReaderLexicalMode k))))
+
+(select-module gauche.internal)
+(define-cproc %port-ungotten-chars (port::<input-port>)
+  Scm_UngottenChars)
+(define-cproc %port-ungotten-bytes (port::<input-port>)
+  Scm_UngottenBytes)
 
 ;; Read time constructor (srfi-10)
 (select-module gauche)
 
 (define-cproc define-reader-ctor (symbol proc :optional (finisher #f))
-  (result (Scm_DefineReaderCtor symbol proc finisher SCM_FALSE)))
+  (return (Scm_DefineReaderCtor symbol proc finisher SCM_FALSE)))
 
 (define-cproc %get-reader-ctor (symbol)
-  (result (Scm_GetReaderCtor symbol SCM_FALSE)))
+  (return (Scm_GetReaderCtor symbol SCM_FALSE)))
 
 (define-cproc define-reader-directive (symbol proc)
   Scm_DefineReaderDirective)
@@ -551,15 +604,22 @@
    "SCM_READ_REFERENCE_P" "SCM_READ_REFERENCE" "")
  )
 
+(define-cproc current-read-context (:optional ctx)
+  (if (SCM_UNBOUNDP ctx)
+    (return (SCM_OBJ (Scm_CurrentReadContext)))
+    (if (SCM_READ_CONTEXT_P ctx)
+      (return (SCM_OBJ (Scm_SetCurrentReadContext (SCM_READ_CONTEXT ctx))))
+      (Scm_Error "<read-context> required, but got:" ctx))))
+
 (define-cproc read-reference? (obj) ::<boolean> SCM_READ_REFERENCE_P)
 
 (define-cproc read-reference-has-value? (ref::<read-reference>)
-  ::<boolean> (result (not (SCM_UNBOUNDP (-> ref value)))))
+  ::<boolean> (return (not (SCM_UNBOUNDP (-> ref value)))))
 
 (define-cproc read-reference-value (ref::<read-reference>)
   (when (SCM_UNBOUNDP (-> ref value))
     (Scm_Error "read reference hasn't been resolved"))
-  (result (-> ref value)))
+  (return (-> ref value)))
 
 ;; srfi-38
 (define-in-module gauche read-with-shared-structure read)
@@ -590,24 +650,58 @@
 ;;; Output
 ;;;
 
+(inline-stub
+ (define-type <write-controls> "ScmWriteControls*" "write controls"
+   "SCM_WRITE_CONTROLS_P" "SCM_WRITE_CONTROLS" "")
+
+ (define-cfn parse-write-optionals (opt1 opt2
+                                    pp::ScmPort**
+                                    pc::(const ScmWriteControls**))
+   ::void :static
+   (let* ([p::ScmPort* SCM_CUROUT]
+          [c::(const ScmWriteControls*) (Scm_DefaultWriteControls)])
+     (unless (SCM_UNBOUNDP opt1)
+       (cond [(SCM_PORTP opt1)
+              (set! p (SCM_PORT opt1))
+              (unless (SCM_UNBOUNDP opt2)
+                (if (SCM_WRITE_CONTROLS_P opt2)
+                  (set! c (SCM_WRITE_CONTROLS opt2))
+                  (Scm_Error "Expected write-controls, but got: %S" opt2)))]
+             [(SCM_WRITE_CONTROLS_P opt1)
+              (set! c (SCM_WRITE_CONTROLS opt1))
+              (unless (SCM_UNBOUNDP opt2)
+                (if (SCM_PORTP opt2)
+                  (set! p (SCM_PORT opt2))
+                  (Scm_Error "Expected port, but got: %S" opt2)))]
+             [else
+              (Scm_Error "Expected port or write-controls, but got: %S" opt1)]))
+     (set! (* pp) p)
+     (set! (* pc) c)))
+ )
 (select-module scheme)
 
-(define-cproc write (obj :optional (port::<output-port> (current-output-port)))
-  ::<void> (Scm_Write obj (SCM_OBJ port) SCM_WRITE_WRITE))
+(define-cproc write (obj :optional port-or-control-1 port-or-control-2)
+  ::<void>
+  (let* ([p::ScmPort*] [c::(const ScmWriteControls*)])
+    (parse-write-optionals port-or-control-1 port-or-control-2 (& p) (& c))
+    (Scm_WriteWithControls obj (SCM_OBJ p) SCM_WRITE_WRITE c)))
 
 (define-cproc write-simple (obj :optional (port::<output-port>
                                            (current-output-port)))
   ::<void>
   (Scm_Write obj (SCM_OBJ port) SCM_WRITE_SIMPLE))
 
-(define-cproc write-shared (obj :optional (port::<output-port>
-                                           (current-output-port)))
+(define-cproc write-shared (obj :optional port-or-control-1 port-or-control-2)
   ::<void>
-  (Scm_Write obj (SCM_OBJ port) SCM_WRITE_SHARED))
+  (let* ([p::ScmPort*] [c::(const ScmWriteControls*)])
+    (parse-write-optionals port-or-control-1 port-or-control-2 (& p) (& c))
+    (Scm_WriteWithControls obj (SCM_OBJ p) SCM_WRITE_SHARED c)))
 
-(define-cproc display
-  (obj :optional (port::<output-port> (current-output-port)))
-  ::<void> (Scm_Write obj (SCM_OBJ port) SCM_WRITE_DISPLAY))
+(define-cproc display (obj :optional port-or-control-1 port-or-control-2)
+  ::<void>
+  (let* ([p::ScmPort*] [c::(const ScmWriteControls*)])
+    (parse-write-optionals port-or-control-1 port-or-control-2 (& p) (& c))
+    (Scm_WriteWithControls obj (SCM_OBJ p) SCM_WRITE_DISPLAY c)))
 
 (define-cproc newline (:optional (port::<output-port> (current-output-port)))
   ::<void> (SCM_PUTC #\newline port))
@@ -623,13 +717,13 @@
                           :optional (port::<output-port> (current-output-port)))
   ::<int>
   (when (or (< byte 0) (> byte 255))
-    (Scm_Error "argument out of range: %d" byte))
+    (Scm_Error "argument out of range: %ld" byte))
   (SCM_PUTB byte port)
-  (result 1))
+  (return 1))
 
 (define-cproc write-limited (obj limit::<fixnum>
                                  :optional (port (current-output-port)))
-  ::<int> (result (Scm_WriteLimited obj port SCM_WRITE_WRITE limit)))
+  ::<int> (return (Scm_WriteLimited obj port SCM_WRITE_WRITE limit)))
 
 (define write* write-shared)
 
@@ -644,7 +738,7 @@
 (select-module gauche.internal)
 
 (define-cproc write-need-recurse? (obj) ::<boolean>
-  (result (not (or (not (SCM_PTRP obj))
+  (return (not (or (not (SCM_PTRP obj))
                    (SCM_NUMBERP obj)
                    (SCM_KEYWORDP obj)
                    (and (SCM_SYMBOLP obj) (SCM_SYMBOL_INTERNED obj)) 
@@ -652,8 +746,8 @@
                    (and (SCM_VECTORP obj) (== (SCM_VECTOR_SIZE obj) 0))))))
 
 (define (write-walk obj port)
-  (if-let1 ctx (%port-recursive-context port)
-    (%write-walk-rec obj port (cdr ctx))))
+  (if-let1 s (%port-write-state port)
+    (%write-walk-rec obj port (~ s shared-table))))
 
 (define (%write-walk-rec obj port tab)
   (when (write-need-recurse? obj)
@@ -687,6 +781,65 @@
 
 (define-in-module gauche (print . args) (for-each display args) (newline))
 
+;;
+;; Write controls
+;;  For performance reasons, we don't make them a srfi-39 parameters.
+;;
+
+(select-module gauche)
+(inline-stub
+ (define-cfn write_controls_allocate (klass::ScmClass* initargs) :static
+   (return (SCM_OBJ (Scm_MakeWriteControls NULL))))
+
+ ;; TODO: We want to treat <write-controls> as immutable structure, but
+ ;; define-cclass doesn't yet handle a slot that's immutable but allowing
+ ;; initialized by init-keywords.
+ (define-cclass <write-controls>
+   "ScmWriteControls*" "Scm_WriteControlsClass"
+   ("Scm_TopClass")
+   ((print-length :type <int>     :c-name "printLength"
+                  :getter "if (obj->printLength < 0) return SCM_FALSE; \
+                           else return SCM_MAKE_INT(obj->printLength);"
+                  :setter "if (SCM_INTP(value) && SCM_INT_VALUE(value) >= 0) \
+                             obj->printLength = SCM_INT_VALUE(value); \
+                           else obj->printLength = -1;")
+    (print-level  :type <int>     :c-name "printLevel"
+                  :getter "if (obj->printLevel < 0) return SCM_FALSE; \
+                           else return SCM_MAKE_INT(obj->printLevel);"
+                  :setter "if (SCM_INTP(value) && SCM_INT_VALUE(value) >= 0) \
+                             obj->printLevel = SCM_INT_VALUE(value); \
+                           else obj->printLevel = -1;")
+    (print-base   :type <int>     :c-name "printBase"
+                  :setter "if (SCM_INTP(value) && SCM_INT_VALUE(value) >= 2 \
+                               && SCM_INT_VALUE(value) <= 36) \
+                             obj->printBase = SCM_INT_VALUE(value); \
+                           else Scm_Error(\"print-base must be an integer \
+                                   between 2 and 36, but got: %S\", value);")
+    (print-radix  :type <boolean> :c-name "printRadix"
+                  :setter "obj->printRadix = !SCM_FALSEP(value);"))
+   (allocator (c "write_controls_allocate")))
+ ;; NB: Printer is defined in libobj.scm via write-object method
+ )
+
+;; The :key and :rest combination is to catch invalid keyword error
+(define (make-write-controls :key print-length print-level
+                                  print-base print-radix
+                             :rest args)
+  (apply make <write-controls> args))
+
+(define (write-controls-copy wc :key print-length print-level
+                                     print-base print-radix)
+  (let-syntax [(select
+                (syntax-rules ()
+                  [(_ k) (if (undefined? k)
+                           (slot-ref wc 'k)
+                           k)]))]
+    (make <write-controls>
+      :print-length (select print-length)
+      :print-level  (select print-level)
+      :print-base   (select print-base)
+      :print-radix  (select print-radix))))
+
 ;;;
 ;;; With-something
 ;;;
@@ -826,10 +979,14 @@
     (values)))
 
 (define-reader-directive 'gauche-legacy
-  (^[sym port ctx] (reader-lexical-mode 'legacy) (values)))
+  (^[sym port ctx]
+    (port-attribute-set! port 'reader-lexical-mode 'legacy)
+    (values)))
 
 (define-reader-directive 'r7rs
-  (^[sym port ctx] (reader-lexical-mode 'strict-r7) (values)))
+  (^[sym port ctx]
+    (port-attribute-set! port 'reader-lexical-mode 'strict-r7)
+    (values)))
 
 ;; HIGHLY EXPERIMENTAL
 (define-reader-directive 'c-expr
