@@ -1,7 +1,7 @@
 ;;;
 ;;; libmod.scm - modules, bindings, GLOCs
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -34,7 +34,9 @@
 (select-module gauche)
 
 (inline-stub
- (declcode (.include <gauche/vminsn.h>)))
+ (declcode (.include "gauche/vminsn.h")
+           (.include "gauche/class.h")
+           (.include "gauche/priv/moduleP.h")))
 
 ;;;
 ;;; Modules API
@@ -43,16 +45,16 @@
 (define-cproc module? (obj) ::<boolean> :constant SCM_MODULEP)
 
 (define-cproc module-name (mod::<module>)
-  (result (SCM_OBJ (-> (SCM_MODULE mod) name))))
+  (return (SCM_OBJ (-> (SCM_MODULE mod) name))))
 
-(define-cproc module-parents (mod::<module>) (result (-> mod parents)))
-(define-cproc module-precedence-list (mod::<module>) (result (-> mod mpl)))
-(define-cproc module-imports (mod::<module>) (result (-> mod imported)))
+(define-cproc module-parents (mod::<module>) (return (-> mod parents)))
+(define-cproc module-precedence-list (mod::<module>) (return (-> mod mpl)))
+(define-cproc module-imports (mod::<module>) (return (-> mod imported)))
 (define-cproc module-exports (mod::<module>) Scm_ModuleExports)
-(define-cproc module-table (mod::<module>) (result (SCM_OBJ (-> mod internal))))
+(define-cproc module-table (mod::<module>) (return (SCM_OBJ (-> mod internal))))
 
 (define-cproc find-module (name::<symbol>) ::<module>?
-  (result (Scm_FindModule name SCM_FIND_MODULE_QUIET)))
+  (return (Scm_FindModule name SCM_FIND_MODULE_QUIET)))
 
 (define-cproc all-modules () Scm_AllModules)
 
@@ -61,7 +63,7 @@
     (cond [(SCM_EQ if-exists ':error) (set! error_if_exists TRUE)]
           [(SCM_FALSEP if-exists)     (set! error_if_exists FALSE)]
           [else (Scm_TypeError ":if-exists" ":error or #f" if_exists)])
-    (result (Scm_MakeModule name error_if_exists))))
+    (return (Scm_MakeModule name error_if_exists))))
 
 ;; (use x.y.z) ==> (requrie "x/y/z") (import x.y.z)
 (define-macro (use module . options)
@@ -91,7 +93,7 @@
                     name)])))
   (with-input-from-string
    (cond [(symbol? name) (symbol->string name)]
-         [(identifier? name) (symbol->string (identifier-name name))]
+         [(identifier? name) (symbol->string (identifier->symbol name))]
          [else (error "symbol or identifier expected, but got:" name)])
    (^[] (let loop ([c (read-char)] [p '()] [ps '()])
           (cond
@@ -121,7 +123,7 @@
 
 ;; Global bindind access, public API
 (define-cproc global-variable-bound? (mod-or-name name::<symbol>) ::<boolean>
-  (result
+  (return
    (not (SCM_UNBOUNDP
          (Scm_GlobalVariableRef (get-module-from-mod-or-name mod-or-name)
                                 name 0)))))
@@ -136,7 +138,7 @@
     (when (SCM_UNBOUNDP r2)
       (Scm_Error "global variable %S is not bound in module %S"
                  name module))
-    (result r2)))
+    (return r2)))
 
 (define-in-module gauche (symbol-bound? name :optional (module #f)) ; Deprecated
   (global-variable-bound? module name))
@@ -155,6 +157,15 @@
       (logior= z SCM_BINDING_INLINABLE))
     (set! SCM_RESULT (SCM_OBJ (Scm_MakeBinding mod name value z)))))
 
+;; Insert binding as a syntactic keyword.  VALUE must be #<macro> or #<syntax>.
+;; Currently Gauche conflates toplevel variable bindings and syntax bindings,
+;; so this is effectively the same as %insert-binding; but we may change it
+;; in future, so use this API to bind syntax.
+(define (%insert-syntax-binding mod name value)
+  (unless (or (syntax? value) (macro? value))
+    (error "Syntax or macro object required, but got" value))
+  (%insert-binding mod name value))
+
 (define-cproc %hide-binding (mod::<module> name::<symbol>) ::<void>
   Scm_HideBinding)
 
@@ -176,6 +187,38 @@
 (define-cproc %import-modules (mod::<module> mods) ;deprecated
   Scm_ImportModules)
 
+;; sealing module.  not public (yet)
+(define-cproc %seal-module! (mod::<module>) ::<void>
+  Scm_ModuleSeal)
+
+(select-module gauche)
+(inline-stub
+ (define-cfn module-print (obj port::ScmPort* ctx::ScmWriteContext*)
+   ::void :static
+   (if (SCM_MODULEP (-> (SCM_MODULE obj) origin))
+     (Scm_Printf port "#<module %A$%A @%p>"
+                 (-> (SCM_MODULE obj) name)
+                 (-> (SCM_MODULE (-> (SCM_MODULE obj) origin)) name)
+                 obj)
+     (Scm_Printf port "#<module %A>" (-> (SCM_MODULE obj) name))))
+ 
+ (define-cclass <module>
+   "ScmModule*" "Scm_ModuleClass"
+   (c "SCM_CLASS_COLLECTION_CPL")
+   ((name :setter #f)
+    (mpl :setter #f)
+    (parents :setter #f)
+    (imports :c-name "imported" :setter #f)
+    (exports :c-spec "Scm_ModuleExports(obj)" :setter #f)
+    (export-all :type <boolean> :c-name "exportAll" :setter #f)
+    (table   :c-name "internal" :setter #f)
+    (depends :c-name "depended" :setter #f)
+    (origin :setter #f)
+    (prefix :setter #f)
+    (info))
+   (printer (c "module_print")))
+ )
+
 ;;;
 ;;; Universal import
 ;;;
@@ -228,33 +271,94 @@
 (define-cproc find-binding (mod::<module> name::<symbol>
                                           stay-in-module::<boolean>)
   ::<gloc>?
-  (result (Scm_FindBinding mod name
+  (return (Scm_FindBinding mod name
                            (?: stay_in_module SCM_BINDING_STAY_IN_MODULE 0))))
 
 ;; This small piece of code encapsulates the common procedure in
 ;; pass1/variable to find whether the variable reference is a constant
 ;; or not.
 (define-cproc find-const-binding (id::<identifier>)
-  (let* ([g::ScmGloc* (Scm_FindBinding (-> id module) (-> id name) 0)])
+  (let* ([g::ScmGloc* (Scm_IdentifierGlobalBinding id)])
     (cond [(or (not g)
                (not (SCM_GLOC_CONST_P g))
                (SCM_VM_COMPILER_FLAG_IS_SET (Scm_VM)
                                             SCM_COMPILE_NOINLINE_CONSTS))
-           (result SCM_FALSE)]
+           (return SCM_FALSE)]
           [else
-           (result (SCM_GLOC_GET g))])))
+           (return (SCM_GLOC_GET g))])))
 
 (define-cproc gloc-bound? (gloc::<gloc>) ::<boolean>
-  (result (not (SCM_UNBOUNDP (SCM_GLOC_GET gloc)))))
+  (return (not (SCM_UNBOUNDP (SCM_GLOC_GET gloc)))))
 (define-cproc gloc-ref (gloc::<gloc> :optional fallback)
   (let* ([v::ScmObj (SCM_GLOC_GET gloc)])
     (if (SCM_UNBOUNDP v)
       (begin
         (when (SCM_UNBOUNDP fallback)
           (Scm_Error "gloc %S doesn't have a value" (SCM_OBJ gloc)))
-        (result fallback))
-      (result v))))
+        (return fallback))
+      (return v))))
 (define-cproc gloc-set! (gloc::<gloc> value) SCM_GLOC_SET)
 (define-cproc gloc-const? (gloc::<gloc>) ::<boolean> Scm_GlocConstP)
 (define-cproc gloc-inlinable? (gloc::<gloc>) ::<boolean> Scm_GlocInlinableP)
 
+;;;
+;;; Identifier and binding
+;;;
+
+;; NB: Identifier procedures are in libsym.scm, for we might integrate
+;; symbols and identifiers (we're still not sure).
+
+;; Returns GLOC if id is bound to one, or #f.  If GLOC is returned,
+;; it is always bound.
+
+;; (define (id->bound-gloc id)
+;;   (and-let* ([gloc (find-binding (identifier-module id)
+;;                                  (identifier-name id) #f)]
+;;              [ (gloc-bound? gloc) ])
+;;     gloc))
+(inline-stub
+ (define-cproc id->bound-gloc (id::<identifier>)
+   (let* ([gloc::ScmGloc* (Scm_IdentifierGlobalBinding id)])
+     (if (and gloc (not (SCM_UNBOUNDP (SCM_GLOC_GET gloc))))
+       (return (SCM_OBJ gloc))
+       (return SCM_FALSE))))
+ )
+
+;; Returns #t if id1 and id2 both refer to the same existing global binding.
+;; Like free-identifier=? but we know id1 and id2 are both toplevel and
+;; at least one is bound, so we skip local binding lookup.
+(define (global-identifier=? id1 id2)
+  (and-let* ([ (identifier? id1) ]
+             [ (identifier? id2) ]
+             [g1 (id->bound-gloc id1)]
+             [g2 (id->bound-gloc id2)])
+    (eq? g1 g2)))
+
+;; Returns #t iff id1 and id2 would resolve to the same binding
+;; (or both are free).
+(define (free-identifier=? id1 id2)
+  (define (lookup id)
+    (env-lookup id (identifier-module id) (identifier-env id)))
+  (define (deep-compare id1 id2)
+    (let ([b1 (lookup id1)]
+          [b2 (lookup id2)])
+      (cond
+       [(or (lvar? b1) (macro? b1))
+        ;;must have the same local variable or syntactic binding
+        (eq? b1 b2)]
+       [(or (lvar? b2) (macro? b2)) #f]
+       [else (let ([g1 (id->bound-gloc id1)]
+                   [g2 (id->bound-gloc id2)])
+               ;; If both has bound in toplevel, they must refer to the
+               ;; same binding, hence (eq? g1 g2).  The name may differ,
+               ;; because of renaming on export/import.
+               ;; If at least either one is unbound, we just compare their
+               ;; names.
+               (if (and g1 g2)
+                 (eq? g1 g2)
+                 (eq? (unwrap-syntax id1)
+                      (unwrap-syntax id2))))])))
+  (and (identifier? id1)
+       (identifier? id2)
+       (or (eq? id1 id2)
+           (deep-compare id1 id2))))
