@@ -1,7 +1,7 @@
 ;;;
 ;;; libnum.stub - builtin number libraries
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -53,23 +53,23 @@
 (define-cproc real? (obj)    ::<boolean> :fast-flonum :constant
   (inliner REALP) SCM_REALP)
 (define-cproc rational? (obj)::<boolean> :fast-flonum :constant
-  (result (and (SCM_REALP obj) (Scm_FiniteP obj))))
+  (return (and (SCM_REALP obj) (Scm_FiniteP obj))))
 (define-cproc integer? (obj) ::<boolean> :fast-flonum :constant
-  (result (and (SCM_NUMBERP obj) (Scm_IntegerP obj))))
+  (return (and (SCM_NUMBERP obj) (Scm_IntegerP obj))))
 
 (define-cproc exact? (obj)   ::<boolean> :fast-flonum :constant SCM_EXACTP)
 (define-cproc inexact? (obj) ::<boolean> :fast-flonum :constant SCM_INEXACTP)
 
 (define-cproc zero? (obj::<number>) ::<boolean> :fast-flonum :constant
-  (result (and (SCM_REALP obj) (== (Scm_Sign obj) 0))))
+  (return (and (SCM_REALP obj) (== (Scm_Sign obj) 0))))
 
 (define-cproc positive? (obj) ::<boolean> :fast-flonum :constant
-  (result (> (Scm_Sign obj) 0)))
+  (return (> (Scm_Sign obj) 0)))
 (define-cproc negative? (obj) ::<boolean> :fast-flonum :constant
-  (result (< (Scm_Sign obj) 0)))
+  (return (and (not (Scm_NanP obj)) (< (Scm_Sign obj) 0))))
 (define-cproc odd? (obj)  ::<boolean> :fast-flonum :constant Scm_OddP)
 (define-cproc even? (obj) ::<boolean> :fast-flonum :constant
-  (result (not (Scm_OddP obj))))
+  (return (not (Scm_OddP obj))))
 
 (select-module gauche)
 ;; fixnum? and bignum? is not :constant, since it is platform-dependent.
@@ -90,9 +90,9 @@
 
 (select-module gauche)
 ;; Names are from R6RS.
-(define-cproc fixnum-width ()    ::<int> (result SCM_SMALL_INT_SIZE))
-(define-cproc least-fixnum ()    ::<long> (result SCM_SMALL_INT_MIN))
-(define-cproc greatest-fixnum () ::<long> (result SCM_SMALL_INT_MAX))
+(define-cproc fixnum-width ()    ::<int> (return (+ SCM_SMALL_INT_SIZE 1)))
+(define-cproc least-fixnum ()    ::<long> (return SCM_SMALL_INT_MIN))
+(define-cproc greatest-fixnum () ::<long> (return SCM_SMALL_INT_MAX))
 
 ;; default-endian is defined in Scm__InitNumber().
 (define-cproc native-endian () Scm_NativeEndian)
@@ -125,19 +125,18 @@
  (define-cise-stmt numcmp
    [(_ compar)
     `(begin
-       (result FALSE)
-       (cond [(not (,compar arg0 arg1))]
-             [(== optcnt 0) (result TRUE)]
-             [(not (,compar arg1 (aref oarg 0)))]
-             [(== optcnt 1) (result TRUE)]
-             [(not (,compar (aref oarg 0) (aref oarg 1)))]
-             [(and (== optcnt 2) (SCM_NULLP args)) (result TRUE)]
+       (cond [(not (,compar arg0 arg1)) (return FALSE)]
+             [(== optcnt 0) (return TRUE)]
+             [(not (,compar arg1 (aref oarg 0))) (return FALSE)]
+             [(== optcnt 1) (return TRUE)]
+             [(not (,compar (aref oarg 0) (aref oarg 1))) (return FALSE)]
+             [(and (== optcnt 2) (SCM_NULLP args)) (return TRUE)]
              [else
               (set! arg0 (aref oarg 1)
                     arg1 (SCM_CAR args)
                     args (SCM_CDR args))
-              (loop (cond [(not (,compar arg0 arg1)) (break)]
-                          [(SCM_NULLP args) (result TRUE) (break)]
+              (loop (cond [(not (,compar arg0 arg1)) (return FALSE)]
+                          [(SCM_NULLP args) (return TRUE)]
                           [else (set! arg0 arg1
                                       arg1 (SCM_CAR args)
                                       args (SCM_CDR args))]))]))])
@@ -178,23 +177,23 @@
 (select-module scheme)
 (define-cproc number->string
   (obj :optional (radix::<fixnum> 10) (use-upper? #f)) :fast-flonum :constant
-  (result (Scm_NumberToString obj radix
+  (return (Scm_NumberToString obj radix
                               (?: (SCM_FALSEP use_upperP)
                                   0
                                   SCM_NUMBER_FORMAT_USE_UPPER))))
 
 (define-cproc string->number (obj::<string> :optional (radix::<fixnum> 10))
-  (result (Scm_StringToNumber obj radix 0)))
+  (return (Scm_StringToNumber obj radix 0)))
 
 (select-module gauche)
 (define-cproc floor->exact (num) :fast-flonum :constant
-  (result (Scm_RoundToExact num SCM_ROUND_FLOOR)))
+  (return (Scm_RoundToExact num SCM_ROUND_FLOOR)))
 (define-cproc ceiling->exact (num) :fast-flonum :constant
-  (result (Scm_RoundToExact num SCM_ROUND_CEIL)))
+  (return (Scm_RoundToExact num SCM_ROUND_CEIL)))
 (define-cproc truncate->exact (num) :fast-flonum :constant
-  (result (Scm_RoundToExact num SCM_ROUND_TRUNC)))
+  (return (Scm_RoundToExact num SCM_ROUND_TRUNC)))
 (define-cproc round->exact (num) :fast-flonum :constant
-  (result (Scm_RoundToExact num SCM_ROUND_ROUND)))
+  (return (Scm_RoundToExact num SCM_ROUND_ROUND)))
 
 (define-cproc decode-float (num)        ;from ChezScheme
   (cond [(SCM_FLONUMP num)
@@ -204,14 +203,14 @@
            (set! (SCM_VECTOR_ELEMENT v 0) f
                  (SCM_VECTOR_ELEMENT v 1) (Scm_MakeInteger exp)
                  (SCM_VECTOR_ELEMENT v 2) (Scm_MakeInteger sign))
-           (result v))]
+           (return v))]
         [(SCM_INTP num)
          (let* ([v (Scm_MakeVector 3 '#f)])
            (set! (SCM_VECTOR_ELEMENT v 0) (Scm_Abs num)
                  (SCM_VECTOR_ELEMENT v 1) (Scm_MakeInteger 0)
                  (SCM_VECTOR_ELEMENT v 2) (Scm_MakeInteger (Scm_Sign num)))
-           (result v))]
-        [else (SCM_TYPE_ERROR num "real number") (result SCM_UNDEFINED)]))
+           (return v))]
+        [else (SCM_TYPE_ERROR num "real number") (return SCM_UNDEFINED)]))
 
 (define-cproc fmod (x::<real> y::<real>) ::<double> fmod)
 
@@ -226,7 +225,7 @@
 (define-cproc log10 (x::<real>) ::<real> log10)
 
 ;; NB: Alternative implemenation of gamma and log-abs-gamma functions are
-;; provided in Scheme (lib/gauche/numeric.scm).
+;; provided in Scheme (lib/gauche/numerical.scm).
 (select-module gauche.internal)
 (inline-stub
  (if "defined(HAVE_TGAMMA)"
@@ -242,38 +241,39 @@
 
 (select-module scheme)
 (define-cproc * (:rest args) ::<number> :fast-flonum
-  (cond [(not (SCM_PAIRP args)) (result (SCM_MAKE_INT 1))]
+  (cond [(not (SCM_PAIRP args)) (return (SCM_MAKE_INT 1))]
         [else (let* ([r::ScmObj (SCM_CAR args)])
                 (dolist [v (SCM_CDR args)] (set! r (Scm_Mul r v)))
-                (result r))]))
+                (return r))]))
 
 (define-cproc + (:rest args) ::<number> :fast-flonum
-  (cond [(not (SCM_PAIRP args)) (result (SCM_MAKE_INT 0))]
+  (cond [(not (SCM_PAIRP args)) (return (SCM_MAKE_INT 0))]
         [else (let* ([r::ScmObj (SCM_CAR args)])
                 (dolist [v (SCM_CDR args)] (set! r (Scm_Add r v)))
-                (result r))]))
+                (return r))]))
 
 (define-cproc - (arg1 :rest args) ::<number> :fast-flonum
   (if (SCM_NULLP args)
-    (result (Scm_VMNegate arg1))
+    (return (Scm_VMNegate arg1))
     (begin (dolist [v args] (set! arg1 (Scm_Sub arg1 v)))
-           (result arg1))))
+           (return arg1))))
 
 (define-cproc / (arg1 :rest args) ::<number> :fast-flonum
   (if (SCM_NULLP args)
-    (result (Scm_VMReciprocal arg1))
+    (return (Scm_VMReciprocal arg1))
     (begin (dolist [v args] (set! arg1 (Scm_Div arg1 v)))
-           (result arg1))))
+           (return arg1))))
 
 (define-cproc abs (obj) :fast-flonum :constant Scm_VMAbs)
 
 (define-cproc quotient (n1 n2) :fast-flonum :constant
-  (result (Scm_Quotient n1 n2 NULL)))
+  (return (Scm_Quotient n1 n2 NULL)))
 (define-cproc remainder (n1 n2) :fast-flonum :constant
-  (result (Scm_Modulo n1 n2 TRUE)))
+  (return (Scm_Modulo n1 n2 TRUE)))
 (define-cproc modulo (n1 n2)    :fast-flonum :constant
-  (result (Scm_Modulo n1 n2 FALSE)))
+  (return (Scm_Modulo n1 n2 FALSE)))
 
+(select-module gauche)
 ;; gcd, lcm: these are the simplest ones.  If you need efficiency, consult
 ;; Knuth: "The Art of Computer Programming" Chap. 4.5.2
 (define-in-module scheme (gcd . args)
@@ -325,13 +325,13 @@
 
 (select-module scheme)
 (define-cproc floor (v) ::<number> :fast-flonum :constant
-  (result (Scm_Round v SCM_ROUND_FLOOR)))
+  (return (Scm_Round v SCM_ROUND_FLOOR)))
 (define-cproc ceiling (v) ::<number> :fast-flonum :constant
-  (result (Scm_Round v SCM_ROUND_CEIL)))
+  (return (Scm_Round v SCM_ROUND_CEIL)))
 (define-cproc truncate (v) ::<number> :fast-flonum :constant
-  (result (Scm_Round v SCM_ROUND_TRUNC)))
+  (return (Scm_Round v SCM_ROUND_TRUNC)))
 (define-cproc round (v) ::<number> :fast-flonum :constant
-  (result (Scm_Round v SCM_ROUND_ROUND)))
+  (return (Scm_Round v SCM_ROUND_ROUND)))
 
 ;; Transcedental functions.   First, real-only versions.
 
@@ -340,19 +340,23 @@
 
 (define-cproc %log (x) ::<number> :fast-flonum :constant
   (unless (SCM_REALP x) (SCM_TYPE_ERROR x "real number"))
+  (when (Scm_InfiniteP x)
+    (if (> (Scm_Sign x) 0)
+      (return SCM_POSITIVE_INFINITY)
+      (return (Scm_MakeComplex SCM_DBL_POSITIVE_INFINITY M_PI))))
   (let* ([d::double (Scm_GetDouble x)]
          [shift::double 0.0])
-    (when (or (== d SCM_DBL_POSITIVE_INFINITY)  ; isinf() may not be available
+    (when (or (== d SCM_DBL_POSITIVE_INFINITY)  ; SCM_IS_INF isn't visible...
               (== d SCM_DBL_NEGATIVE_INFINITY))
-      (SCM_ASSERT (SCM_BIGNUMP x))
+      (SCM_ASSERT (SCM_BIGNUMP x))  ; the arg is too big to represent in double
       (let* ([z::ScmBits* (cast ScmBits* (-> (SCM_BIGNUM x) values))]
              [scale::long (Scm_BitsHighest1 z 0 (* (SCM_BIGNUM_SIZE x) SCM_WORD_BITS))])
         (set! shift (* scale (log 2.0)))
         (set! d (Scm_GetDouble
                  (Scm_DivInexact x (Scm_Ash (SCM_MAKE_INT 1) scale))))))
     (if (< (Scm_FlonumSign d) 0)
-      (result (Scm_MakeComplex (+ (log (- d)) shift) M_PI))
-      (result (Scm_VMReturnFlonum (+ (log d) shift))))))
+      (return (Scm_MakeComplex (+ (log (- d)) shift) M_PI))
+      (return (Scm_VMReturnFlonum (+ (log d) shift))))))
 
 (define-cproc %sin (x::<real>) ::<real> :fast-flonum :constant sin)
 (define-cproc %cos (x::<real>) ::<real> :fast-flonum :constant cos)
@@ -364,24 +368,24 @@
 
 (define-cproc %asin (x::<real>) ::<number> :fast-flonum :constant
   (cond [(> x 1.0)
-         (result (Scm_MakeComplex (/ M_PI 2.0)
+         (return (Scm_MakeComplex (/ M_PI 2.0)
                                   (- (log (+ x (sqrt (- (* x x) 1.0)))))))]
         [(< x -1.0)
-         (result (Scm_MakeComplex (/ (- M_PI) 2.0)
+         (return (Scm_MakeComplex (/ (- M_PI) 2.0)
                                   (- (log (- (- x) (sqrt (- (* x x) 1.0)))))))]
-        [else (result (Scm_VMReturnFlonum (asin x)))]))
+        [else (return (Scm_VMReturnFlonum (asin x)))]))
 
 (define-cproc %acos (x::<real>) ::<number> :fast-flonum :constant
   (cond [(> x 1.0)
-         (result (Scm_MakeComplex 0 (log (+ x (sqrt (- (* x x) 1.0))))))]
+         (return (Scm_MakeComplex 0 (log (+ x (sqrt (- (* x x) 1.0))))))]
         [(< x -1.0)
-         (result (Scm_MakeComplex 0 (log (+ x (sqrt (- (* x x) 1.0))))))]
-        [else (result (Scm_VMReturnFlonum (acos x)))]))
+         (return (Scm_MakeComplex 0 (log (+ x (sqrt (- (* x x) 1.0))))))]
+        [else (return (Scm_VMReturnFlonum (acos x)))]))
 
 (define-cproc %atan (z::<real> :optional x) ::<double> :fast-flonum :constant
-  (cond [(SCM_UNBOUNDP x) (result (atan z))]
+  (cond [(SCM_UNBOUNDP x) (return (atan z))]
         [else (unless (SCM_REALP x) (SCM_TYPE_ERROR x "real number"))
-              (result (atan2 z (Scm_GetDouble x)))]))
+              (return (atan2 z (Scm_GetDouble x)))]))
 
 (define-cproc %sinh (x::<real>) ::<real> :fast-flonum :constant sinh)
 (define-cproc %cosh (x::<real>) ::<real> :fast-flonum :constant cosh)
@@ -390,8 +394,8 @@
 
 (define-cproc %sqrt (x::<real>) :fast-flonum :constant
   (if (< x 0)
-    (result (Scm_MakeComplex 0.0 (sqrt (- x))))
-    (result (Scm_VMReturnFlonum (sqrt x)))))
+    (return (Scm_MakeComplex 0.0 (sqrt (- x))))
+    (return (Scm_VMReturnFlonum (sqrt x)))))
 
 ;; Fast path for typical case of sqrt.  Handles positive flonum
 ;; and exact integer between 0 and 2^52.
@@ -400,7 +404,7 @@
 (select-module gauche.internal)
 (define-cproc %sqrt-fast-path (x) :fast-flonum :constant
   (cond [(and (SCM_FLONUMP x) (>= (Scm_Sign x) 0))
-         (result (Scm_VMReturnFlonum (sqrt (SCM_FLONUM_VALUE x))))]
+         (return (Scm_VMReturnFlonum (sqrt (SCM_FLONUM_VALUE x))))]
         [(and (SCM_INTEGERP x)
               (>= (Scm_Sign x) 0)
               (>= (Scm_NumCmp SCM_2_52 x) 0))
@@ -410,9 +414,9 @@
                 [dd::double (* qq qq)])
            ;; NB: The result is in [0, 2^26], so we know it fits in fixnum.
            (if (== d dd)
-             (result (SCM_MAKE_INT (cast long q)))
-             (result (Scm_VMReturnFlonum q))))]
-        [else (result SCM_FALSE)]))
+             (return (SCM_MAKE_INT (cast long q)))
+             (return (Scm_VMReturnFlonum q))))]
+        [else (return SCM_FALSE)]))
 
 (select-module gauche)
 (define-cproc %expt (x y) :fast-flonum :constant Scm_Expt)
@@ -487,8 +491,11 @@
         [(real? x)
          (cond [(real? y) (%expt x y)]
                [(number? y)
-                (* (%expt x (real-part y))
-                   (exp (* +i (imag-part y) (%log x))))]
+                (let1 ry (real-part y)
+                  (if (and (zero? x) (positive? ry))
+                    (if (exact? x) 0 0.0)
+                    (* (%expt x ry)
+                       (exp (* +i (imag-part y) (%log x))))))]
                [else (error "number required, but got" y)])]
         [(number? x) (exp (* y (log x)))]
         [else (error "number required, but got" x)]))
@@ -662,15 +669,15 @@
 (inline-stub
  (define-cise-stmt logop
    [(_ fn ident)
-    `(cond [(== optcnt 0) (result ,ident)]
+    `(cond [(== optcnt 0) (return ,ident)]
            [(== optcnt 1)
             (unless (SCM_INTEGERP (aref arg2 0))
               (Scm_Error "Exact integer required, but got %S" (aref arg2 0)))
-            (result (aref arg2 0))]
+            (return (aref arg2 0))]
            [else
             (let* ([r (,fn (aref arg2 0) (aref arg2 1))])
               (for-each (lambda (v) (set! r (,fn r v))) args)
-              (result r))])])
+              (return r))])])
  )
 
 (define-cproc logand (:optarray (arg2 optcnt 2) :rest args) :constant
@@ -681,21 +688,21 @@
   (logop Scm_LogXor (SCM_MAKE_INT 0)))
 
 (define-cproc logcount (n) ::<int> :constant
-  (cond [(SCM_EQ n (SCM_MAKE_INT 0)) (result 0)]
+  (cond [(SCM_EQ n (SCM_MAKE_INT 0)) (return 0)]
         [(SCM_INTP n)
          (let* ([z::ScmBits (cast ScmBits (cast long (SCM_INT_VALUE n)))])
            (if (> (SCM_INT_VALUE n) 0)
-             (result (Scm_BitsCount1 (& z) 0 SCM_WORD_BITS))
-             (result (Scm_BitsCount0 (& z) 0 SCM_WORD_BITS))))]
-        [(SCM_BIGNUMP n) (result (Scm_BignumLogCount (SCM_BIGNUM n)))]
-        [else (SCM_TYPE_ERROR n "exact integer") (result 0)]))
+             (return (Scm_BitsCount1 (& z) 0 SCM_WORD_BITS))
+             (return (Scm_BitsCount0 (& z) 0 SCM_WORD_BITS))))]
+        [(SCM_BIGNUMP n) (return (Scm_BignumLogCount (SCM_BIGNUM n)))]
+        [else (SCM_TYPE_ERROR n "exact integer") (return 0)]))
 
 (define-cproc integer-length (n) ::<int> :constant
   (cond [(SCM_INTP n)
          (let* ([z::ScmBits (cast ScmBits (cast long (SCM_INT_VALUE n)))])
            (if (>= (SCM_INT_VALUE n) 0)
-             (result (+ (Scm_BitsHighest1 (& z) 0 SCM_WORD_BITS) 1))
-             (result (+ (Scm_BitsHighest0 (& z) 0 SCM_WORD_BITS) 1))))]
+             (return (+ (Scm_BitsHighest1 (& z) 0 SCM_WORD_BITS) 1))
+             (return (+ (Scm_BitsHighest0 (& z) 0 SCM_WORD_BITS) 1))))]
         [(SCM_BIGNUMP n)
          ;; 2's complement adjustment.
          (when (< (SCM_BIGNUM_SIGN n) 0)
@@ -704,30 +711,30 @@
          (if (SCM_BIGNUMP n)
            (let* ([z::ScmBits* (cast ScmBits* (-> (SCM_BIGNUM n) values))]
                   [k::int (SCM_BIGNUM_SIZE n)])
-             (result (+ (Scm_BitsHighest1 z 0 (* k SCM_WORD_BITS)) 1)))
+             (return (+ (Scm_BitsHighest1 z 0 (* k SCM_WORD_BITS)) 1)))
            ;; If n+1 becomes fixnum, we know n was the least fixnum.
-           (result (+ SCM_SMALL_INT_SIZE 1)))]
-        [else (SCM_TYPE_ERROR n "exact integer") (result 0)]))
+           (return (+ SCM_SMALL_INT_SIZE 1)))]
+        [else (SCM_TYPE_ERROR n "exact integer") (return 0)]))
 
 ;; Returns maximum s where (expt 2 s) is a factor of n.
 ;; This can be (- (integer-length (logxor n (- n 1))) 1), but we can save
 ;; creating intermediate numbers by providing this natively.
 (define-cproc twos-exponent-factor (n) ::<int> :constant
-  (cond [(SCM_EQ n (SCM_MAKE_INT 0)) (result 0)]
+  (cond [(SCM_EQ n (SCM_MAKE_INT 0)) (return 0)]
         [(SCM_INTP n)
          (let* ([z::ScmBits (cast ScmBits (cast long (SCM_INT_VALUE n)))])
-           (result (Scm_BitsLowest1 (& z) 0 SCM_WORD_BITS)))]
+           (return (Scm_BitsLowest1 (& z) 0 SCM_WORD_BITS)))]
         [(SCM_BIGNUMP n)
          (let* ([z::ScmBits* (cast ScmBits* (-> (SCM_BIGNUM n) values))]
                 [k::int (SCM_BIGNUM_SIZE n)])
-           (result (Scm_BitsLowest1 z 0 (* k SCM_WORD_BITS))))]
-        [else (SCM_TYPE_ERROR n "exact integer") (result 0)]))
+           (return (Scm_BitsLowest1 z 0 (* k SCM_WORD_BITS))))]
+        [else (SCM_TYPE_ERROR n "exact integer") (return 0)]))
 
 (define-cproc twos-exponent (n) :constant
   (if (SCM_INTEGERP n) ; exact integer only
     (let* ([i::long (Scm_TwosPower n)])
-      (result (?: (>= i 0) (Scm_MakeInteger i) SCM_FALSE)))
-    (begin (SCM_TYPE_ERROR n "exact integer") (result SCM_FALSE))))
+      (return (?: (>= i 0) (Scm_MakeInteger i) SCM_FALSE)))
+    (begin (SCM_TYPE_ERROR n "exact integer") (return SCM_FALSE))))
 
 ;; As of 0.8.8 we started to support exact rational numbers.  Some existing
 ;; code may count on exact integer division to be coerced to flonum
@@ -739,9 +746,9 @@
 ;; inexact-/.  If the program uses compat.no-rational, '/' is overridden
 ;; by inexact-/ and the old code behaves the same.
 (define-cproc inexact-/ (arg1 :rest args)
-  (cond [(SCM_NULLP args) (result (Scm_ReciprocalInexact arg1))]
+  (cond [(SCM_NULLP args) (return (Scm_ReciprocalInexact arg1))]
         [else (dolist [x args] (set! arg1 (Scm_DivCompat arg1 x)))
-              (result arg1)]))
+              (return arg1)]))
 
 ;; Inexact arithmetics.  Useful for speed-sensitive code to avoid
 ;; accidental use of bignum or ratnum.   We might want to optimize
@@ -749,21 +756,21 @@
 (define-cproc +. (:rest args) :constant
   (let* ([a '0.0])
     (dolist [x args] (set! a (Scm_Add a (Scm_Inexact x))))
-    (result a)))
+    (return a)))
 (define-cproc *. (:rest args) :constant
   (let* ([a '1.0])
     (dolist [x args] (set! a (Scm_Mul a (Scm_Inexact x))))
-    (result a)))
+    (return a)))
 (define-cproc -. (arg1 :rest args) :constant
   (cond
-   [(SCM_NULLP args) (result (Scm_Negate (Scm_Inexact arg1)))]
+   [(SCM_NULLP args) (return (Scm_Negate (Scm_Inexact arg1)))]
    [else (dolist [x args] (set! arg1 (Scm_Sub arg1 (Scm_Inexact x))))
-         (result arg1)]))
+         (return arg1)]))
 (define-cproc /. (arg1 :rest args) :constant
   (cond
-   [(SCM_NULLP args) (result (Scm_Reciprocal (Scm_Inexact arg1)))]
+   [(SCM_NULLP args) (return (Scm_Reciprocal (Scm_Inexact arg1)))]
    [else (dolist [x args] (set! arg1 (Scm_DivInexact arg1 x)))
-         (result arg1)]))
+         (return arg1)]))
 
 (define-cproc clamp (x :optional (min #f) (max #f)) :fast-flonum :constant
   (let* ([r x] [maybe_exact::int (SCM_EXACTP x)])
@@ -797,13 +804,13 @@
 ;; and avoiding extra allocation.
 (define-cproc real-part (z::<number>) :fast-flonum :constant
   (if (SCM_REALP z)
-    (result z)
-    (result (Scm_VMReturnFlonum (SCM_COMPNUM_REAL z)))))
+    (return z)
+    (return (Scm_VMReturnFlonum (SCM_COMPNUM_REAL z)))))
 
 (define-cproc imag-part (z::<number>) :fast-flonum :constant
-  (cond [(SCM_EXACTP z) (result (SCM_MAKE_INT 0))]
-        [(SCM_REALP z)  (result (Scm_VMReturnFlonum 0.0))]
-        [else (result (Scm_VMReturnFlonum (SCM_COMPNUM_IMAG z)))]))
+  (cond [(SCM_EXACTP z) (return (SCM_MAKE_INT 0))]
+        [(SCM_REALP z)  (return (Scm_VMReturnFlonum 0.0))]
+        [else (return (Scm_VMReturnFlonum (SCM_COMPNUM_IMAG z)))]))
 
-(define-cproc magnitude (z) ::<double> :fast-flonum :constant Scm_Magnitude)
+(define-cproc magnitude (z) :fast-flonum :constant Scm_VMAbs)
 (define-cproc angle (z)     ::<double> :fast-flonum :constant Scm_Angle)
