@@ -1,7 +1,7 @@
 ;;;
 ;;; libproc.scm - procedure call & return, and other control stuff
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -49,7 +49,7 @@
 (define-cproc apply (proc arg1 :rest args)
   (inliner TAIL-APPLY)
   (let* ([head::ScmObj] [tail::ScmObj])
-    (cond [(SCM_NULLP args) (result (Scm_VMApply proc arg1))]
+    (cond [(SCM_NULLP args) (return (Scm_VMApply proc arg1))]
           [else (set! head (Scm_Cons arg1 SCM_NIL)
                       tail head)
                 (dopairs [cp args]
@@ -59,7 +59,7 @@
                   (unless (SCM_PAIRP (SCM_CDR cp))
                     (Scm_Error "improper list not allowed: %S" (SCM_CDR cp)))
                   (SCM_APPEND1 head tail (SCM_CAR cp)))
-                (result (Scm_VMApply proc head))])))
+                (return (Scm_VMApply proc head))])))
 
 (define-cproc call-with-current-continuation (proc) Scm_VMCallCC)
 (define-cproc values (:rest args) :constant (inliner VALUES) Scm_Values)
@@ -72,7 +72,7 @@
 
 (select-module gauche.internal)
 ;; for partial continuation.  See lib/gauche/partcont.scm
-(define-cproc %call/pc (proc) (result (Scm_VMCallPC proc)))
+(define-cproc %call/pc (proc) (return (Scm_VMCallPC proc)))
 
 ;;;
 ;;; Extended argument parsing
@@ -229,17 +229,37 @@
 (define-cproc subr? (obj)    ::<boolean> SCM_SUBRP)
 (define-cproc closure? (obj) ::<boolean> SCM_CLOSUREP)
 (define-cproc toplevel-closure? (obj) ::<boolean>
-  (result (and (SCM_CLOSUREP obj) (== (-> (SCM_CLOSURE obj) env) NULL))))
+  (return (and (SCM_CLOSUREP obj) (== (-> (SCM_CLOSURE obj) env) NULL))))
 
-(define-cproc closure-code (clo::<closure>) (result (-> clo code)))
+(define-cproc closure-code (clo::<closure>) (return (-> clo code)))
 (define-cproc method-code (m::<method>)
   (if (-> m func)
     ;; code is not available for C-defined method
-    (result SCM_FALSE)
-    (result (SCM_OBJ (-> m data)))))
+    (return SCM_FALSE)
+    (return (SCM_OBJ (-> m data)))))
 
 (define-cproc procedure-info (proc::<procedure>)
-  (result (SCM_PROCEDURE_INFO proc)))
+  (return (SCM_PROCEDURE_INFO proc)))
+
+;; NB: This takes a list of classes.  But what if we support eqv-specilizer?
+;; One idea is to let the caller wrap a concrete instance.  We'll see...
+(define (applicable? proc . arg-types)
+  (define method-applicable?
+    (with-module gauche.object method-applicable-for-classes?))
+  (let1 c (class-of proc)
+    (cond [(eq? c <procedure>)
+           (let1 nargs (length arg-types)
+             (if-let1 infos (case-lambda-info proc)
+               (any (^[info] (apply [^(reqargs optarg proc)
+                                      ((if optarg >= =) nargs reqargs)]
+                                    info))
+                    infos)
+               ((if (slot-ref proc 'optional) >= =)
+                nargs (slot-ref proc 'required))))]
+          [(eq? c <generic>)
+           (any (^m (apply method-applicable? m arg-types)) (~ proc'methods))]
+          [(eq? c <method>)  (apply method-applicable? m arg-types)]
+          [else (apply applicable? object-apply c arg-types)])))
 
 (select-module gauche.internal)
 ;; Tester procedures
