@@ -1,7 +1,7 @@
 ;;;
 ;;; libsym.scm - built-in symbol and keyword procedures
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -43,18 +43,39 @@
 (define-cproc symbol? (obj) ::<boolean> :fast-flonum :constant
   (inliner SYMBOLP) SCM_SYMBOLP)
 (define-cproc symbol->string (obj::<symbol>) :constant
-  (result (SCM_OBJ (SCM_SYMBOL_NAME obj))))
+  (return (SCM_OBJ (SCM_SYMBOL_NAME obj))))
 (define-cproc string->symbol (obj::<string>) :constant Scm_Intern)
 
 (select-module gauche)
 (define-cproc gensym (:optional (prefix::<string>? #f)) Scm_Gensym)
 (define-cproc symbol-interned? (s::<symbol>) ::<boolean> SCM_SYMBOL_INTERNED)
 (define-cproc string->uninterned-symbol (name::<string>)
-  (result (Scm_MakeSymbol name FALSE)))
+  (return (Scm_MakeSymbol name FALSE)))
 (define-cproc symbol-sans-prefix (s::<symbol> p::<symbol>)
   Scm_SymbolSansPrefix)
 
+;; Bigloo has symbol-append symbol ... -> symbol
+;; We enhance it a bit.
+(select-module gauche.internal)
+(define-in-module gauche symbol-append
+  (letrec ([->string
+            ;; to make it work regardless of keyword-symbol integration
+            (^x (cond [(keyword? x) #":~(keyword->string x)"]
+                      [(identifier? x) (symbol->string (identifier->symbol x))]
+                      [else (x->string x)]))]
+           [do-append
+            (^[objs interned?]
+              ((if interned? string->symbol string->uninterned-symbol)
+               (apply string-append (map ->string objs))))])
+    (case-lambda
+      [() '||] ; edge case
+      [(maybe-flag . syms)
+       (if (boolean? maybe-flag)
+         (do-append syms maybe-flag)
+         (do-append (cons maybe-flag syms) #t))])))
+
 ;; R7RS
+(select-module gauche)
 (define (symbol=? x y . rest)
   (if-let1 z (find ($ not $ symbol? $) (list* x y rest))
     (error "symbol required, but got:" z))
@@ -73,7 +94,7 @@
     (cond [(SCM_STRINGP name) (set! sname (SCM_STRING name))]
           [(SCM_SYMBOLP name) (set! sname (SCM_SYMBOL_NAME name))]
           [else (SCM_TYPE_ERROR name "string or symbol")])
-    (result (Scm_MakeKeyword sname))))
+    (return (Scm_MakeKeyword sname))))
 
 (define-cproc get-keyword (key list :optional fallback) :constant
   Scm_GetKeyword)
@@ -81,10 +102,7 @@
 (define-cproc delete-keyword (key list)  Scm_DeleteKeyword)
 (define-cproc delete-keyword! (key list) Scm_DeleteKeywordX)
 
-(define-cproc keyword->string (key::<keyword>)
-  (.if "defined(GAUCHE_UNIFY_SYMBOL_KEYWORD)"
-       (result (Scm_Substring (SCM_KEYWORD_NAME key) 1 -1 FALSE))
-       (result (SCM_OBJ (SCM_KEYWORD_NAME key)))))
+(define-cproc keyword->string (key::<keyword>) Scm_KeywordToString)
 
 (define-in-module gauche (delete-keywords ks kvlist)
   (define (rec kvs)
@@ -124,15 +142,29 @@
 (define-cproc identifier? (obj) ::<boolean> :constant
   (inliner IDENTIFIERP) SCM_IDENTIFIERP)
 (define-cproc identifier->symbol (obj::<identifier>) :constant
-  (result (SCM_OBJ (-> (SCM_IDENTIFIER obj) name))))
+  (return (SCM_OBJ (Scm_UnwrapIdentifier obj))))
 
 (select-module gauche.internal)
-(define-cproc make-identifier (name::<symbol> mod::<module> env::<list>)
+(define-cproc make-identifier (name mod::<module> env::<list>)
   Scm_MakeIdentifier)
 (define-cproc identifier-module (id::<identifier>)
-  (result (SCM_OBJ (-> id module))))
+  (return (SCM_OBJ (-> id module))))
 (define-cproc identifier-name (id::<identifier>)
-  (result (SCM_OBJ (-> id name))))
+  (return (SCM_OBJ (-> id name))))
 (define-cproc identifier-env (id::<identifier>)
-  (result (-> id env)))
+  (return (-> id env)))
 
+(select-module gauche.internal)
+;; EXPERIMENTAL
+;; Concatenate symbols or identifiers.  If any one of args is an identifier,
+;; the result is also an identifier with the same scope of the first
+;; identifier in the args.
+;; This sounds pretty kludgy.  I suspect this won't be necessary once
+;; we use er-macro for everything low-level (rename procedure should take
+;; care of identifier marking).  It is needed, for now, to add certain level
+;; of hygiene to define-macro.
+(define (identifier-append . args)
+  (let1 r (apply symbol-append #t args)
+    (if-let1 first-id (find identifier? args)
+      (make-identifier r (identifier-module first-id) (identifier-env first-id))
+      r)))
