@@ -1,7 +1,7 @@
 ;;;
 ;;; libvec.scm - builtin vector procedures
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -45,7 +45,7 @@
 (select-module scheme)
 
 (define-cproc vector (:rest args) (inliner VEC)
-  (result (Scm_ListToVector args 0 -1)))
+  (return (Scm_ListToVector args 0 -1)))
 
 (define-cproc vector? (obj) ::<boolean> :fast-flonum :constant
   (inliner VECTORP) SCM_VECTORP)
@@ -61,8 +61,8 @@
              (>= (SCM_INT_VALUE k) (SCM_VECTOR_SIZE vec)))
          (when (SCM_UNBOUNDP fallback)
            (Scm_Error "vector-ref index out of range: %S" k))
-         (result fallback)]
-        [else (result (SCM_VECTOR_ELEMENT vec (SCM_INT_VALUE k)))]))
+         (return fallback)]
+        [else (return (SCM_VECTOR_ELEMENT vec (SCM_INT_VALUE k)))]))
 
 (define-cproc vector-set! (vec::<vector> k::<integer> obj) ::<void>
   (if (or (SCM_BIGNUMP k)
@@ -87,6 +87,52 @@
   (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) fill)
   Scm_VectorCopy)
 
+(define-cproc vector-copy!
+  (t::<vector> tstart::<fixnum> s::<vector>
+               :optional (sstart::<fixnum> 0) (send::<fixnum> -1)) ::<void>
+  (let* ([tsize::long (SCM_VECTOR_SIZE t)])
+    (when (< send 0) (set! send (SCM_VECTOR_SIZE s)))
+    (unless (and (<= 0 tstart) (<= tstart tsize))
+      (Scm_Error "tstart out of range: %ld" tstart))
+    (unless (and (<= 0 sstart) (<= sstart (SCM_VECTOR_SIZE s)))
+      (Scm_Error "sstart out of range: %ld" sstart))
+    (unless (and (<= 0 send) (<= send (SCM_VECTOR_SIZE s)))
+      (Scm_Error "send out of range: %ld" send))
+    (unless (<= (+ tstart (- send sstart)) tsize)
+      (Scm_Error "source vector overruns the target vector: %20.0S [%ld,%ld]"
+                 s sstart send))
+    (unless (<= sstart send)
+      (Scm_Error "send (%ld) must be greater than or equal to sstart (%ld)"
+                 send sstart))
+    (if (<= tstart sstart)
+      (let* ([i::ScmSmallInt sstart] [j::ScmSmallInt tstart])
+        (for [() (and (< i send) (< j tsize)) (begin (post++ i) (post++ j))]
+             (set! (SCM_VECTOR_ELEMENT t j) (SCM_VECTOR_ELEMENT s i))))
+      (let* ([i::ScmSmallInt send]
+             [j::ScmSmallInt (+ tstart (- send sstart))])
+        (when (>= j tsize)
+          (set! i (- i (- j tsize)))
+          (set! j tsize))
+        (for [(begin (post-- i) (post-- j))
+              (and (>= i sstart) (>= j tstart))
+              (begin (post-- i) (post-- j))]
+             (set! (SCM_VECTOR_ELEMENT t j) (SCM_VECTOR_ELEMENT s i)))))))
+
+(define-cproc vector-append (:rest vecs)
+  (let* ([len::long 0])
+    (dolist [v vecs]
+      (unless (SCM_VECTORP v)
+        (Scm_Error "vector required, but got: %S" v))
+      (set! len (+ len (SCM_VECTOR_SIZE v))))
+    (let* ([dst (Scm_MakeVector len SCM_UNDEFINED)]
+           [j::long 0])
+      (dolist [v vecs]
+        (let* ([k::long (SCM_VECTOR_SIZE v)])
+          (memcpy (+ (SCM_VECTOR_ELEMENTS dst) j) (SCM_VECTOR_ELEMENTS v)
+                  (* k (sizeof ScmWord)))
+          (set! j (+ j k))))
+      (return dst))))
+
 (define (vector->string v :optional (start 0) (end -1)) ;;R7RS
   (list->string (vector->list v start end))) ; TODO: can be more efficient
 (define (string->vector s :optional (start 0) (end -1)) ;;R7RS
@@ -101,7 +147,7 @@
 (define-cproc make-weak-vector (size::<fixnum>) Scm_MakeWeakVector)
 
 (define-cproc weak-vector-length (wv::<weak-vector>) ::<int>
-  (result (-> wv size)))
+  (return (-> wv size)))
 
 (define-cproc weak-vector-ref
   (wv::<weak-vector> index::<fixnum> :optional fallback)
@@ -118,19 +164,12 @@
 ;; this one internally).
 (select-module gauche.internal)
 
-(define-cproc %uvector-ref (v::<uvector> t::<int> k::<integer>
+(define-cproc %uvector-ref (v::<uvector> t::<int> k::<fixnum>
                                          :optional fallback)
   :constant
   (unless (== (Scm_UVectorType (SCM_CLASS_OF v)) t)
     (Scm_TypeError "vec" (Scm_UVectorTypeName t) (SCM_OBJ v)))
-  (cond [(or (SCM_BIGNUMP k)
-             (< (SCM_INT_VALUE k) 0)
-             (>= (SCM_INT_VALUE k) (SCM_UVECTOR_SIZE v)))
-         (when (SCM_UNBOUNDP fallback)
-           (Scm_Error "%s-ref index out of range: %S"
-                      (Scm_UVectorTypeName t) k))
-         (result fallback)]
-        [else (result (Scm_VMUVectorRef v t (SCM_INT_VALUE k) fallback))]))
+  (return (Scm_VMUVectorRef v t k fallback)))
 
 (select-module gauche)
 (inline-stub
@@ -151,8 +190,7 @@
   SCM_UVECTOR_SIZE)
 (define-cproc uvector-immutable? (v::<uvector>) ::<boolean>
   SCM_UVECTOR_IMMUTABLE_P)
-
-
-
+(define-cproc uvector? (obj) ::<boolean> :constant
+  SCM_UVECTORP)
 
 
