@@ -1,7 +1,7 @@
 /*
  * list.c - List related functions
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -257,7 +257,7 @@ ScmObj Scm_CopyList(ScmObj list)
 ScmObj Scm_MakeList(ScmSmallInt len, ScmObj fill)
 {
     if (len < 0) {
-        Scm_Error("make-list: negative length given: %d", len);
+        Scm_Error("make-list: negative length given: %ld", len);
     }
     ScmObj start = SCM_NIL, last = SCM_NIL;
     while (len--) {
@@ -709,15 +709,22 @@ ScmObj Scm_PairAttr(ScmPair *pair)
     }
 }
 
-ScmObj Scm_ExtendedCons(ScmObj car, ScmObj cdr)
+ScmObj Scm_MakeExtendedPair(ScmObj car, ScmObj cdr, ScmObj attrs)
 {
     ScmExtendedPair *xp = SCM_NEW(ScmExtendedPair);
     xp->car = car;
     xp->cdr = cdr;
-    xp->attributes = SCM_NIL;
+    xp->attributes = attrs;
     return SCM_OBJ(xp);
 }
 
+/* The common scenario is to use ExtendedCons in place of Cons,
+   then add attributes later.  So we provide this.  */
+ScmObj Scm_ExtendedCons(ScmObj car, ScmObj cdr)
+{
+    return Scm_MakeExtendedPair(car, cdr, SCM_NIL);
+}
+
 ScmObj Scm_PairAttrGet(ScmPair *pair, ScmObj key, ScmObj fallback)
 {
     if (!SCM_EXTENDED_PAIR_P(pair)) {
