@@ -1,7 +1,7 @@
 /*
  * load.c - load a program
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,10 +33,12 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/class.h"
 #include "gauche/port.h"
 #include "gauche/priv/builtin-syms.h"
 #include "gauche/priv/readerP.h"
 #include "gauche/priv/portP.h"
+#include "gauche/priv/moduleP.h"
 
 #include <ctype.h>
 #include <fcntl.h>
@@ -45,9 +47,6 @@
  * Load file.
  */
 
-typedef struct dlobj_rec dlobj;
-typedef struct dlobj_initfn_rec dlobj_initfn;
-
 /* Static parameters */
 static struct {
     /* Load path list */
@@ -74,13 +73,10 @@ static struct {
                                           searched. */
     ScmParameterLoc load_port;         /* current port from which we are
                                           loading */
-    ScmParameterLoc load_main_script;  /* yields #t during loading main script.
-                                          see SCM_LOAD_MAIN_SCRIPT flag
-                                          in load.h. */
     
     /* Dynamic linking */
     ScmObj dso_suffixes;
-    dlobj *dso_list;              /* List of dynamically loaded objects. */
+    ScmHashTable *dso_table;      /* Hashtable path -> <dlobj> */
     ScmObj dso_prelinked;         /* List of 'prelinked' DSOs, that is, they
                                      are already linked but pretened to be
                                      DSOs.  dynamic-load won't do anything.
@@ -95,12 +91,10 @@ static struct {
 static ScmObj key_error_if_not_found = SCM_UNBOUND;
 static ScmObj key_macro              = SCM_UNBOUND;
 static ScmObj key_ignore_coding      = SCM_UNBOUND;
-static ScmObj key_main_script        = SCM_UNBOUND;
 static ScmObj key_paths              = SCM_UNBOUND;
 static ScmObj key_environment        = SCM_UNBOUND;
 
 #define PARAM_REF(vm, loc)      Scm_ParameterRef(vm, &ldinfo.loc)
-#define PARAM_SET(vm, loc, val) Scm_ParameterSet(vm, &ldinfo.loc, val)
 
 /*
  * ScmLoadPacket is the way to communicate to Scm_Load facility.
@@ -129,7 +123,6 @@ void Scm_LoadPacketInit(ScmLoadPacket *p)
 
 /*--------------------------------------------------------------------
  * Scm_LoadFromPort
- * Scm_VMLoadFromPort
  *
  *   The most basic function in the load()-family.  Read an expression
  *   from the given port and evaluates it repeatedly, until it reaches
@@ -146,129 +139,10 @@ void Scm_LoadPacketInit(ScmLoadPacket *p)
  *   extension.  SCM_LOAD_QUIET_NOFILE and SCM_LOAD_IGNORE_CODING
  *   won't have any effect for LoadFromPort; see Scm_Load below.
  *
- *   Scm_VMLoadFromPort returns a Scheme object that encodes result
- *   load.  Right now it is used to propagete ScmLoadProvideStatus
- *   to Scm_LoadFromPort, but its meaning may be changed in future;
- *   the application must treat the returned value opaque, and usually
- *   should discard it.
- *
  *   TODO: if we're using coding-aware port, how should we propagate
  *   locking into the wrapped (original) port?
  */
 
-/* A structure to keep around the transient state of the current loading.
-   Created by Scm_VMLoadFromPort and discarded once the loading ends. */
-struct load_info {
-    ScmPort *port;
-    ScmModule *prev_module;
-    ScmReadContext *prev_ctx;
-    ScmObj prev_port;
-    ScmObj prev_history;
-    ScmObj prev_next;
-    ScmObj prev_main_script;
-    ScmObj prev_reader_lexical_mode;
-    int    prev_situation;
-};
-
-/* Clean up */
-static ScmObj load_after(ScmObj *args, int nargs, void *data)
-{
-    struct load_info *p = (struct load_info *)data;
-    ScmVM *vm = Scm_VM();
-
-#ifdef HAVE_GETTIMEOFDAY
-    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_COLLECT_LOAD_STATS)) {
-        struct timeval t0;
-        gettimeofday(&t0, NULL);
-        vm->stat.loadStat =
-            Scm_Cons(Scm_MakeIntegerU(t0.tv_sec*1000000+t0.tv_usec),
-                     vm->stat.loadStat);
-    }
-#endif /*HAVE_GETTIMEOFDAY*/
-
-    Scm_ClosePort(p->port);
-    PORT_UNLOCK(p->port);
-    Scm_SelectModule(p->prev_module);
-    PARAM_SET(vm, load_port, p->prev_port);
-    PARAM_SET(vm, load_history, p->prev_history);
-    PARAM_SET(vm, load_next, p->prev_next);
-    PARAM_SET(vm, load_main_script, p->prev_main_script);
-    Scm_SetReaderLexicalMode(p->prev_reader_lexical_mode);
-    Scm_SetCurrentReadContext(p->prev_ctx);
-    vm->evalSituation = p->prev_situation;
-    return SCM_UNDEFINED;
-}
-
-/* C-continuation of the loading */
-static ScmObj load_cc(ScmObj result, void **data)
-{
-    struct load_info *p = (struct load_info*)(data[0]);
-    ScmObj expr = Scm_Read(SCM_OBJ(p->port));
-
-    if (!SCM_EOFP(expr)) {
-        Scm_VMPushCC(load_cc, data, 1);
-        return Scm_VMEval(expr, SCM_FALSE);
-    } else {
-        return SCM_TRUE;
-    }
-}
-
-static ScmObj load_body(ScmObj *args, int nargs, void *data)
-{
-    return load_cc(SCM_NIL, &data);
-}
-
-ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths,
-                          ScmObj env, int flags)
-{
-    ScmVM *vm = Scm_VM();
-    ScmModule *module = vm->module;
-
-    /* Sanity check */
-    if (!SCM_IPORTP(port))
-        Scm_Error("input port required, but got: %S", port);
-
-
-    if (SCM_PORT_CLOSED_P(port))
-        Scm_Error("port already closed: %S", port);
-    if (SCM_MODULEP(env)) {
-        module = SCM_MODULE(env);
-    } else if (!SCM_UNBOUNDP(env) && !SCM_FALSEP(env)) {
-        Scm_Error("bad load environment (must be a module or #f): %S", env);
-    }
-
-    struct load_info *p = SCM_NEW(struct load_info);
-    p->port = port;
-    p->prev_module    = vm->module;
-    p->prev_port      = PARAM_REF(vm, load_port);
-    p->prev_history   = PARAM_REF(vm, load_history);
-    p->prev_next      = PARAM_REF(vm, load_next);
-    p->prev_main_script = PARAM_REF(vm, load_main_script);
-    p->prev_reader_lexical_mode = Scm_ReaderLexicalMode();
-    p->prev_situation = vm->evalSituation;
-
-    ScmReadContext *newctx = Scm_MakeReadContext(NULL);
-    newctx->flags |= RCTX_LITERAL_IMMUTABLE | RCTX_SOURCE_INFO;
-    p->prev_ctx = Scm_SetCurrentReadContext(newctx);
-
-    PARAM_SET(vm, load_next, next_paths);
-    PARAM_SET(vm, load_port, SCM_OBJ(port));
-    PARAM_SET(vm, load_main_script, SCM_MAKE_BOOL(flags&SCM_LOAD_MAIN_SCRIPT));
-    vm->module = module;
-    vm->evalSituation = SCM_VM_LOADING;
-    ScmObj port_info;
-    if (SCM_PORTP(p->prev_port)) {
-        port_info = SCM_LIST2(p->prev_port,
-                              Scm_MakeInteger(Scm_PortLine(SCM_PORT(p->prev_port))));
-    } else {
-        port_info = SCM_LIST1(SCM_FALSE);
-    }
-    PARAM_SET(vm,load_history, Scm_Cons(port_info, PARAM_REF(vm,load_history)));
-
-    PORT_LOCK(port, vm);
-    return Scm_VMDynamicWindC(NULL, load_body, load_after, p);
-}
-
 int Scm_LoadFromPort(ScmPort *port, u_long flags, ScmLoadPacket *packet)
 {
     static ScmObj load_from_port = SCM_UNDEFINED;
@@ -276,10 +150,6 @@ int Scm_LoadFromPort(ScmPort *port, u_long flags, ScmLoadPacket *packet)
     SCM_BIND_PROC(load_from_port, "load-from-port", Scm_GaucheModule());
     load_packet_prepare(packet);
 
-    if (flags&SCM_LOAD_MAIN_SCRIPT) {
-        args = Scm_Cons(key_main_script, Scm_Cons(SCM_TRUE, args));
-    }
-
     args = Scm_Cons(SCM_OBJ(port), args);
 
     if (flags&SCM_LOAD_PROPAGATE_ERROR) {
@@ -311,28 +181,6 @@ int Scm_LoadFromPort(ScmPort *port, u_long flags, ScmLoadPacket *packet)
  *  flags      - combination of ScmLoadFlags.
  */
 
-/* internal auxiliary function called from Scheme `load' */
-void Scm__RecordLoadStart(ScmObj load_file_path)
-{
-    ScmVM *vm = Scm_VM();
-#ifdef HAVE_GETTIMEOFDAY
-    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_COLLECT_LOAD_STATS)) {
-        struct timeval t0;
-        gettimeofday(&t0, NULL);
-        vm->stat.loadStat =
-            Scm_Acons(load_file_path,
-                      Scm_MakeIntegerU(t0.tv_sec*1000000+t0.tv_usec),
-                      vm->stat.loadStat);
-    }
-#endif /*HAVE_GETTIMEOFDAY*/
-    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_LOAD_VERBOSE)) {
-        int len = Scm_Length(PARAM_REF(vm, load_history));
-        SCM_PUTZ(";;", 2, SCM_CURERR);
-        while (len-- > 0) SCM_PUTC(' ', SCM_CURERR);
-        Scm_Printf(SCM_CURERR, "Loading %A...\n", load_file_path);
-    }
-}
-
 /* The real `load' function is moved to Scheme.  This is a C stub to
    call it. */
 ScmObj Scm_VMLoad(ScmString *filename, ScmObj paths, ScmObj env, int flags)
@@ -347,9 +195,6 @@ ScmObj Scm_VMLoad(ScmString *filename, ScmObj paths, ScmObj env, int flags)
     if (flags&SCM_LOAD_IGNORE_CODING) {
         opts = Scm_Cons(key_ignore_coding, Scm_Cons(SCM_TRUE, opts));
     }
-    if (flags&SCM_LOAD_MAIN_SCRIPT) {
-        opts = Scm_Cons(key_main_script, Scm_Cons(SCM_TRUE, opts));
-    }
     if (SCM_NULLP(paths) || SCM_PAIRP(paths)) {
         opts = Scm_Cons(key_paths, Scm_Cons(paths, opts));
     }
@@ -372,9 +217,6 @@ int Scm_Load(const char *cpath, u_long flags, ScmLoadPacket *packet)
     if (flags&SCM_LOAD_IGNORE_CODING) {
         opts = Scm_Cons(key_ignore_coding, Scm_Cons(SCM_TRUE, opts));
     }
-    if (flags&SCM_LOAD_MAIN_SCRIPT) {
-        opts = Scm_Cons(key_main_script, Scm_Cons(SCM_TRUE, opts));
-    }
 
     load_packet_prepare(packet);
     if (flags&SCM_LOAD_PROPAGATE_ERROR) {
@@ -394,6 +236,14 @@ int Scm_Load(const char *cpath, u_long flags, ScmLoadPacket *packet)
     }
 }
 
+/* A convenience routine */
+int Scm_LoadFromCString(const char *program, u_long flags, ScmLoadPacket *p)
+{
+    ScmObj ip = Scm_MakeInputStringPort(SCM_STRING(SCM_MAKE_STR(program)), TRUE);
+    return Scm_LoadFromPort(SCM_PORT(ip), flags, p);
+}
+
+
 /*
  * Utilities
  */
@@ -457,7 +307,7 @@ static ScmObj add_list_item(ScmObj orig, ScmObj item, int afterp)
 ScmObj Scm_AddLoadPath(const char *cpath, int afterp)
 {
     ScmObj spath = SCM_MAKE_STR_COPYING(cpath);
-    struct stat statbuf;
+    ScmStat statbuf;
 
     /* check dynload path */
     ScmObj dpath = Scm_StringAppendC(SCM_STRING(spath), "/", 1, 1);
@@ -547,7 +397,7 @@ void Scm_DeleteLoadPathHook(ScmObj proc)
 /* The implementation of dynamic loader is a bit complicated in the presense
    of multiple threads and multiple initialization routines.
 
-   We keep struct dlobj_rec record for each DYNAMIC-LOADed files (keyed
+   We keep ScmDLObj record for each DYNAMIC-LOADed files (keyed
    by pathname including suffix) to track the state of loading.  The thread
    must lock the structure first to operate on the particluar DSO.
 
@@ -562,21 +412,15 @@ void Scm_DeleteLoadPathHook(ScmObj proc)
 
 typedef void (*ScmDynLoadInitFn)(void);
 
-enum dlobj_state {
-    DLOBJ_NONE,             /* dlopen and dlsym haven't completed. */
-    DLOBJ_LOADED,           /* dlopened and initfn found, but not init'ed */
-    DLOBJ_INITIALIZED       /* initialized and ready to use */
-};
-
-struct dlobj_initfn_rec {
+typedef struct dlobj_initfn_rec {
     struct dlobj_initfn_rec *next; /* chain */
     const char *name;           /* name of initfn (always w/ leading '_') */
     ScmDynLoadInitFn fn;        /* function ptr */
     int initialized;            /* TRUE once fn returns */
-};
+} dlobj_initfn;
 
-struct dlobj_rec {
-    dlobj *next;                /* chain */
+struct ScmDLObjRec {
+    SCM_HEADER;
     const char *path;           /* pathname for DSO, including suffix */
     int loaded;                 /* TRUE if this DSO is already loaded.
                                    It may need to be initialized, though.
@@ -589,28 +433,25 @@ struct dlobj_rec {
     ScmInternalCond  cv;
 };
 
-#if 0 /* for debug */
-static void dump_dlobj(dlobj *dlo)
+static void dlobj_print(ScmObj obj, ScmPort *sink, ScmWriteContext *mode)
 {
-    printf("{\n  dlobj \"%s\"%s\n",
-           dlo->path, dlo->loaded? " (loaded)" : " (not loaded)");
-    printf("  initfns:\n");
-    for (dlobj_initfn *ifn = dlo->initfns; ifn; ifn = ifn->next) {
-        printf("    name=\"%s\", fn=%p %s}\n",
-               ifn->name, ifn->fn,
-               ifn->initialized? "(initialized) ":"");
-    }
-    printf("}\n");
+    Scm_Printf(sink, "#<dlobj \"%s\">", SCM_DLOBJ(obj)->path);
 }
 
-static void dump_dlobjs(void)
+SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_DLObjClass, dlobj_print);
+
+static ScmDLObj *make_dlobj(const char *path)
 {
-    printf("Registerd dlobjs:\n");
-    for (dlobj *dlo = ldinfo.dso_list; dlo; dlo = dlo->next) {
-        dump_dlobj(dlo);
-    }
+    ScmDLObj *z = SCM_NEW(ScmDLObj);
+    SCM_SET_CLASS(z, &Scm_DLObjClass);
+    z->path = path;
+    z->loader = NULL;
+    z->loaded = FALSE;
+    z->initfns = NULL;
+    (void)SCM_INTERNAL_MUTEX_INIT(z->mutex);
+    (void)SCM_INTERNAL_COND_INIT(z->cv);
+    return z;
 }
-#endif
 
 /* NB: we rely on dlcompat library for dlopen instead of using dl_darwin.c
    for now; Boehm GC requires dlopen when compiled with pthread, so there's
@@ -624,29 +465,25 @@ static void dump_dlobjs(void)
 #endif
 
 /* Find dlobj with path, creating one if there aren't, and returns it. */
-static dlobj *find_dlobj(const char *path)
+static ScmDLObj *find_dlobj(const char *path)
 {
-    dlobj *z = NULL;
+    ScmObj p;
+    ScmObj spath = SCM_MAKE_STR_COPYING(path);
+    ScmDLObj *z = NULL;
+
     (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.dso_mutex);
-    for (z = ldinfo.dso_list; z; z = z->next) {
-        if (strcmp(z->path, path) == 0) break;
-    }
-    if (z == NULL) {
-        z = SCM_NEW(dlobj);
-        z->path = path;
-        z->loader = NULL;
-        z->loaded = FALSE;
-        z->initfns = NULL;
-        (void)SCM_INTERNAL_MUTEX_INIT(z->mutex);
-        (void)SCM_INTERNAL_COND_INIT(z->cv);
-        z->next = ldinfo.dso_list;
-        ldinfo.dso_list = z;
+    p = Scm_HashTableRef(ldinfo.dso_table, spath, SCM_FALSE);
+    if (SCM_DLOBJP(p)) {
+        z = SCM_DLOBJ(p);
+    } else {
+        z = make_dlobj(path);
+        Scm_HashTableSet(ldinfo.dso_table, spath, SCM_OBJ(z), 0);
     }
     (void)SCM_INTERNAL_MUTEX_UNLOCK(ldinfo.dso_mutex);
     return z;
 }
 
-static void lock_dlobj(dlobj *dlo)
+static void lock_dlobj(ScmDLObj *dlo)
 {
     ScmVM *vm = Scm_VM();
     (void)SCM_INTERNAL_MUTEX_LOCK(dlo->mutex);
@@ -658,7 +495,7 @@ static void lock_dlobj(dlobj *dlo)
     (void)SCM_INTERNAL_MUTEX_UNLOCK(dlo->mutex);
 }
 
-static void unlock_dlobj(dlobj *dlo)
+static void unlock_dlobj(ScmDLObj *dlo)
 {
     (void)SCM_INTERNAL_MUTEX_LOCK(dlo->mutex);
     dlo->loader = NULL;
@@ -693,7 +530,7 @@ static const char *derive_dynload_initfn(const char *filename)
 
 /* find dlobj_initfn from the given dlobj with name.
    Assuming the caller holding the lock of OBJ. */
-static dlobj_initfn *find_initfn(dlobj *dlo, const char *name)
+static dlobj_initfn *find_initfn(ScmDLObj *dlo, const char *name)
 {
     dlobj_initfn *fns = dlo->initfns;
     for (; fns != NULL; fns = fns->next) {
@@ -714,8 +551,10 @@ static const char *find_dso_path(ScmString *dsoname)
     static ScmObj find_file = SCM_UNDEFINED;
     SCM_BIND_PROC(find_file, "find-load-file", Scm_GaucheInternalModule());
 
-    ScmObj spath = Scm_ApplyRec5(find_file, SCM_OBJ(dsoname), Scm_GetDynLoadPath(),
-                                 ldinfo.dso_suffixes, SCM_FALSE, SCM_FALSE);
+    ScmObj spath = Scm_ApplyRec3(find_file,
+                                 SCM_OBJ(dsoname),
+                                 Scm_GetDynLoadPath(),
+                                 ldinfo.dso_suffixes);
     if (SCM_FALSEP(spath)) {
         Scm_Error("can't find dlopen-able module %S", dsoname);
     }
@@ -738,13 +577,13 @@ const char *get_initfn_name(ScmObj initfn, const char *dsopath)
 
 /* Load the DSO.  The caller holds the lock of dlobj.  May throw an error;
    the caller makes sure it releases the lock even in that case. */
-static void load_dlo(dlobj *dlo)
+static void load_dlo(ScmDLObj *dlo)
 {
     ScmVM *vm = Scm_VM();
     if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_LOAD_VERBOSE)) {
         int len = Scm_Length(PARAM_REF(vm, load_history));
         SCM_PUTZ(";;", 2, SCM_CURERR);
-        while (len-- > 0) SCM_PUTC(' ', SCM_CURERR);
+        while (len-- > 0) Scm_Putz("  ", 2, SCM_CURERR);
         Scm_Printf(SCM_CURERR, "Dynamically Loading %s...\n", dlo->path);
     }
     dlo->handle = dl_open(dlo->path);
@@ -762,7 +601,7 @@ static void load_dlo(dlobj *dlo)
 
 /* Call the DSO's initfn.  The caller holds the lock of dlobj, and responsible
    to release the lock even when this fn throws an error. */
-static void call_initfn(dlobj *dlo, const char *name)
+static void call_initfn(ScmDLObj *dlo, const char *name)
 {
     dlobj_initfn *ifn = find_initfn(dlo, name);
 
@@ -823,7 +662,7 @@ void Scm_RegisterPrelinked(ScmString *dsoname,
                            ScmDynLoadInitFn initfns[])
 {
     const char *path = pseudo_pathname_for_prelinked(dsoname);
-    dlobj *dlo = find_dlobj(path);
+    ScmDLObj *dlo = find_dlobj(path);
     dlo->loaded = TRUE;
 
     (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.dso_mutex);
@@ -858,7 +697,7 @@ ScmObj Scm_DynLoad(ScmString *dsoname, ScmObj initfn, u_long flags/*reserved*/)
     const char *dsopath = find_prelinked(dsoname);
     if (!dsopath) dsopath = find_dso_path(dsoname);
     const char *initname = get_initfn_name(initfn, dsopath);
-    dlobj *dlo = find_dlobj(dsopath);
+    ScmDLObj *dlo = find_dlobj(dsopath);
 
     /* Load the dlobj if necessary. */
     lock_dlobj(dlo);
@@ -879,6 +718,54 @@ ScmObj Scm_DynLoad(ScmString *dsoname, ScmObj initfn, u_long flags/*reserved*/)
     return SCM_TRUE;
 }
 
+/* Expose dlobj to Scheme world */
+
+static ScmObj dlobj_path_get(ScmObj obj)
+{
+    return SCM_MAKE_STR_IMMUTABLE(SCM_DLOBJ(obj)->path);
+}
+
+static ScmObj dlobj_loaded_get(ScmObj obj)
+{
+    return SCM_MAKE_BOOL(SCM_DLOBJ(obj)->loaded);
+}
+
+static ScmObj dlobj_initfns_get(ScmObj obj)
+{
+    ScmObj h = SCM_NIL;
+    ScmObj t = SCM_NIL;
+    lock_dlobj(SCM_DLOBJ(obj));
+    dlobj_initfn *ifn = SCM_DLOBJ(obj)->initfns;
+    for (;ifn != NULL; ifn = ifn->next) {
+        ScmObj p = Scm_Cons(SCM_MAKE_STR_IMMUTABLE(ifn->name),
+                            SCM_MAKE_BOOL(ifn->initialized));
+        SCM_APPEND1(h, t, p);
+    }
+    unlock_dlobj(SCM_DLOBJ(obj));
+    return h;
+}
+
+static ScmClassStaticSlotSpec dlobj_slots[] = {
+    SCM_CLASS_SLOT_SPEC("path", dlobj_path_get, NULL),
+    SCM_CLASS_SLOT_SPEC("loaded?", dlobj_loaded_get, NULL),
+    SCM_CLASS_SLOT_SPEC("init-functions", dlobj_initfns_get, NULL),
+    SCM_CLASS_SLOT_SPEC_END()
+};
+
+ScmObj Scm_DLObjs()
+{
+    ScmObj z = SCM_NIL;
+    ScmHashIter iter;
+    ScmDictEntry *e;
+    (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.dso_mutex);
+    Scm_HashIterInit(&iter, SCM_HASH_TABLE_CORE(ldinfo.dso_table));
+    while ((e = Scm_HashIterNext(&iter)) != NULL) {
+        z = Scm_Cons(SCM_OBJ(SCM_DICT_VALUE(e)), z);
+    }
+    (void)SCM_INTERNAL_MUTEX_UNLOCK(ldinfo.dso_mutex);
+    return z;
+}
+
 /*------------------------------------------------------------------
  * Require and provide
  */
@@ -946,20 +833,21 @@ static int do_require(ScmObj, int, ScmModule *, ScmLoadPacket *);
    is loaded at this time or it has already been loaded.  With the same
    reason, it doesn't make much sense to use the current module.
 
-   Until 0.9.4 we didn't actually force 'require' to use #<module gauche>
-   as the base module.  Most of the time, 'require' is called in a module
-   that inherits gauche, so the problem didn't appear.  However, if one
-   requires a file from a module that doesn't inherit gauche, most likely
-   accidentally, he sees confusing errors---because "define-module" form
-   isn't recognized!
+   On 0.9.4 we always set the base module to #<module gauche> to do require,
+   so that we can guarantee the forms like define-module or define-library
+   to be visible from the loaded module (if we use the caller's current
+   module it is not guaranteed.)  However, it had an unexpected side
+   effect: If the loaded module inserts toplevel definitions or imports
+   other modules without first setting its own module, it actually
+   modifies #<module gauche>.
 
-   As of 0.9.4 we fix the base module to #<module gauche> when require is
-   called.   Note that autoload needs different requirements, so we have
-   a subroutine do_require that takes the desired base module.
+   As of 0.9.5, we use an immutable module #<module gauche.require-base>
+   as the base module.  Since it is immutable, any toplevel definitions
+   or imports without first switching modules are rejected.
  */
 int Scm_Require(ScmObj feature, int flags, ScmLoadPacket *packet)
 {
-    return do_require(feature, flags, Scm_GaucheModule(), packet);
+    return do_require(feature, flags, Scm__RequireBaseModule(), packet);
 }
 
 int do_require(ScmObj feature, int flags, ScmModule *base_mod,
@@ -1213,7 +1101,7 @@ ScmObj Scm_ResolveAutoload(ScmAutoload *adata, int flags)
            just in case we raise an error. */
         adata->locker = NULL;
         SCM_INTERNAL_COND_BROADCAST(adata->cv);
-        Scm_Error("Attempted to trigger the same autoload %S#%S recursively.  Maybe circular autoload dependency?\n",
+        Scm_Error("Attempted to trigger the same autoload %S#%S recursively.  Maybe circular autoload dependency?",
                   adata->module, adata->name);
     }
 
@@ -1269,28 +1157,53 @@ ScmObj Scm_ResolveAutoload(ScmAutoload *adata, int flags)
 ScmObj Scm_CurrentLoadHistory() { return PARAM_REF(Scm_VM(), load_history); }
 ScmObj Scm_CurrentLoadNext()    { return PARAM_REF(Scm_VM(), load_next); }
 ScmObj Scm_CurrentLoadPort()    { return PARAM_REF(Scm_VM(), load_port); }
-ScmObj Scm_LoadMainScript()     { return PARAM_REF(Scm_VM(), load_main_script); }
 
 /*------------------------------------------------------------------
  * Compatibility stuff
  */
 
+/* TRANSIENT: Pre-0.9 Compatibility routine.  Kept for the binary compatibility.
+   Will be removed on 1.0 */
 void Scm__LoadFromPortCompat(ScmPort *port, int flags)
 {
     Scm_LoadFromPort(port, flags|SCM_LOAD_PROPAGATE_ERROR, NULL);
 }
 
+/* TRANSIENT: Pre-0.9 Compatibility routine.  Kept for the binary compatibility.
+   Will be removed on 1.0 */
 int  Scm__LoadCompat(const char *file, int flags)
 {
     return (0 == Scm_Load(file, flags|SCM_LOAD_PROPAGATE_ERROR, NULL));
 }
 
+/* TRANSIENT: Pre-0.9 Compatibility routine.  Kept for the binary compatibility.
+   Will be removed on 1.0 */
 ScmObj Scm__RequireCompat(ScmObj feature)
 {
     Scm_Require(feature, SCM_LOAD_PROPAGATE_ERROR, NULL);
     return SCM_TRUE;
 }
 
+/* TRANSIENT: This is entirely moved to Scheme (libeval.scm).  The entry is
+   kept only for the binary compatibility. */
+ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths,
+                          ScmObj env, int flags)
+{
+    Scm_Error("[internal] Scm_VMLoadFromPort() is obsoleted; call load-from-port Scheme procedure.");
+    return SCM_UNDEFINED;
+}
+
+/* TRANSIENT: Kept for the binary compatibility; the feature
+   is in libeval.scm now. */
+void Scm__RecordLoadStart(ScmObj load_file_path/*ARGSUSED*/)
+{
+}
+
+/* TRANSIENT: Kept for the binary compatibility; not used anymore. */
+ScmObj Scm_LoadMainScript()
+{
+    return SCM_UNDEFINED;
+}
 
 /*------------------------------------------------------------------
  * Initialization
@@ -1298,8 +1211,7 @@ ScmObj Scm__RequireCompat(ScmObj feature)
 
 void Scm__InitLoad(void)
 {
-    ScmModule *m = Scm_SchemeModule();
-    ScmVM *vm = Scm_VM();
+    ScmModule *m = Scm_GaucheModule();
     ScmObj t;
 
     ScmObj init_load_path = t = SCM_NIL;
@@ -1313,6 +1225,7 @@ void Scm__InitLoad(void)
     SCM_APPEND1(init_dynload_path, t, Scm_ArchitectureDirectory());
 
     ScmObj init_load_suffixes = t = SCM_NIL;
+    SCM_APPEND1(init_load_suffixes, t, SCM_MAKE_STR(".sld")); /* R7RS library */
     SCM_APPEND1(init_load_suffixes, t, SCM_MAKE_STR(".sci"));
     SCM_APPEND1(init_load_suffixes, t, SCM_MAKE_STR(".scm"));
 
@@ -1324,10 +1237,12 @@ void Scm__InitLoad(void)
     key_error_if_not_found = SCM_MAKE_KEYWORD("error-if-not-found");
     key_macro = SCM_MAKE_KEYWORD("macro");
     key_ignore_coding = SCM_MAKE_KEYWORD("ignore-coding");
-    key_main_script = SCM_MAKE_KEYWORD("main-script");
     key_paths = SCM_MAKE_KEYWORD("paths");
     key_environment = SCM_MAKE_KEYWORD("environment");
 
+    Scm_InitStaticClass(SCM_CLASS_DLOBJ, "<dlobj>",
+                        m, dlobj_slots, 0);
+
 #define DEF(rec, sym, val) \
     rec = SCM_GLOC(Scm_Define(m, SCM_SYMBOL(sym), val))
 
@@ -1336,23 +1251,18 @@ void Scm__InitLoad(void)
     DEF(ldinfo.load_suffixes_rec, SCM_SYM_LOAD_SUFFIXES, init_load_suffixes);
     DEF(ldinfo.load_path_hooks_rec, SCM_SYM_LOAD_PATH_HOOKS, SCM_NIL);
 
-    ldinfo.provided =
-        SCM_LIST5(SCM_MAKE_STR("srfi-2"), /* and-let* */
-                  SCM_MAKE_STR("srfi-6"), /* string ports (builtin) */
-                  SCM_MAKE_STR("srfi-8"), /* receive (builtin) */
-                  SCM_MAKE_STR("srfi-10"), /* #, (builtin) */
-                  SCM_MAKE_STR("srfi-17")  /* set! (builtin) */
-            );
+    /* NB: Some modules are built-in.  We'll register them to the
+       provided list, in libomega.scm. */
+    ldinfo.provided = SCM_NIL;
     ldinfo.providing = SCM_NIL;
     ldinfo.waiting = SCM_NIL;
     ldinfo.dso_suffixes = SCM_LIST2(SCM_MAKE_STR(".la"),
                                     SCM_MAKE_STR("." SHLIB_SO_SUFFIX));
-    ldinfo.dso_list = NULL;
+    ldinfo.dso_table = SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_STRING,0));
     ldinfo.dso_prelinked = SCM_NIL;
 
-#define PARAM_INIT(name, val) Scm_InitParameterLoc(vm, &ldinfo.name, val)
-    PARAM_INIT(load_history, SCM_NIL);
-    PARAM_INIT(load_next, SCM_NIL);
-    PARAM_INIT(load_port, SCM_FALSE);
-    PARAM_INIT(load_main_script, SCM_FALSE);
+#define PARAM_INIT(var, name, val) Scm_DefinePrimitiveParameter(m, name, val, &ldinfo.var)
+    PARAM_INIT(load_history, "current-load-history", SCM_NIL);
+    PARAM_INIT(load_next, "current-load-next", SCM_NIL);
+    PARAM_INIT(load_port, "current-load-port", SCM_FALSE);
 }
