@@ -1,7 +1,7 @@
 /*
  * macro.c - macro implementation
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,9 +33,9 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/macro.h"
 #include "gauche/code.h"
 #include "gauche/vminsn.h"
+#include "gauche/priv/macroP.h"
 #include "gauche/priv/builtin-syms.h"
 
 /* avoid C++ reserved name conflict.
@@ -51,7 +51,8 @@
 
 static void syntax_print(ScmObj obj, ScmPort *port, ScmWriteContext *mode)
 {
-    Scm_Printf(port, "#<syntax %A>", SCM_SYNTAX(obj)->name);
+    ScmSymbol *name = SCM_SYNTAX(obj)->name;
+    Scm_Printf(port, "#<syntax %A>", (name ? SCM_OBJ(name) : SCM_FALSE));
 }
 
 SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SyntaxClass, syntax_print);
@@ -71,22 +72,26 @@ ScmObj Scm_MakeSyntax(ScmSymbol *name, ScmObj handler)
 
 static void macro_print(ScmObj obj, ScmPort *port, ScmWriteContext *mode)
 {
-    Scm_Printf(port, "#<macro %A>", SCM_MACRO(obj)->name);
+    ScmSymbol *name = SCM_MACRO(obj)->name;
+    Scm_Printf(port, "#<macro %A>", (name ? SCM_OBJ(name) : SCM_FALSE));
 }
 
 SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_MacroClass, macro_print);
 
-ScmObj Scm_MakeMacro(ScmSymbol *name, ScmTransformerProc transformer,
-                     void *data)
+ScmObj Scm_MakeMacro(ScmSymbol *name, ScmObj transformer)
 {
     ScmMacro *s = SCM_NEW(ScmMacro);
     SCM_SET_CLASS(s, SCM_CLASS_MACRO);
     s->name = name;
     s->transformer = transformer;
-    s->data = data;
     return SCM_OBJ(s);
 }
 
+ScmObj Scm_MacroTransformer(ScmMacro *mac)
+{
+    return mac->transformer;
+}
+
 /*===================================================================
  * SyntaxPattern object
  * Repesents a repeatable subpattern
@@ -167,55 +172,6 @@ ScmSyntaxRules *make_syntax_rules(int nr)
     return r;
 }
 
-/*===================================================================
- * Macro for the new compiler
- */
-
-/* In the new compiler, macro transformers for hygienic and traditional
- * macros are integrated.
- * The lowest-level macro transformer can be introduced by define-syntax,
- * let-syntax and letrec-syntax (but not syntax-case or syntax-rules; they
- * are built on top of it).
- *
- *   (define-syntax foo <transformer>)
- *
- * Where <transformer> is a procedure that takes one argument, a syntactic
- * closure.  It must return a syntactic closure as the result of trans
- * formation.
- *
- * From the point of the compiler, define-syntax triggers the following
- * actions.
- *
- *  - evaluate <transformer> in the compiler environment.
- *  - encapsulate it into <macro> object, and insert it to the compiler
- *    environment.
- *  - insert the binding to foo in the runtime toplevel environment.
- *
- * Define-macro is also built on top of define-syntax.  Concepturally,
- * it is transformed as follows.
- *
- *  (define-macro foo procedure)
- *   => (define-syntax foo
- *        (lambda (x)
- *          (let ((env  (slot-ref x 'env))
- *                (form (slot-ref x 'expr)))
- *            (make-syntactic-closure
- *              env () (apply procedure form)))))
- */
-
-static ScmObj macro_transform(ScmObj self, ScmObj form, ScmObj env,
-                              void *data)
-{
-    ScmObj proc = SCM_OBJ(data);
-    SCM_ASSERT(SCM_SYNTACTIC_CLOSURE_P(form));
-    return Scm_ApplyRec(proc, SCM_LIST1(form));
-}
-
-ScmObj Scm_MakeMacroTransformer(ScmSymbol *name, ScmObj proc)
-{
-    return Scm_MakeMacro(name, macro_transform, (void*)proc);
-}
-
 /*===================================================================
  * Traditional Macro
  */
@@ -224,9 +180,10 @@ ScmObj Scm_MakeMacroTransformer(ScmSymbol *name, ScmObj proc)
 /* TODO: better error message on syntax error (macro invocation with
    bad number of arguments) */
 
-static ScmObj macro_transform_old(ScmObj self, ScmObj form,
-                                  ScmObj env, void *data)
+static ScmObj macro_transform_old(ScmObj *argv, int argc, void *data)
 {
+    SCM_ASSERT(argc == 2);
+    ScmObj form = argv[0];      /* we ignore env (argv[1]) */
     ScmObj proc = SCM_OBJ(data);
     SCM_ASSERT(SCM_PAIRP(form));
     return Scm_VMApply(proc, SCM_CDR(form));
@@ -234,7 +191,9 @@ static ScmObj macro_transform_old(ScmObj self, ScmObj form,
 
 ScmObj Scm_MakeMacroTransformerOld(ScmSymbol *name, ScmProcedure *proc)
 {
-    return Scm_MakeMacro(name, macro_transform_old, (void*)proc);
+    ScmObj transformer = Scm_MakeSubr(macro_transform_old, proc, 2, 0,
+                                      SCM_FALSE);
+    return Scm_MakeMacro(name, transformer);
 }
 
 static ScmMacro *resolve_macro_autoload(ScmAutoload *adata)
@@ -249,15 +208,23 @@ static ScmMacro *resolve_macro_autoload(ScmAutoload *adata)
     return SCM_MACRO(mac);
 }
 
-static ScmObj macro_autoload(ScmObj self, ScmObj form, ScmObj env, void *data)
+static ScmObj macro_autoload(ScmObj *argv, int argc, void *data)
 {
+    SCM_ASSERT(argc == 2);
+    ScmObj form = argv[0];
+    ScmObj env = argv[1];
+    /* Important to save form and env before calling resolve_macro_autoload,
+       for it may overwrite stack region pointed by argv. */
+    SCM_ASSERT(SCM_AUTOLOADP(data));
     ScmMacro *mac = resolve_macro_autoload(SCM_AUTOLOAD(data));
-    return mac->transformer(SCM_OBJ(mac), form, env, mac->data);
+    return Scm_CallMacroExpander(mac, form, env);
 }
 
 ScmObj Scm_MakeMacroAutoload(ScmSymbol *name, ScmAutoload *adata)
 {
-    return Scm_MakeMacro(name, macro_autoload, (void*)adata);
+    ScmObj transformer = Scm_MakeSubr(macro_autoload, adata,
+                                      2, 0, SCM_FALSE);
+    return Scm_MakeMacro(name, transformer);
 }
 
 /*===================================================================
@@ -372,8 +339,17 @@ static int isEllipsis(PatternContext *ctx, ScmObj obj)
 {
     if (SCM_FALSEP(ctx->ellipsis)) return FALSE;
     if (SCM_IDENTIFIERP(obj)) {
-        if (!SCM_NULLP(SCM_IDENTIFIER(obj)->env)) return FALSE;
-        return SCM_EQ(ctx->ellipsis, SCM_OBJ(SCM_IDENTIFIER(obj)->name));
+        if (SCM_IDENTIFIERP(ctx->ellipsis)) {
+            static ScmObj free_identifier_eq_proc = SCM_UNDEFINED;
+            SCM_BIND_PROC(free_identifier_eq_proc, "free-identifier=?",
+                          Scm_GaucheInternalModule());
+            if (!SCM_NULLP(SCM_IDENTIFIER(obj)->env)) return FALSE;
+            return !SCM_FALSEP(Scm_ApplyRec2(free_identifier_eq_proc,
+                                             ctx->ellipsis, obj));
+        } else {
+            return SCM_EQ(ctx->ellipsis,
+                          SCM_OBJ(Scm_UnwrapIdentifier(SCM_IDENTIFIER(obj))));
+        }
     } else {
         return SCM_EQ(ctx->ellipsis, obj);
     }
@@ -395,7 +371,7 @@ static ScmObj preprocess_literals(ScmObj literals, ScmModule *mod, ScmObj env)
         if (SCM_IDENTIFIERP(lit))
             SCM_APPEND1(h, t, lit);
         else if (SCM_SYMBOLP(lit))
-            SCM_APPEND1(h, t, Scm_MakeIdentifier(SCM_SYMBOL(lit), mod, env));
+            SCM_APPEND1(h, t, Scm_MakeIdentifier(lit, mod, env));
         else if (SCM_KEYWORDP(lit))
             /* This branch is to allow r7rs-compliant code to go through
                with legacy mode.  If keyword-symbol integration is turned on,
@@ -521,8 +497,7 @@ static ScmObj compile_rule1(ScmObj form,
                 if (SCM_IDENTIFIERP(form)) {
                     id = form;
                 } else {
-                    id = Scm_MakeIdentifier(SCM_SYMBOL(form),
-                                            ctx->mod, ctx->env);
+                    id = Scm_MakeIdentifier(form, ctx->mod, ctx->env);
                 }
                 ctx->tvars = Scm_Cons(id, ctx->tvars);
                 return id;
@@ -672,7 +647,7 @@ static void print_matchvec(MatchVar *mvec, int numPvars, ScmPort *port)
 }
 #endif
 
-static int match_synrule(ScmObj form, ScmObj pattern, ScmObj env,
+static int match_synrule(ScmObj form, ScmObj pattern, ScmObj mod, ScmObj env,
                          MatchVar *mvec);
 
 #define SPROUT  Scm_Cons(SCM_NIL, SCM_NIL)
@@ -749,22 +724,31 @@ static inline void match_insert(ScmObj pvref, ScmObj matched, MatchVar *mvec)
 }
 
 /* see if literal identifier ID in the pattern matches the given object */
-static inline int match_identifier(ScmIdentifier *id, ScmObj obj, ScmObj env)
+static inline int match_identifier(ScmIdentifier *id, ScmObj obj,
+                                   ScmObj mod, ScmObj env)
 {
     if (SCM_SYMBOLP(obj)) {
-        return (id->name == SCM_SYMBOL(obj)
-                && Scm_IdentifierBindingEqv(id, SCM_SYMBOL(obj), env));
+        /* This is temporary: We don't want to allocate identifier for
+           every bare symbol we want to match.  But this is the shortcut
+           to do the right thing (using free-identifier=? to comapre literals.
+           Let's think optimization later. */
+        SCM_ASSERT(SCM_MODULEP(mod));
+        obj = Scm_MakeIdentifier(obj, SCM_MODULE(mod), env);
     }
     if (SCM_IDENTIFIERP(obj)) {
-        /*TODO: module?*/
-        return (id->name == SCM_IDENTIFIER(obj)->name
-                && id->env == SCM_IDENTIFIER(obj)->env);
+        /* free-identifier=? is defined in Scheme (libmod.scm)  */
+        static ScmObj free_identifier_eq_proc = SCM_UNDEFINED;
+        SCM_BIND_PROC(free_identifier_eq_proc, "free-identifier=?",
+                      Scm_GaucheInternalModule());
+        return !SCM_FALSEP(Scm_ApplyRec2(free_identifier_eq_proc,
+                                         SCM_OBJ(id), obj));
     }
     return FALSE;
 }
 
 static inline int match_subpattern(ScmObj form, ScmSyntaxPattern *pat,
-                                   ScmObj rest, ScmObj env, MatchVar *mvec)
+                                   ScmObj rest, ScmObj mod, ScmObj env,
+                                   MatchVar *mvec)
 {
     /* TODO: If pat->numFollowingItems == 0, we don't need to calculate
        length beforehand.  Some optimization opportunity. */
@@ -776,19 +760,19 @@ static inline int match_subpattern(ScmObj form, ScmSyntaxPattern *pat,
 
     enter_subpattern(pat, mvec);
     while (limit > 0) {
-        if (!match_synrule(SCM_CAR(form), pat->pattern, env, mvec))
+        if (!match_synrule(SCM_CAR(form), pat->pattern, mod, env, mvec))
             return FALSE;
         form = SCM_CDR(form);
         limit--;
     }
     exit_subpattern(pat, mvec);
-    return match_synrule(form, rest, env, mvec);
+    return match_synrule(form, rest, mod, env, mvec);
 }
 
 /* See if form matches pattern.  If match, add matched syntax variable
    bindings to match vector and return TRUE; otherwise, return FALSE
 */
-static int match_synrule(ScmObj form, ScmObj pattern, ScmObj env,
+static int match_synrule(ScmObj form, ScmObj pattern, ScmObj mod, ScmObj env,
                          MatchVar *mvec)
 {
     if (PVREF_P(pattern)) {
@@ -796,11 +780,11 @@ static int match_synrule(ScmObj form, ScmObj pattern, ScmObj env,
         return TRUE;
     }
     if (SCM_IDENTIFIERP(pattern)) {
-        return match_identifier(SCM_IDENTIFIER(pattern), form, env);
+        return match_identifier(SCM_IDENTIFIER(pattern), form, mod, env);
     }
     if (SCM_SYNTAX_PATTERN_P(pattern)) {
         return match_subpattern(form, SCM_SYNTAX_PATTERN(pattern),
-                                SCM_NIL, env, mvec);
+                                SCM_NIL, mod, env, mvec);
     }
     if (SCM_PAIRP(pattern)) {
         while (SCM_PAIRP(pattern)) {
@@ -808,18 +792,18 @@ static int match_synrule(ScmObj form, ScmObj pattern, ScmObj env,
             if (SCM_SYNTAX_PATTERN_P(elt)) {
                 return match_subpattern(form, SCM_SYNTAX_PATTERN(elt),
                                         SCM_CDR(pattern),
-                                        env, mvec);
+                                        mod, env, mvec);
             } else if (!SCM_PAIRP(form)) {
                 return FALSE;
             } else {
-                if (!match_synrule(SCM_CAR(form), elt, env, mvec))
+                if (!match_synrule(SCM_CAR(form), elt, mod, env, mvec))
                     return FALSE;
                 pattern = SCM_CDR(pattern);
                 form = SCM_CDR(form);
             }
         }
         if (!SCM_NULLP(pattern))
-            return match_synrule(form, pattern, env, mvec);
+            return match_synrule(form, pattern, mod, env, mvec);
         else
             return SCM_NULLP(form);
     }
@@ -842,7 +826,7 @@ static int match_synrule(ScmObj form, ScmObj pattern, ScmObj env,
         for (int i=0; i < elli; i++) {
             if (!match_synrule(SCM_VECTOR_ELEMENT(form, i),
                                SCM_VECTOR_ELEMENT(pattern, i),
-                               env, mvec))
+                               mod, env, mvec))
                 return FALSE;
         }
         if (elli < flen) {
@@ -850,7 +834,7 @@ static int match_synrule(ScmObj form, ScmObj pattern, ScmObj env,
             ScmObj prest = Scm_VectorToList(SCM_VECTOR(pattern), elli+1, plen);
             ScmObj frest = Scm_VectorToList(SCM_VECTOR(form), elli, flen);
             return match_subpattern(frest, SCM_SYNTAX_PATTERN(pat),
-                                    prest, env, mvec);
+                                    prest, mod, env, mvec);
         } else {
             return TRUE;
         }
@@ -928,7 +912,7 @@ static ScmObj realize_template_rec(ScmObj template,
         return Scm_ListToVector(h, 0, -1);
     }
     if (SCM_IDENTIFIERP(template)) {
-        /* we copy the identifier, so that the symbol bindings introduced
+        /* we wrap the identifier, so that the symbol bindings introduced
            by recursive macro call won't interfere each other.
            (e.g. the macro definitions of "letrec" and "do" shown in R5RS
            use the fact that the symbol "newtemp" introduced in each
@@ -936,7 +920,7 @@ static ScmObj realize_template_rec(ScmObj template,
         ScmObj p = Scm_Assq(template, *idlist);
         if (SCM_PAIRP(p)) return SCM_CDR(p);
         else {
-            ScmObj id = Scm_CopyIdentifier(SCM_IDENTIFIER(template));
+            ScmObj id = Scm_WrapIdentifier(SCM_IDENTIFIER(template));
             *idlist = Scm_Acons(template, id, *idlist);
             return id;
         }
@@ -959,7 +943,7 @@ static ScmObj realize_template(ScmSyntaxRuleBranch *branch,
     return realize_template_rec(branch->template, mvec, 0, indices, &idlist, &exlev);
 }
 
-static ScmObj synrule_expand(ScmObj form, ScmObj env, ScmSyntaxRules *sr)
+static ScmObj synrule_expand(ScmObj form, ScmObj mod, ScmObj env, ScmSyntaxRules *sr)
 {
     MatchVar *mvec = alloc_matchvec(sr->maxNumPvars);
 
@@ -971,7 +955,7 @@ static ScmObj synrule_expand(ScmObj form, ScmObj env, ScmSyntaxRules *sr)
         Scm_Printf(SCM_CUROUT, "pattern #%d: %S\n", i, sr->rules[i].pattern);
 #endif
         init_matchvec(mvec, sr->rules[i].numPvars);
-        if (match_synrule(SCM_CDR(form), sr->rules[i].pattern, env, mvec)) {
+        if (match_synrule(SCM_CDR(form), sr->rules[i].pattern, mod, env, mvec)) {
 #ifdef DEBUG_SYNRULE
             Scm_Printf(SCM_CUROUT, "success #%d:\n", i);
             print_matchvec(mvec, sr->rules[i].numPvars, SCM_CUROUT);
@@ -987,11 +971,16 @@ static ScmObj synrule_expand(ScmObj form, ScmObj env, ScmSyntaxRules *sr)
     return SCM_NIL;
 }
 
-static ScmObj synrule_transform(ScmObj self, ScmObj form, ScmObj env,
-                                void *data)
+static ScmObj synrule_transform(ScmObj *argv, int argc, void *data)
 {
+    SCM_ASSERT(argc == 2);
+    ScmObj form = argv[0];
+    ScmObj cenv = argv[1];
+    SCM_ASSERT(SCM_VECTORP(cenv));
+    ScmObj module = SCM_VECTOR_ELEMENT(cenv, 0); /* macro use env */
+    ScmObj frames = SCM_VECTOR_ELEMENT(cenv, 1); /* macro use env */
     ScmSyntaxRules *sr = (ScmSyntaxRules *)data;
-    return synrule_expand(form, env, sr);
+    return synrule_expand(form, module, frames, sr);
 }
 
 /* NB: a stub for the new compiler (TEMPORARY) */
@@ -1003,64 +992,29 @@ ScmObj Scm_CompileSyntaxRules(ScmObj name, ScmObj ellipsis,
     if (!SCM_MODULEP(mod)) Scm_Error("module required, but got %S", mod);
     ScmSyntaxRules *sr = compile_rules(name, ellipsis, literals, rules,
                                        SCM_MODULE(mod), env);
-    return Scm_MakeMacro(SCM_SYMBOL(name), synrule_transform, (void*)sr);
+    ScmObj sr_xform = Scm_MakeSubr(synrule_transform, sr,
+                                   2, 0, SCM_FALSE);
+    return Scm_MakeMacro(SCM_SYMBOL(name), sr_xform);
 }
 
 /*===================================================================
  * macro-expand
  */
 
-ScmObj macro_expand_cc(ScmObj result, void **data)
-{
-    ScmObj env = SCM_OBJ(data[0]);
-    return Scm_VMMacroExpand(result, env, FALSE);
-}
-
+/* TRANSIENT
+   Now it's in compile.scm (%internal-macro-expand).  This is kept
+   for ABI compatibility, but nobody is supposed to call this.
+ */
 ScmObj Scm_VMMacroExpand(ScmObj expr, ScmObj env, int oncep)
 {
-    ScmMacro *mac;
-
-    if (!SCM_PAIRP(expr)) return expr;
-    ScmObj op = SCM_CAR(expr);
-    if (SCM_MACROP(op)) {
-        mac = SCM_MACRO(op);
-    } else if (!SCM_SYMBOLP(op) && !SCM_IDENTIFIERP(op)) {
-        return expr;
-    } else {
-        mac = NULL;
-        ScmObj sym = op;
-        if (SCM_MACROP(sym)) {
-            /* local syntactic binding */
-            mac = SCM_MACRO(sym);
-        } else {
-            ScmGloc *g = NULL;
-            if (SCM_IDENTIFIERP(sym)) {
-                g = Scm_FindBinding(SCM_IDENTIFIER(sym)->module,
-                                    SCM_IDENTIFIER(sym)->name, 0);
-            } else if (SCM_SYMBOLP(sym)) {
-                g = Scm_FindBinding(Scm_VM()->module,
-                                    SCM_SYMBOL(sym), 0);
-            }
-            if (g) {
-                ScmObj gv = SCM_GLOC_GET(g);
-                if (SCM_MACROP(gv)) mac = SCM_MACRO(gv);
-            }
-        }
-    }
-    if (mac) {
-        if (!oncep) {
-            void *data[1];
-            data[0] = env;
-            Scm_VMPushCC(macro_expand_cc, data, 1);
-        }
-        expr = Scm_CallMacroExpander(mac, expr, env);
-    }
-    return expr;
+    Scm_Error("Scm_VMMacroExpand is obsoleted.");
+    return SCM_UNDEFINED;
 }
 
-ScmObj Scm_CallMacroExpander(ScmMacro *mac, ScmObj expr, ScmObj env)
+ScmObj Scm_CallMacroExpander(ScmMacro *mac, ScmObj expr, ScmObj cenv)
 {
-    return mac->transformer(SCM_OBJ(mac), expr, env, mac->data);
+    SCM_ASSERT(SCM_VECTORP(cenv));
+    return Scm_ApplyRec2(mac->transformer, expr, cenv);
 }
 
 /*===================================================================
