@@ -1,7 +1,7 @@
 /*
  * main.c - interpreter main program
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -89,9 +89,8 @@ void usage(void)
             "           name of the module in which the 'main' procedure is defined.\n"
             "           By default, the 'main' procedure in the user module is called\n"
             "           after loading the script (srfi-22).  This option allows to call\n"
-            "           a main procedure in the different module\n"
-            "  -p<type> Turns on the profiler.  Currently <type> can only be\n"
-            "           'time'.\n"
+            "           a main procedure in the different module.\n"
+            "  -p<type> Turns on the profiler.  <Type> can be 'time' or 'load'.\n"
             "  -F<feature> Makes <feature> available in cond-expand forms\n"
             "  -r<standard>  Starts gosh with the default environment defined\n"
             "           in RnRS, where n is determined by <standard>.  The following\n"
@@ -100,13 +99,22 @@ void usage(void)
             "  -f<flag> Sets various flags\n"
             "      case-fold       uses case-insensitive reader (as in R5RS)\n"
             "      load-verbose    report while loading files\n"
+            "      include-verbose report while including files\n"
+            "      warn-legacy-syntax\n"
+            "                      print warning when legacy Gauche syntax is encountered\n"
             "      no-inline       don't inline procedures & constants (combined\n"
             "                      no-inline-globals, no-inline-locals, and\n"
             "                      no-inline-constants.)\n"
-            "      no-inline-globals don't inline global procedures.\n"
-            "      no-inline-locals  don't inline local procedures.\n"
-            "      no-inline-constants don't inline constants.\n"
-            "      no-post-inline-pass  don't run post-inline optimization pass.\n"
+            "      no-inline-globals\n"
+            "                      don't inline global procedures.\n"
+            "      no-inline-locals\n"
+            "                      don't inline local procedures.\n"
+            "      no-inline-constants\n"
+            "                      don't inline constants.\n"
+            "      no-lambda-lifting-pass\n"
+            "                      don't run lambda lifting pass.\n"
+            "      no-post-inline-pass\n"
+            "                      don't run post-inline optimization pass.\n"
             "      no-source-info  don't preserve source information for debugging\n"
             "      test            test mode, to run gosh inside the build tree\n"
             );
@@ -193,7 +201,7 @@ void further_options(const char *optarg)
     }
     else {
         fprintf(stderr, "unknown -f option: %s\n", optarg);
-        fprintf(stderr, "supported options are: -fcase-fold or -fload-verbose, -fno-inline, -fno-inline-globals, -fno-inline-locals, -fno-inline-constants, -fno-source-info, -fno-post-inline-pass, -fno-lambda-lifting-pass, -ftest\n");
+        fprintf(stderr, "supported options are: -fcase-fold, -fload-verbose, -finclude-verbose, -fno-inline, -fno-inline-globals, -fno-inline-locals, -fno-inline-constants, -fno-source-info, -fno-post-inline-pass, -fno-lambda-lifting-pass, -fwarn-legacy-syntax, or -ftest\n");
         exit(1);
     }
 }
@@ -210,7 +218,7 @@ void profiler_options(const char *optarg)
     }
     else {
         fprintf(stderr, "unknown -p option: %s\n", optarg);
-        fprintf(stderr, "supported profiling options are: -ptime\n");
+        fprintf(stderr, "supported profiling options are: -ptime or -pload\n");
     }
 }
 
@@ -272,17 +280,15 @@ static void sig_setup(void)
 #ifdef SIGBUS
     sigdelset(&set, SIGBUS);
 #endif /*SIGBUS*/
-#if defined(GC_LINUX_THREADS)
-    /* some signals are used in the system */
-    sigdelset(&set, SIGPWR);  /* used in gc */
-    sigdelset(&set, SIGXCPU); /* used in gc */
-    sigdelset(&set, SIGUSR1); /* used in linux threads */
-    sigdelset(&set, SIGUSR2); /* used in linux threads */
-#endif /*GC_LINUX_THREADS*/
-#if defined(GC_FREEBSD_THREADS)
-    sigdelset(&set, SIGUSR1); /* used by GC to stop the world */
-    sigdelset(&set, SIGUSR2); /* used by GC to restart the world */
-#endif /*GC_FREEBSD_THREADS*/
+
+    /* Exclude signals used by GC to stop and restart the world. */
+#ifdef GC_THREADS
+    int sig_suspend = GC_get_suspend_signal();
+    if (sig_suspend >= 0) sigdelset(&set, sig_suspend);
+    int sig_restart = GC_get_thr_restart_signal();
+    if (sig_restart >= 0) sigdelset(&set, sig_restart);
+#endif /*GC_THREADS*/
+
     Scm_SetMasterSigmask(&set);
 }
 
@@ -367,10 +373,13 @@ static int init_console(void)
     char buf[100];
     int in_fd, out_fd;
 #define ERR(msg) do {sprintf(buf, msg, strerror(errno));goto fail;} while(0)
-
-    if ((in_fd = open("NUL", O_RDONLY)) < 0) ERR("couldn't open NUL: %s");
-    if ((out_fd = open("NUL", O_WRONLY))< 0) ERR("couldn't open NUL: %s");
-    if (_dup2(in_fd, 0) < 0)  ERR("dup2(0) failed (%s)");
+    /* If we don't have console, we'll start off with stdio to be redirected
+       to NUL; but whenever Scheme program tries to do I/O from/to
+       standard Scheme ports, we open the console and reconnect the
+       ports. */
+    if ((in_fd  = open("NUL", O_RDONLY)) < 0) ERR("couldn't open NUL: %s");
+    if ((out_fd = open("NUL", O_WRONLY)) < 0) ERR("couldn't open NUL: %s");
+    if (_dup2(in_fd,  0) < 0) ERR("dup2(0) failed (%s)");
     if (_dup2(out_fd, 1) < 0) ERR("dup2(1) failed (%s)");
     if (_dup2(out_fd, 2) < 0) ERR("dup2(2) failed (%s)");
     close(in_fd);
@@ -513,21 +522,45 @@ int execute_script(const char *scriptfile, ScmObj args)
     return 0;
 }
 
+static int has_terminal()
+{
+    if (isatty(0)) return TRUE;
+#if defined(GAUCHE_WINDOWS)
+    /* MinGW console (mintty) communicates with running processes via pipe,
+       so isatty(0) returns false even if we're running on mintty.
+       We use pipe name to determine whether we're talking to mintty.
+       %sys-mintty? is defined in libsys.scm. */
+    ScmEvalPacket pac;
+    int r = Scm_EvalCString("(%sys-mintty? 0)",
+			    SCM_OBJ(Scm_GaucheInternalModule()),
+			    &pac);
+    if (r == 1 && !SCM_FALSEP(pac.results[0])) return TRUE;
+#endif /*GAUCHE_WINDOWS*/
+    return FALSE;
+}
+
 /* When no script is given, enter REPL. */
 void enter_repl()
 {
-    /* We're in interactive mode. (use gauche.interactive) */
+    /* Load gauche.interactive, for we're not executing a script. */
     if (load_initfile) {
         ScmLoadPacket lpak;
         if (Scm_Require(SCM_MAKE_STR("gauche/interactive"), 0, &lpak) < 0) {
-            Scm_Warn("couldn't load gauche.interactive\n");
+            Scm_ReportError(lpak.exception);
+            Scm_Warn("Encountered an error during loading gauche.interactive.  Falling back to basic REPL.");
         } else {
             Scm_ImportModule(SCM_CURRENT_MODULE(),
                              SCM_INTERN("gauche.interactive"), SCM_FALSE, 0);
         }
     }
 
-    if (batch_mode || (!isatty(0) && !interactive_mode)) {
+    /* If -fcase-fold flag is given, switch the stdin port to the
+       case folding mode. */
+    if (SCM_VM_RUNTIME_FLAG_IS_SET(Scm_VM(), SCM_CASE_FOLD)) {
+        Scm_SetPortCaseFolding(SCM_PORT(Scm_Stdin()), TRUE);
+    }
+    
+    if (batch_mode || (!has_terminal() && !interactive_mode)) {
         Scm_LoadFromPort(SCM_PORT(Scm_Stdin()), SCM_LOAD_PROPAGATE_ERROR, NULL);
     } else {
         /* Call read-eval-print-loop.  If gauche.interactive is loaded,
@@ -586,7 +619,7 @@ int main(int ac, char **av)
     /* prepare *program-name* and *argv* */
     if (argind < argc) {
         /* We have a script file specified. */
-        struct stat statbuf;
+        ScmStat statbuf;
 
         /* if the script name is given in relative pathname, see if
            it exists from the current directory.  if not, leave it
@@ -635,7 +668,9 @@ int main(int ac, char **av)
 
     /* Following is the main dish. */
     if (scriptfile != NULL) exit_code = execute_script(scriptfile, args);
+#if !defined(GAUCHE_WINDOWS_NOCONSOLE)
     else                    enter_repl();
+#endif /*!defined(GAUCHE_WINDOWS_NOCONSOLE)*/
 
     /* All is done.  */
     Scm_Exit(exit_code);
