@@ -3,10 +3,38 @@
 # You need MinGW and MSYS.  (You no longer need Cygwin.)
 # The compiled binary is installed in ../Gauche-mingw-dist.  You can just
 # zip it and distribute.
-# See mingw-memo.txt for the details.
+# See doc/HOWTO-mingw.txt for the details.
+
+set -e
 
 # Set MINGWDIR if MinGW is installed in different place.
-mingwdir=${MINGWDIR:-/mingw}
+case "$MSYSTEM" in
+  MINGW64)
+    mingwdir=${MINGWDIR:-/mingw64}
+    mingwarch="x86_64"
+    ;;
+  MINGW32)
+    mingwdir=${MINGWDIR:-/mingw32}
+    mingwarch="i686"
+    ;;
+  *)
+    echo 'WARNING: $MSYSTEM is neither "MINGW32" or "MINGW64".'
+    mingwdir=${MINGWDIR:-/mingw}
+    mingwarch="unknown"
+    ;;
+esac
+
+# Setting PATH to make sure the build process find right tools.
+# NB: We don't use CC=/path/to/gcc trick for ./configure, since
+# the gcc path embedded in gauche-config should be Windows path.
+PATH=$mingwdir/bin:$PATH
+
+# We assume we run this script on 64bit platform.  For 32bit distribution,
+# though, we want the resulting gosh to run as if it's built on 32bit platform.
+case "$MSYSTEM" in
+  MINGW32)
+    buildopt=--build=i686-pc-mingw32
+esac
 
 # Process Options:
 while [ "$#" -gt 0 ]; do
@@ -47,10 +75,12 @@ fi
 if [ "$INSTALLER" = yes ]; then
   distdir=`pwd`/winnt/wix/Gauche
 else
-  distdir=`pwd`/../Gauche-mingw-dist/Gauche
+  distdir=`pwd`/../Gauche-mingw-dist/Gauche-${mingwarch}
 fi
 rm -rf $distdir
-./configure --enable-multibyte=utf8 --prefix=$distdir
+./configure --prefix=$distdir --enable-threads=win32 \
+            --enable-multibyte=utf8 --enable-ipv6=no \
+            --with-dbm=ndbm,odbm $buildopt
 make
 
 if [ $? -ne 0 ]; then
@@ -63,25 +93,32 @@ make install
 (cd src; make install-mingw)
 make install-examples
 rm -rf $distdir/lib/libgauche.dll*
-cp $mingwdir/bin/mingwm10.dll $distdir/bin
-for dll in libcharset-1.dll libiconv-2.dll libz-1.dll; do
-  if [ -f $mingwdir/bin/$dll ]; then
-    cp $mingwdir/bin/$dll $distdir/bin
-  fi
-done
+case "$MSYSTEM" in
+  MINGW64|MINGW32)
+    for dll in libwinpthread-1.dll; do
+      if [ -f $mingwdir/bin/$dll ]; then
+        cp $mingwdir/bin/$dll $distdir/bin
+      fi
+    done
+    ;;
+  *)
+    cp $mingwdir/bin/mingwm10.dll $distdir/bin
+    ;;
+esac
 
 # Build GL
+# NB: 
 if [ "$WITH_GL" = "yes" ]; then
   PATH=$distdir/bin:$PATH
   (cd ../Gauche-gl; ./DIST gen; \
-   if test -f Makefile; then make clean; fi; \
    ./configure --prefix=$distdir --with-glut=mingw-static; \
-   make; make install; make install-examples)
+   make clean; make; make install; make install-examples)
 fi
 
 # Build installer
 if [ "$INSTALLER" = "yes" ]; then
   (cd winnt/wix; make)
+  cp winnt/wix/Gauche-mingw-*.msi ..
 fi
 
 # 'zip' isn't included in MinGW.
