@@ -1,7 +1,7 @@
 /*
  * module.c - module implementation
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,8 +33,9 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/priv/builtin-syms.h"
 #include "gauche/class.h"
+#include "gauche/priv/builtin-syms.h"
+#include "gauche/priv/moduleP.h"
 
 /*
  * Modules
@@ -61,9 +62,9 @@
  *  it, recovering its resouces.
  */
 
-/* Mutex of module operation
+/* Note on mutex of module operation
  *
- * [SK] Each module used to have a mutex for accesses to it.  I changed it
+ * Each module used to have a mutex for accesses to it.  I changed it
  * to use a single global lock (modules.mutex), based on the following
  * observations:
  *
@@ -77,21 +78,46 @@
  * Benchmark showed the change made program loading 30% faster.
  */
 
-static void module_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
-{
-    if (SCM_MODULEP(SCM_MODULE(obj)->origin)) {
-        Scm_Printf(port, "#<module %A$%A @%p>",
-                   SCM_MODULE(obj)->name,
-                   SCM_MODULE(SCM_MODULE(obj)->origin)->name,
-                   obj);
-    } else {
-        Scm_Printf(port, "#<module %A>", SCM_MODULE(obj)->name);
-    }
-}
+/* Special treatment of keyword modules.
+ * We need to achieve two goals:
+ *
+ * (1) For ordinary Gauche programs (modules that uses implicit inheritance
+ *   of #<module gauche>), we want to see all keywords being bound to
+ *   itself by default.  It can be achieved by inheriting a module that
+ *   has such keyword bindings.
+ * (2) For R7RS programs we want to see the default keyword bindings only
+ *   when the programmer explicitly asks so - notably, by importing some
+ *   special module.  It can be achieved by a module that has all keywords
+ *   each bound to itself, *and* exports all of such bindings.
+ *
+ * It turned out we can't use one 'keyword' module for both purpose; if
+ * we have a keyword module that exports all bindings, they are automatically   
+ * exported from modules that inherits them.  This means if a R7RS program
+ * imports any of Gauche modules, it carries all of keyword bindings.  It's
+ * not only nasty, but also dangerous for it can shadow bindings to the
+ * symbols starting with a colon inadvertently.
+ *
+ * So we have two modules, #<module gauche.keyword> and
+ * #<module keyword>.  The former have export-all flag, and to be imported
+ * from R7RS programs as needed.  The latter doesn't export anything, and
+ * to be inherited to Gauche modules by default.  Whenever a keyword
+ * is created, its default binding is inserted to both - actually,
+ * to prevent two modules from being out of sync, we specially wire them
+ * to share a single hashtable for their internal bindings.
+ */
 
-SCM_DEFINE_BUILTIN_CLASS(Scm_ModuleClass,
-                         module_print, NULL, NULL, NULL,
-                         SCM_CLASS_COLLECTION_CPL);
+/* Note on gauche.require-base module
+   It is an immutable module to which 'require' loads code.
+   We need a base module where 'define-module' and 'define-library' are
+   visible in order for requiring modules using them to work, so
+   loading into the current module won't cut it.  However, we don't
+   want to use a specific mutable module (such as #<module gauche>) as
+   a base, since if the required module has toplevel defines without
+   switching the module, it will modify the base module.
+   By using immutable module as a base, we can reject the latter case;
+   requiring a code that inserts toplevel binding without specifying
+   a module is simply a bad idea and shouldn't be allowed.
+ */
 
 /* Global module table */
 static struct {
@@ -110,9 +136,9 @@ DEFINE_STATIC_MODULE(gaucheModule);   /* #<module gauche> */
 DEFINE_STATIC_MODULE(internalModule); /* #<module gauche.internal> */
 DEFINE_STATIC_MODULE(gfModule);       /* #<module gauche.gf> */
 DEFINE_STATIC_MODULE(userModule);     /* #<module user> */
-#if GAUCHE_UNIFY_SYMBOL_KEYWORD
 DEFINE_STATIC_MODULE(keywordModule);  /* #<module keyword> */
-#endif /*!GAUCHE_UNIFY_SYMBOL_KEYWORD*/
+DEFINE_STATIC_MODULE(gkeywordModule); /* #<module gauche.keyword> */
+DEFINE_STATIC_MODULE(reqbaseModule);  /* #<module gauche.require-base> */
 
 static ScmObj defaultParents = SCM_NIL; /* will be initialized */
 static ScmObj defaultMpl =     SCM_NIL; /* will be initialized */
@@ -121,24 +147,29 @@ static ScmObj defaultMpl =     SCM_NIL; /* will be initialized */
  * Constructor
  */
 
-static void init_module(ScmModule *m, ScmObj name)
+static void init_module(ScmModule *m, ScmObj name, ScmHashTable *internal)
 {
     m->name = name;
     m->imported = m->depended = SCM_NIL;
     m->exportAll = FALSE;
     m->parents = defaultParents;
     m->mpl = Scm_Cons(SCM_OBJ(m), defaultMpl);
-    m->internal = SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQ, 0));
+    if (internal) {
+        m->internal = internal;
+    } else {
+        m->internal = SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQ, 0));
+    }
     m->external = SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQ, 0));
     m->origin = m->prefix = SCM_FALSE;
+    m->sealed = FALSE;
 }
 
 /* Internal */
-static ScmObj make_module(ScmObj name)
+static ScmObj make_module(ScmObj name, ScmHashTable *internal)
 {
     ScmModule *m = SCM_NEW(ScmModule);
     SCM_SET_CLASS(m, SCM_CLASS_MODULE);
-    init_module(m, name);
+    init_module(m, name, internal);
     return SCM_OBJ(m);
 }
 
@@ -160,7 +191,7 @@ static ScmModule *lookup_module_create(ScmSymbol *name, int *created)
                                          (intptr_t)name,
                                          SCM_DICT_CREATE);
     if (e->value == 0) {
-        (void)SCM_DICT_SET_VALUE(e, make_module(SCM_OBJ(name)));
+        (void)SCM_DICT_SET_VALUE(e, make_module(SCM_OBJ(name), NULL));
         *created = TRUE;
     } else {
         *created = FALSE;
@@ -172,7 +203,7 @@ static ScmModule *lookup_module_create(ScmSymbol *name, int *created)
 ScmObj Scm_MakeModule(ScmSymbol *name, int error_if_exists)
 {
     if (name == NULL) {
-        return make_module(SCM_FALSE);
+        return make_module(SCM_FALSE, NULL);
     }
     int created;
     ScmObj r = SCM_OBJ(lookup_module_create(name, &created));
@@ -189,7 +220,7 @@ ScmObj Scm_MakeModule(ScmSymbol *name, int error_if_exists)
 /* internal API to create an anonymous wrapper module */
 ScmObj Scm__MakeWrapperModule(ScmModule *origin, ScmObj prefix)
 {
-    ScmModule *m = SCM_MODULE(make_module(SCM_FALSE));
+    ScmModule *m = SCM_MODULE(make_module(SCM_FALSE, NULL));
     m->parents = SCM_LIST1(SCM_OBJ(origin));
     m->mpl = Scm_Cons(SCM_OBJ(m), origin->mpl);
     m->prefix = prefix;
@@ -200,6 +231,30 @@ ScmObj Scm__MakeWrapperModule(ScmModule *origin, ScmObj prefix)
     return SCM_OBJ(m);
 }
 
+/* Common code to report error on sealed module.
+   One common error is to try to 'use' or 'require' a file that doesn't
+   have module definitions in it.  The default error message in that case
+   is perplexing, so we use more helpful message in that case.
+ */
+static void err_sealed(ScmObj source, ScmModule *target)
+{
+    const char *what = "";
+    if (SCM_MODULEP(source)) what = "import a module";
+    else                     what = "create a binding";
+    
+    if (target == Scm__RequireBaseModule()) {
+        Scm_Error("Attempted to %s (%S) into gauche.require-base. "
+                  "This may be caused by trying to 'use' or 'require' a file"
+                  " in which no module is defined.  Make sure the file has"
+                  " define-module/select-module or define-library at the"
+                  " beginning.",
+                  what, source, SCM_OBJ(target));
+    } else {
+        Scm_Error("Attempted to %s (%S) in a sealed module: %S",
+                  what, source, SCM_OBJ(target));
+    }
+}
+
 /*----------------------------------------------------------------------
  * Finding and modifying bindings
  */
@@ -221,23 +276,23 @@ static inline void init_module_cache(module_cache *c)
     c->more_searched = SCM_NIL;
 }
 
-static inline int module_visited_p(module_cache *c, ScmObj m)
+static inline int module_visited_p(module_cache *c, ScmModule *m)
 {
     for (int i=0; i<c->num_searched; i++) {
-        if (SCM_EQ(m, c->searched[i])) return TRUE;
+        if (SCM_EQ(SCM_OBJ(m), c->searched[i])) return TRUE;
     }
     if (!SCM_NULLP(c->more_searched)) {
-        if (!SCM_FALSEP(Scm_Memq(m, c->more_searched))) return TRUE;
+        if (!SCM_FALSEP(Scm_Memq(SCM_OBJ(m), c->more_searched))) return TRUE;
     }
     return FALSE;
 }
 
-static inline void module_add_visited(module_cache *c, ScmObj m)
+static inline void module_add_visited(module_cache *c, ScmModule *m)
 {
     if (c->num_searched < SEARCHED_ARRAY_SIZE) {
-        c->searched[c->num_searched++] = m;
+        c->searched[c->num_searched++] = SCM_OBJ(m);
     } else {
-        c->more_searched = Scm_Cons(m, c->more_searched);
+        c->more_searched = Scm_Cons(SCM_OBJ(m), c->more_searched);
     }
 }
 
@@ -267,32 +322,35 @@ static ScmGloc *search_binding(ScmModule *module, ScmSymbol *symbol,
                    when we search inherited modules we look into it's
                    internal bindings. */
                 external_only = FALSE;
+                symbol = SCM_GLOC(v)->name; /* in case it is renamed on export */
             } else {
                 return SCM_GLOC(v);
             }
         }
         if (stay_in_module) return NULL;
-        module_add_visited(&searched, SCM_OBJ(module));
+        module_add_visited(&searched, module);
     }
 
     ScmObj p, mp;
-    /* Next, search from imported modules */
+    /* Next, search from imported modules
+       If the import is prefixed, we avoid caching the result. */
     SCM_FOR_EACH(p, module->imported) {
         ScmObj elt = SCM_CAR(p);
         ScmObj sym = SCM_OBJ(symbol);
+        int prefixed = FALSE;
 
         SCM_ASSERT(SCM_MODULEP(elt));
         SCM_FOR_EACH(mp, SCM_MODULE(elt)->mpl) {
             ScmGloc *g;
 
             SCM_ASSERT(SCM_MODULEP(SCM_CAR(mp)));
-
-            if (module_visited_p(&searched, SCM_CAR(mp))) continue;
             ScmModule *m = SCM_MODULE(SCM_CAR(mp));
+            if (!prefixed && module_visited_p(&searched, m)) continue;
             if (SCM_SYMBOLP(m->prefix)) {
                 sym = Scm_SymbolSansPrefix(SCM_SYMBOL(sym),
                                            SCM_SYMBOL(m->prefix));
                 if (!SCM_SYMBOLP(sym)) break;
+                prefixed = TRUE;
             }
 
             ScmObj v = Scm_HashTableRef(m->external, SCM_OBJ(sym), SCM_FALSE);
@@ -300,13 +358,13 @@ static ScmGloc *search_binding(ScmModule *module, ScmSymbol *symbol,
                 g = SCM_GLOC(v);
                 if (g->hidden) break;
                 if (SCM_GLOC_PHANTOM_BINDING_P(g)) {
-                    g = search_binding(m, symbol, FALSE, FALSE, TRUE);
+                    g = search_binding(m, g->name, FALSE, FALSE, TRUE);
                     if (g) return g;
                 } else {
                     return g;
                 }
             }
-            module_add_visited(&searched, SCM_OBJ(m));
+            if (!prefixed) module_add_visited(&searched, m);
         }
     }
 
@@ -367,6 +425,8 @@ ScmObj Scm_GlobalVariableRef(ScmModule *module,
 ScmGloc *Scm_MakeBinding(ScmModule *module, ScmSymbol *symbol,
                          ScmObj value, int flags)
 {
+    if (module->sealed) err_sealed(SCM_OBJ(symbol), module);
+    
     ScmGloc *g;
     ScmObj oldval = SCM_UNDEFINED;
     int prev_kind = 0;
@@ -388,7 +448,7 @@ ScmGloc *Scm_MakeBinding(ScmModule *module, ScmSymbol *symbol,
         g = SCM_GLOC(Scm_MakeGloc(symbol, module));
         Scm_HashTableSet(module->internal, SCM_OBJ(symbol), SCM_OBJ(g), 0);
         /* If module is marked 'export-all', export this binding by default */
-        if (module->exportAll) {
+        if (module->exportAll && SCM_SYMBOL_INTERNED(symbol)) {
             Scm_HashTableSet(module->external, SCM_OBJ(symbol), SCM_OBJ(g), 0);
         }
     }
@@ -398,8 +458,8 @@ ScmGloc *Scm_MakeBinding(ScmModule *module, ScmSymbol *symbol,
     Scm_GlocMark(g, kind);
 
     if (prev_kind != 0) {
-        /* TODO: value and oldval may have circular structure, and we should
-           avoid diverging. */
+        /* NB: Scm_EqualP may throw an error.  It won't leave the state
+           inconsistent, but be aware. */
         if (prev_kind != kind || !Scm_EqualP(value, oldval)) {
             Scm_Warn("redefining %s %S::%S",
                      (prev_kind == SCM_BINDING_CONST)? "constant" : "inlinable",
@@ -434,6 +494,8 @@ ScmObj Scm_DefineConst(ScmModule *module, ScmSymbol *symbol, ScmObj value)
  */
 void Scm_HideBinding(ScmModule *module, ScmSymbol *symbol)
 {
+    if (module->sealed) err_sealed(SCM_OBJ(symbol), module);
+
     int err_exists = FALSE;
 
     (void)SCM_INTERNAL_MUTEX_LOCK(modules.mutex);
@@ -466,7 +528,7 @@ void Scm_HideBinding(ScmModule *module, ScmSymbol *symbol)
  *   CAVEATS:
  *
  *   - gloc's module remains the same.
- *   - autoload won't resolved.
+ *   - autoload won't be resolved.
  *   - TARGETNAME shouldn't be bound in TARGET beforehand.  We don't check
  *     it and just insert the gloc.  If there is an existing binding,
  *     it would become orphaned, possibly causing problems.
@@ -480,6 +542,8 @@ void Scm_HideBinding(ScmModule *module, ScmSymbol *symbol)
 int Scm_AliasBinding(ScmModule *target, ScmSymbol *targetName,
                      ScmModule *origin, ScmSymbol *originName)
 {
+    if (target->sealed) err_sealed(SCM_OBJ(targetName), target);
+
     ScmGloc *g = Scm_FindBinding(origin, originName, SCM_BINDING_EXTERNAL);
     if (g == NULL) return FALSE;
     SCM_INTERNAL_MUTEX_SAFE_LOCK_BEGIN(modules.mutex);
@@ -497,13 +561,15 @@ ScmObj Scm_ImportModule(ScmModule *module,
                         ScmObj prefix,
                         u_long flags) /* reserved for future use */
 {
+    if (module->sealed) err_sealed(SCM_OBJ(imported), module);
+
     ScmModule *imp = NULL;
     if (SCM_MODULEP(imported)) {
         imp = SCM_MODULE(imported);
     } else if (SCM_SYMBOLP(imported)) {
         imp = Scm_FindModule(SCM_SYMBOL(imported), 0);
     } else if (SCM_IDENTIFIERP(imported)) {
-        imp = Scm_FindModule(SCM_IDENTIFIER(imported)->name, 0);
+        imp = Scm_FindModule(Scm_UnwrapIdentifier(SCM_IDENTIFIER(imported)), 0);
     } else {
         Scm_Error("module name or module required, but got %S", imported);
     }
@@ -520,12 +586,15 @@ ScmObj Scm_ImportModule(ScmModule *module,
     {
         ScmObj ms, prev = p;
         SCM_SET_CDR(p, module->imported);
-        /* Remove duplicate module, if any. */
+        /* Remove duplicate module, if any.
+           NB: We allow to import the same module multiple times if they are
+           qualified by :only, :prefix, etc.  Theoretically we should check
+           exactly same qualifications, but we hope that kind of duplication
+           is rare.
+        */
         SCM_FOR_EACH(ms, SCM_CDR(p)) {
             ScmModule *m = SCM_MODULE(SCM_CAR(ms));
-            ScmObj b0 = SCM_MODULEP(m->origin)? m->origin : SCM_OBJ(m);
-            ScmObj b1 = SCM_MODULEP(imp->origin)? imp->origin : SCM_OBJ(imp);
-            if (!SCM_EQ(b0, b1)) {
+            if (!SCM_EQ(SCM_OBJ(m), SCM_OBJ(imp))) {
                 prev = ms;
                 continue;
             }
@@ -690,8 +759,11 @@ ScmObj Scm_ModuleExports(ScmModule *module)
 
 ScmObj Scm_ExtendModule(ScmModule *module, ScmObj supers)
 {
+    if (module->sealed) {
+        Scm_Error("Attempt to extend a sealed module: %S", SCM_OBJ(module));
+    }
+    
     ScmObj seqh = SCM_NIL, seqt = SCM_NIL;
-
     ScmObj sp;
     SCM_FOR_EACH(sp, supers) {
         if (!SCM_MODULEP(SCM_CAR(sp))) {
@@ -710,6 +782,18 @@ ScmObj Scm_ExtendModule(ScmModule *module, ScmObj supers)
     return module->mpl;
 }
 
+/*----------------------------------------------------------------------
+ * Module sealing
+ */
+
+/* NB: In general it is a bad idea to "unseal" module, so we only
+   provide an API to make module sealed.  However, unsealing might
+   be useful for debugging.  */
+void Scm_ModuleSeal(ScmModule *module)
+{
+    module->sealed = TRUE;
+}
+
 /*----------------------------------------------------------------------
  * Finding modules
  */
@@ -775,74 +859,6 @@ ScmObj Scm_PathToModuleName(ScmString *path)
     return Scm_ApplyRec1(path_to_module_name_proc, SCM_OBJ(path));
 }
 
-/*----------------------------------------------------------------------
- * Module introspection
- */
-
-static ScmObj module_name(ScmObj m)
-{
-    return SCM_MODULE(m)->name;
-}
-
-static ScmObj module_imported(ScmObj m)
-{
-    return SCM_MODULE(m)->imported;
-}
-
-static ScmObj module_exported(ScmObj m)
-{
-    return Scm_ModuleExports(SCM_MODULE(m));
-}
-
-static ScmObj module_exportAll(ScmObj m)
-{
-    return SCM_MAKE_BOOL(SCM_MODULE(m)->exportAll);
-}
-
-static ScmObj module_parents(ScmObj m)
-{
-    return SCM_MODULE(m)->parents;
-}
-
-static ScmObj module_mpl(ScmObj m)
-{
-    return SCM_MODULE(m)->mpl;
-}
-
-static ScmObj module_depended(ScmObj m)
-{
-    return SCM_MODULE(m)->depended;
-}
-
-static ScmObj module_table(ScmObj m)
-{
-    return SCM_OBJ(SCM_MODULE(m)->internal);
-}
-
-static ScmObj module_origin(ScmObj m)
-{
-    return SCM_MODULE(m)->origin;
-}
-
-static ScmObj module_prefix(ScmObj m)
-{
-    return SCM_MODULE(m)->prefix;
-}
-
-static ScmClassStaticSlotSpec module_slots[] = {
-    SCM_CLASS_SLOT_SPEC("name", module_name, NULL),
-    SCM_CLASS_SLOT_SPEC("mpl", module_mpl, NULL),
-    SCM_CLASS_SLOT_SPEC("parents", module_parents, NULL),
-    SCM_CLASS_SLOT_SPEC("imports", module_imported, NULL),
-    SCM_CLASS_SLOT_SPEC("exports", module_exported, NULL),
-    SCM_CLASS_SLOT_SPEC("export-all", module_exportAll, NULL),
-    SCM_CLASS_SLOT_SPEC("table", module_table, NULL),
-    SCM_CLASS_SLOT_SPEC("depends", module_depended, NULL),
-    SCM_CLASS_SLOT_SPEC("origin", module_origin, NULL),
-    SCM_CLASS_SLOT_SPEC("prefix", module_prefix, NULL),
-    SCM_CLASS_SLOT_SPEC_END()
-};
-
 /*----------------------------------------------------------------------
  * Predefined modules and initialization
  */
@@ -872,12 +888,20 @@ ScmModule *Scm_UserModule(void)
     return &userModule;
 }
 
-#if GAUCHE_UNIFY_SYMBOL_KEYWORD
-ScmModule *Scm_KeywordModule(void)
+ScmModule *Scm__KeywordModule(void) /* internal */
 {
     return &keywordModule;
 }
-#endif /*!GAUCHE_UNIFY_SYMBOL_KEYWORD*/
+
+ScmModule *Scm__GaucheKeywordModule(void) /* internal */
+{
+    return &gkeywordModule;
+}
+
+ScmModule *Scm__RequireBaseModule(void) /* internal */
+{
+    return &reqbaseModule;
+}
 
 ScmModule *Scm_CurrentModule(void)
 {
@@ -885,10 +909,10 @@ ScmModule *Scm_CurrentModule(void)
 }
 
 /* NB: we don't need to lock the global module table in initialization */
-#define INIT_MOD(mod, mname, mpl)                                           \
+#define INIT_MOD(mod, mname, mpl, inttab)                                   \
     do {                                                                    \
       SCM_SET_CLASS(&mod, SCM_CLASS_MODULE);                                \
-      init_module(&mod, mname);                                             \
+      init_module(&mod, mname,  inttab);                                    \
       Scm_HashTableSet(modules.table, (mod).name, SCM_OBJ(&mod), 0);        \
       mod.parents = (SCM_NULLP(mpl)? SCM_NIL : SCM_LIST1(SCM_CAR(mpl)));    \
       mpl = mod.mpl = Scm_Cons(SCM_OBJ(&mod), mpl);                         \
@@ -897,13 +921,16 @@ ScmModule *Scm_CurrentModule(void)
 void Scm__InitModule(void)
 {
     /* List of builtin modules.  We create these so that 'use' or r7rs 'import'
-       won't try to search the file. */
+       won't try to search the file.
+       The modules listed here are marked "provided" at the startup, so it can
+       no longer be loaded by 'use' or 'require'.  Don't list modules that
+       needs to be loaded.
+    */
     static const char *builtin_modules[] = {
         "srfi-2", "srfi-6", "srfi-8", "srfi-10", "srfi-16", "srfi-17",
-        "srfi-22", "srfi-23", "srfi-26", "srfi-28", "srfi-31", "srfi-34",
-        "srfi-35", "srfi-36", "srfi-38", "srfi-45", "srfi-55", "srfi-61",
-        "srfi-62", "srfi-87", "srfi-95",
-        "gauche.vm.debugger",   /* for the backward compatibility */
+        "srfi-22", "srfi-23", "srfi-28", "srfi-34",
+        "srfi-35", "srfi-36", "srfi-38", "srfi-45", "srfi-61",
+        "srfi-62", "srfi-87", "srfi-95", "srfi-111",
         NULL };
     const char **modname;
 
@@ -912,14 +939,12 @@ void Scm__InitModule(void)
 
     /* standard module chain */
     ScmObj mpl = SCM_NIL;
-    INIT_MOD(nullModule, SCM_SYM_NULL, mpl);
-    INIT_MOD(schemeModule, SCM_SYM_SCHEME, mpl);
-#if GAUCHE_UNIFY_SYMBOL_KEYWORD
-    INIT_MOD(keywordModule, SCM_SYM_KEYWORD, mpl);
-#endif /*!GAUCHE_UNIFY_SYMBOL_KEYWORD*/
-    INIT_MOD(gaucheModule, SCM_SYM_GAUCHE, mpl);
-    INIT_MOD(gfModule, SCM_SYM_GAUCHE_GF, mpl);
-    INIT_MOD(userModule, SCM_SYM_USER, mpl);
+    INIT_MOD(nullModule, SCM_SYM_NULL, mpl, NULL);
+    INIT_MOD(schemeModule, SCM_SYM_SCHEME, mpl, NULL);
+    INIT_MOD(keywordModule, SCM_SYM_KEYWORD, mpl, NULL);
+    INIT_MOD(gaucheModule, SCM_SYM_GAUCHE, mpl, NULL);
+    INIT_MOD(gfModule, SCM_SYM_GAUCHE_GF, mpl, NULL);
+    INIT_MOD(userModule, SCM_SYM_USER, mpl, NULL);
 
     mpl = SCM_CDR(mpl);  /* default mpl doesn't include user module */
     defaultParents = SCM_LIST1(SCM_CAR(mpl));
@@ -927,19 +952,17 @@ void Scm__InitModule(void)
 
     /* other modules */
     mpl = defaultMpl;
-    INIT_MOD(internalModule, SCM_SYM_GAUCHE_INTERNAL, mpl);
+    INIT_MOD(internalModule, SCM_SYM_GAUCHE_INTERNAL, mpl, NULL);
+    INIT_MOD(reqbaseModule, SCM_INTERN("gauche.require-base"), mpl, NULL);
+    Scm_ModuleSeal(&reqbaseModule);
+
+    mpl = keywordModule.mpl;
+    INIT_MOD(gkeywordModule, SCM_INTERN("gauche.keyword"), mpl,
+             keywordModule.internal);
+    gkeywordModule.exportAll = TRUE;
 
     /* create predefined moudles */
     for (modname = builtin_modules; *modname; modname++) {
         (void)SCM_FIND_MODULE(*modname, SCM_FIND_MODULE_CREATE);
     }
 }
-
-void Scm__InitModulePost(void)
-{
-    Scm_InitStaticClassWithMeta(&Scm_ModuleClass, "<module>", &gaucheModule,
-                                NULL, /* auto-generate meta */
-                                SCM_FALSE, /* calculate supers from cpl */
-                                module_slots,
-                                0);
-}
