@@ -13,14 +13,17 @@
 (use gauche.sequence)
 (use gauche.hook)
 (use gauche.parameter)
-(use gauche.parseopt) ; until 0.9.4 release
 (use gauche.uvector)
 (use gauche.regexp)
+(use gauche.record)
+(use gauche.generator)
 (use srfi-1)
 (use srfi-13)
 (use file.util)
 (use text.tr)
 (use util.match)
+(use rfc.822)
+(use rfc.mime)
 
 
 
