@@ -1,7 +1,7 @@
 /*
  * read.c - reader
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -56,7 +56,12 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx);
 static ScmObj read_word(ScmPort *port, ScmChar initial, ScmReadContext *ctx,
                         int temp_case_fold, int include_hash_sign);
 static ScmObj read_symbol(ScmPort *port, ScmChar initial, ScmReadContext *ctx);
-static ScmObj read_number(ScmPort *port, ScmChar initial, ScmReadContext *ctx);
+static ScmObj read_immediate_symbol(ScmPort *port, ScmChar initial,
+                                    int interned, const char *preceding,
+                                    ScmReadContext *ctx);
+static ScmObj read_number(ScmPort *port, ScmChar initial,
+                          int radix, /* #<radix>r case */
+                          ScmReadContext *ctx);
 static ScmObj read_symbol_or_number(ScmPort *port, ScmChar initial, ScmReadContext *ctx);
 static ScmObj read_escaped_symbol(ScmPort *port, ScmChar delim, int interned,
                                   ScmReadContext *ctx);
@@ -67,12 +72,9 @@ static ScmObj read_sharp_comma(ScmPort *port, ScmReadContext *ctx);
 static ScmObj process_sharp_comma(ScmPort *port, ScmObj key, ScmObj args,
                                   ScmReadContext *ctx, int has_ref);
 static ScmObj read_shebang(ScmPort *port, ScmReadContext *ctx);
-static ScmObj read_reference(ScmPort *port, ScmChar ch, ScmReadContext *ctx);
+static ScmObj read_num_prefixed(ScmPort *port, ScmChar ch, ScmReadContext *ctx);
 static ScmObj read_sharp_word(ScmPort *port, char c, ScmReadContext *ctx);
 
-/* Special hook for SRFI-4 syntax */
-static ScmObj (*read_uvector_hook)(ScmPort *, const char*, ScmReadContext *) = NULL;
-
 /* Table of 'read-time constructor' in SRFI-10 */
 static struct {
     ScmHashTable *table;
@@ -88,9 +90,6 @@ static struct {
 /* Parameter location for default reader mode */
 ScmParameterLoc defaultReadContext;
 
-/* Parameter location for the current reader lexical mode */
-ScmParameterLoc readerLexicalMode;
-
 /*----------------------------------------------------------------
  * Entry points
  *   Note: Entire read operation are done while locking the input port.
@@ -227,29 +226,6 @@ int Scm_ReadContextLiteralImmutable(ScmReadContext *ctx)
     return (ctx->flags & RCTX_LITERAL_IMMUTABLE);
 }
 
-/*
- * Reader lexical mode
- */
-ScmObj Scm_SetReaderLexicalMode(ScmObj mode)
-{
-    if (!(SCM_EQ(mode, SCM_SYM_LEGACY)
-          || SCM_EQ(mode, SCM_SYM_WARN_LEGACY)
-          || SCM_EQ(mode, SCM_SYM_PERMISSIVE)
-          || SCM_EQ(mode, SCM_SYM_STRICT_R7))) {
-        Scm_Error("reader-lexical-mode must be one of the following symbols:"
-                  " legacy, warn-legacy, permissive, strict-r7, but got %S",
-                  mode);
-    }
-    ScmObj prev_mode = Scm_ParameterRef(Scm_VM(), &readerLexicalMode);
-    Scm_ParameterSet(Scm_VM(), &readerLexicalMode, mode);
-    return prev_mode;
-}
-
-ScmObj Scm_ReaderLexicalMode()
-{
-    return Scm_ParameterRef(Scm_VM(), &readerLexicalMode);
-}
-
 /*----------------------------------------------------------------
  * Error
  */
@@ -504,7 +480,7 @@ static int skipws(ScmPort *port, ScmReadContext *ctx)
 
 static void reject_in_r7(ScmPort *port, ScmReadContext *ctx, const char *token)
 {
-    if (SCM_EQ(Scm_ReaderLexicalMode(), SCM_SYM_STRICT_R7)) {
+    if (SCM_EQ(Scm_GetPortReaderLexicalMode(port), SCM_SYM_STRICT_R7)) {
         Scm_ReadError(port,
                       "lexical syntax %s isn't allowed in strict R7RS mode",
                       token);
@@ -537,7 +513,7 @@ static ScmObj read_internal(ScmPort *port, ScmReadContext *ctx)
             case 'b':; case 'B':; case 'd':; case 'D':;
             case 'e':; case 'E':; case 'i':; case 'I':;
                 Scm_UngetcUnsafe(c1, port);
-                return read_number(port, c, ctx);
+                return read_number(port, c, 0, ctx); /* let StringToNumber handle radix prefix */
             case '!':
                 /* #! is either a script shebang or a reader directive */
                 return read_shebang(port, ctx);
@@ -586,6 +562,11 @@ static ScmObj read_internal(ScmPort *port, ScmReadContext *ctx)
                     ScmObj form = read_item(port, ctx);
                     return SCM_LIST2(SCM_SYM_DEBUG_PRINT, form);
                 }
+                case ',': {
+                    /* #?,form - debug funcall */
+                    ScmObj form = read_item(port, ctx);
+                    return SCM_LIST2(SCM_SYM_DEBUG_FUNCALL, form);
+                }
                 case EOF:
                     return SCM_EOF;
                 default:
@@ -594,8 +575,8 @@ static ScmObj read_internal(ScmPort *port, ScmReadContext *ctx)
             }
             case '0': case '1': case '2': case '3': case '4':
             case '5': case '6': case '7': case '8': case '9':
-                /* #N# or #N= form */
-                return read_reference(port, c1, ctx);
+                /* #N#, #N= or #Nr form */
+                return read_num_prefixed(port, c1, ctx);
             case '*': {
                 reject_in_r7(port, ctx, "#*");
                 /* #*"...." byte string
@@ -607,13 +588,8 @@ static ScmObj read_internal(ScmPort *port, ScmReadContext *ctx)
             case ':': {
                 reject_in_r7(port, ctx, "#:");
                 /* #:name - uninterned symbol */
-                int c2 = Scm_GetcUnsafe(port);
-                if (c2 == '|') {
-                    return read_escaped_symbol(port, c2, FALSE, ctx);
-                } else {
-                    ScmObj name = read_word(port, c2, ctx, FALSE, FALSE);
-                    return Scm_MakeSymbol(SCM_STRING(name), FALSE);
-                }
+                return read_immediate_symbol(port, Scm_GetcUnsafe(port),
+                                             FALSE, "#:", ctx);
             }
             case ';': {
                 /* #;expr - comment out sexpr */
@@ -631,7 +607,7 @@ static ScmObj read_internal(ScmPort *port, ScmReadContext *ctx)
     case '`': return read_quoted(port, SCM_SYM_QUASIQUOTE, ctx);
     case ':':
         /* Would be removed once we make keywords a subtype of symbols. */
-        if (SCM_EQ(Scm_ReaderLexicalMode(), SCM_SYM_STRICT_R7)) {
+        if (SCM_EQ(Scm_GetPortReaderLexicalMode(port), SCM_SYM_STRICT_R7)) {
             return read_symbol(port, c, ctx);
         } else {
             return read_keyword(port, ctx);
@@ -706,6 +682,7 @@ static ScmObj read_item(ScmPort *port, ScmReadContext *ctx)
 /*--------------------------------------------------------
  * Common routine to handle hex-digit escape \xNN etc.
  */
+/* TODO: Redesign these for clearer and simpler API. */
 
 /*
    Hex-escape sequence can appear in the following places:
@@ -841,6 +818,73 @@ ScmObj Scm_ReadXdigitsFromPort(ScmPort *port, int key, ScmObj mode,
     }
 }
 
+/* Another utility procedure.   Read sequence of digits from PORT and made it
+   long and return.  CH can be prefetched digit char; if you don't
+   have one, pass SCM_CHAR_INVALID.  The first nondigit char read
+   is stored in NEXT, and number of digits read is stored in NUMREAD.
+   If the number overflows LONG_MAX or there's no digits to be read,
+   returns -1 (you can check NUMREAD to see which is the case).
+   NB: This may throw an error if ch is neither a valid digit char
+   nor SCM_CHAR_INVALID. */
+/* NB: Call for refactoring, using Scm_ParseDigitsAsLong  */
+long Scm_ReadDigitsAsLong(ScmPort *port, ScmChar ch, int radix,
+                          ScmChar *next, /*out*/
+                          int *numread /*out*/)
+{
+    u_long val = 0;
+    int nchars = 0;
+    if (ch != SCM_CHAR_INVALID) {
+        int v = Scm_DigitToInt(ch, radix, FALSE);
+        if (v < 0) {
+            /* This is for safety net.  The caller should filter out
+               this case. */
+            Scm_ReadError(port, "Digit char expected, but got %C", ch);
+        }
+        val = v;
+    }
+    for (;;) {
+        ch = Scm_Getc(port);
+        nchars++;
+        int v = Scm_DigitToInt(ch, radix, FALSE);
+        if (v < 0) {            /* EOF case is covered here */
+            *next = ch;
+            *numread = nchars;
+            return val;
+        }
+        if (val >= (LONG_MAX/radix+1)) {
+            /* we'll overflow */
+            *next = ch;
+            *numread = nchars;
+            return -1;
+        }
+        val = val*radix+v;
+    }
+    /* NOTREACHED */
+}
+
+/* Read long digits from BUF up to LEN.  BUF must only contain single-byte
+   chars. */
+long Scm_ParseDigitsAsLong(const char *buf, size_t len, int radix,
+                           int *numread) /*out*/
+{
+    u_long val = 0;
+    int nchars = 0;
+    for (; nchars < len; nchars++, buf++) {
+        int v = Scm_DigitToInt((ScmChar)*buf, radix, FALSE);
+        if (v < 0) {
+            *numread = nchars;
+            return val;
+        }
+        if (val >= (LONG_MAX/radix+1)) {
+            /* we'll overflow */
+            *numread = nchars;
+            return -1;
+        }
+        val = val*radix+v;
+    }
+    *numread = nchars;
+    return (nchars == 0)? -1 : val;
+}
 
 /*----------------------------------------------------------------
  * List
@@ -862,10 +906,12 @@ static ScmObj read_list_int(ScmPort *port, ScmChar closer,
             return start;
         }
 
-        if (dot_seen) goto baddot;
-
+        /* NB: We need to keep reading after we see the dotted pair, since
+           there can be a case like (a . b #;c), which is valid syntax.
+           We only reject when we see a next item after reading dotted pair. */
         ScmObj item;
         if (c == '.') {
+            if (dot_seen) goto baddot;
             int c2 = Scm_GetcUnsafe(port);
             if (c2 == closer) {
                 goto baddot;
@@ -886,7 +932,8 @@ static ScmObj read_list_int(ScmPort *port, ScmChar closer,
         } else {
             Scm_UngetcUnsafe(c, port);
             item = read_internal(port, ctx);
-            if (SCM_UNDEFINEDP(item)) continue;
+            if (SCM_UNDEFINEDP(item)) continue; /* it was just a comment */
+            if (dot_seen) goto baddot;
             if (SCM_READ_REFERENCE_P(item)) ref_seen = TRUE;
         }
         SCM_APPEND1(start, last, item);
@@ -967,7 +1014,8 @@ static ScmObj read_quoted(ScmPort *port, ScmObj quoter, ScmReadContext *ctx)
 static void read_string_xdigits(ScmPort *port, int key,
                                 int incompletep, ScmDString *buf)
 {
-    ScmObj bad = Scm_ReadXdigitsFromPort(port, key, Scm_ReaderLexicalMode(),
+    ScmObj bad = Scm_ReadXdigitsFromPort(port, key,
+                                         Scm_GetPortReaderLexicalMode(port),
                                          incompletep, buf);
     if (SCM_STRINGP(bad)) {
         /* skip chars to the end of string, so that the reader will read
@@ -1008,11 +1056,7 @@ static ScmObj read_string(ScmPort *port, int incompletep,
         FETCH(c);
         switch (c) {
         case EOF: goto eof_exit;
-        case '"': {
-            int flags = ((incompletep? SCM_STRING_INCOMPLETE : 0)
-                         | SCM_STRING_IMMUTABLE);
-            return Scm_DStringGet(&ds, flags);
-        }
+        case '"': goto finish;
         backslash:
         case '\\': {
             FETCH(c);
@@ -1031,7 +1075,8 @@ static ScmObj read_string(ScmPort *port, int incompletep,
                 break;
             }
             case 'u': case 'U': {
-                if (SCM_EQ(Scm_ReaderLexicalMode(), SCM_SYM_STRICT_R7)) {
+                if (SCM_EQ(Scm_GetPortReaderLexicalMode(port),
+                           SCM_SYM_STRICT_R7)) {
                     Scm_ReadError(port, "\\%c in string literal isn't allowed "
                                   "in strinct-r7rs mode", c);
                 }
@@ -1044,7 +1089,7 @@ static ScmObj read_string(ScmPort *port, int incompletep,
             case '\t': {
                 for (;;) {
                     FETCH(c);
-                    if (c == EOF) goto eof_exit;
+                    if (c == EOF)  goto eof_exit;
                     if (c == '\r') goto cont_cr;
                     if (c == '\n') goto cont_lf;
                     if (!INTRALINE_WS(c)) goto cont_err;
@@ -1055,6 +1100,7 @@ static ScmObj read_string(ScmPort *port, int incompletep,
             case '\r': {
                 FETCH(c);
                 if (c == EOF)  goto eof_exit;
+                if (c == '"')  goto finish;
                 if (c == '\\') goto backslash;
                 if (c != '\n' && !INTRALINE_WS(c)) {
                     ACCUMULATE(c);
@@ -1066,7 +1112,8 @@ static ScmObj read_string(ScmPort *port, int incompletep,
             case '\n': {
                 for (;;) {
                     FETCH(c);
-                    if (c == EOF) goto eof_exit;
+                    if (c == EOF)  goto eof_exit;
+                    if (c == '"')  goto finish;
                     if (c == '\\') goto backslash;
                     if (!INTRALINE_WS(c)) {
                         ACCUMULATE(c);
@@ -1090,8 +1137,9 @@ static ScmObj read_string(ScmPort *port, int incompletep,
  cont_err:
     Scm_ReadError(port, "Invalid line continuation sequence in a string literal: %S...",
                   Scm_DStringGet(&ds, 0));
-    /* NOTREACHED */
-    return SCM_FALSE;
+ finish:;
+    int flags = ((incompletep? SCM_STRING_INCOMPLETE:0) | SCM_STRING_IMMUTABLE);
+    return Scm_DStringGet(&ds, flags);
 }
 
 /*----------------------------------------------------------------
@@ -1139,10 +1187,10 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx)
         /* need to read word to see if it is a character name */
         name = SCM_STRING(read_word(port, c, ctx, TRUE, FALSE));
 
-        if (SCM_EQ(Scm_ReaderLexicalMode(), SCM_SYM_STRICT_R7)) {
+        if (SCM_EQ(Scm_GetPortReaderLexicalMode(port), SCM_SYM_STRICT_R7)) {
             ScmChar following = Scm_GetcUnsafe(port);
             if (!char_is_delimiter(following)) {
-                Scm_Error("Character literal isn't delimited: #\\%s%C ...",
+                Scm_Error("Character literal isn't delimited: #\\%A%C ...",
                           name, following);
             }
             Scm_UngetcUnsafe(following, port);
@@ -1160,7 +1208,7 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx)
             goto unknown;
         }
 
-        ScmObj lexmode = Scm_ReaderLexicalMode();
+        ScmObj lexmode = Scm_GetPortReaderLexicalMode(port);
 
         /* handle #\x1f etc. */
         if (cname[0] == 'x' && isxdigit(cname[1])) {
@@ -1216,7 +1264,7 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx)
 static ScmObj read_word(ScmPort *port, ScmChar initial, ScmReadContext *ctx,
                         int temp_case_fold, int include_hash_sign)
 {
-    int case_fold = temp_case_fold || SCM_PORT_CASE_FOLD(port);
+    int case_fold = temp_case_fold || SCM_PORT_CASE_FOLDING(port);
     ScmDString ds;
     Scm_DStringInit(&ds);
     if (initial != SCM_CHAR_INVALID) {
@@ -1248,6 +1296,7 @@ static void check_valid_symbol(ScmString *s)
     }
 }
 
+/* Read a symbol starting with INITIAL (assuming unescaped), interned. */
 static ScmObj read_symbol(ScmPort *port, ScmChar initial, ScmReadContext *ctx)
 {
     ScmString *s = SCM_STRING(read_word(port, initial, ctx, FALSE, TRUE));
@@ -1255,12 +1304,50 @@ static ScmObj read_symbol(ScmPort *port, ScmChar initial, ScmReadContext *ctx)
     return Scm_Intern(s);
 }
 
-static ScmObj read_number(ScmPort *port, ScmChar initial, ScmReadContext *ctx)
+/* This is called when a symbol (either bare or escaped) is expected
+   starting with INITIAL char; e.g. after #: or #!.  An error is thrown
+   if a symbol isn't immediately followed.  PRECEDING is used in the
+   error message. */
+static ScmObj read_immediate_symbol(ScmPort *port, ScmChar initial,
+                                    int interned, const char *preceding,
+                                    ScmReadContext *ctx)
+{
+    if (initial == '|') {
+        return read_escaped_symbol(port, initial, interned, ctx);
+    } else if (char_word_constituent(initial, FALSE)) {
+        ScmString *s = SCM_STRING(read_word(port, initial, ctx, FALSE, TRUE));
+        check_valid_symbol(s);
+        /* we have to exclude numbers.  a bit ugly - call for cleanup */
+        if (!(isdigit(initial) || initial == '+' || initial == '-')
+            || SCM_FALSEP(Scm_StringToNumber(s, 10, 0))) {
+            return Scm_MakeSymbol(s, interned);
+        }
+    }
+    /* If we come here, we have invalid syntax. */
+    if (initial == EOF) {
+        Scm_ReadError(port, "'%s' followed by nothing", preceding);
+    } else {
+        Scm_ReadError(port, "invalid %s syntax near '%s%C'", preceding,
+                      preceding, initial);
+    }
+    return SCM_UNDEFINED;   /* dummy */
+}
+
+static ScmObj read_number(ScmPort *port, ScmChar initial, int radix,
+                          ScmReadContext *ctx)
 {
     ScmString *s = SCM_STRING(read_word(port, initial, ctx, FALSE, TRUE));
-    ScmObj num = Scm_StringToNumber(s, 10, 0);
-    if (num == SCM_FALSE)
-        Scm_ReadError(port, "bad numeric format: %S", s);
+    u_long flags = radix >=2 ? SCM_NUMBER_FORMAT_ALT_RADIX : 0;
+    int default_radix = radix >= 2? radix : 10;
+    ScmObj num = Scm_StringToNumber(s, default_radix, flags);
+    if (num == SCM_FALSE) {
+        if (radix >= 2) {
+            /* In this case, we've read #<radix>r syntax */
+            Scm_ReadError(port, "bad numeric format: \"#%dr%A\"", radix, s);
+        } else {
+            Scm_ReadError(port, "bad numeric format: %S", s);
+        }
+    }
     return num;
 }
 
@@ -1292,7 +1379,7 @@ static ScmObj read_escaped_symbol(ScmPort *port, ScmChar delim, int interned,
 {
     ScmDString ds;
     Scm_DStringInit(&ds);
-    ScmObj xmode = Scm_ReaderLexicalMode();
+    ScmObj xmode = Scm_GetPortReaderLexicalMode(port);
 
     for (;;) {
         int c = Scm_GetcUnsafe(port);
@@ -1385,65 +1472,72 @@ static ScmObj read_charset(ScmPort *port)
 }
 
 /*----------------------------------------------------------------
- * Back reference (#N# and #N=)
+ * Numeric prefixed (#N#, #N=, #Nr)
  */
 
-static ScmObj read_reference(ScmPort *port, ScmChar ch, ScmReadContext *ctx)
+static ScmObj read_num_prefixed(ScmPort *port, ScmChar ch, ScmReadContext *ctx)
 {
     ScmObj e = SCM_UNBOUND;
-    int refnum = Scm_DigitToInt(ch, 10, FALSE);
+    ScmChar ch2;
+    int nread;
+    long prefix = Scm_ReadDigitsAsLong(port, ch, 10, &ch2, &nread);
 
-    for (;;) {
-        ch = Scm_GetcUnsafe(port);
-        if (ch == EOF) {
-            Scm_ReadError(port, "unterminated reference form (#digits)");
-        }
-        if (SCM_CHAR_ASCII_P(ch) && isdigit(ch)) {
-            refnum = refnum*10+Scm_DigitToInt(ch, 10, FALSE);
-            if (refnum < 0) Scm_ReadError(port, "reference number overflow");
-            continue;
-        }
-        if (ch != '#' && ch != '=') {
-            Scm_ReadError(port, "invalid reference form (must be either #digits# or #digits=) : #%d%A", refnum, SCM_MAKE_CHAR(ch));
-        }
-        break;
-    }
-    if (ch == '#') {
+    if (ch2 == EOF) Scm_ReadError(port, "unterminated reference form (#digits)");
+    if (prefix < 0) Scm_ReadError(port, "#-prefix number overflow");
+
+    switch (ch2) {
+    case '#':
         /* #digit# - back reference */
         if (ctx->table == NULL
             || SCM_UNBOUNDP(e = Scm_HashTableRef(ctx->table,
-                                                 Scm_MakeInteger(refnum),
+                                                 Scm_MakeInteger(prefix),
                                                  SCM_UNBOUND))) {
-            Scm_ReadError(port, "invalid reference number in #%d#", refnum);
+            Scm_ReadError(port, "invalid reference number in #%d#", prefix);
         }
         if (SCM_READ_REFERENCE_P(e) && SCM_READ_REFERENCE_REALIZED(e)) {
             return SCM_READ_REFERENCE(e)->value;
         } else {
             return e;
         }
-    } else {
+    case '=':
         /* #digit= - register */
-        ScmObj ref = Scm_MakeReadReference();
-
-        if (ctx->table == NULL) {
-            ctx->table =
-                SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQV, 0));
+        {
+            ScmObj ref = Scm_MakeReadReference();
+            if (ctx->table == NULL) {
+                ctx->table =
+                    SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQV, 0));
+            }
+            if (!SCM_UNBOUNDP(Scm_HashTableRef(ctx->table,
+                                               Scm_MakeInteger(prefix),
+                                               SCM_UNBOUND))) {
+                Scm_ReadError(port, "duplicate back-reference number in #%d=",
+                              prefix);
+            }
+            ref_register(ctx, ref, prefix);
+            ScmObj val = read_item(port, ctx);
+            if (ref == val) {
+                /* an edge case: #0=#0# */
+                Scm_ReadError(port, "indeterminate read reference: #%d=#%d#",
+                              prefix, prefix);
+            }
+            SCM_READ_REFERENCE(ref)->value = val;
+            return val;
         }
-        if (!SCM_UNBOUNDP(Scm_HashTableRef(ctx->table,
-                                           Scm_MakeInteger(refnum),
-                                           SCM_UNBOUND))) {
-            Scm_ReadError(port, "duplicate back-reference number in #%d=",
-                          refnum);
+    case 'r': case 'R':
+        /* #digitR - radix */
+        if (prefix < 2 || prefix > 36) {
+            Scm_ReadError(port, "Radix prefix out of range: radix in #<radix>R must be between 2 and 36 inclusive, but got: %d", prefix);
+        } else {
+            int ch = Scm_GetcUnsafe(port);
+            if (ch == EOF) {
+                Scm_ReadError(port, "Premature end of radix-prefixed number: #%dr",
+                              prefix);
+            }
+            return read_number(port, ch, prefix, ctx);
         }
-        ref_register(ctx, ref, refnum);
-        ScmObj val = read_item(port, ctx);
-        if (ref == val) {
-            /* an edge case: #0=#0# */
-            Scm_ReadError(port, "indeterminate read reference: #%d=#%d#",
-                          refnum, refnum);
-        }
-        SCM_READ_REFERENCE(ref)->value = val;
-        return val;
+    default:
+        Scm_ReadError(port, "invalid numeric prefix (#, =, r or R is expected) : #%d%A", prefix, SCM_MAKE_CHAR(ch2));
+        return SCM_UNDEFINED;
     }
 }
 
@@ -1459,7 +1553,7 @@ ScmObj Scm_DefineReaderCtor(ScmObj symbol, ScmObj proc, ScmObj finisher,
                             ScmObj module /*reserved*/)
 {
     if (!SCM_PROCEDUREP(proc)) {
-        Scm_Error("procedure required, but got %S\n", proc);
+        Scm_Error("procedure required, but got %S", proc);
     }
     ScmObj pair = Scm_Cons(proc, finisher);
     (void)SCM_INTERNAL_MUTEX_LOCK(readCtorData.mutex);
@@ -1554,8 +1648,7 @@ static ScmObj read_shebang(ScmPort *port, ScmReadContext *ctx)
         }
         /*NOTREACHED*/
     } else {
-        ScmObj id = read_symbol(port, c2, ctx);
-        SCM_ASSERT(SCM_SYMBOLP(id));
+        ScmObj id = read_immediate_symbol(port, c2, TRUE, "#!", ctx);
         (void)SCM_INTERNAL_MUTEX_LOCK(hashBangData.mutex);
         ScmObj e = Scm_HashTableRef(hashBangData.table, id, SCM_FALSE);
         (void)SCM_INTERNAL_MUTEX_UNLOCK(hashBangData.mutex);
@@ -1624,16 +1717,7 @@ static ScmObj read_sharp_word_legacy(ScmPort *port, char ch, ScmReadContext *ctx
         *bufp = '\0';
         Scm_ReadError(port, "invalid uniform vector tag: %s", buf);
     }
-    if (read_uvector_hook == NULL) {
-        /* Require srfi-4 (gauche/uvector)
-           NB: we don't need mutex here, for the loading of srfi-4 is
-           serialized in Scm_Require. */
-        Scm_Require(SCM_MAKE_STR("gauche/uvector"),
-                    SCM_LOAD_PROPAGATE_ERROR, NULL);
-        if (read_uvector_hook == NULL)
-            Scm_ReadError(port, "couldn't load srfi-4 module");
-    }
-    return read_uvector_hook(port, tag, ctx);
+    return Scm_ReadUVector(port, tag, ctx);
 }
 
 /* A 'new' version, friendly to R7RS */
@@ -1677,32 +1761,23 @@ static ScmObj read_sharp_word_1(ScmPort *port, char ch, ScmReadContext *ctx)
     if (tag == NULL) {
         Scm_ReadError(port, "invalid #-token: #%s", w);
     }
-    /* Uvector support is implemented by gauche.uvector.  When it
-       is loaded, it sets up the pointer read_uvector_hook. */
-    if (read_uvector_hook == NULL) {
-        /* Require srfi-4 (gauche/uvector)
-           NB: we don't need mutex here, for the loading of srfi-4 is
-           serialized in Scm_Require. */
-        Scm_Require(SCM_MAKE_STR("gauche/uvector"),
-                    SCM_LOAD_PROPAGATE_ERROR, NULL);
-        if (read_uvector_hook == NULL)
-            Scm_ReadError(port, "couldn't load srfi-4 module");
-    }
-    return read_uvector_hook(port, tag, ctx);
+    return Scm_ReadUVector(port, tag, ctx);
 }
 
 static ScmObj read_sharp_word(ScmPort *port, char ch, ScmReadContext *ctx)
 {
-    if (SCM_EQ(Scm_ReaderLexicalMode(), SCM_SYM_LEGACY)) {
+    if (SCM_EQ(Scm_GetPortReaderLexicalMode(port), SCM_SYM_LEGACY)) {
         return read_sharp_word_legacy(port, ch, ctx);
     } else {
         return read_sharp_word_1(port, ch, ctx);
     }
 }
 
+/* OBSOLETED: gauche.uvector used to call this to set up reader pointer.
+   Now it is read in src/vector.c.   We keep this entry for ABI compatibility.
+   Remove on 1.0 release. */
 void Scm__InstallReadUvectorHook(ScmObj (*f)(ScmPort*, const char*, ScmReadContext *))
 {
-    read_uvector_hook = f;
 }
 
 
@@ -1726,6 +1801,4 @@ void Scm__InitRead(void)
 
     Scm_InitParameterLoc(Scm_VM(), &defaultReadContext,
                          SCM_OBJ(make_read_context(NULL)));
-    Scm_InitParameterLoc(Scm_VM(), &readerLexicalMode,
-                         SCM_OBJ(SCM_SYM_PERMISSIVE));
 }
