@@ -1,7 +1,7 @@
 /*
  * signal.c - signal handling
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -311,8 +311,7 @@ void sigset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
 
 ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmSysSigset *s = SCM_ALLOCATE(ScmSysSigset, klass);
-    SCM_SET_CLASS(s, klass);
+    ScmSysSigset *s = SCM_NEW_INSTANCE(ScmSysSigset, klass);
     sigemptyset(&s->set);
     return SCM_OBJ(s);
 }
@@ -804,10 +803,9 @@ sigset_t Scm_GetMasterSigmask(void)
 /*
  * Convenience routines hiding platform-dependent stuff
  *
- * NB: These used to be used in vm.c, but no longer.  As of 0.9.3 these
- * aren't used anywhere.  We keep these for a while just in case if
- * we find them useful in some day, but if no such case arises, we may
- * drop them eventually.
+ * TRANSIENT: These used to be used in vm.c, but no longer.  As of 0.9.3 these
+ * aren't used anywhere.  Scm_SysSigmask covers those functionalities,
+ * so we'll drop them by 1.0.
  */
 void Scm_GetSigmask(sigset_t *mask)
 {
