@@ -1,7 +1,7 @@
 /*
  * vector.c - vector implementation
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -53,8 +53,30 @@ static void vector_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
     SCM_PUTZ(")", -1, port);
 }
 
-SCM_DEFINE_BUILTIN_CLASS(Scm_VectorClass, vector_print, NULL, NULL, NULL,
-                         SCM_CLASS_SEQUENCE_CPL);
+static int vector_compare(ScmObj x, ScmObj y, int equalp)
+{
+    if (equalp) {
+        /* Vector equality is handled in Scm_Eq* and will never come
+           here, but just in case. */
+        return Scm_EqualP(x, y)? 0 : 1;
+    }
+    /* Follow srfi-114 */
+    ScmWord xlen = SCM_VECTOR_SIZE(x);
+    ScmWord ylen = SCM_VECTOR_SIZE(y);
+    if (xlen < ylen) return -1;
+    if (xlen > ylen) return 1;
+    for (int i=0; i<xlen; i++) {
+        int r = Scm_Compare(SCM_VECTOR_ELEMENT(x, i),
+                            SCM_VECTOR_ELEMENT(y, i));
+        if (r != 0) return r;
+    }
+    return 0;
+}
+
+
+SCM_DEFINE_BUILTIN_CLASS_FLAGS(Scm_VectorClass, vector_print, vector_compare,
+                               NULL, NULL, SCM_CLASS_SEQUENCE_CPL,
+                               SCM_CLASS_AGGREGATE);
 
 static ScmVector *make_vector(ScmSmallInt size)
 {
@@ -145,7 +167,7 @@ ScmObj Scm_VectorCopy(ScmVector *vec,
     ScmVector *v = NULL;
     if (end < 0) end = len;
     if (end < start) {
-        Scm_Error("vector-copy: start (%d) is greater than end (%d)",
+        Scm_Error("vector-copy: start (%ld) is greater than end (%ld)",
                   start, end);
     } else if (end == start) {
         v = make_vector(0);
@@ -185,10 +207,11 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_UVectorClass, NULL, NULL, NULL, NULL,
 static void SCM_CPP_CAT3(print_,tag,vector)(ScmObj obj, ScmPort *out,   \
                                             ScmWriteContext *ctx);      \
 static int SCM_CPP_CAT3(compare_,tag,vector)(ScmObj x, ScmObj y, int equalp); \
-SCM_DEFINE_BUILTIN_CLASS(SCM_CPP_CAT3(Scm_,TAG,VectorClass),            \
-                         SCM_CPP_CAT3(print_,tag,vector),               \
-                         SCM_CPP_CAT3(compare_,tag,vector),             \
-                         NULL, NULL, uvector_cpl);
+SCM_DEFINE_BUILTIN_CLASS_FLAGS(SCM_CPP_CAT3(Scm_,TAG,VectorClass),      \
+                               SCM_CPP_CAT3(print_,tag,vector),         \
+                               SCM_CPP_CAT3(compare_,tag,vector),       \
+                               NULL, NULL, uvector_cpl,                 \
+                               SCM_CLASS_AGGREGATE);
 
 DEF_UVCLASS(S8, s8)
 DEF_UVCLASS(U8, u8)
@@ -283,6 +306,68 @@ ScmObj Scm_MakeUVector(ScmClass *klass, ScmSmallInt size, void *init)
     return Scm_MakeUVectorFull(klass, size, init, FALSE, NULL);
 }
 
+ScmObj Scm_ListToUVector(ScmClass *klass, ScmObj list, int clamp)
+{
+    ScmUVectorType type = Scm_UVectorType(klass);
+    if (type < 0) Scm_Error("uvector class required, but got: %S", klass);
+    ScmSmallInt length = Scm_Length(list);
+    if (length < 0) Scm_Error("improper list not allowed: %S", list);
+
+    ScmUVector *v = (ScmUVector*)Scm_MakeUVector(klass, length, NULL);
+    ScmObj cp = list;
+    for (int i=0; i<length; i++, cp = SCM_CDR(cp)) {
+        switch (type) {
+        case SCM_UVECTOR_S8:
+            SCM_S8VECTOR_ELEMENTS(v)[i] =
+                (signed char)Scm_GetInteger8Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_U8:
+            SCM_U8VECTOR_ELEMENTS(v)[i] =
+                (unsigned char)Scm_GetIntegerU8Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_S16:
+            SCM_S16VECTOR_ELEMENTS(v)[i] =
+                (short)Scm_GetInteger16Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_U16:
+            SCM_U16VECTOR_ELEMENTS(v)[i] =
+                (u_short)Scm_GetIntegerU16Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_S32:
+            SCM_S32VECTOR_ELEMENTS(v)[i] =
+                (ScmInt32)Scm_GetInteger32Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_U32:
+            SCM_U32VECTOR_ELEMENTS(v)[i] =
+                (ScmUInt32)Scm_GetIntegerU32Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_S64:
+            SCM_S64VECTOR_ELEMENTS(v)[i] =
+                (ScmInt64)Scm_GetInteger64Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_U64:
+            SCM_U64VECTOR_ELEMENTS(v)[i] =
+                (ScmUInt64)Scm_GetIntegerU64Clamp(SCM_CAR(cp), clamp, NULL);
+            break;
+        case SCM_UVECTOR_F16:
+            SCM_F16VECTOR_ELEMENTS(v)[i] =
+                (ScmHalfFloat)Scm_DoubleToHalf(Scm_GetDouble(SCM_CAR(cp)));
+            break;
+        case SCM_UVECTOR_F32:
+            SCM_F32VECTOR_ELEMENTS(v)[i] =
+                (float)Scm_GetDouble(SCM_CAR(cp));
+            break;
+        case SCM_UVECTOR_F64:
+            SCM_F64VECTOR_ELEMENTS(v)[i] =
+                Scm_GetDouble(SCM_CAR(cp));
+            break;
+        default:
+            Scm_Error("[internal error] unknown uvector type given to Scm_ListToUVector");
+        }
+    }
+    return SCM_OBJ(v);
+}
+
 /* Generic accessor, intended to be called from VM loop.
    (As the 'VM' in the name suggests, the return value of this API
    should immediately be passed to VM.  See comments on FFX in gauche/number.h)
@@ -290,7 +375,13 @@ ScmObj Scm_MakeUVector(ScmClass *klass, ScmSmallInt size, void *init)
 ScmObj Scm_VMUVectorRef(ScmUVector *v, int t, ScmSmallInt k, ScmObj fallback)
 {
     SCM_ASSERT(Scm_UVectorType(SCM_CLASS_OF(v)) == t);
-    if (k < 0 || k >= SCM_UVECTOR_SIZE(v)) return fallback;
+    if (k < 0 || k >= SCM_UVECTOR_SIZE(v)) {
+        if (SCM_UNBOUNDP(fallback)) {
+            Scm_Error("%s-ref index out of range: %ld",
+                      Scm_UVectorTypeName(t), k);
+        }
+        return fallback;
+    }
     switch (t) {
     case SCM_UVECTOR_S8:  return SCM_MAKE_INT(SCM_S8VECTOR_ELEMENT(v, k));
     case SCM_UVECTOR_U8:  return SCM_MAKE_INT(SCM_U8VECTOR_ELEMENT(v, k));
@@ -383,6 +474,38 @@ DEF_UVCTOR_ARRAY(F16, ScmHalfFloat)
 DEF_UVCTOR_ARRAY(F32, float)
 DEF_UVCTOR_ARRAY(F64, double)
 
+/*
+ * Reader
+ */
+ScmObj Scm_ReadUVector(ScmPort *port, const char *tag, ScmReadContext *ctx)
+{
+    ScmChar c;
+    SCM_GETC(c, port);
+    if (c != '(') Scm_Error("bad uniform vector syntax for %s", tag);
+    ScmObj list = Scm_ReadList(SCM_OBJ(port), ')');
+    ScmClass *klass = NULL;
+    if (strcmp(tag, "s8") == 0)       klass = SCM_CLASS_S8VECTOR;
+    else if (strcmp(tag, "u8") == 0)  klass = SCM_CLASS_U8VECTOR;
+    else if (strcmp(tag, "s16") == 0) klass = SCM_CLASS_S16VECTOR;
+    else if (strcmp(tag, "u16") == 0) klass = SCM_CLASS_U16VECTOR;
+    else if (strcmp(tag, "s32") == 0) klass = SCM_CLASS_S32VECTOR;
+    else if (strcmp(tag, "u32") == 0) klass = SCM_CLASS_U32VECTOR;
+    else if (strcmp(tag, "s64") == 0) klass = SCM_CLASS_S64VECTOR;
+    else if (strcmp(tag, "u64") == 0) klass = SCM_CLASS_U64VECTOR;
+    else if (strcmp(tag, "f16") == 0) klass = SCM_CLASS_F16VECTOR;
+    else if (strcmp(tag, "f32") == 0) klass = SCM_CLASS_F32VECTOR;
+    else if (strcmp(tag, "f64") == 0) klass = SCM_CLASS_F64VECTOR;
+    else Scm_Error("invalid unform vector tag: %s", tag);
+
+    ScmObj uv = Scm_ListToUVector(klass, list, 0);
+    
+    /* If we are reading source file, let literal uvectors be immutable. */
+    if (Scm_ReadContextLiteralImmutable(ctx)) {
+        SCM_UVECTOR_IMMUTABLE_P(uv) = TRUE;
+    }
+    return uv;
+}
+
 /*
  * Class-dependent functions
  */
@@ -394,10 +517,16 @@ static void SCM_CPP_CAT3(print_,tag,vector)(ScmObj obj,                 \
                                             ScmPort *out,               \
                                             ScmWriteContext *ctx)       \
 {                                                                       \
+    const ScmWriteControls *wp =                                        \
+        Scm_GetWriteControls(ctx, out->writeState);                     \
     Scm_Printf(out, "#"#tag"(");                                        \
     for (int i=0; i<SCM_CPP_CAT3(SCM_,TAG,VECTOR_SIZE)(obj); i++) {     \
         T elt = SCM_CPP_CAT3(SCM_,TAG,VECTOR_ELEMENTS)(obj)[i];         \
         if (i != 0) Scm_Printf(out, " ");                               \
+        if (wp->printLength >= 0 && i >= wp->printLength) {             \
+            Scm_Printf(out, "...");                                     \
+            break;                                                      \
+        }                                                               \
         pr(out, elt);                                                   \
     }                                                                   \
     Scm_Printf(out, ")");                                               \
@@ -453,22 +582,33 @@ DEF_PRINT(F64, f64, double, fpr)
 
 /* comparer */
 
-#define DEF_CMP(TAG, tag, T, eq)                                        \
+#define DEF_CMP(TAG, tag, T, eq, lt)                                    \
 static int SCM_CPP_CAT3(compare_,tag,vector)(ScmObj x, ScmObj y, int equalp) \
 {                                                                       \
-    ScmSmallInt len = SCM_CPP_CAT3(SCM_,TAG,VECTOR_SIZE)(x);            \
-    if (SCM_CPP_CAT3(SCM_,TAG,VECTOR_SIZE)(y) != len) return -1;        \
-    for (ScmSmallInt i=0; i<len; i++) {                                 \
-        T xx = SCM_CPP_CAT3(SCM_,TAG,VECTOR_ELEMENTS)(x)[i];            \
-        T yy = SCM_CPP_CAT3(SCM_,TAG,VECTOR_ELEMENTS)(y)[i];            \
-        if (!eq(xx,yy)) {                                               \
-            return -1;                                                  \
+    ScmSmallInt xlen = SCM_CPP_CAT3(SCM_,TAG,VECTOR_SIZE)(x);           \
+    ScmSmallInt ylen = SCM_CPP_CAT3(SCM_,TAG,VECTOR_SIZE)(y);           \
+    if (equalp) {                                                       \
+        if (xlen != ylen) return -1;                                    \
+        for (ScmSmallInt i=0; i<xlen; i++) {                            \
+            T xx = SCM_CPP_CAT3(SCM_,TAG,VECTOR_ELEMENTS)(x)[i];        \
+            T yy = SCM_CPP_CAT3(SCM_,TAG,VECTOR_ELEMENTS)(y)[i];        \
+            if (!eq(xx,yy)) return -1;                                  \
         }                                                               \
+        return 0;                                                       \
+    } else {                                                            \
+        if (xlen != ylen) return (xlen < ylen) ? -1 : 1;                \
+        for (ScmSmallInt i=0; i<xlen; i++) {                            \
+            T xx = SCM_CPP_CAT3(SCM_,TAG,VECTOR_ELEMENTS)(x)[i];        \
+            T yy = SCM_CPP_CAT3(SCM_,TAG,VECTOR_ELEMENTS)(y)[i];        \
+            if (lt(xx, yy)) return -1;                                  \
+            if (!eq(xx,yy)) return 1;                                   \
+        }                                                               \
+        return 0;                                                       \
     }                                                                   \
-    return 0;                                                           \
 }
 
 #define common_eqv(x, y)  ((x)==(y))
+#define common_lt(x, y)   ((x)<(y))
 
 static inline int int64eqv(ScmInt64 x, ScmInt64 y)
 {
@@ -488,16 +628,35 @@ static inline int uint64eqv(ScmUInt64 x, ScmUInt64 y)
 #endif
 }
 
-#define f16eqv(a, b) SCM_HALF_FLOAT_CMP(==, a, b)
+static inline int int64lt(ScmInt64 x, ScmInt64 y)
+{
+#if SCM_EMULATE_INT64
+    return (x.hi < y.hi || (x.hi == y.hi && x.lo < y.lo));
+#else
+    return x < y;
+#endif
+}
 
-DEF_CMP(S8, s8, signed char, common_eqv)
-DEF_CMP(U8, u8, unsigned char, common_eqv)
-DEF_CMP(S16, s16, short, common_eqv)
-DEF_CMP(U16, u16, u_short, common_eqv)
-DEF_CMP(S32, s32, ScmInt32, common_eqv)
-DEF_CMP(U32, u32, ScmUInt32, common_eqv)
-DEF_CMP(S64, s64, ScmInt64, int64eqv)
-DEF_CMP(U64, u64, ScmUInt64, uint64eqv)
-DEF_CMP(F16, f16, ScmHalfFloat, f16eqv)
-DEF_CMP(F32, f32, float, common_eqv)
-DEF_CMP(F64, f64, double, common_eqv)
+static inline int uint64lt(ScmUInt64 x, ScmUInt64 y)
+{
+#if SCM_EMULATE_INT64
+    return (x.hi < y.hi || (x.hi == y.hi && x.lo < y.lo));
+#else
+    return x < y;
+#endif
+}
+
+#define f16eqv(a, b) SCM_HALF_FLOAT_CMP(==, a, b)
+#define f16lt(a, b)  SCM_HALF_FLOAT_CMP(<, a, b)
+
+DEF_CMP(S8, s8, signed char, common_eqv, common_lt)
+DEF_CMP(U8, u8, unsigned char, common_eqv, common_lt)
+DEF_CMP(S16, s16, short, common_eqv, common_lt)
+DEF_CMP(U16, u16, u_short, common_eqv, common_lt)
+DEF_CMP(S32, s32, ScmInt32, common_eqv, common_lt)
+DEF_CMP(U32, u32, ScmUInt32, common_eqv, common_lt)
+DEF_CMP(S64, s64, ScmInt64, int64eqv, int64lt)
+DEF_CMP(U64, u64, ScmUInt64, uint64eqv, uint64lt)
+DEF_CMP(F16, f16, ScmHalfFloat, f16eqv, f16lt)
+DEF_CMP(F32, f32, float, common_eqv, common_lt)
+DEF_CMP(F64, f64, double, common_eqv, common_lt)
