@@ -1,7 +1,7 @@
 ;;;
 ;;; vminsn.scm - Virtual machine instruction definition
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -135,14 +135,12 @@
               `(let* ([,r :: double ,expr])
                  ($result (Scm_VMReturnFlonum ,r))))])
 
-;; Extract local value.  If it is boxed, we need dereference.
+;; Extract local value.
+;; If local var nees unboxing, necessary UNBOX insn will be generated
+;; by the compiler, so we don't need to worry about it.
 (define-cise-stmt $lref!
   [(_ var env off)
-   (let1 v (gensym)
-     `(let* ([,v (ENV-DATA ,env ,off)])
-        (CHECK_UNBOX ,v)
-        (set! ,var ,v)
-        ))])
+   `(set! ,var (ENV-DATA ,env ,off))])
 
 ;;
 ;; ($w/argr <val> <expr> ...)
@@ -177,7 +175,6 @@
                                                'ENV
                                                `(-> ,(loop (- d 1)) up)))
                                           ,o)))
-                     (CHECK_UNBOX ,val)
                      ,@body)])])
 
 (define-cise-stmt $w/argr               ; use VAL0 default
@@ -314,14 +311,12 @@
 ;;
 (define-cise-stmt $lrefNN
   [(_ depth offset)
-   ;; TODO 0.9.2: Extra check for backward compatibility
    (let1 v (gensym)
      `(let* ([,v :: ScmObj (ENV_DATA ,(let loop ((d depth))
                                         (case d
                                           [(0) 'ENV]
                                           [else `(-> ,(loop (- d 1)) up)]))
                                      ,offset)])
-        (CHECK_UNBOX ,v)
         ($result ,v)))])
 
 ;;
@@ -407,7 +402,7 @@
     (CHECK-STACK reqstack)
     NEXT))
 
-;; CALL(NARGS)
+;; CALL(nargs)
 ;;  Call procedure in val0.  The continuation of this call is already
 ;;  pushed by PRE-CALL, so this instruction is always the end of a graph.
 ;;
@@ -420,9 +415,9 @@
     ($define APPLY_CALL)
     ($include "./vmcall.c")))
 
-;; TAIL-CALL(NARGS)
+;; TAIL-CALL(nargs)
 ;;  Call procedure in val0.  Same as CALL except this discards the
-;;  caller's arugment frame and shift the callee's argument frame.
+;;  caller's argument frame and shift the callee's argument frame.
 ;;
 (define-insn TAIL-CALL 1 none #f
   (begin (DISCARD-ENV) ($goto-insn CALL)))
@@ -453,8 +448,9 @@
     (VM_ASSERT (SCM_IDENTIFIERP var))
     (SCM_FLONUM_ENSURE_MEM val)
     INCR-PC
-    (let* ([mod::ScmModule* (-> (SCM_IDENTIFIER var) module)]
-           [name::ScmSymbol* (-> (SCM_IDENTIFIER var) name)])
+    (let* ([id::ScmIdentifier* (Scm_OutermostIdentifier (SCM_IDENTIFIER var))]
+           [mod::ScmModule* (-> id module)]
+           [name::ScmSymbol* (SCM_SYMBOL (-> id name))])
       (case (SCM_VM_INSN_ARG code)        ;flag
         [(0)                    (Scm_MakeBinding mod name val 0)]
         [(1 SCM_BINDING_CONST)  (Scm_MakeBinding mod name val SCM_BINDING_CONST)]
@@ -471,7 +467,7 @@
     INCR-PC
     ($result (Scm_MakeClosure body (get_env vm)))))
 
-;; LOCAL-ENV(NLOCALS)
+;; LOCAL-ENV(nlocals)
 ;;  Create a new environment frame from the current arg frame.
 ;;  Used for let.
 (define-insn LOCAL-ENV  1 none #f
@@ -523,52 +519,18 @@
 (define-insn POP-LOCAL-ENV 0 none #f
   (begin (set! ENV (-> ENV up)) NEXT))
 
-;; LOCAL-ENV-JUMP(DEPTH) <addr>
-;;  This instruction appears when local function call is optimized.
-;;  The stack already has NLOCALS values.  This instruction creates an
-;;  env frame with them (just like LOCAL-ENV), then jump to <addr>.
-;;  (# of arguments can be known by SP - ARGP).
+;; LOCAL-ENV-JUMP(depth) <addr>
+;;  Combination of LOCAL-ENV-SHIFT + JUMP.
+;;  We can use this when none of the new environment needs boxing.
 (define-insn LOCAL-ENV-JUMP 1 addr #f
-  (let* ([nargs::int (cast int (- SP ARGP))]
-         [env_depth::int (SCM_VM_INSN_ARG code)]
-         [to::ScmObj*] [tenv::ScmEnvFrame* ENV])
-    ;; We can discard env_depth environment frames.
-    ;; There are several cases:
-    ;;  - if the target env frame (TENV) is in stack:
-    ;;   -- if the current cont frame is over TENV
-    ;;      => shift argframe on top of the current cont frame
-    ;;   -- otherwise => shift argframe on top of TENV
-    ;;  - if TENV is in heap:
-    ;;   -- if the current cont frame is in stack
-    ;;      => shift argframe on top of the current cont frame
-    ;;   -- otherwise => shift argframe at the stack base
-    (while (> (post-- env_depth) 0)
-      (SCM_ASSERT tenv)
-      (set! tenv (-> tenv up)))
-    (cond [(IN-STACK-P (cast ScmObj* tenv))
-           (if (and (IN-STACK-P (cast ScmObj* CONT))
-                    (> (cast ScmObj* CONT) (cast ScmObj* tenv)))
-             (set! to (CONT-FRAME-END CONT))
-             (set! to (+ (cast ScmObj* tenv) ENV_HDR_SIZE)))]
-          [else
-           (if (IN-STACK-P (cast ScmObj* CONT))
-             (set! to (CONT-FRAME-END CONT))
-             (set! to (-> vm stackBase)))]) ; continuation has been saved
-    (when (and (> nargs 0) (!= to ARGP))
-      (let* ([t::ScmObj* to] [a::ScmObj* ARGP])
-        (dotimes [c nargs]
-          (set! (* (post++ t)) (* (post++ a))))))
-    (set! ARGP to)
-    (set! SP (+ to nargs))
-    (if (> nargs 0)
-      (FINISH-ENV SCM_FALSE tenv)
-      (set! ENV tenv))
+  (begin
+    (local_env_shift vm (SCM_VM_INSN_ARG code))
     (FETCH-LOCATION PC)
     CHECK-INTR
     NEXT))
 
-;; LOCAL-ENV-CALL(DEPTH)
-;; LOCAL-ENV-TAIL-CALL(DEPTH)
+;; LOCAL-ENV-CALL(depth)
+;; LOCAL-ENV-TAIL-CALL(depth)
 ;;  This instruction appears when local function call is optimized.
 ;;  VAL0 has a closure to call, and the stack already has the arguments.
 ;;
@@ -779,7 +741,9 @@
     (VM-ASSERT (!= e NULL))
     (VM-ASSERT (> (-> e size) off))
     (SCM_FLONUM_ENSURE_MEM VAL0)
-    (CHECK_SET_BOX (ENV-DATA e off) VAL0)
+    (let* ([box (ENV-DATA e off)])
+      (VM_ASSERT (SCM_BOXP box))
+      (SCM_BOX_SET box VAL0))
     (set! (-> vm numVals) 1)
     NEXT))
 
@@ -796,16 +760,17 @@
       ;; If runtime flag LIMIT_MODULE_MUTATION is set,
       ;; we search only for the id's module, so that set! won't
       ;; mutate bindings in the other module.
-      (let* ([id::ScmIdentifier* (SCM_IDENTIFIER loc)]
+      (let* ([id::ScmIdentifier* (Scm_OutermostIdentifier (SCM_IDENTIFIER loc))]
+             [name::ScmSymbol* (SCM_SYMBOL (-> id name))]
              [limit::int
               (SCM_VM_RUNTIME_FLAG_IS_SET vm SCM_LIMIT_MODULE_MUTATION)]
              [gloc::ScmGloc*
-              (Scm_FindBinding (-> id module) (-> id name)
+              (Scm_FindBinding (-> id module) name
                                (?: limit SCM_BINDING_STAY_IN_MODULE 0))])
         (when (== gloc NULL)
           ;; Do search again for meaningful error message
           (when limit
-            (set! gloc (Scm_FindBinding (-> id module) (-> id name) 0))
+            (set! gloc (Scm_FindBinding (-> id module) name 0))
             (when (!= gloc NULL)
               ($vm-err "can't mutate binding of %S, \
                         which is in another module"
@@ -827,13 +792,10 @@
 (define-insn LREF        2 none #f
   (let* ([dep::int (SCM_VM_INSN_ARG0 code)]
          [off::int (SCM_VM_INSN_ARG1 code)]
-         [e::ScmEnvFrame* ENV]
-         [v])
+         [e::ScmEnvFrame* ENV])
     (for [() (> dep 0) (post-- dep)]
          (set! e (-> e up)))
-    (set! v (ENV-DATA e off))
-    (CHECK_UNBOX v)
-    ($result v)))
+    ($result (ENV-DATA e off))))
 
 ;; Shortcut for the frequent depth/offset.
 ;; From statistics, we found that the following depth/offset combinations
@@ -853,7 +815,7 @@
 (define-insn LREF21 0 none #f ($lrefNN 2 1))
 (define-insn LREF30 0 none #f ($lrefNN 3 0))
 
-;; combined instrction
+;; combined instruction
 (define-insn LREF-PUSH   2 none  (LREF PUSH))
 (define-insn LREF0-PUSH  0 none  (LREF0 PUSH))
 (define-insn LREF1-PUSH  0 none  (LREF1 PUSH))
@@ -897,7 +859,7 @@
 (define-insn PROMISE  0 none #f
   ($result (Scm_MakePromise FALSE VAL0)))
 
-;; VALUES_APPLY(nargs) <args>
+;; VALUES-APPLY(nargs) <args>
 ;;  This instruction only appears in the code generated dynamically
 ;;  by Scm_Apply(Rec).  This is used to pass the application information
 ;;  across the boundary frame (see user_eval_inner() in vm.c).
@@ -1296,11 +1258,6 @@
              ($result:f (- (cast double imm) (SCM_FLONUM_VALUE arg)))]
             [else           ($result (Scm_Sub (SCM_MAKE_INT imm) arg))]))))
 
-
-;(define-insn NUMQUOT     0 none)        ; quotient
-;(define-insn NUMMOD      0 none)        ; modulo
-;(define-insn NUMREM      0 none)        ; remainder
-
 (define-insn ASHI         1 none #f
   (let* ([cnt::ScmSmallInt (SCM_VM_INSN_ARG code)])
     ($w/argr arg ($result (Scm_Ash arg cnt)))))
@@ -1421,23 +1378,21 @@
 (define-insn-lref* LREF-RET 0 none (LREF RET))
 
 ;; if param == 0, VAL0 <- box(VAL0)
-;; else ENV[param-1] = box(ENV[param-1])
-;; The latter is for arguments that are mutated.
-;; NB: Prior to 0.9.4, BOX insn takes no parameters.  To allow 0.9.4 to be
-;; compiled by 0.9.3.3 compiler, we specify alternative num-params.
-(define-insn BOX (1 0) none #f
+;; else if param > 0,  ENV[param-1] <- box(ENV[param-1])
+;; The second case is for arguments that are mutated.
+(define-insn BOX 1 none #f
   (let* ([param::int (SCM_VM_INSN_ARG code)])
-    (if (== param 0)
-      (begin
-        (SCM_FLONUM_ENSURE_MEM VAL0)
-        (let* ([b::ScmBox* (Scm_MakeBox VAL0)])
-          (set! VAL0 (SCM_OBJ b))))
-      (let* ([off::int (- param 1)])
-        (VM-ASSERT (> (-> ENV size) off))
-        (let* ([v (ENV-DATA ENV off)])
-          (SCM_FLONUM_ENSURE_MEM v)
-          (let* ([b::ScmBox* (Scm_MakeBox v)])
-            (set! (ENV-DATA ENV off) (SCM_OBJ b))))))
+    (cond [(== param 0)
+           (SCM_FLONUM_ENSURE_MEM VAL0)
+           (let* ([b::ScmBox* (Scm_MakeBox VAL0)])
+             (set! VAL0 (SCM_OBJ b)))]
+          [(> param 0)
+           (let* ([off::int (- param 1)])
+             (VM-ASSERT (> (-> ENV size) off))
+             (let* ([v (ENV-DATA ENV off)])
+               (SCM_FLONUM_ENSURE_MEM v)
+               (let* ([b::ScmBox* (Scm_MakeBox v)])
+                 (set! (ENV-DATA ENV off) (SCM_OBJ b)))))])
     NEXT))
 
 ;; ENV-SET(offset)
@@ -1457,9 +1412,26 @@
 ;;  VAL0 <- unbox(VAL0)
 (define-insn UNBOX 0 none #f
   ($w/argr v
-    (DO_UNBOX v)
-    (set! VAL0 v)
+    (set! VAL0 (SCM_BOX_VALUE v))
     NEXT))
 
 (define-insn LREF-UNBOX 2 none (LREF UNBOX) #f :fold-lref)
 
+;; LOCAL-ENV-SHIFT(depth)
+;;  This instruction appears when local function call is optimized.
+;;  The stack already has NLOCALS values.  We discard DEPTH env frames,
+;;  and creates a new local env from the stack value.
+(define-insn LOCAL-ENV-SHIFT 1 none #f
+  (begin
+    (local_env_shift vm (SCM_VM_INSN_ARG code))
+    NEXT))
+
+;; TRANSIENT: These are better to be with ASHI; we put here to keep
+;; binary compatibility.  Should be moved on 1.0 release.
+(define-insn NUMMODI      1 none #f
+  (let* ([divisor::ScmSmallInt (SCM_VM_INSN_ARG code)])
+    ($w/argr arg ($result (Scm_Modulo arg (SCM_MAKE_INT divisor) FALSE)))))
+(define-insn NUMREMI      1 none #f
+  (let* ([divisor::ScmSmallInt (SCM_VM_INSN_ARG code)])
+    ($w/argr arg ($result (Scm_Modulo arg (SCM_MAKE_INT divisor) TRUE)))))
+
