@@ -1,7 +1,7 @@
 /*
  * write.c - writer
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,7 +33,10 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/class.h"
+#include "gauche/writer.h"
 #include "gauche/priv/builtin-syms.h"
+#include "gauche/priv/macroP.h" /* PVREF stuff.  Will go in future */
 #include "gauche/priv/writerP.h"
 #include "gauche/priv/portP.h"
 #include "gauche/char_attr.h"
@@ -47,6 +50,8 @@ static void write_object(ScmObj obj, ScmPort *out, ScmWriteContext *ctx);
 static ScmObj write_object_fallback(ScmObj *args, int nargs, ScmGeneric *gf);
 SCM_DEFINE_GENERIC(Scm_GenericWriteObject, write_object_fallback, NULL);
 
+static const ScmWriteControls *defaultWriteControls;
+
 /*============================================================
  * Writers
  */
@@ -103,14 +108,82 @@ static void write_context_init(ScmWriteContext *ctx, int mode, int flags, int li
     if (SCM_WRITE_CASE(ctx) == 0) ctx->mode |= DEFAULT_CASE;
     ctx->flags = flags;
     ctx->limit = limit;
+    ctx->controls = NULL;
     if (limit > 0) ctx->flags |= WRITE_LIMITED;
 }
 
+/*
+ * WriteControls
+ */
+ScmWriteControls *Scm_MakeWriteControls(const ScmWriteControls *proto)
+{
+    ScmWriteControls *p = SCM_NEW(ScmWriteControls);
+    SCM_SET_CLASS(p, SCM_CLASS_WRITE_CONTROLS);
+    if (proto) {
+        *p = *proto;
+    } else {
+        p->printLength = -1;
+        p->printLevel = -1;
+        p->printBase = 10;
+        p->printRadix = FALSE;
+    }
+    return p;
+}
+
+const ScmWriteControls *Scm_DefaultWriteControls(void)
+{
+    return defaultWriteControls;
+}
+
+const ScmWriteControls *Scm_GetWriteControls(ScmWriteContext *ctx,
+                                             ScmWriteState *st)
+{
+    if (ctx && ctx->controls) return ctx->controls;
+    if (st && st->controls) return st->controls;
+    return defaultWriteControls;
+}
+
+/*
+ * WriteState
+ */
+/* The class definition is in libio.scm  */
+
+/* NB: For the time being, proto argument is ignored. */
+ScmWriteState *Scm_MakeWriteState(ScmWriteState *proto)
+{
+    ScmWriteState *z = SCM_NEW(ScmWriteState);
+    SCM_SET_CLASS(z, SCM_CLASS_WRITE_STATE);
+    z->sharedTable = NULL;
+    z->sharedCounter = 0;
+    z->currentLevel = 0;
+    z->controls = NULL;
+    return z;
+}
+
 /* Cleanup transient data attached to the port. */
-static void cleanup_port_context(ScmPort *port)
+static void cleanup_port_write_state(ScmPort *port)
 {
     port->flags &= ~(SCM_PORT_WALKING|SCM_PORT_WRITESS);
-    port->recursiveContext = SCM_FALSE;
+    if (port->writeState != NULL) {
+        ScmWriteState *s = port->writeState;
+        port->writeState = NULL;
+        /* The table for recursive/shared detection should be GC'ed after
+           we drop the reference to it.  However, the table can be quite big
+           after doing write-shared on a large graph, and our implementation of
+           big hashtables isn't particularly friendly to GC---it is prone
+           to be a victim of false pointers, especially on 32bit architecture.
+           It becomes an issue if the app repeatedly use write-shared on
+           large graph, for an incorrectly retained hashtable may have false
+           pointers to other incorrectly retained hashtable, making the amount
+           of retained garbage unbounded.  So, we take extra step to clear
+           the table to avoid the risk.  In vast majority of the case, the
+           table is used only for circle detection, in which case the table
+           is small and it won't add much overhead.
+        */
+        if (s && s->sharedTable) {
+            Scm_HashCoreClear(SCM_HASH_TABLE_CORE(s->sharedTable));
+        }
+    }
 }
 
 /*
@@ -125,31 +198,59 @@ static void cleanup_port_context(ScmPort *port)
  */
 
 /*
- * Scm_Write - Standard Write.
+ * Scm_Write - Standard Write, convenience version.  Using default controls.
  */
 void Scm_Write(ScmObj obj, ScmObj p, int mode)
+{
+    Scm_WriteWithControls(obj, p, mode, NULL);
+}
+
+/*
+ * Scm_WriteWithControls - the general entry
+ */
+void Scm_WriteWithControls(ScmObj obj, ScmObj p, int mode,
+                           const ScmWriteControls *ctrl)
 {
     if (!SCM_OPORTP(p)) Scm_Error("output port required, but got %S", p);
 
     ScmPort *port = SCM_PORT(p);
-    ScmWriteContext ctx;
-    write_context_init(&ctx, mode, 0, 0);
     ScmVM *vm = Scm_VM();
+    if (ctrl == NULL) ctrl = Scm_DefaultWriteControls();
 
     if (PORT_LOCK_OWNER_P(port, vm) && PORT_RECURSIVE_P(port)) {
-        if (PORT_WALKER_P(port)) write_walk(obj, port);
-        else                     write_rec(obj, port, &ctx);
-        return;
-    }
+        /* We're in the recursive call, so we just recurse into write_walk
+           or write_rec, according to the phase.   NB: The controls passed
+           into the argument CTRL is ignored; the "root" control, passed
+           to the toplevel write API, will be used.  */
+        if (PORT_WALKER_P(port)) {
+            /* Special treatment - if we're "display"-ing a string, we'll
+               bypass walk path even if we're in the middle of write/ss.
+               Using srfi-38 notation to show displayed strings doesn't
+               make sense at all. */
+            if (!((mode == SCM_WRITE_DISPLAY) && SCM_STRINGP(obj))) {
+                write_walk(obj, port);
+            }
+        } else {
+            ScmWriteContext ctx;
+            write_context_init(&ctx, mode, 0, 0);
+            write_rec(obj, port, &ctx);
+        }
 
-    PORT_LOCK(port, vm);
-    if (WRITER_NEED_2PASS(&ctx)) {
-        PORT_SAFE_CALL(port, write_ss(obj, port, &ctx),
-                       cleanup_port_context(port));
     } else {
-        PORT_SAFE_CALL(port, write_rec(obj, port, &ctx), /*no cleanup*/);
+        /* We're in the toplevel call.*/
+        ScmWriteContext ctx;
+        write_context_init(&ctx, mode, 0, 0);
+        PORT_LOCK(port, vm);
+        if (WRITER_NEED_2PASS(&ctx)) {
+            ctx.controls = ctrl;
+            PORT_SAFE_CALL(port, write_ss(obj, port, &ctx),
+                           cleanup_port_write_state(port));
+        } else {
+            /* write-simple case.  CTRL is ignored. */
+            PORT_SAFE_CALL(port, write_rec(obj, port, &ctx), /*no cleanup*/);
+        }
+        PORT_UNLOCK(port);
     }
-    PORT_UNLOCK(port);
 }
 
 /*
@@ -180,7 +281,7 @@ int Scm_WriteLimited(ScmObj obj, ScmObj p, int mode, int width)
     }
 
     ScmObj out = Scm_MakeOutputStringPort(TRUE);
-    SCM_PORT(out)->recursiveContext = SCM_PORT(port)->recursiveContext;
+    SCM_PORT(out)->writeState = SCM_PORT(port)->writeState;
     ScmWriteContext ctx;
     write_context_init(&ctx, mode, 0, width);
 
@@ -322,6 +423,8 @@ static size_t write_char(ScmChar ch, ScmPort *port, ScmWriteContext *ctx)
  */
 ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 {
+    const ScmWriteControls *wp = Scm_GetWriteControls(ctx, port->writeState);
+    
 #define CASE_ITAG_RET(obj, str)                 \
     case SCM_ITAG(obj):                         \
         Scm_PutzUnsafe(str, -1, port);          \
@@ -339,7 +442,8 @@ ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
             Scm_Panic("write: unknown itag object: %08x", SCM_WORD(obj));
         }
     }
-    else if (SCM_INTP(obj)) {
+    else if (SCM_INTP(obj) && wp->printBase == 10 && !wp->printRadix) {
+        /* Shortcut to avoid allocation */
         char buf[SPBUFSIZ];
         int k = snprintf(buf, SPBUFSIZ, "%ld", SCM_INT_VALUE(obj));
         Scm_PutzUnsafe(buf, -1, port);
@@ -350,7 +454,20 @@ ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
         return SCM_MAKE_INT(k);
     }
     else if (SCM_NUMBERP(obj)) {
-        return SCM_MAKE_INT(Scm_PrintNumber(port, obj, NULL));
+        ScmNumberFormat fmt;
+        Scm_NumberFormatInit(&fmt);
+        fmt.radix = wp->printBase;
+        if (wp->printRadix) fmt.flags |= SCM_NUMBER_FORMAT_ALT_RADIX;
+        return SCM_MAKE_INT(Scm_PrintNumber(port, obj, &fmt));
+    }
+    /* PVREF only appears in pattern temlate in the current macro expander.
+       It will be go away once we rewrite the expander. */
+    else if (SCM_PVREF_P(obj)) {
+        char buf[SPBUFSIZ];
+        int k = snprintf(buf, SPBUFSIZ, "#<pvar %" PRIdPTR ".%" PRIdPTR ">",
+                         SCM_PVREF_LEVEL(obj), SCM_PVREF_COUNT(obj));
+        Scm_PutzUnsafe(buf, -1, port);
+        return SCM_MAKE_INT(k);
     }
     return SCM_FALSE;
 }
@@ -389,8 +506,9 @@ ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
    overhaul to fix that.  However, just preventing blowup by lists
    and vectors is still useful.
 
-   The stack is a list, whose element can be (#t . list) or
-   (index . vector).  In the first case, the list part is the
+   The stack is a list, whose element can be (#t count . list) or
+   (index . vector).  In the first case, the count part keeps
+   track of how many siblings we already printed, and the list part is the
    rest of the list we should process after the current item is
    written out.  In the second case, we're processing the vector,
    and the next item we should process is pointed by index.
@@ -407,7 +525,9 @@ ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 static void write_walk(ScmObj obj, ScmPort *port)
 {
     static ScmObj proc = SCM_UNDEFINED;
-    ScmHashTable *ht = SCM_HASH_TABLE(SCM_CDR(port->recursiveContext));
+    SCM_ASSERT(port->writeState);
+    ScmHashTable *ht = port->writeState->sharedTable;
+    SCM_ASSERT(ht != NULL);
     SCM_BIND_PROC(proc, "%write-walk-rec", Scm_GaucheInternalModule());
     Scm_ApplyRec3(proc, obj, SCM_OBJ(port), SCM_OBJ(ht));
 }
@@ -427,9 +547,11 @@ static void write_walk(ScmObj obj, ScmPort *port)
 static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 {
     char numbuf[50];  /* enough to contain long number */
-    ScmHashTable *ht = NULL;
     ScmObj stack = SCM_NIL;
-    int stack_depth = 0;
+    ScmWriteState *st = port->writeState;
+    ScmHashTable *ht = (st? st->sharedTable : NULL);
+    const ScmWriteControls *wp = Scm_GetWriteControls(ctx, st);
+    int stack_depth = 0;        /* only used when !ht */
 
 #define PUSH(elt)                                       \
     do {                                                \
@@ -442,13 +564,20 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 #define POP()                                   \
     do {                                        \
         stack = SCM_CDR(stack);                 \
-        if (ht) stack_depth--;                  \
+        if (!ht) stack_depth--;                 \
     } while (0)
-
-
-    if (SCM_PAIRP(port->recursiveContext) && SCM_HASH_TABLE_P(SCM_CDR(port->recursiveContext))) {
-        ht = SCM_HASH_TABLE(SCM_CDR(port->recursiveContext));
-    }
+#define CHECK_LEVEL()                                                   \
+    do {                                                                \
+        if (st) {                                                       \
+            if (wp->printLevel >= 0 && st->currentLevel >= wp->printLevel) { \
+                Scm_PutcUnsafe('#', port);                              \
+                goto next;                                              \
+            } else {                                                    \
+                if (st) st->currentLevel++;                             \
+            }                                                           \
+        }                                                               \
+    } while (0)
+    
 
     for (;;) {
     write1:
@@ -470,6 +599,7 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
             goto next;
         }
 
+        /* obj is heap allocated and we may use label notation. */
         if (ht) {
             ScmObj e = Scm_HashTableRef(ht, obj, SCM_MAKE_INT(1));
             long k = SCM_INT_VALUE(e);
@@ -480,16 +610,19 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
                 goto next;
             } else if (k > 1) {
                 /* This object will be seen again. Put a reference tag. */
-                int count = SCM_INT_VALUE(SCM_CAR(port->recursiveContext));
-                snprintf(numbuf, 50, "#%d=", count);
-                Scm_HashTableSet(ht, obj, SCM_MAKE_INT(-count), 0);
-                SCM_SET_CAR(port->recursiveContext, SCM_MAKE_INT(count+1));
+                ScmWriteState *s = port->writeState;
+                snprintf(numbuf, 50, "#%d=", s->sharedCounter);
+                Scm_HashTableSet(ht, obj, SCM_MAKE_INT(-s->sharedCounter), 0);
+                s->sharedCounter++;
                 Scm_PutzUnsafe(numbuf, -1, port);
             }
         }
 
         /* Writes aggregates */
         if (SCM_PAIRP(obj)) {
+
+            CHECK_LEVEL();
+
             /* special case for quote etc.
                NB: we need to check if we've seen SCM_CDR(obj), otherwise we'll
                get infinite recursion for the case like (cdr '#1='#1#). */
@@ -513,18 +646,40 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
                 }
             }
 
+            if (wp->printLength == 0) {
+                /* in this case we don't print the elements at all, so we need
+                   to treat this specially. */
+                Scm_PutzUnsafe("(...)", -1, port);
+                if (st) st->currentLevel--;
+                goto next;
+            }
+
             /* normal case */
             Scm_PutcUnsafe('(', port);
-            PUSH(Scm_Cons(SCM_TRUE, SCM_CDR(obj)));
+            PUSH(Scm_Cons(SCM_TRUE, Scm_Cons(SCM_MAKE_INT(1), SCM_CDR(obj))));
             obj = SCM_CAR(obj);
             goto write1;
         } else if (SCM_VECTORP(obj)) {
+
+            CHECK_LEVEL();
+
+            if (wp->printLength == 0) {
+                /* in this case we don't print the elements at all, so we need
+                   to treat this specially. */
+                Scm_PutzUnsafe("#(...)", -1, port);
+                if (st) st->currentLevel--;
+                goto next;
+            }
             Scm_PutzUnsafe("#(", -1, port);
             PUSH(Scm_Cons(SCM_MAKE_INT(1), obj));
             obj = SCM_VECTOR_ELEMENT(obj, 0);
             goto write1;
+        } else if (Scm_ClassOf(obj)->flags & SCM_CLASS_AGGREGATE) {
+            CHECK_LEVEL();
+            write_general(obj, port, ctx);
+            if (st) st->currentLevel--;
+            goto next;
         } else {
-            /* string or user-defined object */
             write_general(obj, port, ctx);
             goto next;
         }
@@ -542,6 +697,9 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
                 if (i == len) { /* we've done this vector */
                     Scm_PutcUnsafe(')', port);
                     POP();
+                } else if (wp->printLength >= 0 && wp->printLength <= i) {
+                    Scm_PutzUnsafe(" ...)", -1, port);
+                    POP();
                 } else {
                     Scm_PutcUnsafe(' ', port);
                     obj = SCM_VECTOR_ELEMENT(v, i);
@@ -550,63 +708,89 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
                 }
             } else {
                 /* we're processing a list */
-                ScmObj v = SCM_CDR(top);
-                ScmObj e;
+                SCM_ASSERT(SCM_PAIRP(SCM_CDR(top)));
+                long count = SCM_INT_VALUE(SCM_CADR(top));
+                ScmObj v = SCM_CDDR(top);
                 if (SCM_NULLP(v)) { /* we've done with this list */
                     Scm_PutcUnsafe(')', port);
                     POP();
                 } else if (!SCM_PAIRP(v)) {
+                    /* Improper list.  We treat aggregate types specially,
+                       since such object at this position shouldn't increment
+                       "level" - its content is regarded as the same level of
+                       the current list.
+                     */
                     Scm_PutzUnsafe(" . ", -1, port);
-                    obj = v;
-                    SCM_SET_CDR(top, SCM_NIL);
-                    goto write1;
+                    if (Scm_ClassOf(v)->flags & SCM_CLASS_AGGREGATE) {
+                        if (st) st->currentLevel--;
+                        write_rec(v, port, ctx);
+                        if (st) st->currentLevel++;
+                        Scm_PutcUnsafe(')', port);
+                        POP();
+                    } else {
+                        obj = v;
+                        SCM_SET_CAR(SCM_CDR(top), SCM_MAKE_INT(count+1));
+                        SCM_SET_CDR(SCM_CDR(top), SCM_NIL);
+                        goto write1;
+                    }
+                } else if (wp->printLength >= 0 && wp->printLength <= count) {
+                    /* print-length limit reached */
+                    Scm_PutzUnsafe(" ...)", -1, port);
+                    POP();
                 } else if (ht && !SCM_EQ(Scm_HashTableRef(ht, v, SCM_MAKE_INT(1)), SCM_MAKE_INT(1)))  {
                     /* cdr part is shared */
                     Scm_PutzUnsafe(" . ", -1, port);
                     obj = v;
-                    SCM_SET_CDR(top, SCM_NIL);
+                    SCM_SET_CAR(SCM_CDR(top), SCM_MAKE_INT(count+1));
+                    SCM_SET_CDR(SCM_CDR(top), SCM_NIL);
                     goto write1;
                 } else {
                     Scm_PutcUnsafe(' ', port);
                     obj = SCM_CAR(v);
-                    SCM_SET_CDR(top, SCM_CDR(v));
+                    SCM_SET_CAR(SCM_CDR(top), SCM_MAKE_INT(count+1));
+                    SCM_SET_CDR(SCM_CDR(top), SCM_CDR(v));
                     goto write1;
                 }
             }
+            if (st) st->currentLevel--;
         }
         break;
     }
 
 #undef PUSH
 #undef POP
+#undef CHECK_DEPTH
 }
 
 /* Write/ss main driver
    This should never be called recursively.
-   We modify port->flags and port->recursiveContext; they are cleaned up
+   We modify port->flags and port->writeState; they are cleaned up
    by the caller even if we throw an error during write. */
 static void write_ss(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 {
-    SCM_ASSERT(SCM_FALSEP(port->recursiveContext));
+    SCM_ASSERT(port->writeState == NULL);
 
     /* pass 1 */
     port->flags |= SCM_PORT_WALKING;
     if (SCM_WRITE_MODE(ctx)==SCM_WRITE_SHARED) port->flags |= SCM_PORT_WRITESS;
-    port->recursiveContext = Scm_Cons(SCM_MAKE_INT(0),
-                                      Scm_MakeHashTableSimple(SCM_HASH_EQ, 0));
+    ScmWriteState *s = Scm_MakeWriteState(NULL);
+    s->sharedTable = SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQ, 0));
+    s->controls = ctx->controls;
+    port->writeState = s;
+
     write_walk(obj, port);
     port->flags &= ~(SCM_PORT_WALKING|SCM_PORT_WRITESS);
 
     /* pass 2 */
     write_rec(obj, port, ctx);
-    port->recursiveContext = SCM_FALSE;
+    cleanup_port_write_state(port);
 }
 
 /*OBSOLETED*/
 /*format is now in Scheme (libfmt.scm).*/
 void Scm_Format(ScmPort *out, ScmString *fmt, ScmObj args, int sharedp)
 {
-    Scm_Error("Scm_Format is obsoleted\n");
+    Scm_Error("Scm_Format is obsoleted");
 }
 
 /*
@@ -716,7 +900,7 @@ static ScmObj vprintf_pass1(ScmPort *out, const char *fmt, va_list ap)
                 }
             case 'C':
                 {
-                    int c = va_arg(ap, int);
+                    ScmChar c = va_arg(ap, ScmChar);
                     SCM_APPEND1(h, t, Scm_MakeInteger(c));
                     break;
                 }
@@ -974,4 +1158,5 @@ void Scm__InitWrite(void)
 {
     Scm_InitBuiltinGeneric(&Scm_GenericWriteObject, "write-object",
                            Scm_GaucheModule());
+    defaultWriteControls = Scm_MakeWriteControls(NULL);
 }
