@@ -138,14 +138,15 @@
   (test* "port-attributes" '((my-attribute . ok)
                              (my-attribute2 . 103)
                              (my-attribute3 . 12))
-         (sort (port-attributes p)
-               string<?
-               ($ symbol->string $ car $)))
+         (alist-delete 'reader-lexical-mode
+                       (sort (port-attributes p)
+                             string<?
+                             ($ symbol->string $ car $))))
   (test* "port-attribute-delete!" '((my-attribute3 . 13))
          (begin
            (port-attribute-delete! p 'my-attribute)
            (port-attribute-delete! p 'my-attribute2)
-           (port-attributes p)))
+           (alist-delete 'reader-lexical-mode (port-attributes p))))
   )
 
 ;;-------------------------------------------------------------------
@@ -472,6 +473,41 @@
            (let1 c1 (peek-char)
              (port-seek (current-input-port) 0 SEEK_CUR)
              (list c1 (peek-char))))))
+(test* "seek (istr, with peek-char)" '(#\a 0 #\b)
+       (with-input-from-string "ab"
+         (^()
+           (let* ((c0 (peek-char))
+                  (pos (port-tell (current-input-port)))
+                  (c1 (begin
+                        (port-seek  (current-input-port) 1 SEEK_CUR)
+                        (read-char))))
+             (list c0 pos c1)))))
+(test* "seek (istr, with peek-char, negative offset)" '(#\c 2 #\b)
+       (with-input-from-string "abc"
+         (^()
+           (read-char)
+           (read-char)
+           (let* ((c0 (peek-char))
+                  (pos (port-tell (current-input-port)))
+                  (c1 (begin
+                        (port-seek  (current-input-port) -1 SEEK_CUR)
+                        (read-char))))
+             (list c0 pos c1)))))
+(test* "seek (istr, with peek-byte)" '(#x61 #x62)
+       (with-input-from-string (rlet1 s (make-byte-string 2 #x61)
+                                 (string-byte-set! s 1 #x62))
+         (^()
+           (let1 b1 (peek-byte)
+             (port-seek (current-input-port) 1)
+             (list b1 (peek-byte))))))
+(test* "seek (istr, with peek-byte)" '(#x61 0 #x62)
+       (with-input-from-string (rlet1 s (make-byte-string 2 #x61)
+                                 (string-byte-set! s 1 #x62))
+         (^()
+           (let* ([b1 (peek-byte)]
+                  [pos (port-tell (current-input-port))])
+             (port-seek (current-input-port) 1 SEEK_CUR)
+             (list b1 pos (peek-byte))))))
 
 ;; NB: in the following four test, each ifile-ofile test is a pair
 ;;     (the ofile test depends on the previous state by ifile).  do not
@@ -573,6 +609,29 @@
                       [second (read-zstring p)])
                  (list first second)))))))
 
+(test* "seek (ifile, with peek-byte)" '(#x61 #x62)
+       (begin
+         (sys-unlink "test.o")
+         (with-output-to-file "test.o"
+           (cut display (rlet1 s (make-byte-string 2 #x61)
+                          (string-byte-set! s 1 #x62))))
+         (call-with-input-file "test.o"
+           (^p
+             (let1 b1 (peek-byte p)
+               (port-seek p 1)
+               (list b1 (peek-byte p)))))))
+
+(test* "seek (with appending output)" '(50 100)
+       (begin
+         (sys-unlink "test.o")
+         (with-output-to-file "test.o"
+           (cut display (make-string 50 #\a)))
+         (call-with-output-file "test.o"
+           (^p (let1 a (port-tell p)
+                 (display (make-string 50 #\b) p)
+                 (list a (port-tell p))))
+           :if-exists :append)))
+
 (sys-unlink "test.o")
 
 ;;-------------------------------------------------------------------
