@@ -66,15 +66,17 @@
 (test "reload after error"
       1
       (^[]
-        (with-error-handler
-         (^e #t)
-         (^[] (eval '(require "test.o/d") (interaction-environment))))
-        (with-output-to-file "test.o/d.scm"
-          (^[]
-            (write '(define z 1))
-            (write '(provide "tset.o/d"))))
-        (eval '(require "test.o/d") (interaction-environment))
-        (eval 'z (interaction-environment))))
+        (let ((m (make-module 'reload-after-error)))
+          (with-error-handler
+              (^e #t)
+            (^[] (eval '(require "test.o/d") (interaction-environment))))
+          (with-output-to-file "test.o/d.scm"
+            (^[]
+              (write '(select-module reload-after-error))
+              (write '(define z 1))
+              (write '(provide "tset.o/d"))))
+          (eval '(require "test.o/d") (interaction-environment))
+          (eval 'z m))))
 
 ;; :environment arg -------------------------------------
 (test-section "load environment")
@@ -182,6 +184,14 @@
        (let ((inc-var 2))
          (include "test.o/inc0")
          inc-var))
+(test* "include (relative to current, dots, expands inside let)" 4
+       (let ((inc-var 2))
+         (include "../src/test.o/inc0")
+         inc-var))
+(test* "include (relative to current, dots, expands inside let)" 4
+       (let ((inc-var 2))
+         (include "./test.o/inc0")
+         inc-var))
 (test* "include (absolute, expands into toplevel)" 4
        (include-into-toplevel
         (sys-normalize-pathname "test.o/inc0" :absolute #t)))
@@ -378,7 +388,7 @@
        (not (not (library-exists? 'gauche :paths paths-b))))
 (test* "library-exists? gauche, force-search" #f
        (not (not (library-exists? 'gauche :paths paths-b :force-search? #t))))
-(test* "library-exists? gauche" #f
+(test* "library-exists? gauche" #t
        (not (not (library-exists? "gauche" :paths paths-b))))
 ;;NB: this no longer work since gauche.object is compiled in.
 ;(test* "library-exists? gauche/object" #t
