@@ -18,6 +18,212 @@
 (define-macro (test-macro msg expect form)
   `(test ,msg ',expect (lambda () (unident (%macroexpand ,form)))))
 
+;;----------------------------------------------------------------------
+;; 
+
+(test-section "primitive macro transformer")
+
+(define-syntax defmac  ; (defmac name args . body)
+  (primitive-macro-transformer
+   (lambda (form def-env use-env)
+     (let ([name (cadr form)]
+           [args (caddr form)]
+           [body (cdddr form)])
+       `(define-syntax ,name
+          (let ((,name (lambda ,args ,@body)))
+            (primitive-macro-transformer
+             (lambda (f d u) (apply ,name (cdr f))))))))))
+
+(defmac myif (test then else) `(if ,test ,then ,else))
+
+(test-macro "defmac if" (if a b c) (myif a b c))
+
+;; explicit renaming macros
+
+(test-section "ER macro basics")
+
+(define-syntax er-when
+  (er-macro-transformer
+   (^[f r c]
+     (let ([test (cadr f)]
+           [exprs (cddr f)])
+       `(,(r 'if) ,test (,(r 'begin) ,@exprs))))))
+
+(test "when - basic" #t (^[] (let ((x #f)) (er-when #t (set! x #t)) x)))
+(test "when - basic" #f (^[] (let ((x #f)) (er-when #f (set! x #t)) x)))
+
+(test "when - hygene" 3
+      (^[] (let ([if list]
+                 [begin list])
+             (er-when #t 1 2 3))))
+
+(define-syntax er-aif
+  (er-macro-transformer
+   (^[f r c]
+     (let ([test (cadr f)]
+           [then (caddr f)]
+           [else (cadddr f)])
+       `(,(r 'let) ((it ,test))
+           (,(r 'if) it ,then ,else))))))
+
+(test "aif - basic" 4 (^[] (er-aif (+ 1 2) (+ it 1) #f)))
+(test "aif - basic" 5 (^[] (let ((it 999)) (er-aif (+ 1 2) (+ it 2) #f))))
+
+(test "aif - hygene" 6
+      (^[] (let ((it 999)
+                 (let list))
+             (er-aif (+ 1 2) (+ it 3) #f))))
+(test "aif - nesting" #t
+      (^[] (let ([it 999])
+             (er-aif (+ 1 2) (er-aif (odd? it) it #f) #f))))
+
+(test-section "ER macro local scope")
+
+(let ([if list])
+  (let-syntax ([fake-if (er-macro-transformer
+                         (^[f r c] `(,(r 'if) ,@(cdr f))))])
+    (test "fake-if" '(1 2 3) (^[] (fake-if 1 2 3)))
+    (let ([if +])
+      (test "fake-if" '(4 5 6) (^[] (fake-if 4 5 6))))))
+
+(test-section "ER compare literals")
+
+;; from Clinger "Hygienic Macros Through Explicit Renaming"
+(define-syntax er-cond
+  (er-macro-transformer
+   (^[f r c]
+     (let1 clauses (cdr f)
+       (if (null? clauses)
+         `(,(r 'quote) ,(r 'unspecified))
+         (let* ([first (car clauses)]
+                [rest  (cdr clauses)]
+                [test  (car first)])
+           (cond [(and (or (symbol? test)
+                           (identifier? test))
+                       (c test (r 'else)))
+                  `(,(r 'begin) ,@(cdr first))]
+                 [else `(,(r 'if) ,test
+                         (,(r 'begin) ,@(cdr first))
+                         (er-cond ,@rest))])))))))
+
+(define (er-cond-tester1 x)
+  (er-cond [(odd? x) 'odd] [else 'even]))
+
+(test "er-cond 1" '(even odd)
+      (^[] (list (er-cond-tester1 0) (er-cond-tester1 1))))
+
+(let ([else #f])
+  (define (er-cond-tester2 x)
+    (er-cond [(odd? x) 'odd] [else 'even]))
+  (test "er-cond 2" '(unspecified odd)
+        (^[] (list (er-cond-tester2 0) (er-cond-tester2 1)))))
+
+(define-module er-test-mod
+  (export er-cond2)
+  (define-syntax er-cond2
+    (er-macro-transformer
+     (^[f r c]
+       (let1 clauses (cdr f)
+         (if (null? clauses)
+           `(,(r 'quote) ,(r 'unspecified))
+           (let* ([first (car clauses)]
+                  [rest  (cdr clauses)]
+                  [test  (car first)])
+             (cond [(and (or (symbol? test)
+                             (identifier? test))
+                         (c test (r 'else)))
+                    `(,(r 'begin) ,@(cdr first))]
+                   [else `(,(r 'if) ,test
+                           (,(r 'begin) ,@(cdr first))
+                           (er-cond2 ,@rest))]))))))))
+
+(define-module er-test-mod2
+  (use gauche.test)
+  (import er-test-mod)
+  (define (er-cond-tester1 x)
+    (er-cond2 [(odd? x) 'odd] [else 'even]))
+  (test "er-cond (cross-module)" '(even odd)
+        (^[] (list (er-cond-tester1 0) (er-cond-tester1 1)))))
+
+;; Introducing local bindings
+(let ((x 3))
+  (let-syntax ([foo (er-macro-transformer
+                     (^[f r c]
+                       (let1 body (cdr f)
+                         `(,(r 'let) ([,(r 'x) (,(r '+) ,(r 'x) 2)])
+                           (,(r '+) ,(r 'x) ,@body)))))])
+    (let ((x -1))
+      (test* "er-macro introducing local bindings" 4
+             (foo x)))))
+
+;; er-macro and nested identifier
+;; cf. http://saito.hatenablog.jp/entry/2014/11/18/233209
+(define (er-test-traverse proc obj)
+  (let loop ((obj obj))
+    (cond [(or (symbol? obj) (identifier? obj))    (proc obj)]
+          [(pair? obj)   (cons (loop (car obj)) (loop (cdr obj)))]
+          [(vector? obj) (vector-map loop obj)]
+          [else obj])))
+
+(define-syntax er-test-let/scope
+  (er-macro-transformer
+   (lambda (form rename _)
+     (let ([scope (cadr form)]
+           [body (cddr form)])
+       `(let-syntax ((,scope
+                      (,(rename 'er-macro-transformer)
+                       (,(rename 'lambda) (f r _)
+                        (,(rename 'let) ((form2 (,(rename 'cdr) f)))
+                         (,(rename 'cons)
+                          ',(rename 'begin)
+                          (,(rename 'er-test-traverse) r form2)))))))
+          ,@body)))))
+
+(test "er-macro and nested identifier"
+      '(2 2 3 4)
+      (lambda ()
+        (let ([x 1])
+          (er-test-let/scope scope-1
+            (let ([x 2])
+              (er-test-let/scope scope-2
+                (let ([x 3])
+                  (er-test-let/scope scope-1
+                    (let ([x 4])
+                      (list (scope-2 (scope-1 x))
+                            (scope-2 x)
+                            (scope-1 x)
+                            x))))))))))
+
+;; passing form rename procedure
+(let ([a 1] [b 2])
+  (let-syntax ([foo (er-macro-transformer
+                     (lambda (f r c)
+                       (r '(cons (list a b) `#(,a ,b)))))])
+    (let ([a -1] [b -2] [list *])
+      (test* "list arg for rename procedure"
+             '((1 2) . #(1 2))
+             (foo)))))
+
+;; quasirename
+;; Note: currently quasirename depends on util.match, too.
+(let ((unquote list)
+      (x 1)
+      (y 2))
+  (let-syntax ([foo (er-macro-transformer
+                     (^[f r c]
+                       (let ([a (cadr f)]
+                             [b (caddr f)]
+                             [all (cdr f)])
+                         (quasirename r
+                           (list x ,a y ,b ,@all
+                                 '#(x ,a y ,b) ,@(reverse all))))))])
+    (let ((list vector)
+          (x 10)
+          (y 20))
+      (test* "er-macro and quasirename"
+             '(1 3 2 4 3 4 #(x 3 y 4) 4 3)
+             (foo 3 4)))))
+
 ;;----------------------------------------------------------------------
 ;; basic tests
 
@@ -550,6 +756,26 @@
       (lambda ()
         (let-syntax ([if fi]) (if #f 'ok 'ng))))
 
+;; Macro-generating-macro scoping
+;; Currently it's not working.
+(define-syntax mgm-bar
+  (syntax-rules ()
+    ((_ . xs) '(bad . xs))))
+
+(define-syntax mgm-foo
+  (syntax-rules ()
+    ((_ xs)
+     (letrec-syntax ((mgm-bar
+                      (syntax-rules ()
+                        ((_ (%x . %xs) %ys)
+                         (mgm-bar %xs (%x . %ys)))
+                        ((_ () %ys)
+                         '%ys))))
+       (mgm-bar xs ())))))
+
+(test "macro-generating-macro scope" '(z y x)
+      (lambda () (mgm-foo (x y z))))
+
 ;;----------------------------------------------------------------------
 ;; macro and internal define
 
@@ -756,15 +982,14 @@
     ((mdm-foo3 y)
      (letrec-syntax ((o? (syntax-rules ()
                            ((o? ()) #f)
-                           ((o? (x . xs)) (not (e? xs)))))
+                           ((o? (x . xs)) (e? xs))))
                      (e? (syntax-rules ()
                            ((e? ()) #t)
-                           ((e? (x . xs)) (not (o? xs))))))
-       (%macroexpand (e? y))))))
+                           ((e? (x . xs)) (o? xs)))))
+       (e? y)))))
 
-;; this doesn't work for now, due to the bug of macro expander
-'(test "define-syntax - letrec-syntax" #t
-      (lambda () (mdm-foo3 (a))))
+(test "define-syntax - letrec-syntax" #t
+      (lambda () (mdm-foo3 (a b c d))))
 
 ;; Examples from "Two pitfalls in programming nested R5RS macros"
 ;; by Oleg Kiselyov
@@ -806,9 +1031,86 @@
 (test "passing by parameters" 5
       (lambda () ((mdm-bar-m3 z (+ z 1)) 4)))
 
+;; Macro defining toplevel macros.
+(define-syntax defMyQuote
+  (syntax-rules ()
+    ((_ name)
+     (begin
+       (define-syntax TEMP
+         (syntax-rules ()
+           ((_ arg)
+            `arg)))
+       (define-syntax name
+         (syntax-rules ()
+           ((_ arg)
+            (TEMP arg))))))))
+
+(defMyQuote MyQuote)
+
+(test "macro defining a toplevel macro" '(1 2 3)
+      (lambda () (MyQuote (1 2 3))))
+
+;; Macro inserting toplevel identifier
+(define-module defFoo-test
+  (export defFoo)
+  (define-syntax defFoo
+    (syntax-rules ()
+      [(_ accessor)
+       (begin
+         (define foo-toplevel 42)
+         (define (accessor) foo-toplevel))])))
+
+(import defFoo-test)
+(defFoo get-foo)
+
+(test "macro injecting toplevel definition" '(#f #f 42)
+      (lambda ()
+        (list (global-variable-ref (current-module) 'foo-toplevel #f)
+              (global-variable-ref (find-module 'defFoo-test) 'foo-toplevel #f)
+              (get-foo))))
+
+;; recursive reference in macro-defined-macro
+;; https://gist.github.com/ktakashi/03ae059f804a723a9589
+(define-syntax assocm
+  (syntax-rules ()
+    ((_ key (alist ...))
+     (letrec-syntax ((fooj (syntax-rules (key)
+			    ((_ (key . e) res (... ...)) '(key . e))
+			    ((_ (a . d) res (... ...)) (fooj res (... ...))))))
+       (fooj alist ...)))))
+
+(test "recursive reference in macro-defined-macro" '(c . d)
+      (lambda () (assocm c ((a . b) (b . d) (c . d) (d . d)))))
+
+;; literal identifier comparison with renamed identifier
+;; https://gist.github.com/ktakashi/fa4ee23da88151536619
+(define-module literal-id-test-sub
+  (export car))
+
+(define-module literal-id-test
+  (use gauche.test)
+  (import (literal-id-test-sub :rename ((car car-alias))))
+
+  (define-syntax free-identifier=??
+    (syntax-rules ()
+      ((_ a b)
+       (let-syntax ((foo (syntax-rules (a)
+                           ((_ a) #t)
+                           ((_ _) #f))))
+         (foo b)))))
+
+  (test "literal identifier comparison a a" #t
+        (lambda () (free-identifier=?? a a)))
+  (test "literal identifier comparison b a" #f
+        (lambda () (free-identifier=?? b a)))
+  (test "literal identifier comparison car car-alias" #t
+        (lambda () (free-identifier=?? car car-alias))))
+
 ;;----------------------------------------------------------------------
 ;; identifier comparison
 
+(test-section "identifier comparison")
+
 ;; This is EXPERIMENTAL: may be changed in later release.
 
 (define-syntax hoge (syntax-rules () ((hoge foo ...) (cdr b))))
@@ -817,6 +1119,30 @@
 (test "comparison of identifiers" (macroexpand '(hoge bar))
       (lambda () (macroexpand '(hoge bar))))
 
+;;----------------------------------------------------------------------
+;; keyword and extended lambda list
+
+(test-section "keyword inserted by macro")
+
+(define-syntax define-extended-1
+  (syntax-rules ()
+    [(_ name)
+     (define (name a :key (b #f))
+       (list a b))]))
+
+(define-extended-1 extended-1)
+(test "macro expands to extended lambda list" '(1 2)
+      (lambda () (extended-1 1 :b 2)))
+
+(define-syntax define-extended-2
+  (syntax-rules ()
+    [(_ name)
+     (define (name a :key ((:b boo) #f))
+       (list a boo))]))
+(define-extended-2 extended-2)
+(test "macro expands to extended lambda list" '(3 4)
+      (lambda () (extended-2 3 :b 4)))
+
 ;;----------------------------------------------------------------------
 ;; common-macros
 
