@@ -599,9 +599,6 @@
 (test* "object-compare" -1 (compare #\a #\b))
 (test* "object-compare" 0  (compare #\a #\a))
 (test* "object-compare" 1  (compare #\b #\a))
-(test* "object-compare" (test-error) (compare #\b 4))
-(test* "object-compare" (test-error) (compare "zzz" 4))
-(test* "object-compare" (test-error) (compare 2+i 3+i))
 (test* "object-compare" -1 (compare (make <cmp> :x 3) (make <cmp> :x 4)))
 (test* "object-compare" 0 (compare (make <cmp> :x 3) (make <cmp> :x 3)))
 (test* "object-compare" 1 (compare (make <cmp> :x 4) (make <cmp> :x 3)))
@@ -892,6 +889,22 @@
        '(A b C x Y z)
        (map (cut slot-ref *docu-sub* <>) '(a b c x y z)))
 
+;;----------------------------------------------------------------
+(test-section "metaclass/bound-slot")
+
+(use gauche.mop.bound-slot)
+(test-module 'gauche.mop.bound-slot)
+
+(define-class <bound-slot-A> (<bound-slot-mixin>)
+  ((a :init-keyword :a)
+   (b :init-keyword :b :init-value 1)))
+
+(test* "<bound-slot-A> check" (test-error)
+       (make <bound-slot-A>))
+
+(test* "<bound-slot-A> check" #t
+       (is-a? (make <bound-slot-A> :a 3) <bound-slot-A>))
+
 ;;----------------------------------------------------------------
 (test-section "metaclass/singleton")
 
@@ -1398,10 +1411,8 @@
 ;;----------------------------------------------------------------
 (test-section "applicable?")
 
-;; The procedure applicable? is in lib/gauche/procedure.scm and
-;; it's supposed to be tested in ./procedure.scm, but internally
-;; it depends on libobj's method-applicable-for-clases, so we test
-;; it here.
+;; The procedure applicable? depends on libobj's method-applicable-for-clases,
+;; so we test it here.
 
 (define-syntax atest*
   (syntax-rules (=>)
@@ -1415,11 +1426,18 @@
 (atest* list [() => #t] [(<top>) => #t] [(<top> <top>) => #t])
 (atest* apply [() => #f] [(<top>) => #f] [(<top> <top>) => #t])
 
+(atest* (case-lambda ([a] 1) ([a b c] 2) ([a b c d e . f] 3))
+        [() => #f] [(<top>) => #t] [(<top> <top>) => #f]
+        [(<top> <top> <top>) => #t] [(<top> <top> <top> <top>) => #f]
+        [(<top> <top> <top> <top> <top>) => #t]
+        [(<top> <top> <top> <top> <top> <top>) => #t]
+        [(<top> <top> <top> <top> <top> <top> <top>) => #t])
+
+;; NB: The result may change if loaded module adds new method to 'ref'.
 (atest* ref [() => #f] [(<boolean>) => #f] [(<boolean> <top>) => #f]
         [(<top>) => #f] [(<top> <symbol>) => #t] [(<top> <symbol> <top>) => #t]
         [(<vector>) => #f] [(<vector> <integer>) => #t]
-        [(<vector> <integer> <integer>) => #f]
-        [(<vector> <integer> <top>) => #f])
+        [(<vector> <vector> <integer> <integer>) => #f])
 
 (atest* #/a/ [() => #f] [(<string>) => #t] [(<integer>) => #f])
 
