@@ -42,11 +42,15 @@
                (for-each print (sys-readdir "."))
                0))
      (write '(define (grep args)
-               (let ([pat (string->regexp (car args))]
-                     [hit 1])
+               (let* ([opt-v (equal? (car args) "-v")]
+                      [pat (string->regexp (if opt-v (cadr args) (car args)))]
+                      [hit 1])
                  (generator-for-each
                   (lambda (line)
-                    (when (rxmatch pat line) (set! hit 0) (print line)))
+                    (when ((if opt-v not identity)
+                           (rxmatch pat line))
+                      (set! hit 0)
+                      (print line)))
                   read-line)
                  hit)))
      ))]
@@ -253,6 +257,36 @@
 (test* "process-list" '()
        (process-list))
 
+;;-------------------------------
+(test-section "pipeline")
+
+(let ()
+  (with-output-to-file "test.o"
+    (^[] (for-each print '("banana" "habana" "tabata" "cabara"))))
+
+  (test* "pipelining 1" "banana\nhabana\n"
+         (let1 ps (run-process-pipeline `(,(cmd cat "test.o")
+                                          ,(cmd grep "bana"))
+                                        :output :pipe)
+           (begin0 (port->string (process-output (last ps)))
+             (for-each process-wait ps))))
+
+  (test* "pipelining 2" "tabata\ncabara\n"
+         (let1 ps (run-process-pipeline `(,(cmd cat)
+                                          ,(cmd grep "-v" "bana"))
+                                        :input "test.o" :output :pipe)
+           (begin0 (port->string (process-output (last ps)))
+             (for-each process-wait ps))))
+
+  (test* "pipelining 3" "banana\ncabara\n"
+         (let1 ps (run-process-pipeline `(,(cmd cat)
+                                          ,(cmd grep "-v" "ta")
+                                          ,(cmd grep "-v" "ha"))
+                                        :input "test.o" :output :pipe)
+           (begin0 (port->string (process-output (last ps)))
+             (for-each process-wait ps))))  
+  )
+
 ;;-------------------------------
 (test-section "process ports")
 
@@ -488,31 +522,78 @@
              (s (call-with-input-file "test.o" port->string-list)))
          (equal? r s)))
 
-(cond-expand
- [gauche.os.windows
-  (test* "shell-escape-string" "\"\"" (shell-escape-string ""))
-  (test* "shell-escape-string" "abc" (shell-escape-string "abc"))
-  (test* "shell-escape-string" "\"a b c\"" (shell-escape-string "a b c"))
-  (test* "shell-escape-string" "\"a \"\"b\"\" c\""
-         (shell-escape-string "a \"b\" c"))
-  ]
- [else
-  (test* "shell-escape-string" "''" (shell-escape-string ""))
-  (test* "shell-escape-string" "abc" (shell-escape-string "abc"))
-  (test* "shell-escape-string" "'a b c'" (shell-escape-string "a b c"))
-  (test* "shell-escape-string" "'$abc'" (shell-escape-string "$abc"))
-  (test* "shell-escape-string" "'[abc]'" (shell-escape-string "[abc]"))
-  (test* "shell-escape-string" "'\\abc'" (shell-escape-string "\\abc"))
-  (test* "shell-escape-string" "'>abc'" (shell-escape-string ">abc"))
-  (test* "shell-escape-string" "'<abc'" (shell-escape-string "<abc"))
-  (test* "shell-escape-string" "'a\"b\"c'" (shell-escape-string "a\"b\"c"))
-  (test* "shell-escape-string" "'a(b)c'" (shell-escape-string "a(b)c"))
-  (test* "shell-escape-string" "'a{b}c'" (shell-escape-string "a{b}c"))
-  (test* "shell-escape-string" "'a'\"'\"'c'" (shell-escape-string "a'c"))
-  ])
-
 (rmrf "testc.o")
 
+(let ()
+  (define (t input expected flavor)
+    (test* #"shell-escape-string(~flavor) - ~input" expected
+           (shell-escape-string input flavor)))
+
+  (t ""          "\"\""              'windows)
+  (t "abc"       "abc"               'windows)
+  (t "a b c"     "\"a b c\""         'windows)
+  (t "a \"b\" c" "\"a \\\"b\\\" c\"" 'windows)
+  (t "a\\ \\\""  "\"a\\ \\\\\\\"\""  'windows)
+  (t ""          "''"                'posix)
+  (t "abc"       "abc"               'posix)
+  (t "a b c"     "'a b c'"           'posix)
+  (t "$abc"      "'$abc'"            'posix)
+  (t "[abc]"     "'[abc]'"           'posix)
+  (t "\\abc"     "'\\abc'"           'posix)
+  (t ">abc"      "'>abc'"            'posix)
+  (t "<abc"      "'<abc'"            'posix)
+  (t "a\"b\"c"   "'a\"b\"c'"         'posix)
+  (t "a(b)c"     "'a(b)c'"           'posix)
+  (t "a{b}c"     "'a{b}c'"           'posix)
+  (t "a'c"       "'a'\"'\"'c'"       'posix)
+  )
+
+(let ()
+  (define (t input expected flavor)
+    (test* #"shell-tokenize-string(~flavor) - ~input" expected
+           (shell-tokenize-string input flavor)))
+
+  (t ""                    '()                         'posix)
+  (t "   "                 '()                         'posix)
+  (t "abc"                 '("abc")                    'posix)
+  (t "  abc "              '("abc")                    'posix)
+  (t "ab   c d"            '("ab" "c" "d")             'posix)
+  (t "ab\\ c\\ d"          '("ab c d")                 'posix)
+  (t "ab\\'c\\\"d"         '("ab'c\"d")                'posix)
+  (t "'ab c d' ef \"g h\"" '("ab c d" "ef" "g h")      'posix)
+  (t "'ab\\\\cd' \"ab\\\\cd\"" '("ab\\\\cd" "ab\\cd")  'posix)
+  (t "'ab\\xcd' \"ab\\xcd\"" '("ab\\xcd" "ab\\xcd")    'posix)
+
+  (t "$abc"                (test-error)                'posix)
+  (t "\"$abc\""            (test-error)                'posix)
+  (t "'$abc'"              '("$abc")                   'posix)
+  (t "abc >z"              (test-error)                'posix)
+  (t "abc \">z\""          '("abc" ">z")               'posix) 
+  (t "abc \"`z`\""         (test-error)                'posix) 
+  (t "abc '`z`'"           '("abc" "`z`")              'posix)
+
+  (t ""                    '()                         'windows)
+  (t "   "                 '()                         'windows)
+  (t "abc"                 '("abc")                    'windows)
+  (t "  abc "              '("abc")                    'windows)
+  (t "ab   c d"            '("ab" "c" "d")             'windows)
+  (t "\"a b c\" d e"       '("a b c" "d" "e")          'windows)
+  (t "\"ab\\\"c\" \"\\\\\" d" '("ab\"c" "\\" "d")      'windows)
+  (t "a\\\\\\b d\"e f\"g h" '("a\\\\\\b" "de fg" "h")  'windows)
+  (t "a\\\\\\\"b c d"      '("a\\\"b" "c" "d")         'windows)
+  (t "a\\\\\\\\\"b c\" d e" '("a\\\\b c" "d" "e")      'windows)
+  (t "\""                  '("")                       'windows)
+  (t "\"\""                '("")                       'windows)
+  (t "\"\"\""              '("\"")                     'windows)
+  (t "\"\"\"\""            '("\"")                     'windows)
+  (t "\"\"\"\"\""          '("\"")                     'windows)
+  (t "\"\"\"\"\"\""        '("\"\"")                   'windows)
+  (t "\"\"\"\"\"\"\""      '("\"\"")                   'windows)
+  (t "a\"b\"\" c"          '("ab\"" "c")               'windows)
+  (t "a\"b\"\"\" c"        '("ab\" c")                 'windows)
+  (t "a\"b\"\"\"\" c"      '("ab\"" "c")               'windows)
+  )
+
 ;;-------------------------------
 (test-section "unwind-protect upon exit")
 
@@ -530,10 +611,8 @@
                   (sys-unlink "test1.o")))))
 
 (test* "unwind-protect upon exit" '(1 #f)
-       ;; we haven't tested file.util yet, so using a clumsy way.
-       (let* ([gosh (string-append (sys-dirname (current-load-path))
-                                   "/../src/gosh")]
-              [p (run-process `(,gosh "-ftest" "test.o") :wait #t)])
+       ;; assuming we're running gosh under $top_builddir/src
+       (let1 p (run-process `("./gosh" "-ftest" "test.o") :wait #t)
          (list (sys-wait-exit-status (process-exit-status p))
                (file-exists? "test1.o"))))
 
