@@ -198,8 +198,8 @@
 (test* "string-every" #t (string-every #[-] ""))
 (test* "string-every" #f (string-every #[-] "a"))
 (test* "string-every" #t (string-every #[-] ""))
-(test* "string-every" #t (string-every (lambda (x) (char-ci=? x #\)) ""))
-(test* "string-every" #f (string-every (lambda (x) (char-ci=? x #\)) ""))
+(test* "string-every" #t (string-every (^x (char-ci=? x #\)) ""))
+(test* "string-every" #f (string-every (^x (char-ci=? x #\)) ""))
 
 (test* "string-any" #t (string-any #\ ""))
 (test* "string-any" #f (string-any #\ ""))
@@ -207,8 +207,8 @@
 (test* "string-any" #t (string-any #[-] "["))
 (test* "string-any" #f (string-any #[-] "XL["))
 (test* "string-any" #f (string-any #[-] ""))
-(test* "string-any" #t (string-any (lambda (x) (char-ci=? x #\)) "炠"))
-(test* "string-any" #f (string-any (lambda (x) (char-ci=? x #\)) "A"))
+(test* "string-any" #t (string-any (^x (char-ci=? x #\)) "炠"))
+(test* "string-any" #f (string-any (^x (char-ci=? x #\)) "A"))
 (test* "string-tabulate" "ABCDE"
        (string-tabulate (lambda (code)
                           (integer->char (+ code
@@ -421,19 +421,19 @@
 
 (define (byte-list->port p bytes)
   (set! *flusher-out* '())
-  (for-each (lambda (b) (write-byte b p)) bytes)
+  (for-each (^b (write-byte b p)) bytes)
   (close-output-port p)
   *flusher-out*)
 
 (define (char-list->port p chars)
   (set! *flusher-out* '())
-  (for-each (lambda (c) (write-char c p)) chars)
+  (for-each (^c (write-char c p)) chars)
   (close-output-port p)
   *flusher-out*)
 
 (define (string-list->port p strs)
   (set! *flusher-out* '())
-  (for-each (lambda (s) (display s p)) strs)
+  (for-each (^s (display s p)) strs)
   (close-output-port p)
   *flusher-out*)
 
