@@ -3,6 +3,8 @@
 ;;
 
 (use gauche.test)
+(use gauche.parameter)
+(use compat.chibi-test)
 
 (test-start "SRFIs")
 
@@ -459,7 +461,7 @@
 (cond-expand
  ;; Only test in utf-8, for the literal notation is interpreted differently
  ;; in other encodings.
- [gauche.ces-utf-8
+ [gauche.ces.utf8
   (test* "char-set-delete!" #[\x81\x83\x84\x86]
          (char-set-delete! (->char-set '(#\x81 #\x82 #\x83 #\x84 #\x85 #\x86 #\x87))
                            #\x82 #\x87 #\x85)
@@ -483,7 +485,7 @@
 (cond-expand
  ;; Only test in utf-8, for the literal notation is interpreted differently
  ;; in other encodings.
- [gauche.ces.utf-8
+ [gauche.ces.utf8
   (test* "char-set-union!" #[\x81-\x89]
          (char-set-union! (->char-set '(#\x81 #\x83 #\x84 #\x86 #\x87))
                           (->char-set '(#\x82 #\x85 #\x86 #\x88 #\x89)))
@@ -1365,9 +1367,437 @@
 (test* "booleans->integer" 9 (booleans->integer #f #f #t #f #f #t))
 
 ;;-----------------------------------------------------------------------
-(test-section "srfi-98")
+(test-section "srfi-69")
+(use srfi-69)
+(test-module 'srfi-69)
 
+;; These test is contributed by Takashi Kato @tk_riple
+(letrec-syntax
+    ([test-error
+      (syntax-rules ()
+        ((_ expr)
+         (test* 'expr #t (guard (e (else #t)) expr #f))))]
+     [test-equal
+      (syntax-rules ()
+        ((_ expect expr)
+         (test* 'expr expect expr)))]
+     [test-assert
+      (syntax-rules ()
+        ((_ expr)
+         (test-assert 'expr expr))
+        ((_ name expr)
+         (test* name #t expr (lambda (a r) (and a r)))))]
+     [test-not
+      (syntax-rules ()
+        ((_ expr)
+         (test-not 'expr expr))
+        ((_ name expr)
+         (test-assert name (not expr))))]
+     [test-lset-eq?
+      (syntax-rules ()
+        ((test-lset= a b)
+         (test-assert 'a (lset= eq? a b))))]
+     [test-lset-equal?
+      (syntax-rules ()
+        ((test-lset-equal? a b)
+         (test-assert 'a (lset= equal? a b))))])
+
+  (let ((ht (make-hash-table eq?)))
+    ;; 3 initial elements
+
+    (test-equal 0 (hash-table-size ht))
+    (hash-table-set! ht 'cat 'black)
+    (hash-table-set! ht 'dog 'white)
+    (hash-table-set! ht 'elephant 'pink)
+
+    (test-equal 3 (hash-table-size ht))
+    (test-assert (hash-table-exists? ht 'dog))
+    (test-assert (hash-table-exists? ht 'cat))
+    (test-assert (hash-table-exists? ht 'elephant))
+    (test-not (hash-table-exists? ht 'goose))
+    (test-equal 'white (hash-table-ref ht 'dog))
+    (test-equal 'black (hash-table-ref ht 'cat))
+    (test-equal 'pink (hash-table-ref ht 'elephant))
+    (test-error (hash-table-ref ht 'goose))
+    (test-equal 'grey (hash-table-ref ht 'goose (lambda () 'grey)))
+    (test-equal 'grey (hash-table-ref/default ht 'goose 'grey))
+    (test-lset-eq? '(cat dog elephant) (hash-table-keys ht))
+    (test-lset-eq? '(black white pink) (hash-table-values ht))
+    (test-lset-equal? '((cat . black) (dog . white) (elephant . pink))
+                      (hash-table->alist ht))
+
+    ;; remove an element
+    (hash-table-delete! ht 'dog)
+    (test-equal 2 (hash-table-size ht))
+    (test-not (hash-table-exists? ht 'dog))
+    (test-assert (hash-table-exists? ht 'cat))
+    (test-assert (hash-table-exists? ht 'elephant))
+    (test-error (hash-table-ref ht 'dog))
+    (test-equal 'black (hash-table-ref ht 'cat))
+    (test-equal 'pink (hash-table-ref ht 'elephant))
+    (test-lset-eq? '(cat elephant) (hash-table-keys ht))
+    (test-lset-eq? '(black pink) (hash-table-values ht))
+    (test-lset-equal? '((cat . black) (elephant . pink)) (hash-table->alist ht))
+
+    ;; remove a non-existing element
+    (hash-table-delete! ht 'dog)
+    (test-equal 2 (hash-table-size ht))
+    (test-not (hash-table-exists? ht 'dog))
+
+    ;; overwrite an existing element
+    (hash-table-set! ht 'cat 'calico)
+    (test-equal 2 (hash-table-size ht))
+    (test-not (hash-table-exists? ht 'dog))
+    (test-assert (hash-table-exists? ht 'cat))
+    (test-assert (hash-table-exists? ht 'elephant))
+    (test-error (hash-table-ref ht 'dog))
+    (test-equal 'calico (hash-table-ref ht 'cat))
+    (test-equal 'pink (hash-table-ref ht 'elephant))
+    (test-lset-eq? '(cat elephant) (hash-table-keys ht))
+    (test-lset-eq? '(calico pink) (hash-table-values ht))
+    (test-lset-equal? '((cat . calico) (elephant . pink)) (hash-table->alist ht))
+
+    ;; walk and fold
+    (test-lset-equal?
+     '((cat . calico) (elephant . pink))
+     (let ((a '()))
+       (hash-table-walk ht (lambda (k v) (set! a (cons (cons k v) a))))
+       a))
+    (test-lset-equal? '((cat . calico) (elephant . pink))
+                      (hash-table-fold ht (lambda (k v a) (cons (cons k v) a)) '()))
+
+    ;; copy
+    (let ((ht2 (hash-table-copy ht)))
+      (test-equal 2 (hash-table-size ht2))
+      (test-not (hash-table-exists? ht2 'dog))
+      (test-assert (hash-table-exists? ht2 'cat))
+      (test-assert (hash-table-exists? ht2 'elephant))
+      (test-error (hash-table-ref ht2 'dog))
+      (test-equal 'calico (hash-table-ref ht2 'cat))
+      (test-equal 'pink (hash-table-ref ht2 'elephant))
+      (test-lset-eq? '(cat elephant) (hash-table-keys ht2))
+      (test-lset-eq? '(calico pink) (hash-table-values ht2))
+      (test-lset-equal? '((cat . calico) (elephant . pink))
+                        (hash-table->alist ht2)))
+
+    ;; merge
+    (let ((ht2 (make-hash-table eq?)))
+      (hash-table-set! ht2 'bear 'brown)
+      (test-equal 1 (hash-table-size ht2))
+      (test-not (hash-table-exists? ht2 'dog))
+      (test-assert (hash-table-exists? ht2 'bear))
+      (hash-table-merge! ht2 ht)
+      (test-equal 3 (hash-table-size ht2))
+      (test-assert (hash-table-exists? ht2 'bear))
+      (test-assert (hash-table-exists? ht2 'cat))
+      (test-assert (hash-table-exists? ht2 'elephant))
+      (test-not (hash-table-exists? ht2 'goose))
+      (test-equal 'brown (hash-table-ref ht2 'bear))
+      (test-equal 'calico (hash-table-ref ht2 'cat))
+      (test-equal 'pink (hash-table-ref ht2 'elephant))
+      (test-error (hash-table-ref ht2 'goose))
+      (test-equal 'grey (hash-table-ref/default ht2 'goose 'grey))
+      (test-lset-eq? '(bear cat elephant) (hash-table-keys ht2))
+      (test-lset-eq? '(brown calico pink) (hash-table-values ht2))
+      (test-lset-equal? '((cat . calico) (bear . brown) (elephant . pink))
+                        (hash-table->alist ht2)))
+
+    ;; alist->hash-table
+    (test-lset-equal? (hash-table->alist ht)
+                      (hash-table->alist
+                       (alist->hash-table
+                        '((cat . calico) (elephant . pink)))))
+    )
+
+  ;; update
+  (let ((ht (make-hash-table eq?))
+        (add1 (lambda (x) (+ x 1))))
+    (hash-table-set! ht 'sheep 0)
+    (hash-table-update! ht 'sheep add1)
+    (hash-table-update! ht 'sheep add1)
+    (test-equal 2 (hash-table-ref ht 'sheep))
+    (hash-table-update!/default ht 'crows add1 0)
+    (hash-table-update!/default ht 'crows add1 0)
+    (hash-table-update!/default ht 'crows add1 0)
+    (test-equal 3 (hash-table-ref ht 'crows)))
+
+  ;; string keys
+  (let ((ht (make-hash-table equal?)))
+    (hash-table-set! ht "cat" 'black)
+    (hash-table-set! ht "dog" 'white)
+    (hash-table-set! ht "elephant" 'pink)
+    (hash-table-ref/default ht "dog" #f)
+    (test-equal 'white (hash-table-ref ht "dog"))
+    (test-equal 'black (hash-table-ref ht "cat"))
+    (test-equal 'pink (hash-table-ref ht "elephant"))
+    (test-error (hash-table-ref ht "goose"))
+    (test-equal 'grey (hash-table-ref/default ht "goose" 'grey))
+    (test-lset-equal? '("cat" "dog" "elephant") (hash-table-keys ht))
+    (test-lset-equal? '(black white pink) (hash-table-values ht))
+    (test-lset-equal?
+     '(("cat" . black) ("dog" . white) ("elephant" . pink))
+     (hash-table->alist ht)))
+
+  ;; string-ci keys
+  (let ((ht (make-hash-table string-ci=? string-ci-hash)))
+    (hash-table-set! ht "cat" 'black)
+    (hash-table-set! ht "dog" 'white)
+    (hash-table-set! ht "elephant" 'pink)
+    (hash-table-ref/default ht "DOG" #f)
+    (test-equal 'white (hash-table-ref ht "DOG"))
+    (test-equal 'black (hash-table-ref ht "Cat"))
+    (test-equal 'pink (hash-table-ref ht "eLePhAnT"))
+    (test-error (hash-table-ref ht "goose"))
+    (test-lset-equal? '("cat" "dog" "elephant") (hash-table-keys ht))
+    (test-lset-equal? '(black white pink) (hash-table-values ht))
+    (test-lset-equal?
+     '(("cat" . black) ("dog" . white) ("elephant" . pink))
+     (hash-table->alist ht)))
+
+  ;; stress test
+  (test-equal 625
+              (let ((ht (make-hash-table)))
+                (do ((i 0 (+ i 1))) ((= i 1000))
+                  (hash-table-set! ht i (* i i)))
+                (hash-table-ref/default ht 25 #f)))
+  )
+
+;;-----------------------------------------------------------------------
+(test-section "srfi-98")
 (use srfi-98)
 (test-module 'srfi-98)
 
+;;-----------------------------------------------------------------------
+(test-section "srfi-111")
+(use srfi-111)
+(test-module 'srfi-111)
+
+;; srfi-111 is built-in.
+(test* "box primitives"
+       '(#t #f 2 3)
+       (let1 b (box 2)
+         (list (box? b) (box? 2) (unbox b)
+               (begin(set-box! b 3)
+                     (unbox b)))))
+
+(test* "box compare"
+       '(#t #f #f #t #f)
+       (let ([b1 (box 2)]
+             [b2 (box 2)]
+             [b3 (box 3)])
+         (list (equal? b1 b2)
+               (equal? b2 b3)
+               (equal? b1 2)
+               (eqv?   b1 b1)
+               (eqv?   b1 b2))))
+
+;;-----------------------------------------------------------------------
+;; srfi-113 depends on srfi-114, so we test this first.
+(test-section "srfi-114")
+(use srfi-114)
+(test-module 'srfi-114)
+
+;; builtin comparators are tested in test/compare.scm
+(let ()
+  (define (test-cmp msg cmpr data) ; data = ((a b result) ...)
+    (test* msg data
+           (map (^x (list (car x) (cadr x)
+                          (comparator-compare cmpr (car x) (cadr x))))
+                data)))
+
+  (test-cmp "inexact1"
+            (make-inexact-real-comparator 0.25 'round 'min)
+            '((1.0 1.1 0)
+              (1.1 1.2 -1)
+              (1.2 1.1 1)
+              (1.1 1.0 0)
+              (1.2 1.3 0)
+              (+nan.0 1.1 -1)
+              (1.1 +nan.0 1)
+              (+nan.0 +nan.0 0)))
+
+  (test-cmp "inexact2"
+            (make-inexact-real-comparator 0.25 'round 'max)
+            '((+nan.0 1.1 1)
+              (1.1 +nan.0 -1)
+              (+nan.0 +nan.0 0)))
+
+  (test-cmp "inexact3 - nan handling"
+            ($ make-inexact-real-comparator 0.25 'round
+               (^x (comparator-compare default-comparator 0 x)))
+            '((+nan.0 1.1 -1)
+              (+nan.0 0 0)
+              (+nan.0 -1.1 1)
+              (+nan.0 +nan.0 0)))
+
+  (test-cmp "list" (make-list-comparator
+                    (make-inexact-real-comparator 0.25 'round 'error))
+            '(((1 2 3) (1 2 3) 0)
+              ((1 2 3) (1.1 2 3) 0)
+              ((1 2 3) (1.1 2) 1)
+              ((1 2 3) (1.2 2) -1)
+              ((1 2 3) (1.2 2 3) -1)
+              (() () 0)
+              (() (1) -1)
+              ((1 2 3) (1 2 2.9) 0)
+              ((1 2 3) (1 2.2 2.9) -1)
+              ((1 2.2 2.9) (1 2 3) 1)
+              ))
+            
+  (test-cmp "vector" (make-vector-comparator
+                      (make-inexact-real-comparator 0.25 'round 'error))
+            '((#(1 2 3) #(1 2 3) 0)
+              (#(1 2 3) #(1.1 2 3) 0)
+              (#(1 2 3) #(1.1 2) 1)
+              (#(1 2 3) #(1.2 2) 1)
+              (#(1 2 3) #(1.2 2 3) -1)
+              (#() #() 0)
+              (#() #(1) -1)
+              (#(1 2 3) #(1 2 2.9) 0)
+              (#(1 2 3) #(1 2.2 2.9) -1)
+              (#(1 2.2 2.9) #(1 2 3) 1)
+              ))
+            
+  )
+
+(let ()
+  (define ca (make-car-comparator string-comparator))
+  (define cd (make-cdr-comparator integer-comparator))
+  (define ta ((with-module gauche make-hash-table) ca))
+  (define td ((with-module gauche make-hash-table) cd))
+  (define sa '("orange" "apple" "mango" "papaya"))
+  (define sd '(29 34 98 1001))
+  (define data (append-map (^a (map (^d (cons a d)) sd)) sa))
+
+  (test* "make-car-comparator" (sort sa)
+         (begin (dolist [x data]
+                  (hash-table-put! ta x #t))
+                (sort (map car (hash-table-keys ta)))))
+
+  (test* "make-cdr-comparator" (sort sd)
+         (begin (dolist [x data]
+                  (hash-table-put! td x #t))
+                (sort (map cdr (hash-table-keys td)))))
+  )
+
+(let ([lw (make-listwise-comparator vector?
+                                    (make-reverse-comparator number-comparator)
+                                    (^x (= (vector-length x) 0))
+                                    (^x (vector-ref x 0))
+                                    (^x (vector-copy x 1)))])
+  (define (test-lw a b expect)
+    (test* (format "listwise ~s ~s" a b)
+           (list (zero? expect) expect (- expect) #t)
+           (list (comparator-equal? lw a b)
+                 (comparator-compare lw a b)
+                 (comparator-compare lw b a)
+                 (integer? (comparator-hash lw a)))))
+  (test-lw '#(1) '#(2) 1)
+  (test-lw '#(1 4) '#(1 3 5) -1)
+  (test-lw '#(1 2) '#(1 3 5) 1)
+  (test-lw '#(1 3 5) '#(1 3 5) 0)
+  (test-lw '#(1 3 6) '#(1 3 5) -1)
+  (test-lw '#() '#() 0)
+  (test-lw '#() '#(1) -1))
+
+(let ([vw (make-vectorwise-comparator list?
+                                      (make-reverse-comparator number-comparator)
+                                      length
+                                      list-ref)])
+  (define (test-vw a b expect)
+    (test* (format "vectorwise ~s ~s" a b)
+           (list (zero? expect) expect (- expect) #t)
+           (list (comparator-equal? vw a b)
+                 (comparator-compare vw a b)
+                 (comparator-compare vw b a)
+                 (integer? (comparator-hash vw a)))))
+  (test-vw '(1) '(2) 1)
+  (test-vw '(1 4) '(1 3 5) -1)
+  (test-vw '(1 2) '(1 3 5) -1)
+  (test-vw '(1 3 5) '(1 3 5) 0)
+  (test-vw '(1 3 6) '(1 3 5) -1)
+  (test-vw '() '() 0)
+  (test-vw '() '(1) -1))
+
+(let ()
+  (define (test-minmax msg cmp min-item max-item args)
+    (test* (format "test-minmax ~a" msg)
+           (list min-item max-item max-item min-item)
+           (list (apply comparator-min cmp args)
+                 (apply comparator-max cmp args)
+                 (apply comparator-min (make-reverse-comparator cmp) args)
+                 (apply comparator-max (make-reverse-comparator cmp) args))))
+
+  (test-minmax "default" default-comparator 1 9
+               '(3 1 4 5 9 2 6 8 7))
+  )
+
+;;-----------------------------------------------------------------------
+(test-section "srfi-113")
+(use srfi-113)
+(test-module 'srfi-113)
+
+(chibi-test
+ (include "include/sets-test"))
+
+;; collection framework test
+;; (we use srfi-1 with prefix not to mess gauche.collection#fold etc.)
+(use srfi-1 :prefix srfi-1:)
+(test* "coerce-to" '(a b)
+       (coerce-to <list> (set eq-comparator 'a 'b 'a 'b))
+       (cut srfi-1:lset= eq? <> <>))
+(test* "coerce-to" '(a a b b)
+       (coerce-to <list> (bag eq-comparator 'a 'b 'a 'b))
+       (cut srfi-1:lset= eq? <> <>))
+
+;;-----------------------------------------------------------------------
+(test-section "srfi-117")
+(use srfi-117)
+(test-module 'srfi-117)
+
+(chibi-test
+ (include "include/list-queues-test"))
+
+;;-----------------------------------------------------------------------
+(test-section "srfi-118")
+(use srfi-118)
+(test-module 'srfi-118)
+
+(test* "string-append!" "abcdef"
+       (rlet1 s (string-copy "ab")
+         (string-append! s "cd" #\e "f")))
+
+(test* "string-replace!" "abcXYZfgh"
+       (rlet1 s (string-copy "abcdefgh")
+         (string-replace! s 3 5 "XYZ")))
+
+(test* "string-replace!" "abcXYZfgh"
+       (rlet1 s (string-copy "abcdefgh")
+         (string-replace! s 3 5 "WXYZw" 1 4)))
+
+;;-----------------------------------------------------------------------
+(test-section "srfi-121")
+(use srfi-121) ; this is in fact gauche.generator
+(use srfi-11)
+(use scheme.base :prefix r7rs:)
+(test-module 'srfi-121)
+
+(let ([string-for-each r7rs:string-for-each])
+  (chibi-test
+   (include "include/generators-test")))
+
+(test* "generator->vector!" '(3 #(1 2 3) 4 0)
+       (let ((g (list->generator '(1 2 3 4 5)))
+             (v (make-vector 3)))
+         (list (generator->vector! v 0 g)
+               v
+               (g)
+               (generator->vector! v 3 g))))
+
 (test-end)
+
+;;-----------------------------------------------------------------------
+(test-section "srfi-128")
+(use srfi-128)
+(test-module 'srfi-128)
+
