@@ -44,11 +44,25 @@
 (test* "symbol reader uninterned" "foo"
        (let1 s (read-from-string "#:foo")
          (and (symbol? s) (not (eq? 'foo s)) (symbol->string s))))
+(test* "symbol reader uninterned" "+"
+       (let1 s (read-from-string "#:+")
+         (and (symbol? s) (not (eq? '+ s)) (symbol->string s))))
+(test* "symbol reader uninterned" "-a"
+       (let1 s (read-from-string "#:-a")
+         (and (symbol? s) (not (eq? '-a s)) (symbol->string s))))
 (test* "symbol reader uninterned" "foo bar"
        (let1 s (read-from-string "#:|foo bar|")
          (and (symbol? s) (not (eq? '|foo bar| s)) (symbol->string s))))
 (test* "symbol reader uninterned" #f
        (eq? (read-from-string "#:foo") (read-from-string "#:foo")))
+(test* "symbol reader uninterned invalid" (test-error)
+       (read-from-string "#:"))
+(test* "symbol reader uninterned invalid" (test-error)
+       (read-from-string "#: yoo"))
+(test* "symbol reader uninterned invalid" (test-error)
+       (read-from-string "#:1234"))
+(test* "symbol reader uninterned invalid" (test-error)
+       (read-from-string "#:+inf.0"))
 
 (test* "symbol writer uninterned" "#:foo"
        (write-to-string (string->uninterned-symbol "foo")))
@@ -66,6 +80,14 @@
 (test* "prefix" 'bar (symbol-sans-prefix 'foo:bar 'foo:))
 (test* "prefix" #f   (symbol-sans-prefix 'foo:bar 'bar:))
 
+(test* "symbol-append" 'ab:c45 (symbol-append 'ab ':c 45))
+(test* "symbol-append" 'quux (symbol-append #t 'qu "ux"))
+(test* "symbol-append" #t
+       (let1 x (symbol-append #f 'qu "ux")
+         (and (not (symbol-interned? x))
+              (equal? (symbol->string x) "quux"))))
+(test* "symbol-append" '|| (symbol-append))
+(test* "symbol-append" '|| (symbol-append #t))
 
 ;;----------------------------------------------------------------
 (test-section "keywords")
@@ -80,7 +102,9 @@
 (test* "keyword->string" "abc" (keyword->string (make-keyword 'abc)))
 (test* "keyword->string" "a b" (keyword->string (read-from-string ":|a b|")))
 (test* "writer" ":abc" (write-to-string (make-keyword "abc")))
-(test* "writer" ":|a b c|" (write-to-string (make-keyword "a b c")))
+(test* "writer"
+       (if (symbol? :x) "|:a b c|" ":|a b c|");transient during symbol-keyword integration
+       (write-to-string (make-keyword "a b c")))
 (test* "writer" ":" (write-to-string (make-keyword "")))
 (test* "writer" "::" (write-to-string (make-keyword ":")))
 (test* "writer" ":3" (write-to-string (make-keyword "3")))
