@@ -374,6 +374,29 @@
   ]
  [else])
 
+;; sys-mkstemp/sys-mkdtemp
+(let ()
+  ;; Check if multiple temporary files or directories are created.
+  (define (test-mkxtemp name creator tester remover)
+    (test* name '(#t #t)
+       (let1 temps (map creator (iota 3))
+         (begin0 (list (every tester temps)
+                       (let1 sorted (sort temps)
+                         (every (complement equal?) sorted (cdr sorted))))
+           (for-each remover temps)))))
+
+  (test-mkxtemp "sys-mkstemp"
+                (^_ (receive (p n) (sys-mkstemp "test.o")
+                      (close-output-port p)
+                      n))
+                file-is-regular?
+                sys-unlink)
+
+  (test-mkxtemp "sys-mkdtemp"
+                (^_ (sys-mkdtemp "test.dir"))
+                file-is-directory?
+                sys-rmdir))
+
 ;;-------------------------------------------------------------------
 (test-section "time")
 
@@ -484,27 +507,20 @@
                (close-input-port in) (close-output-port out)
                (list f1 f2 f3))))))
 
-;; Kludge: MinGW32 seems not to support :none, :line buffering,
-;; so we flush and close the output pipe before reading from it.
-
 (test* "pipe and read-block(none)" 2
        (receive (in out) (sys-pipe :buffering :none)
          (display "ab" out)
-         (cond-expand (gauche.os.windows (close-output-port out)) (else))
          (let1 r (string-size (read-block 1000 in))
            (close-input-port in)
-           (cond-expand ((not gauche.os.windows) (close-output-port out))
-                        (else))
+           (close-output-port out)
            r)))
 
 (test* "pipe and read-block(line)" 2
        (receive (in out) (sys-pipe :buffering :line)
          (display "a\n" out)
-         (cond-expand (gauche.os.windows (close-output-port out)) (else))
          (let1 r (string-size (read-block 1000 in))
            (close-input-port in)
-           (cond-expand ((not gauche.os.windows) (close-output-port out))
-                        (else))
+           (close-output-port out)
            r)))
 
 ;;-------------------------------------------------------------------
@@ -618,28 +634,28 @@
       (lambda ()
         (write '(define (main args)
                   (let1 expected-ppid (x->integer (cadr args))
-                    (let loop ()
+                    (let loop ([c 0])
                       (sys-nanosleep #e2e8)
-                      (if (eqv? (sys-getppid) expected-ppid)
-                        (write (list (sys-getpid) (sys-getpgrp)))
-                        (loop))))))))
-    (receive (in out) (sys-pipe :buffering :none)
-      (define (run-and-read ppid detached)
-        (let1 pid (sys-fork-and-exec "./gosh"
-                                     `("./gosh" "-ftest" "./test.out"
-                                       ,(x->string ppid))
-                                     :iomap `((1 . ,out))
-                                     :detached detached)
-          (begin0 (read in) (sys-waitpid pid))))
-      (test* "fork, exec and detached process (not detached)"
-             #f
-             (let1 r (run-and-read (sys-getpid) #f)
-               (= (car r) (cadr r))))
-      (test* "fork, exec and detached process (detached)"
-             #t
-             (let1 r (run-and-read 1 #t)
-               (= (car r) (cadr r))))
-      ))
+                      (cond [(or (eqv? (sys-getppid) expected-ppid)
+                                 (>= c 50)) ; takes about 10s
+                             (write (list (sys-getpid) (sys-getpgrp)))]
+                            [else (loop (+ c 1))])))))))
+    (define (run-and-read-test msg ppid detached)
+      (test* msg detached
+             (receive (in out) (sys-pipe :buffering :none)
+               (let1 pid (sys-fork-and-exec "./gosh"
+                                            `("./gosh" "-ftest" "./test.out"
+                                              ,(x->string ppid))
+                                            :iomap `((1 . ,out))
+                                            :detached detached)
+                 (close-port out)
+                 (let1 result (read in)
+                   (sys-waitpid pid)
+                   (eqv? (car result) (cadr result)))))))
+    (run-and-read-test "fork, exec and detached process (not detached)" 
+                       (sys-getpid) #f)
+    (run-and-read-test "fork, exec and detached process (detached)"
+                       1 #t))
   (cmd-rmrf "test.out")
 
   ;; Testing GC in forked process---we don't explicitly spawn a thread here,
