@@ -5,6 +5,11 @@
 (use gauche.test)
 (test-start "text utilities")
 
+;;-------------------------------------------------------------------
+(test-section "console")
+(use text.console)
+(test-module 'text.console)
+
 ;;-------------------------------------------------------------------
 (test-section "csv")
 (use text.csv)
@@ -68,6 +73,68 @@
          (lambda (out)
            ((make-csv-writer #\,) out '()))))
 
+;; middle-level API
+
+(let ([data '(("" "" "" "" "" "" "" "" "")
+              ("Exported data" "" "" "" "" "" "" "" "")
+              ("" "" "" "" "" "" "" "" "")
+              ("" "" "Year" "Country" "" "Population" "GDP" "" "Note")
+              ("" "" "1958" "Land of Lisp" "" "39994" "551,435,453" "" "")
+              ("" "" "1957" "United States of Formula Translators" "" "115333"
+               "4,343,225,434" "" "Estimated")
+              ("" "" "1959" "People's Republic of COBOL" ""
+               "82524" "3,357,551,143" "" "")
+              ("" "" "1970" "Kingdom of Pascal" "" "3785" "" "" "GDP missing")
+              ("" "" "" "" "" "" "" "" "")
+              ("" "" "1962" "APL Republic" "" "1545" "342,335,151" "" ""))]
+      [header-slots1  '("Country" "Year" "GDP" "Population")]
+      [header-slots2 '(#/country/i #/year/i #/gdp/i #/popu/i)])
+  (test* "make-csv-header-parser (strings)" '#(3 2 6 5)
+         (any (make-csv-header-parser header-slots1) data))
+
+  (test* "make-csv-header-parser (regexps)" '#(3 2 6 5)
+         (any (make-csv-header-parser header-slots2) data))
+  
+  (test* "make-csv-record-parser (strings)"
+         '(("Land of Lisp" "1958" "551,435,453" "39994")
+           ("United States of Formula Translators" "1957" "4,343,225,434"
+            "115333")
+           ("People's Republic of COBOL" "1959" "3,357,551,143" "82524")
+           ("APL Republic" "1962" "342,335,151" "1545"))
+         (filter-map (make-csv-record-parser header-slots1 '#(3 2 6 5)
+                                             '(("Year" #/^\d+$/)
+                                               "Country" "Population" "GDP"))
+                     data))
+
+  (test* "make-csv-record-parser (regexps)"
+         '(("Land of Lisp" "1958" "551,435,453" "39994")
+           ("United States of Formula Translators" "1957" "4,343,225,434"
+            "115333")
+           ("People's Republic of COBOL" "1959" "3,357,551,143" "82524")
+           ("APL Republic" "1962" "342,335,151" "1545"))
+         (filter-map (make-csv-record-parser header-slots2 '#(3 2 6 5)
+                                             '((#/year/i #/^\d+$/)
+                                               #/country/i #/popu/i #/gdp/i))
+                     data))
+  
+  (test* "csv-rows->tuples (allow-gap? #f)"
+         '(("Land of Lisp" "1958" "551,435,453" "39994")
+           ("United States of Formula Translators" "1957" "4,343,225,434"
+            "115333")
+           ("People's Republic of COBOL" "1959" "3,357,551,143" "82524")
+           ("Kingdom of Pascal" "1970" "" "3785"))
+         (csv-rows->tuples data header-slots1))
+
+  (test* "csv-rows->tuples (allow-gap? #t)"
+         '(("Land of Lisp" "1958" "551,435,453" "39994")
+           ("United States of Formula Translators" "1957" "4,343,225,434"
+            "115333")
+           ("People's Republic of COBOL" "1959" "3,357,551,143" "82524")
+           ("Kingdom of Pascal" "1970" "" "3785")
+           ("APL Republic" "1962" "342,335,151" "1545"))
+         (csv-rows->tuples data header-slots1 :allow-gap? #t))
+  )
+
 ;;-------------------------------------------------------------------
 (test-section "diff")
 (use text.diff)
@@ -102,6 +169,110 @@ fuga
        (with-output-to-string
          (lambda () (diff-report diff-a diff-b))))
 
+;;-------------------------------------------------------------------
+(test-section "gap-buffer")
+(use text.gap-buffer)
+(use gauche.uvector)
+
+(test-module 'text.gap-buffer)
+
+;; stringify gap-buffer for test purpose
+(define (gap-buffer-visualize gbuf)
+  (with-output-to-string
+    (^[]
+      (let1 buf (~ gbuf'buffer) ;accessing internal
+        (dotimes [i (gap-buffer-gap-start gbuf)]
+          (display (integer->char (~ buf i))))
+        (dotimes [i (- (gap-buffer-gap-end gbuf) (gap-buffer-gap-start gbuf))]
+          (display #\_))
+        (dotimes [i (- (gap-buffer-capacity gbuf) (gap-buffer-gap-end gbuf))]
+          (display (integer->char (~ buf (+ i (gap-buffer-gap-end gbuf))))))))))
+
+(test* "constuct" "abcde___"
+       (gap-buffer-visualize (string->gap-buffer "abcde")))
+(test* "constuct" "___abcde"
+       (gap-buffer-visualize (string->gap-buffer "abcde" 0 'beginning)))
+(test* "constuct" "bcd_"
+       (gap-buffer-visualize (string->gap-buffer "abcde" 0 'end 1 4)))
+  
+(let1 gbuf (string->gap-buffer "abcde")
+  (test* "move" "abcd___e"
+         (gap-buffer-visualize (gap-buffer-move! gbuf -1 'current)))
+  (test* "move" "a___bcde"
+         (gap-buffer-visualize (gap-buffer-move! gbuf 1)))
+  (test* "move" "abc___de"
+         (gap-buffer-visualize (gap-buffer-move! gbuf -2 'end)))
+  (test* "insert" "abcZ__de"
+         (gap-buffer-visualize (gap-buffer-insert! gbuf #\Z)))
+  (test* "insert" "abcZxyde"
+         (gap-buffer-visualize (gap-buffer-insert! gbuf "xy")))
+  (test* "insert" "abcZxyw_______de"
+         (gap-buffer-visualize (gap-buffer-insert! gbuf #\w)))
+  (test* "insert"
+         "abcZxyw012345678901234567890123456789_________________________de"
+         (gap-buffer-visualize
+          (gap-buffer-insert! gbuf "012345678901234567890123456789")))
+  (test* "delete"
+         "abcZxyw012345678901234567890123456789__________________________e"
+         (gap-buffer-visualize
+          (gap-buffer-delete! gbuf 1)))
+  (test* "delete"
+         "abc____________________________________678901234567890123456789e"
+         (begin (gap-buffer-move! gbuf 3)
+                (gap-buffer-visualize
+                 (gap-buffer-delete! gbuf 10))))
+  (test* "delete"
+         (test-error)
+         (gap-buffer-visualize (gap-buffer-delete! gbuf 100)))
+  (test* "change"
+         "abcdef______________________________________1234567890123456789e"
+         (gap-buffer-visualize (gap-buffer-change! gbuf 5 "def")))
+  )
+
+(let1 gbuf (string->gap-buffer "abcde")
+  (test* "->string (gap at end)" "abcde"
+         (gap-buffer->string gbuf))
+  (test* "->string (gap at beginning)" "abcde"
+         (begin (gap-buffer-move! gbuf 0)
+                (gap-buffer->string gbuf)))
+  (test* "->string (gap at middle)" "abcde"
+         (begin (gap-buffer-move! gbuf 3 'current)
+                (gap-buffer->string gbuf)))
+  )
+
+(let1 gbuf (gap-buffer-copy (string->gap-buffer "abcde"))
+  (test* "undo and redo"
+         '("abXYZcde" "abXYde" "aQRSYde" "abXYde" "abXYZcde"
+           "abcde" "aQRSYde" "abcde")
+         (let* ([u0 (begin (gap-buffer-move! gbuf 2)
+                           (gap-buffer-edit! gbuf '(i #f "XYZ")))]
+                [s0 (gap-buffer->string gbuf)] ; abXYZcde
+                [u1 (begin (gap-buffer-move! gbuf -1 'current)
+                           (gap-buffer-edit! gbuf '(d #f 2)))]
+                [s1 (gap-buffer->string gbuf)] ; abXYde
+                [u2 (begin (gap-buffer-move! gbuf -3 'current)
+                           (gap-buffer-edit! gbuf '(c #f 2 "QRS")))]
+                [s2 (gap-buffer->string gbuf)] ; aQRSYde
+                [r2 (begin (gap-buffer-move! gbuf 0)
+                           (gap-buffer-edit! gbuf u2))] ; undo change!
+                [s3 (gap-buffer->string gbuf)] ; abXYde
+                [r1 (gap-buffer-edit! gbuf u1)] ; undo delete!
+                [s4 (gap-buffer->string gbuf)] ; abXYZcde
+                [r0 (begin (gap-buffer-move! gbuf -1 'end)
+                           (gap-buffer-edit! gbuf u0))] ; undo insert!
+                [s5 (gap-buffer->string gbuf)] ; abcde
+                [u0 (gap-buffer-edit! gbuf r0)] ; redo insert!
+                [u1 (gap-buffer-edit! gbuf r1)] ; redo delete!
+                [u2 (gap-buffer-edit! gbuf r2)] ; redo change!
+                [s6 (gap-buffer->string gbuf)] ; aQRSYde
+                [s7 (begin (gap-buffer-edit! gbuf u2)
+                           (gap-buffer-edit! gbuf u1)
+                           (gap-buffer-edit! gbuf u0)
+                           (gap-buffer->string gbuf))] ; abcde
+                )
+           (list s0 s1 s2 s3 s4 s5 s6 s7)))
+  )
+
 ;;-------------------------------------------------------------------
 (test-section "html-lite")
 (use text.html-lite)
@@ -401,6 +572,53 @@ fuga
                                      (parameter 2))
        (sql-tokenize "?,?,:foo, :bar , ?"))
 
+;;-------------------------------------------------------------------
+(test-section "line-edit")
+(use text.line-edit)
+(test-module 'text.line-edit)
+
+;; Test internal functions
+(let ()
+  (define (t s pos expect)
+    (test* #"buffer-find-matching-paren: ~s ~pos" expect
+           ((with-module text.line-edit buffer-find-matching-paren)
+            ((with-module text.gap-buffer string->gap-buffer) s)
+            0 pos)))
+
+  (t "(abc)" 4 0)
+  (t "aa(abc)zz" 6 2)
+  (t "(a(abc)z)" 6 2)
+  (t "(a)(abc)z" 7 3)
+  (t "(a\\(b)z" 5 0)
+  (t "(a\\(\\))z" 6 0)
+  (t "(a\\)\\()z" 6 0)
+  (t "(a(abc)(def)z)" 11 7)
+  (t "(a(abc)(def)z)" 13 0)
+  (t "aaaabc)zz" 6 #f)
+  (t "(a{b}c)" 4 2)
+  (t "(a{b}c)" 6 0)
+  (t "(a{b(}c)" 5 #f)
+  (t "(a{b(}c)" 7 4)
+  (t "(a{b)}c)" 4 #f)
+  (t "(a{b)}c)" 5 2)
+  (t "(a\"(\"b)" 6 0)
+  (t "(a\")\"b)" 6 0)
+  (t "(a\"\\\"(\"b)" 8 0)
+  (t "(a#/bc([d\\//)]" 12 0)
+  (t "(a#/bc([d\\//)]" 13 #f)
+  (t "(a#(b)c)" 5 3)
+  (t "(a#(b)c)" 7 0)
+  (t "(a#[b[:space:]]c)" 14 3)
+  (t "(a#[b[:space:]]c)" 13 5)
+  (t "(a#[b[:space:]]c)" 16 0)
+  (t "(a#{b}c)" 5 3)
+  (t "(a#{b)c}" 5 #f)
+  (t "(a#{b)c}" 7 3)
+  (t "(a\\#(b)c)" 6 4)
+  (t "(a#\\(b)c)" 6 0)
+  )
+
+
 ;;-------------------------------------------------------------------
 (test-section "tree")
 (use text.tree)
@@ -412,7 +630,14 @@ fuga
 (test* "tree->string" "ab" (tree->string 'ab))
 (test* "tree->string" "ab" (tree->string '(a . b)))
 (test* "tree->string" "ab" (tree->string '(a b)))
-(test* "tree->string" "Ab" (tree->string '(|A| . :b)))
 (test* "tree->string" "ab" (tree->string '((((() ())) . a) ((((b)))))))
+(test* "tree->string"
+       (if (symbol? :b) "A:b" "Ab") ; transient during symbol-keyword integration
+       (tree->string '(|A| . :b)))
+
+;;-------------------------------------------------------------------
+(test-section "unicode.ucd")
+(use text.unicode.ucd)
+(test-module 'text.unicode.ucd)
 
 (test-end)
