@@ -206,8 +206,8 @@
 (test "string-every" #t (lambda () (string-every #[あ-ん] "ああいあ")))
 (test "string-every" #f (lambda () (string-every #[あ-ん] "ああaあ")))
 (test "string-every" #t (lambda () (string-every #[あ-ん] "")))
-(test "string-every" #t (lambda () (string-every (lambda (x) (char-ci=? x #\あ)) "ああああ")))
-(test "string-every" #f (lambda () (string-every (lambda (x) (char-ci=? x #\あ)) "あいあい")))
+(test "string-every" #t (lambda () (string-every (^x (char-ci=? x #\あ)) "ああああ")))
+(test "string-every" #f (lambda () (string-every (^x (char-ci=? x #\あ)) "あいあい")))
 
 (test "string-any" #t (lambda () (string-any #\あ "ああああ")))
 (test "string-any" #f (lambda () (string-any #\あ "いうえお")))
@@ -215,8 +215,8 @@
 (test "string-any" #t (lambda () (string-any #[あ-ん] "すきーむ")))
 (test "string-any" #f (lambda () (string-any #[あ-ん] "スキーム")))
 (test "string-any" #f (lambda () (string-any #[あ-ん] "")))
-(test "string-any" #t (lambda () (string-any (lambda (x) (char-ci=? x #\あ)) "らららあ")))
-(test "string-any" #f (lambda () (string-any (lambda (x) (char-ci=? x #\あ)) "ラララア")))
+(test "string-any" #t (lambda () (string-any (^x (char-ci=? x #\あ)) "らららあ")))
+(test "string-any" #f (lambda () (string-any (^x (char-ci=? x #\あ)) "ラララア")))
 (test "string-tabulate" "アィイゥウ"
       (lambda ()
         (string-tabulate (lambda (code)
@@ -456,19 +456,19 @@
 
 (define (byte-list->port p bytes)
   (set! *flusher-out* '())
-  (for-each (lambda (b) (write-byte b p)) bytes)
+  (for-each (^b (write-byte b p)) bytes)
   (close-output-port p)
   *flusher-out*)
 
 (define (char-list->port p chars)
   (set! *flusher-out* '())
-  (for-each (lambda (c) (write-char c p)) chars)
+  (for-each (^c (write-char c p)) chars)
   (close-output-port p)
   *flusher-out*)
 
 (define (string-list->port p strs)
   (set! *flusher-out* '())
-  (for-each (lambda (s) (display s p)) strs)
+  (for-each (^s (display s p)) strs)
   (close-output-port p)
   *flusher-out*)
 
