@@ -81,6 +81,50 @@
 (test* "version>=?" #t (version>=? "1.1.12" "1.1.12"))
 (test* "version>=?" #t (version>=? "1.1.2" "1.1.1"))
 
+(let ()
+  (define (valid? expected spec)
+    (test* #"valid-version-spec? ,|spec|" expected (valid-version-spec? spec)))
+  (valid? #t "1")
+  (valid? #t "1a")
+  (valid? #t "1.3")
+  (valid? #t '(= "1.3"))
+  (valid? #t '(< "1.3"))
+  (valid? #t '(> "1.3"))
+  (valid? #t '(<= "1.3"))
+  (valid? #t '(>= "1.3"))
+  (valid? #t '(and))
+  (valid? #t '(and "1.3" "1.4"))
+  (valid? #t '(or))
+  (valid? #t '(or "1.3" "1.4"))
+  (valid? #t '(not "1.3"))
+  (valid? #t '(and (not "1.3.1")
+                   (>= "1.1")))
+  (valid? #t '(or (>= "2")
+                  (and (< "1.8.1")
+                       (>= "1.6.7"))))
+  (valid? #f 3.14)
+  (valid? #f '(= "1.3" "1.4"))
+  (valid? #f '(<< "1.3"))
+  (valid? #f '(not))
+  (valid? #f '(and not "3.3"))
+  )
+
+(let ()
+  (define (ok? spec tests)
+    (dolist [test tests]
+      (test* #"version ~(cadr test) satisfies ~|spec|" (car test)
+             (version-satisfy? spec (cadr test)))))
+
+  (ok? "1.2" '((#t "1.2") (#f "1.2.1") (#f "2.2")))
+  (ok? '(= "1.2") '((#t "1.2") (#f "1.2.1") (#f "2.2")))
+  (ok? '(< "1.2") '((#f "1.2") (#f "1.2.1") (#f "1.9") (#t "1.1") (#t "0.9")))
+  (ok? '(and (not "1.3.1") (>= "1.1"))
+       '((#t "1.1") (#t "1.2.1") (#t "1.3") (#f "1.3.1") (#t "1.4")
+         (#f "1.0.9")))
+  (ok? '(or (>= "2") (and (< "1.8.1") (>= "1.6.7")))
+       '((#f "1.6") (#t "1.7.1") (#f "1.8.1") (#f "1.10") (#t "2.0")))
+  )
+
 (test-end)
 
 
